/* This file is part of MoFEM.
 * MoFEM is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * MoFEM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MoFEM. If not, see <http://www.gnu.org/licenses/>. */

#ifdef WITH_TETGEN

#include <tetgen.h>
#ifdef REAL
  #undef REAL
#endif

#endif

// #include <moab/BVHTree.hpp>
#include <moab/SpatialLocator.hpp>
#include <MoFEM.hpp>
using namespace MoFEM;
#include <BasicFiniteElements.hpp>
#include <NeoHookean.hpp>
#include <Hooke.hpp>
#include <CrackPropagation.hpp>

ErrorCode rval;
PetscErrorCode ierr;

static char help[] =
"Testing moving weighted least approximation"
"n\n";

int main(int argc, char *argv[]) {
  PetscInitialize(&argc,&argv,(char *)0,help);

  try {

    PetscBool flg = PETSC_TRUE;
    char mesh_file_name[255];
    ierr = PetscOptionsGetString(PETSC_NULL,"","-my_file",mesh_file_name,255,&flg); CHKERRQ(ierr);

    if(flg != PETSC_TRUE) {
      SETERRQ(PETSC_COMM_SELF,1,"*** ERROR -my_file (MESH FILE NEEDED)");
    }

    moab::Core mb_instance;
    moab::Interface& moab = mb_instance;
    const char *option;
    option = "";
    rval = moab.load_file(mesh_file_name, 0, option); CHKERRQ_MOAB(rval);
    ParallelComm* pcomm = ParallelComm::get_pcomm(&moab,MYPCOMM_INDEX);
    if(pcomm == NULL) pcomm =  new ParallelComm(&moab,PETSC_COMM_WORLD);

    MoFEM::Core core(moab);
    MoFEM::Interface& m_field = core;

    CrackPropagation cp(m_field,3,2);
    ierr = cp.getOptions(); CHKERRQ(ierr);

    MeshsetsManager* block_manager_ptr;
    ierr = m_field.query_interface(block_manager_ptr); CHKERRQ(ierr);
    EntityHandle approx_meshset = 0;
    if(cp.residualStressBlock!=-1) {
      ierr = block_manager_ptr->getMeshset(
        cp.residualStressBlock,BLOCKSET,approx_meshset
      ); CHKERRQ(ierr);
    }

    Range tets;
    rval = moab.get_entities_by_type(approx_meshset,MBTET,tets,true); CHKERRQ_MOAB(rval);
    // cerr << tets << endl;
    // Range edges;
    // rval = moab.get_adjacencies(tets,1,true,edges,moab::Interface::UNION); CHKERRQ_MOAB(rval);
    EntityHandle meshset;
    rval = moab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset,tets); CHKERRQ_MOAB(rval);
    // rval = moab.convert_entities(meshset,true,false,false); CHKERRQ_MOAB(rval);

    CrackPropagation::MWLSApprox mwls(m_field);
    ierr = mwls.loadMWLSMesh(cp.mwlsApproxFile.c_str());

    // Create tahts to approximate data
    std::vector<Tag> tag_handles;
    rval = mwls.mwlsMoab.tag_get_tags(tag_handles);
    ublas::vector<Tag> tag_approx_handles(tag_handles.size());
    ublas::matrix<Tag> tag_approx_handles_diff(3,tag_handles.size());
    double def_vals[9];
    fill(def_vals,&def_vals[9],0);
    for(int t = 0;t!=tag_handles.size();t++) {
      std::string name;
      rval = mwls.mwlsMoab.tag_get_name(tag_handles[t],name); CHKERRQ_MOAB(rval);
      // cerr << name << endl;
      if(name.compare(0,3,"MED")==0) {
        PetscPrintf(PETSC_COMM_WORLD,"Pre-process tag data < %s >\n",name.c_str());
        ierr = mwls.getValuesToNodes(tag_handles[t]); CHKERRQ(ierr);
        std::string approx_name = "APPROX_" + name;
        std::string approx_diff_name[] = {
          "APPROX_DX_" + name,"APPROX_DY_" + name,"APPROX_DZ_" + name
        };
        int length;
        rval = mwls.mwlsMoab.tag_get_length(tag_handles[t],length); CHKERRQ_MOAB(rval);
        DataType data_type;
        rval = mwls.mwlsMoab.tag_get_data_type(tag_handles[t],data_type); CHKERRQ_MOAB(rval);
        Tag th_approx;
        rval = moab.tag_get_handle(
          approx_name.c_str(),length,data_type,th_approx,MB_TAG_CREAT|MB_TAG_SPARSE,def_vals
        ); CHKERRQ_MOAB(rval);
        tag_approx_handles[t]=th_approx;
        for(int d = 0;d!=3;d++) {
          Tag th_approx_diff;
          rval = moab.tag_get_handle(
            approx_diff_name[d].c_str(),length,data_type,th_approx_diff,MB_TAG_CREAT|MB_TAG_SPARSE,def_vals
          ); CHKERRQ_MOAB(rval);
          tag_approx_handles_diff(d,t) =  th_approx_diff;
        }
      }
    }

    Range nodes;
    rval = moab.get_connectivity(tets,nodes,false); CHKERRQ_MOAB(rval);
    // cerr << nodes << endl;

    // Approximate tags
    for(Range::iterator nit = nodes.begin();nit!=nodes.end();nit++) {
      double coords[3];
      rval = moab.get_coords(&*nit,1,coords); CHKERRQ_MOAB(rval);
      ierr = mwls.getInfluenceNodes(coords); CHKERRQ(ierr);
      ierr = mwls.caculateApproxFun(coords); CHKERRQ(ierr);
      for(int t = 0;t!=tag_handles.size();t++) {
        std::string name;
        rval = mwls.mwlsMoab.tag_get_name(tag_handles[t],name); CHKERRQ_MOAB(rval);
        if(name.compare(0,3,"MED")==0) {
          if(nit == nodes.begin()) {
            PetscPrintf(PETSC_COMM_WORLD,"Approxmate tag < %s >\n",name.c_str());
          }
          ierr = mwls.approxTagData(tag_handles[t]); CHKERRQ(ierr);
          CrackPropagation::MWLSApprox::VecVals vals = mwls.getData();
          CrackPropagation::MWLSApprox::MatDiffVals diff_vals = mwls.getDiffData();
          // cerr << diff_vals << endl;
          rval = moab.tag_set_data(tag_approx_handles[t],&*nit,1,&*vals.data().begin()); CHKERRQ_MOAB(rval);
          for(int d = 0;d!=3;d++) {
            rval = moab.tag_set_data(
              tag_approx_handles_diff(d,t),&*nit,1,&diff_vals(d,0)
            ); CHKERRQ_MOAB(rval);
          }
        }
      }
    }

    rval = moab.write_file("out_mwls.h5m","MOAB","",&meshset,1); CHKERRQ_MOAB(rval);
    rval = mwls.mwlsMoab.write_file("out_mwls_moab.h5m","MOAB"); CHKERRQ_MOAB(rval);


  } catch (MoFEMException const &e) {
    SETERRQ(PETSC_COMM_SELF,e.errorCode,e.errorMessage);
  }

  PetscFinalize();
}
