/* This file is part of MoFEM.
 * MoFEM is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * MoFEM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MoFEM. If not, see <http://www.gnu.org/licenses/>. */

#ifndef __CONSTANT_AREA_HPP__
#define __CONSTANT_AREA_HPP__

#ifndef WITH_ADOL_C
  #error "MoFEM need to be compiled with ADOL-C"
#endif

/** \brief Constant area constrains
*/
struct ConstantArea {

  MoFEM::Interface &mField;
  ConstantArea(MoFEM::Interface &m_field):
  mField(m_field),
  commonData(m_field) {
  }

  struct CommonData {

    MoFEM::Interface &mField;
    CommonData(MoFEM::Interface &m_field):
    mField(m_field) {}

    ublas::vector<double> crackAreaIncrement;
    ublas::matrix<double> tangentCrackAreaIncrement;
    vector<double*> tangentCrackAreaIncrementRowPtr;

    ublas::vector<double> crackFrontTangent;
    ublas::matrix<double> tangentCrackFrontTangent;
    vector<double*> tangentCrackFrontTangentRowPtr;

    map<int,Vec> mapV;

    Tag thInterfaceSide;


  };
  CommonData commonData;

  struct MyTriangleFE: public FaceElementForcesAndSourcesCore {

    CommonData &commonData;
    Mat petscB;
    Vec petscF;
    Mat petscQ;

    MyTriangleFE(MoFEM::Interface &m_field,CommonData &common_data):
    FaceElementForcesAndSourcesCore(m_field),
    commonData(common_data),
    petscB(PETSC_NULL),
    petscF(PETSC_NULL),
    petscQ(PETSC_NULL)
    {}
    int getRule(int order) { return order; };

    PetscErrorCode preProcess() {
      PetscFunctionBegin;
      PetscErrorCode ierr;

      ierr = FaceElementForcesAndSourcesCore::preProcess(); CHKERRQ(ierr);

      if(petscB != PETSC_NULL) {
        snes_B = petscB;
      }

      if(petscF != PETSC_NULL) {
        snes_f = petscF;
      }

      switch (ts_ctx) {
        case CTX_TSSETIFUNCTION: {
          if(!petscF) {
            snes_ctx = CTX_SNESSETFUNCTION;
            snes_f = ts_F;
          }
          break;
        }
        case CTX_TSSETIJACOBIAN: {
          if(!petscB) {
            snes_ctx = CTX_SNESSETJACOBIAN;
            snes_B = ts_B;
          }
          break;
        }
        default:
        break;
      }

      // Tag to detect whcih side is it
      rval = mField.get_moab().tag_get_handle(
        "INTERFACE_SIDE",commonData.thInterfaceSide
      ); CHKERRQ_MOAB(rval);

      // If petscQ matrix is defined appply projection to each row
      if(petscQ != PETSC_NULL) {
        if(petscB == PETSC_NULL) {
          SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Matrix B should be set");
        }
        for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_NAME_FOR_LOOP_(problemPtr,"LAMBDA_CRACKFRONT_AREA",dit)) {
          #if PETSC_VERSION_GE(3,5,3)
          ierr = MatCreateVecs(
            petscB,&commonData.mapV[dit->get()->getPetscGlobalDofIdx()],PETSC_NULL
          ); CHKERRQ(ierr);
          #else
          ierr = MatGetVecs(petscB,&commonData.mapV[dit->get()->getPetscGlobalDofIdx()],PETSC_NULL); CHKERRQ(ierr);
          #endif
          ierr = VecZeroEntries(commonData.mapV[dit->get()->getPetscGlobalDofIdx()]); CHKERRQ(ierr);
        }
      }

      PetscFunctionReturn(0);
    }

    PetscErrorCode postProcess() {
      PetscFunctionBegin;
      ierr = FaceElementForcesAndSourcesCore::postProcess(); CHKERRQ(ierr);

      if(petscQ != PETSC_NULL) {
        if(petscB == PETSC_NULL) {
          SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Matrix B should be set");
        }
        Vec Qv;
        ierr = mField.VecCreateGhost(problemPtr->getName(),COL,&Qv); CHKERRQ(ierr);
        for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_NAME_FOR_LOOP_(problemPtr,"LAMBDA_CRACKFRONT_AREA",dit)) {
          if(commonData.mapV.find(dit->get()->getPetscGlobalDofIdx())==commonData.mapV.end()) {
            SETERRQ(PETSC_COMM_SELF,MOFEM_DATA_INCONSISTENCY,"data inconsistency");
          }
          ierr = VecAssemblyBegin(commonData.mapV[dit->get()->getPetscGlobalDofIdx()]); CHKERRQ(ierr);
          ierr = VecAssemblyEnd(commonData.mapV[dit->get()->getPetscGlobalDofIdx()]); CHKERRQ(ierr);
          ierr = MatMult(petscQ,commonData.mapV[dit->get()->getPetscGlobalDofIdx()],Qv); CHKERRQ(ierr);
          ierr = VecGhostUpdateBegin(Qv,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
          ierr = VecGhostUpdateEnd(Qv,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
          double *array;
          ierr = VecGetArray(Qv,&array); CHKERRQ(ierr);
          vector<int> glob_idx;
          vector<double> vals;
          for(_IT_NUMEREDDOFMOFEMENTITY_COL_BY_ENT_FOR_LOOP_(problemPtr,dit->get()->getEnt(),diit)) {
            if(diit->get()->getPart()!=mField.get_comm_rank()) continue;
            if(diit->get()->getName()!="MESH_NODE_POSITIONS") continue;
            glob_idx.push_back(diit->get()->getPetscGlobalDofIdx());
            if(diit->get()->getPetscLocalDofIdx()<0) {
              SETERRQ(PETSC_COMM_SELF,MOFEM_DATA_INCONSISTENCY,"data inconsistency");
            }
            vals.push_back(array[diit->get()->getPetscLocalDofIdx()]);
          }
          int row = dit->get()->getPetscGlobalDofIdx();
          ierr = MatSetValues(
            petscB,
            1,&row,glob_idx.size(),
            glob_idx.empty()?PETSC_NULL:&glob_idx[0],
            vals.empty()?PETSC_NULL:&vals[0],
            ADD_VALUES
          ); CHKERRQ(ierr);
          ierr = VecRestoreArray(Qv,&array); CHKERRQ(ierr);
          ierr = VecDestroy(&commonData.mapV[dit->get()->getPetscGlobalDofIdx()]); CHKERRQ(ierr);
        }
        ierr = VecDestroy(&Qv); CHKERRQ(ierr);
        commonData.mapV.clear();
      }

      PetscFunctionReturn(0);
    }


  };

  boost::shared_ptr<MyTriangleFE> feRhsPtr;
  boost::shared_ptr<MyTriangleFE> feLhsPtr;

  struct OpAreaJacobian: public FaceElementForcesAndSourcesCore::UserDataOperator {

    int tAg;
    CommonData &commonData;
    bool isTapeRecorded;

    OpAreaJacobian(int tag,CommonData &common_data):
    FaceElementForcesAndSourcesCore::UserDataOperator("MESH_NODE_POSITIONS",UserDataOperator::OPROW),
    tAg(tag),
    commonData(common_data),
    isTapeRecorded(false) {}

    /**

    \f[
    A = \| \mathbf{N} \| =
      \left\|
        \textrm{Spin}\left[
          \frac{\partial\mathbf{N}\mathbf{X}}{\partial\xi}
        \right]
        \frac{\partial\mathbf{N}\mathbf{X}}{\partial\eta}
        \right\|
    \f]

    */
    template<typename TYPE>
    struct AuxFunctions {

      ublas::matrix<double> N,NTN;
      ublas::matrix<double> Bksi;
      ublas::matrix<double> Beta;

      ublas::vector<TYPE> referenceCoords;

      ublas::vector<TYPE> referenceXdKsi;
      ublas::vector<TYPE> referenceXdEta;
      ublas::matrix<TYPE> referenceSpinKsi;
      ublas::matrix<TYPE> referenceSpinEta;
      ublas::vector<TYPE> referenceNormal;
      TYPE referenceArea;

      ublas::vector<TYPE> currentCoords;
      ublas::vector<TYPE> currentXdKsi;
      ublas::vector<TYPE> currentXdEta;
      ublas::matrix<TYPE> currentSpinKsi;
      ublas::matrix<TYPE> currentSpinEta;
      ublas::vector<TYPE> currentNormal;
      TYPE currentArea;

      ublas::matrix<TYPE> A;
      ublas::vector<TYPE> crackAreaIncrement;
      ublas::vector<TYPE> crackFrontTangent;

      PetscErrorCode sPin(ublas::matrix<TYPE> &spin,ublas::vector<TYPE> &vec) {
        PetscFunctionBegin;
        spin.resize(3,3,false);
        spin.clear();
        spin(0,1) = -vec[2];
        spin(0,2) = +vec[1];
        spin(1,0) = +vec[2];
        spin(1,2) = -vec[0];
        spin(2,0) = -vec[1];
        spin(2,1) = +vec[0];
        PetscFunctionReturn(0);
      }

      PetscErrorCode matrixB(int gg,DataForcesAndSurcesCore::EntData &data) {
        PetscFunctionBegin;
        Bksi.resize(3,9);
        Bksi.clear();
        for(int ii = 0;ii<3;ii++) {
          for(int jj = 0;jj<3;jj++) {
            Bksi(jj,ii*3+jj) = data.getDiffN(gg)(ii,0);
          }
        }
        Beta.resize(3,9);
        Beta.clear();
        for(int ii = 0;ii<3;ii++) {
          for(int jj = 0;jj<3;jj++) {
            Beta(jj,ii*3+jj) = data.getDiffN(gg)(ii,1);
          }
        }
        PetscFunctionReturn(0);
      }

      PetscErrorCode dIffX() {
        PetscFunctionBegin;
        currentXdKsi.resize(3,false);
        currentXdEta.resize(3,false);
        noalias(currentXdKsi) = prod(Bksi,currentCoords);
        noalias(currentXdEta) = prod(Beta,currentCoords);
        PetscFunctionReturn(0);
      }

      PetscErrorCode nOrmal() {
        PetscFunctionBegin;
        PetscErrorCode ierr;
        ierr = sPin(currentSpinKsi,currentXdKsi); CHKERRQ(ierr);
        ierr = sPin(currentSpinEta,currentXdEta); CHKERRQ(ierr);
        currentNormal.resize(3,false);
        noalias(currentNormal) = 0.5*prod(currentSpinKsi,currentXdEta);
        currentArea = 0;
        for(int dd = 0;dd!=3;dd++) {
          currentArea += currentNormal[dd]*currentNormal[dd];
        }
        currentArea = sqrt(currentArea);
        PetscFunctionReturn(0);
      }

      PetscErrorCode matrixA() {
        PetscFunctionBegin;
        A.resize(3,9,false);
        noalias(A) = 0.5*(prod(currentSpinKsi,Beta) - prod(currentSpinEta,Bksi));
        PetscFunctionReturn(0);
      }

      PetscErrorCode calculateAreaIncrementDirection(double beta) {
        PetscFunctionBegin;
        for(int dd = 0;dd!=9;dd++) {
          for(int ii = 0;ii!=3;ii++) {
            crackAreaIncrement[dd] += beta*A(ii,dd)*currentNormal[ii]/currentArea;
          }
        }
        PetscFunctionReturn(0);
      }

      // FIXME Need to implement this with antisymmetric verison need fix in FTensor

      FTensor::Tensor2<TYPE,3,3> tSpin;

      PetscErrorCode calculateFrontTangent(double beta) {
        PetscFunctionBegin;
        FTensor::Index<'i',3> i;
        FTensor::Index<'j',3> j;
        tSpin(i,j) = 0;
        tSpin(0,1) = -currentNormal[2];
        tSpin(0,2) = +currentNormal[1];
        tSpin(1,0) = +currentNormal[2];
        tSpin(1,2) = -currentNormal[0];
        tSpin(2,0) = -currentNormal[1];
        tSpin(2,1) = +currentNormal[0];
        tSpin(i,j) /= beta*currentArea;
        crackFrontTangent.resize(9,false);
        FTensor::Tensor1<TYPE*,3> t_crack_front_tangent(
          &crackFrontTangent[0],&crackFrontTangent[1],&crackFrontTangent[2],3
        );
        FTensor::Tensor1<TYPE*,3> t_area_increment(
          &crackAreaIncrement[0],&crackAreaIncrement[1],&crackAreaIncrement[2],3
        );
        for(int ii = 0;ii!=3;ii++) {
          t_crack_front_tangent(i) = tSpin(i,j)*t_area_increment(j);
          ++t_area_increment;
          ++t_crack_front_tangent;
        }
        PetscFunctionReturn(0);
      }


    };

    AuxFunctions<adouble> auxFun;

    PetscErrorCode doWork(int side,EntityType type,DataForcesAndSurcesCore::EntData &data) {
      PetscFunctionBegin;

      if(isTapeRecorded) {
        PetscFunctionReturn(0);
      }
      if(type != MBVERTEX) {
        PetscFunctionReturn(0);
      }

      try {

        PetscErrorCode ierr;
        int nb_dofs = data.getFieldData().size();
        int nb_gauss_pts = data.getN().size1();

        auxFun.referenceCoords.resize(9,false);
        auxFun.currentCoords.resize(9,false);

        trace_on(tAg);

        for(int dd = 0;dd!=nb_dofs;dd++) {
          auxFun.referenceCoords[dd] <<= getCoords()[dd];
        }
        for(int dd = 0;dd!=nb_dofs;dd++) {
          auxFun.currentCoords[dd] <<= data.getFieldData()[dd];
        }

        auxFun.crackAreaIncrement.resize(9,false);
        auxFun.crackAreaIncrement.clear();

        for(int gg = 0;gg!=nb_gauss_pts;gg++) {

          double val = getGaussPts()(2,gg)*0.5;

          ierr = auxFun.matrixB(gg,data); CHKERRQ(ierr);
          ierr = auxFun.dIffX(); CHKERRQ(ierr);
          ierr = auxFun.nOrmal(); CHKERRQ(ierr);
          ierr = auxFun.matrixA(); CHKERRQ(ierr);
          ierr = auxFun.calculateAreaIncrementDirection(val); CHKERRQ(ierr);

        }

        //cerr << "Griffith Force " << auxFun.griffithForce << endl;

        commonData.crackAreaIncrement.resize(nb_dofs,false);
        for(int dd = 0;dd!=nb_dofs;dd++) {
          auxFun.crackAreaIncrement[dd] >>= commonData.crackAreaIncrement[dd];
        }

        trace_off();

        isTapeRecorded = true;


      } catch (const std::exception& ex) {
        ostringstream ss;
        ss << "throw in method: " << ex.what() << endl;
        SETERRQ(PETSC_COMM_SELF,1,ss.str().c_str());
      }

      PetscFunctionReturn(0);
    }

  };

  struct AuxOp {

    int tAg;
    CommonData &commonData;
    AuxOp(int tag,CommonData &common_data):
    tAg(tag),
    commonData(common_data) {
    }

    ublas::vector<int> rowIndices;
    ublas::vector<double> activeVariables;

    PetscErrorCode setVariables(
      FaceElementForcesAndSourcesCore::UserDataOperator *fe_ptr,
      int side,
      EntityType type,
      DataForcesAndSurcesCore::EntData &data
    ) {
      PetscFunctionBegin;
      int nb_dofs = data.getIndices().size();
      activeVariables.resize(18);
      for(int dd = 0;dd!=nb_dofs;dd++) {
        activeVariables[dd] = fe_ptr->getCoords()[dd];
      }
      for(int dd = 0;dd!=nb_dofs;dd++) {
        activeVariables[9+dd] = data.getFieldData()[dd];
      }
      PetscFunctionReturn(0);
    }

  };

  struct OpAreaC: public FaceElementForcesAndSourcesCore::UserDataOperator,AuxOp {

    OpAreaC(int tag,CommonData &common_data):
    FaceElementForcesAndSourcesCore::UserDataOperator(
      "LAMBDA_CRACKFRONT_AREA",
      "MESH_NODE_POSITIONS",
      UserDataOperator::OPROWCOL
    ),
    AuxOp(tag,common_data) {
    }

    MatrixDouble cMat;

    PetscErrorCode doWork(
      int row_side,int col_side,
      EntityType row_type,EntityType col_type,
      DataForcesAndSurcesCore::EntData &row_data,
      DataForcesAndSurcesCore::EntData &col_data
    ) {
      PetscErrorCode ierr;
      PetscFunctionBegin;

      if(row_side != MBVERTEX) {
        PetscFunctionReturn(0);
      }
      VectorInt &row_indices = row_data.getIndices();
      const int nb_rows = row_indices.size();
      if(!nb_rows) PetscFunctionReturn(0);
      VectorInt &col_indices = col_data.getIndices();
      const int nb_cols = col_indices.size();
      if(!nb_cols) PetscFunctionReturn(0);
      cMat.resize(3,9,false);
      cMat.clear();

      ierr = setVariables(this,col_side,col_type,col_data); CHKERRQ(ierr);

      int nb_dofs = col_data.getIndices().size();
      commonData.crackAreaIncrement.resize(nb_dofs,false);
      int r;
      //play recorder for values
      r = ::function(tAg,nb_dofs,18,&activeVariables[0],&commonData.crackAreaIncrement[0]);
      if(r<3) {
        SETERRQ1(PETSC_COMM_SELF,MOFEM_OPERATION_UNSUCCESSFUL,"ADOL-C function evaluation with error r = %d",r);
      }

      for(int rr = 0;rr!=nb_rows;rr++) {
        if(row_indices[rr] != -1) {
          cMat(rr,3*rr+0) = commonData.crackAreaIncrement(3*rr+0);
          cMat(rr,3*rr+1) = commonData.crackAreaIncrement(3*rr+1);
          cMat(rr,3*rr+2) = commonData.crackAreaIncrement(3*rr+2);
        }
      }

      if(commonData.mapV.size()==0) {
        ierr = MatSetValues(
          getFEMethod()->snes_B,
          nb_rows,&row_indices[0],
          nb_cols,&col_indices[0],
          &cMat(0,0),
          ADD_VALUES
        ); CHKERRQ(ierr);
      } else {
        for(int rr = 0;rr!=nb_rows;rr++) {
          if(row_indices[rr] != -1) {
            if(commonData.mapV.find(row_indices[rr])==commonData.mapV.end()) {
              SETERRQ(PETSC_COMM_SELF,MOFEM_DATA_INCONSISTENCY,"data inconsistency");
            }
            ierr = VecSetValues(
              commonData.mapV[row_indices[rr]],
              nb_cols,&col_indices[0],
              &cMat(rr,0),ADD_VALUES
            ); CHKERRQ(ierr);
          }
        }
      }

      PetscFunctionReturn(0);
    }

  };

  struct OpTangentJacobian: public OpAreaJacobian {

    OpTangentJacobian(int tag,CommonData &common_data):
    OpAreaJacobian(tag,common_data) {}

    PetscErrorCode doWork(int side,EntityType type,DataForcesAndSurcesCore::EntData &data) {
      PetscFunctionBegin;

      if(isTapeRecorded) {
        PetscFunctionReturn(0);
      }
      if(type != MBVERTEX) {
        PetscFunctionReturn(0);
      }

      try {

        PetscErrorCode ierr;
        int nb_dofs = data.getFieldData().size();
        int nb_gauss_pts = data.getN().size1();

        auxFun.referenceCoords.resize(9,false);
        auxFun.currentCoords.resize(9,false);

        trace_on(tAg);

        for(int dd = 0;dd!=nb_dofs;dd++) {
          auxFun.referenceCoords[dd] <<= getCoords()[dd];
        }
        for(int dd = 0;dd!=nb_dofs;dd++) {
          auxFun.currentCoords[dd] <<= data.getFieldData()[dd];
        }

        auxFun.crackAreaIncrement.resize(9,false);
        auxFun.crackAreaIncrement.clear();
        auxFun.crackFrontTangent.resize(9,false);
        auxFun.crackFrontTangent.clear();

        for(int gg = 0;gg!=nb_gauss_pts;gg++) {

          double val = getGaussPts()(2,gg)*0.5;

          ierr = auxFun.matrixB(gg,data); CHKERRQ(ierr);
          ierr = auxFun.dIffX(); CHKERRQ(ierr);
          ierr = auxFun.nOrmal(); CHKERRQ(ierr);
          ierr = auxFun.matrixA(); CHKERRQ(ierr);
          ierr = auxFun.calculateAreaIncrementDirection(1); CHKERRQ(ierr);
          ierr = auxFun.calculateFrontTangent(val); CHKERRQ(ierr);

        }

        commonData.crackFrontTangent.resize(nb_dofs,false);
        for(int dd = 0;dd!=nb_dofs;dd++) {
          auxFun.crackFrontTangent[dd] >>= commonData.crackFrontTangent[dd];
        }

        trace_off();

        isTapeRecorded = true;


      } catch (const std::exception& ex) {
        ostringstream ss;
        ss << "throw in method: " << ex.what() << endl;
        SETERRQ(PETSC_COMM_SELF,1,ss.str().c_str());
      }

      PetscFunctionReturn(0);
    }

  };

  struct OpTangentC: public FaceElementForcesAndSourcesCore::UserDataOperator,AuxOp {

    OpTangentC(int tag,CommonData &common_data):
    FaceElementForcesAndSourcesCore::UserDataOperator(
      "LAMBDA_CRACKFRONT_AREA_TANGENT",
      "MESH_NODE_POSITIONS",
      UserDataOperator::OPROWCOL
    ),
    AuxOp(tag,common_data) {
    }

    MatrixDouble cMat;

    PetscErrorCode doWork(
      int row_side,int col_side,
      EntityType row_type,EntityType col_type,
      DataForcesAndSurcesCore::EntData &row_data,
      DataForcesAndSurcesCore::EntData &col_data
    ) {
      PetscErrorCode ierr;
      MoABErrorCode rval;
      PetscFunctionBegin;

      if(row_side != MBVERTEX) {
        PetscFunctionReturn(0);
      }

      VectorInt &row_indices = row_data.getIndices();
      const int nb_rows = row_indices.size();
      if(!nb_rows) PetscFunctionReturn(0);
      VectorInt &col_indices = col_data.getIndices();
      const int nb_cols = col_indices.size();
      if(!nb_cols) PetscFunctionReturn(0);
      cMat.resize(3,9,false);
      cMat.clear();

      ierr = setVariables(this,col_side,col_type,col_data); CHKERRQ(ierr);

      int nb_dofs = col_data.getIndices().size();
      commonData.crackFrontTangent.resize(nb_dofs,false);
      int r;
      //play recorder for values
      r = ::function(tAg,nb_dofs,18,&activeVariables[0],&commonData.crackFrontTangent[0]);
      if(r<3) {
        SETERRQ1(PETSC_COMM_SELF,MOFEM_OPERATION_UNSUCCESSFUL,"ADOL-C function evaluation with error r = %d",r);
      }

      // Set face orientation
      {
        // Get orientation in respect to adjacent tet
        const EntityHandle face = getFEEntityHandle();
        const BitRefLevel bit = getFEMethod()->problemPtr->getBitRefLevel();
        Range adj_side_elems;
        ierr = commonData.mField.get_adjacencies(
          bit,&face,1,3,adj_side_elems
        ); CHKERRQ(ierr);
        adj_side_elems = adj_side_elems.subset_by_type(MBTET);
        if(adj_side_elems.size()!=1) {
          SETERRQ1(
            commonData.mField.get_comm(),
            MOFEM_DATA_INCONSISTENCY,
            "expect 1 tet but is %u",
            adj_side_elems.size()
          );
        }
        EntityHandle side_elem = *adj_side_elems.begin();
        int side_number,sense,offset;
        rval = commonData.mField.get_moab().side_number(
          side_elem,face,side_number,sense,offset
        ); CHKERRQ_MOAB(rval);
        if(sense == -1) {
          commonData.crackFrontTangent *= -1;
        }
        // Get orientaton of face in respect to face in the prism
        int side;
        rval = commonData.mField.get_moab().tag_get_data(
          commonData.thInterfaceSide,&face,1,&side
        ); CHKERRQ_MOAB(rval);
        if(side == 1) {
          commonData.crackFrontTangent *= -1;
        }
      }

      for(int rr = 0;rr!=nb_rows;rr++) {
        if(row_indices[rr] != -1) {
          // cerr << commonData.crackFrontTangent(3*rr+0) << " "
          // << commonData.crackFrontTangent(3*rr+1) << " "
          // << commonData.crackFrontTangent(3*rr+2) << endl;
          cMat(rr,3*rr+0) = commonData.crackFrontTangent[3*rr+0];
          cMat(rr,3*rr+1) = commonData.crackFrontTangent[3*rr+1];
          cMat(rr,3*rr+2) = commonData.crackFrontTangent[3*rr+2];
        }
      }

      ierr = MatSetValues(
        getFEMethod()->snes_B,
        nb_rows,&row_indices[0],
        nb_cols,&col_indices[0],
        &cMat(0,0),
        ADD_VALUES
      ); CHKERRQ(ierr);

      PetscFunctionReturn(0);
    }

  };

};


#endif //__CONSTANT_AREA_HPP__
