/** \file CrackPropagation.hpp
*/

/* This file is part of MoFEM.
 * MoFEM is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * MoFEM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MoFEM. If not, see <http://www.gnu.org/licenses/>. */


#ifndef __CRACK_PROAGATION_HPP__
#define __CRACK_PROAGATION_HPP__

struct CrackPropagation {

  MoFEM::Interface &mField;
  int approxOrder;
  int geometryOrder;
  double gC;
  PetscBool isPartitioned;
  int refAtCrackTip;
  int refOrderAtTip;

  std::string mwlsApproxFile;
  int residualStressBlock;
  std::string medFileName;
  std::string configFile;

  CrackPropagation(MoFEM::Interface &m_field,const int approx_order = 2,const int geometry_order = 1);
  virtual ~CrackPropagation();

  /**
   * \brief Get opyions form command line
   * @return error code
   */
  PetscErrorCode getOptions();

  inline int& getNbLoadSteps() { return nbLoadSteps; }
  inline double& getLoadScale() { return loadScale; }
  inline int getNbLoadSteps() const { return nbLoadSteps; }
  inline double getLoadScale() const { return loadScale; }

  PetscErrorCode readMedFile();

  PetscErrorCode readConfigFileOptions();

  PetscErrorCode getElastcBlockTets();

  PetscErrorCode resolveShared(
    const Range &tets,
    Range &proc_ents,
    const int verb = 1,
    const bool debug = false
  );

  PetscErrorCode resolveSharedBlockTest(
    const int verb = 1,
    const bool debug = false
  );

  PetscErrorCode resolveSharedBitRefLevel(
    const BitRefLevel bit,
    const int verb = 1,
    const bool debug = false
  );

  PetscErrorCode findBodySkin(
    const BitRefLevel bit = BitRefLevel().set(0),
    const int id = 102,
    const int verb = 1,const bool debug = false
  );

  PetscErrorCode getCrack(
    const BitRefLevel bit = BitRefLevel().set(0),
    const int crack_id = 200,
    const int front_id = 201,
    const int verb = 1,
    const bool debug = false
  );

  /**
   * \brief Declate fields for elastic analysis
   * @param  bit bit refinment level
   * @return     error code
   */
  PetscErrorCode buildElasticFields(
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set(),
    const bool proc_only = true,
    const bool build_fields  = true
  );

  /**
   * \brief Declate field for arc-length
   * @param  bit bit refinment level
   * @return     error code
   */
  PetscErrorCode buildArcLengthField(
    const bool build_fields  = true
  );


  PetscErrorCode declareElasticFE(
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set(),
    const bool add_forces = true,
    const bool proc_only = true
  );
  PetscErrorCode declareMaterialFE(
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set(),
    const bool add_forces = true,
    const bool proc_only = true
  );
  PetscErrorCode buildSurfaceFields(
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set(),
    const int id = 102,
    const bool proc_only = true,
    const bool build_fields  = true,
    const int verb = 1,
    const bool debug = false
  );
  PetscErrorCode declareSurfaceFE(
    std::string fe_name = "SURFACE",
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set(),
    const int id = 102,
    const bool proc_only = true,
    const int verb = 1,
    const bool debug = false
  );
  PetscErrorCode buildCrackSurfaceFieldId(
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set(),
    const int surf_id = 200,
    const bool proc_only = true,
    const bool build_fields  = true,
    const int verb = 1,
    const bool debug = false
  );
  PetscErrorCode declareCrackSurfaceFE(
    std::string fe_name = "CRACK_SURFACE",
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set(),
    const int id = 102,
    const bool proc_only = true,
    const int verb = 1,
    const bool debug = false
  );

  PetscErrorCode crackCut(
    const BitRefLevel bit1 = BitRefLevel().set(1),
    const BitRefLevel bit2 = BitRefLevel().set(2),
    const int verb = 1,
    const bool debug = false
  );

  PetscErrorCode splitFaces(
    BitRefLevel bit = BitRefLevel().set(0),
    BitRefLevel bit_split = BitRefLevel().set(1),
    const int crack_id = 200,
    const int verb = 1,
    const bool debug = false
  );

  PetscErrorCode refineCrackTip(
    BitRefLevel bit = BitRefLevel().set(0),
    BitRefLevel bit_ref = BitRefLevel().set(1),
    const int front_id = 201,
    const int verb = 1,
    const bool debug = false
  );

  PetscErrorCode createElasticDM(
    DM *dm,
    const std::string prb_name,
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set()
  );
  PetscErrorCode createBcDM(
    DM *dm,
    const std::string prb_name,
    const BitRefLevel bit = BitRefLevel().set(0),
    const BitRefLevel mask = BitRefLevel().set()
  );
  PetscErrorCode createMaterialDM(
    DM *dm,
    const std::string prb_name,
    const BitRefLevel bit,
    const BitRefLevel mask,
    const std::vector<std::string> fe_list
  );
  PetscErrorCode createMaterialForcesDM(
    DM *dm,
    DM dm_material,
    const std::string prb_name
  );
  PetscErrorCode createSurfaceProjectionDM(
    DM *dm,
    DM dm_material,
    const std::string prb_name,
    const std::vector<int> surface_ids,
    const std::vector<std::string> fe_list
  );

  PetscErrorCode createCrackFrontAreaDM(
    DM *dm,DM dm_material,
    const std::string prb_name
  );

  PetscErrorCode assembleElasticDM(
    DM dm,Mat m,Vec q,Vec f,const int verb = 1,const bool debug = false
  );
  PetscErrorCode assembleMaterialForcesDM(
    DM dm,const int verb = 1,const bool debug = false
  );
  PetscErrorCode calculateElasticEnergy(DM dm);
  PetscErrorCode calculateMaterialFocesDM(
    DM dm,Vec q,Vec f,const int verb = 1,const bool debug = false
  );
  PetscErrorCode calculateSurfaceProjectionMatrix(
    DM dm_front,
    DM dm_project,
    std::vector<int> &ids,
    const int verb  = 1,
    const bool debug = false
  );
  PetscErrorCode calculateFrontProjectionMatrix(
    DM dm_surface,
    DM dm_project,
    const int verb  = 1,
    const bool debug = false
  );
  PetscErrorCode projectMaterialForcesDM(
    DM dm_project,
    Vec f,
    Vec f_proj,
    const int verb  = 1,
    const bool debug = false
  );
  PetscErrorCode calculateGriffithForce(
    DM dm,const double gc,
    Vec f_griffith,
    const int verb = 1,
    const bool debug = false
  );
  PetscErrorCode projectGriffithForce(
    DM dm,
    Vec f_griffith,
    Vec f_griffith_proj,
    const int verb = 1,
    const bool debug = false
  );
  PetscErrorCode calculateReleaseEneregy(
    DM dm,
    Vec f_material_proj,
    Vec f_griffith_proj,
    Vec f_lambda,
    const double gc,
    const int verb = 1,
    const bool debug = true
  );

  PetscErrorCode savePositionsOnCrackFrontDM(
    DM dm,Vec q,const int verb = 1,const bool debug = false
  );
  PetscErrorCode writeCrackFont(const int step);
  PetscErrorCode solveElasticDM(
    DM dm,SNES snes,Mat m,Vec q,Vec f,bool snes_set_up
  );
  PetscErrorCode postProcessElasticDM(DM dm,const int step);

  PetscErrorCode setPositionFromCoords(const std::string field_name);
  PetscErrorCode setMatrialPositionFromCoords();
  PetscErrorCode setSpatialPositionFromCoords();

  /**
   *
   * Use idea from \cite barsoum1976use to approximate singularity at crack tip.
   * The idea is to shift mid noted to quarter edge length to crack tip on edges
   * adjacent to crack front.
   *
   */
  PetscErrorCode getFrontEdgesAndElements(
    const BitRefLevel bit = BitRefLevel().set(0),
    const int verb = 1
  );

  static bool setSingularCoordinatesSet;

  PetscErrorCode setSingularDofs(const string field_name,const int verb = 1);
  PetscErrorCode setSingularElementMatrialPositions(const int verb = 1);
  PetscErrorCode unsetSingularElementMatrialPositions();
  PetscErrorCode cleanSingularElementMatrialPositions();

  PetscErrorCode setCrackFrontBitLevel(
    BitRefLevel from_bit = BitRefLevel().set(1),
    BitRefLevel bit = BitRefLevel().set(2)
  );

  boost::shared_ptr<DofEntity> arcLengthDof;
  PetscErrorCode getArcLengthDof();

  struct CrackFrontFE: public NonlinearElasticElement::MyVolumeFE {

    Range crackFrontNodes;
    Range crackFrontNodesEdges;
    Range crackFrontElements;

    MatrixDouble sJac;
    MatrixDouble invSJac;
    MatrixDouble singularCoords;

    bool singularElement;
    bool singularEdges[6];
    bool singularNodes[4];

    int getRule(int order) {
      return -1;
    }

    CrackFrontFE(
      MoFEM::Interface &m_field,
      const Range &crack_front_nodes,
      const Range &crack_front_nodes_edges,
      const Range &crack_front_elements
    ):
    NonlinearElasticElement::MyVolumeFE(m_field),
    crackFrontNodes(crack_front_nodes),
    crackFrontNodesEdges(crack_front_nodes_edges),
    crackFrontElements(crack_front_elements) {
    }

    CrackFrontFE(MoFEM::Interface &m_field):
    NonlinearElasticElement::MyVolumeFE(m_field) {
    }

    MatrixDouble refCoords;
    MatrixDouble refGuassCoords;

    PetscErrorCode getSpaceBaseAndOrderOnElement();
    PetscErrorCode setGaussPts(int order);
    PetscErrorCode calculateHoJacobian();

  };


  struct OpGetCrackFrontDataAtGaussPts: public NonlinearElasticElement::OpGetDataAtGaussPts {

    bool &singularElement;
    MatrixDouble &invSJac;

    OpGetCrackFrontDataAtGaussPts(
      const std::string field_name,
      std::vector<VectorDouble > &values_at_gauss_pts,
      std::vector<MatrixDouble > &gardient_at_gauss_pts,
      bool &singular_element,
      MatrixDouble &inv_s_jac
    ):
    NonlinearElasticElement::OpGetDataAtGaussPts(field_name,values_at_gauss_pts,gardient_at_gauss_pts),
    singularElement(singular_element),
    invSJac(inv_s_jac) {
    }
    PetscErrorCode doWork(
      int side,EntityType type,DataForcesAndSurcesCore::EntData &data
    );
  };

  struct OpGetCrackFrontCommonDataAtGaussPts: public OpGetCrackFrontDataAtGaussPts {
    OpGetCrackFrontCommonDataAtGaussPts(
      const std::string field_name,
      NonlinearElasticElement::CommonData &common_data,
      bool &singular_element,
      MatrixDouble &inv_s_jac
    ):
    OpGetCrackFrontDataAtGaussPts(
      field_name,
      common_data.dataAtGaussPts[field_name],
      common_data.gradAtGaussPts[field_name],
      singular_element,
      inv_s_jac
    ) {}
  };

  struct OpAddSingularCoords: public MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator {
    bool &singularElement;
    MatrixDouble &singularCoords;
    moab::Interface &postProcMesh;
    std::vector<EntityHandle> &mapGaussPts;
    OpAddSingularCoords(
      bool &singular_element,
      MatrixDouble &singular_coords,
      moab::Interface &post_proc_mesh,
      std::vector<EntityHandle> &map_gauss_pts
    ):
    MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator("MESH_NODE_POSITIONS",UserDataOperator::OPROW),
    singularElement(singular_element),
    singularCoords(singular_coords),
    postProcMesh(post_proc_mesh),
    mapGaussPts(map_gauss_pts) {
    }
    PetscErrorCode doWork(
      int side,EntityType type,DataForcesAndSurcesCore::EntData &data
    );
  };

  struct OpTransfromSingularBaseFunctions: public MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator {
    bool &singularElement;
    MatrixDouble &invSJac;
    OpTransfromSingularBaseFunctions(bool &singular_element,MatrixDouble &inv_s_jac):
    MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator(
      "SPATIAL_POSITION",UserDataOperator::OPROW
    ),
    singularElement(singular_element),
    invSJac(inv_s_jac) {
    }
    MatrixDouble diffNinvJac;
    PetscErrorCode doWork(
      int side,EntityType type,DataForcesAndSurcesCore::EntData &data
    );
  };

  /**
  * \brief Calulate explicit direvative of enrgya

  \f[
  \left(
  \frac{\partial \Psi}{\partial \mathbf{X}}
  \right)_\textrm{expl.}=
  \left.\left(
  \frac{\partial \Psi}{\partial \mathbf{X}}
  \right)\right|_{\mathbf{u}=const.,\mathbf{F}=const.}
  \f]
  ]

  */
  struct OpRhsBoneExplicitDerivarive: public MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator {
    NonlinearElasticElement::BlockData &dAta;
    NonlinearElasticElement::CommonData &commonData;
    boost::shared_ptr<MatrixDouble> rhoGradAtGaussPts;
    ublas::vector<int> iNdices;
    OpRhsBoneExplicitDerivarive(
      NonlinearElasticElement::BlockData &data,
      NonlinearElasticElement::CommonData &common_data,
      boost::shared_ptr<MatrixDouble> rho_grad_at_gauss_pts
    );
    VectorDouble nF;
    PetscErrorCode doWork(
      int row_side,EntityType row_type,DataForcesAndSurcesCore::EntData &row_data
    );
  };

  /**
   * \brief implement NeoHookean material for bone remodeling
   */
  template<typename TYPE>
  struct BoneHooke: public Hooke<TYPE> {

    // NOTE: Need to be set to make this work

    boost::shared_ptr<VectorDouble> rhoAtGaussPts;
    double rHo0;
    double n;

    BoneHooke(): Hooke<TYPE>() {
    }

    TYPE rHo;

    PetscErrorCode setUserActiveVariables(int &nb_active_variables) {
      PetscFunctionBegin;
      SETERRQ(
        PETSC_COMM_SELF,
        MOFEM_NOT_IMPLEMENTED,
        "Shuld be implemented by member function specialization"
      );
      PetscFunctionReturn(0);
    }

    int nbActiveVariables0;

    PetscErrorCode setUserActiveVariables(VectorDouble &activeVariables) {
      PetscFunctionBegin;
      int shift = this->nbActiveVariables0;
      activeVariables[shift+0] = (*rhoAtGaussPts)[this->gG];
      PetscFunctionReturn(0);
    }

    /**
     * \brief Calculate Piola Kirchoff stress for bone model

     \f[
     \mathbf{P} = \left(\frac{\rho}{\rho_0}\right)^n\mathbf{P}_{0}]
     \f]

     * @param  block_data block data structure
     * @param  fe_ptr     pointer to base finite element
     * @return            error code
     */
    PetscErrorCode calculateP_PiolaKirchhoffI(
      const NonlinearElasticElement::BlockData block_data,
      boost::shared_ptr<const NumeredEntFiniteElement> fe_ptr
    ) {
      PetscErrorCode ierr;
      PetscFunctionBegin;
      ierr = Hooke<TYPE>::calculateP_PiolaKirchhoffI(block_data,fe_ptr); CHKERRQ(ierr);
      this->P *= pow(rHo/rHo0,n);
      PetscFunctionReturn(0);
    }


    /**
     * \brief Calculate Piola Kirchoff stress for bone model

     \f[
     \Psi = \left(\frac{\rho}{\rho_0}\right)^n\Psi_{0}
     \f]

     * @param  block_data block data structure
     * @param  fe_ptr     pointer to base finite element
     * @return            error code
     */
    PetscErrorCode calculateElasticEnergy(
      const NonlinearElasticElement::BlockData block_data,
      boost::shared_ptr<const NumeredEntFiniteElement> fe_ptr
    ) {
      PetscErrorCode ierr;
      PetscFunctionBegin;
      ierr = Hooke<TYPE>::calculateElasticEnergy(block_data,fe_ptr); CHKERRQ(ierr);
      this->eNergy *= pow(rHo/rHo0,n);
      PetscFunctionReturn(0);
    }

  };


  struct PostProcVertexMethod: public MoFEM::EntMethod {

    MoFEM::Interface &mField;
    Vec F;
    std::string tagName;
    PostProcVertexMethod(
      MoFEM::Interface &m_field,
      Vec f,
      std::string tag_name
    ):
    mField(m_field),
    F(f),
    tagName(tag_name) {
    }

    Tag tH;
    double *aRray;
    PetscErrorCode preProcess();

    PetscErrorCode postProcess();

    PetscErrorCode operator()();

  };

  struct FaceOrientation: public SurfaceSlidingConstrains::DriverElementOrientation {
    bool useProjectionFromCrackFront;
    FaceOrientation(bool crack_fornt):
    useProjectionFromCrackFront(crack_fornt) {
    }
    PetscErrorCode getElementOrientation(
      MoFEM::Interface &m_field,const FEMethod *fe_method_ptr
    );
  };

  /**

  Moving least weighted square approximation (MWLS)

  \f[
  u^h(x) = p_k(x)a_k(x)\\
  J(x)=\frac{1}{2} \sum_i w(x-x_i)\left(p_j(x_i)a_j(x)-u_i\right)^2 = \\
  J(x)=\frac{1}{2} \sum_i w(x-x_i)\left( p_j(x_i)a_j(x) p_k(x_i)a_k(x) - 2p_j(x_i)a_j(x)u_i + u_i^2 \right) \\
  \frac{\partial J}{\partial a_j} =
  \sum_i w(x_i-x)\left(p_j(x_i)p_k(x_i) a_k(x) -p_j(x_j)u_i\right)\\
  A_{jk}(x) = \sum_i w(x_i-x)p_j(x_i)p_k(x_i)\\
  B_{ji}(x) = w(x-x_i) p_j(x_i)\\
  A_{jk}(x) a_k(x) = B_{ji}(x)u_i\\
  a_k(x) = \left( A^{-1}_{kj}(x) B_{ji}(x) \right) u_i\\
  u^h(x) = p_k(x) \left( A^{-1}_{kj}(x) B_{ji}(x) \right) u_i \\
  u^h(x) = \left[ p_k(x) \left( A^{-1}_{kj}(x) B_{ji}(x) \right) \right] u_i \\
  \phi_i = p_k(x) \left( A^{-1}_{kj}(x) B_{ji}(x) \right) \\
  \phi_{i,l} =
  p_{k,l}(x) \left( A^{-1}_{kj}(x) B_{ji}(x) \right)+
  p_k(x) \left( A^{-1}_{kj,l}(x) B_{ji}(x) + A^{-1}_{kj}(x) B_{ji,l}(x)  \right)\\
  \phi_{i,l} =
  p_{k,l}(x) \left( A^{-1}_{kj}(x) B_{ji}(x) \right)+
  p_k(x) \left( A^{-1}_{kj,l}(x) B_{ji}(x) + A^{-1}_{kj}(x) B_{ji,l}(x)  \right)\\
  \left(\mathbf{A}\mathbf{A}^{-1}\right)_{,l} = \mathbf{0} \to \mathbf{A}^{-1}_{,l} = -\mathbf{A}^{-1} \mathbf{A}_{,l} \mathbf{A}^{-1}\\
  A_{jk,l} = \sum_i w_{,l}(x_i-x)p_j(x_i)p_k(x_i) \\
  B_{ij,l} = w_{,l}(x_i-x)p_j(x_i)
  \f]

   */
  struct MWLSApprox {

    MoFEM::Interface &mField;
    moab::Core mwlsCore;
    moab::Interface& mwlsMoab;

    MWLSApprox(
      MoFEM::Interface &m_field
    ):
    mField(m_field),
    mwlsMoab(mwlsCore),
    nbBasePolynomials(10),
    dmFactor(1.25) {
    }

    ~MWLSApprox() {
      // if(myTree) {
      //   myTree->reset_tree();
      // }
    }

    /**
     * Load mesh with data and do basic calulation
     * @param  file_name Filne name
     * @return           Error code
     */
    PetscErrorCode loadMWLSMesh(std::string file_name);

    /**
     * \brief get values form elements to nodes
     * @param  th tag with approximated data
     * @return    error code
     */
    PetscErrorCode getValuesToNodes(Tag th);

    /**
     * \brief Get noded in influence domain
     * @param  material_coords material position
     * @return error code
     */
    PetscErrorCode getInfluenceNodes(double material_coords[3],bool search_point = true);

    /**
     * \brief calculate approximation f9ntcion at material point
     * @param  material_coords material position
     * @return          error code

     NOTE: This function remove form influenceNodes entities which
     are beyound influece radius.

     */
    PetscErrorCode caculateApproxFun(
      double material_coords[3],
      bool calc_base = true,
      bool calc_diff_base = true,
      bool trim_influence_nodes = true
    );

    /**
     * Get tag data on influence nodes
     * @param  th tag
     * @return    error code
     */
    PetscErrorCode approxTagData(Tag th);

    typedef ublas::vector<double,ublas::bounded_array<double,9> > VecVals;
    typedef ublas::matrix<double,ublas::row_major,ublas::bounded_array<double,27> > MatDiffVals;

    /**
     * Get values at point
     * @return error code
     */
    VecVals& getData();

    /**
     * Get dirveatives of values at points
     * @return error code
     */
    MatDiffVals& getDiffData();

    // Stress at gauss point on finite element
    MatrixDouble stressAtGaussPts;
    Range tetsInBlock;

    /**
     * \brief Evaluare stress at integration points
     */
    struct OpMWLSStressAtGaussPts: public MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator {
      std::string tagName;
      boost::shared_ptr<MWLSApprox> mwlsApprox;
      OpMWLSStressAtGaussPts(boost::shared_ptr<MWLSApprox>  mwls_approx):
      MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator(
        "SPATIAL_POSITION",UserDataOperator::OPROW
      ),
      mwlsApprox(mwls_approx),
      tagName("MED_SIGP") {
      }
      // Range influenceNodes;
      PetscErrorCode doWork(
        int side,EntityType type,DataForcesAndSurcesCore::EntData &data
      );
    };

    /**
     * \bruef Integrate force vector
     */
    struct OpMWLSStressRhs: public MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator {
      boost::shared_ptr<MWLSApprox> mwlsApprox;
      OpMWLSStressRhs(boost::shared_ptr<MWLSApprox>  mwls_approx):
      MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator(
        "SPATIAL_POSITION",UserDataOperator::OPROW
      ),
      mwlsApprox(mwls_approx) {
      }
      ublas::vector<double> nF;
      PetscErrorCode doWork(
        int side,EntityType type,DataForcesAndSurcesCore::EntData &data
      );
    };

  private:

    boost::shared_ptr<AdaptiveKDTree> myTree;
    boost::shared_ptr<ElemEvaluator> elemEval;

    // 1 (1) x y z (4) xx yy zz xy xz yz (10)
    const int nbBasePolynomials;
    const double dmFactor;

    EntityHandle treeRoot;
    double maxEdgeL;

    Range levelNodes,levelEdges,levelTets;
    Range influenceNodes;

    ublas::symmetric_matrix<double,ublas::lower> A;
    ublas::matrix<double> invA;
    ublas::triangular_matrix<double,ublas::lower> L;
    ublas::matrix<double> B;
    ublas::vector<double> baseFun;
    ublas::vector<double> approxFun;

    ublas::symmetric_matrix<double,ublas::lower> diffA[3];
    ublas::symmetric_matrix<double,ublas::lower> diffInvA[3];
    ublas::matrix<double> diffB[3];
    ublas::vector<double> diffBaseFun[3];
    ublas::matrix<double> diffApproxFun;

    ublas::matrix<double> influenceNodesData;
    VecVals outData;
    MatDiffVals outDiffData;

    FTensor::Index<'i',3> i;

    inline double evalWeight(const double r) {
      const double rr = r*r;
      const double rrr = rr*r;
      const double r4 = rrr*r;
      if(r<1) {
        return 1-6*rr+8*rrr-3*r4;
      } else {
        return 0;
      }
      // if(r<=0.5) {
      //   return 2./3-4*rr+4*rrr;
      // } else if(r<=1) {
      //   return 4./3-4*r+4*rr-(4./3)*rrr;
      // } else return 0;
    }

    inline double evalDiffWeight(const double r) {
      const double rr = r*r;
      const double rrr = rr*r;
      if(r<1) {
         return -12*r+24*rr-12*rrr;
      } else {
        return 0;
      }
      // if(r<=0.5) {
      //   return -8*r+12*rr;
      // } else if(r<=1) {
      //   return -4+8*r-4*rr;
      // } else return 0;
    }

    double boxSize;
    PetscErrorCode getBox(
      double material_coords[3],
      std::vector<double>& coords
    );

  };

private:

  enum AdolcTags {
    ELASTIC_TAG = 1,
    GRIFFITH_FORCE_TAG = 2,
    CONSTANT_AREA_TAG = 3,
    FRONT_TANGENT = 4,
    EXTERIOR_DIREVATIVE_TAG = 5
  };
  Range elasticBlockTets;
  Range elasticBlockProcTets;
  Range bitEnts;
  Range bitProcEnts;

  boost::shared_ptr<NonlinearElasticElement> elasticFe;
  boost::shared_ptr<CrackFrontFE> feMaterialRhs;
  boost::shared_ptr<CrackFrontFE> feMaterialLhs;
  boost::shared_ptr<CrackFrontFE> feEnergy;
  boost::shared_ptr<ConstrainMatrixCtx> projSurfaceCtx;
  boost::shared_ptr<ConstrainMatrixCtx> projFrontCtx;
  boost::shared_ptr<MWLSApprox> mwlsApprox;

  Range bodySkin;
  Range crackFaces;
  Range oneSideCrackFaces;
  Range crackFront;
  Range crackFrontNodes;
  Range crackFrontNodesEdges;
  Range crackFrontElements;

  EntityHandle arcLengthVertex;   ///< Vertex associated with dof for arc-length control

  int defaultMaterial;
  boost::shared_ptr<VectorDouble> rhoAtGaussPt;      ///< density at integration points
  boost::shared_ptr<MatrixDouble> rhoGradAtGaussPt;  ///< gradient of density at integration point

  double rHo0;  ///< Reference density if bone is analyzed
  double nBone; ///< Exponent parameter in bone density

  int nbLoadSteps;
  double loadScale;

};

template<>
PetscErrorCode CrackPropagation::BoneHooke<adouble>::setUserActiveVariables(int &nb_active_variables);
template<>
PetscErrorCode CrackPropagation::BoneHooke<double>::setUserActiveVariables(int &nb_active_variables);

#endif // __CRACK_PROAGATION_HPP__
