/* This file is part of MoFEM.
 * MoFEM is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * MoFEM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MoFEM. If not, see <http://www.gnu.org/licenses/>. */

#ifdef WITH_TETGEN

#include <tetgen.h>
#ifdef REAL
  #undef REAL
#endif

#endif

#include <MoFEM.hpp>
using namespace MoFEM;

#include <MethodForForceScaling.hpp>
#include <DirichletBC.hpp>

#include <Projection10NodeCoordsOnField.hpp>

#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include <MethodForForceScaling.hpp>
#include <SurfacePressure.hpp>
#include <NodalForce.hpp>
#include <EdgeForce.hpp>

#include <FEMethod_LowLevelStudent.hpp>
#include <FEMethod_UpLevelStudent.hpp>

#include <PostProcVertexMethod.hpp>

#include <complex>
// extern "C" {
//   #include <complex_for_lazy.h>
// }

#include <ArcLengthTools.hpp>
#include <ConstrainMatrixCtx.hpp>
// #include <FEMethod_ComplexForLazy.hpp>
// #include <FEMethod_DriverComplexForLazy.hpp>

#include <SurfacePressureComplexForLazy.hpp>

#include <moab/Skinner.hpp>
#include <moab/AdaptiveKDTree.hpp>

#include <petsctime.h>

#include <ComplexConstArea.hpp>
#include <FaceSplittingTool.hpp>
#include <ConfigurationalFractureMechanics.hpp>

#include <adolc/adolc.h>
#include <NonLinearElasticElement.hpp>
#include <Hooke.hpp>
//#include <NeoHookean.hpp>
#include <VolumeLengthQuality.hpp>

#include <GriffithForceElement.hpp>
#include <Smoother.hpp>
#include <SurfaceSlidingConstrains.hpp>

#include <PostProcOnRefMesh.hpp>
#include <PostProcStresses.hpp>

#include <AnalyticalDirichlet.hpp>
#include <AnalyticalFun.hpp>

using namespace ObosleteUsersModules;

struct FaceOrientation: public SurfaceSlidingConstrains::DriverElementOrientation {

  bool useProjectionFromCrackFront;
  FaceOrientation(bool crack_fornt):
  useProjectionFromCrackFront(crack_fornt)
  {}

  PetscErrorCode getElementOrientation(MoFEM::Interface &m_field,const FEMethod *fe_method_ptr) {
    PetscFunctionBegin;
    elementOrientation = +1;
    ErrorCode rval;
    PetscErrorCode ierr;
    Range adj_side_elems;
    EntityHandle face = fe_method_ptr->numeredEntFiniteElementPtr->getEnt();
    BitRefLevel bit = fe_method_ptr->problemPtr->getBitRefLevel();
    ierr = m_field.get_adjacencies(bit,&face,1,3,adj_side_elems,moab::Interface::INTERSECT,0); CHKERRQ(ierr);
    adj_side_elems = adj_side_elems.subset_by_type(MBTET);
    if(adj_side_elems.size()==0) {
      Range adj_tets_on_surface;
      BitRefLevel bit_tet_on_surface;
      bit_tet_on_surface.set(BITREFLEVEL_SIZE-2);
      ierr = m_field.get_adjacencies(bit_tet_on_surface,&face,1,3,adj_tets_on_surface,moab::Interface::INTERSECT,0); CHKERRQ(ierr);
      adj_side_elems.insert(*adj_tets_on_surface.begin());
    }
    if(adj_side_elems.size()!=1) {
      adj_side_elems.clear();
      ierr = m_field.get_adjacencies(bit,&face,1,3,adj_side_elems,moab::Interface::INTERSECT,5); CHKERRQ(ierr);
      Range::iterator it = adj_side_elems.begin();
      for(;it!=adj_side_elems.end();it++) {
        Range nodes;
        rval = m_field.get_moab().get_connectivity(&*it,1,nodes,true); CHKERRQ_MOAB(rval);
        PetscPrintf(PETSC_COMM_WORLD,"Connectivity %lu %lu %lu %lu\n",nodes[0],nodes[1],nodes[2],nodes[3]);
      }
      int rank;
      MPI_Comm_rank(m_field.get_comm(),&rank);
      if(rank==0) {
        EntityHandle out_meshset;
        rval = m_field.get_moab().create_meshset(MESHSET_SET|MESHSET_TRACK_OWNER,out_meshset); CHKERRQ_MOAB(rval);
        rval = m_field.get_moab().add_entities(out_meshset,adj_side_elems); CHKERRQ_MOAB(rval);
        rval = m_field.get_moab().add_entities(out_meshset,&face,1); CHKERRQ_MOAB(rval);
        rval = m_field.get_moab().write_file("debug_error.vtk","VTK","",&out_meshset,1); CHKERRQ_MOAB(rval);
        rval = m_field.get_moab().delete_entities(&out_meshset,1); CHKERRQ_MOAB(rval);
      }
      SETERRQ1(PETSC_COMM_SELF,1,"Expect 1 tet but is %u",adj_side_elems.size());
    }
    EntityHandle side_elem = *adj_side_elems.begin();
    if(side_elem!=0) {
      int side_number,sense,offset;
      rval = m_field.get_moab().side_number(side_elem,face,side_number,sense,offset); CHKERRQ_MOAB(rval);
      if(sense == -1) {
        elementOrientation = -1;
      }
    }
    if(useProjectionFromCrackFront) {
      Tag th_interface_side;
      rval = m_field.get_moab().tag_get_handle("INTERFACE_SIDE",th_interface_side); CHKERRQ_MOAB(rval);
      int side;
      rval = m_field.get_moab().tag_get_data(th_interface_side,&face,1,&side); CHKERRQ_MOAB(rval);
      if(side == 1) {
        elementOrientation = -1;
      }
    }
    PetscFunctionReturn(0);
  }

};

//phisical_equation_volume eq_solid = hooke; /*stvenant_kirchhoff;*/

struct CrackFrontData {

  Range crackFrontEdgeNodes;

  PetscErrorCode initCrackFrontData(MoFEM::Interface& m_field) {
    PetscFunctionBegin;
    ErrorCode rval;
    PetscErrorCode ierr;
    Range crack_corners_edges;
    ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_corners_edges,true); CHKERRQ(ierr);
    rval = m_field.get_moab().get_connectivity(crack_corners_edges,crackFrontEdgeNodes,true); CHKERRQ_MOAB(rval);
    PetscFunctionReturn(0);
  }

  PetscErrorCode setCrackFrontIndices(FEMethod *fe_ptr,string &material_field_name,vector<DofIdx>& GlobIndices,bool not_at_crack_front) {
    PetscFunctionBegin;
    if(!crackFrontEdgeNodes.empty()) {
      for(_IT_GET_FEROW_BY_NAME_DOFS_FOR_LOOP_(fe_ptr,material_field_name,dof)) {
        Range::iterator nit = find(crackFrontEdgeNodes.begin(),crackFrontEdgeNodes.end(),dof->get()->getEnt());
        if(not_at_crack_front) {
          //if nit is not a part of crack front set
          if(nit != crackFrontEdgeNodes.end()) continue;
        } else {
          //if nit is part of crack front set
          if(nit == crackFrontEdgeNodes.end()) continue;
        }
        vector<DofIdx>::iterator it = find(GlobIndices.begin(),GlobIndices.end(),dof->get()->getPetscGlobalDofIdx());
        if(it != GlobIndices.end()) {
          *it = -1;
        }
      }
    }
    PetscFunctionReturn(0);
  }

};

struct TangentWithMeshSmoothingFrontConstrain: public C_CONSTANT_AREA {

  Vec frontF;
  Vec tangentFrontF;
  const double eps;

  TangentWithMeshSmoothingFrontConstrain(
    MoFEM::Interface& m_field,
    string _lambda_field_name,
    int _verbose = 0
  ):
  C_CONSTANT_AREA(m_field,PETSC_NULL,PETSC_NULL,_lambda_field_name,_verbose),
  frontF(PETSC_NULL),
  tangentFrontF(PETSC_NULL),
  eps(1e-10) {}

    Tag thFrontTangent;
    PetscErrorCode preProcess() {
      PetscFunctionBegin;

      switch (ts_ctx) {
        case CTX_TSSETIFUNCTION: {
          snes_ctx = CTX_SNESSETFUNCTION;
          snes_f = ts_F;
          break;
        }
        case CTX_TSSETIJACOBIAN: {
          snes_ctx = CTX_SNESSETJACOBIAN;
          snes_B = ts_B;
          break;
        }
        default:
        break;
      }

      ierr = C_CONSTANT_AREA::preProcess(); CHKERRQ(ierr);
      /*//TAG  - only for one proc analysis
      double def[] = {0,0,0};
      rval = mField.get_moab().tag_get_handle("FRONT_TANGENT",3,MB_TYPE_DOUBLE,
      thFrontTangent,MB_TAG_CREAT|MB_TAG_SPARSE,&def); MOAB_THROW(rval);*/
      switch(snes_ctx) {
        case CTX_SNESSETFUNCTION: {
          ierr = VecAssemblyBegin(snes_f); CHKERRQ(ierr);
          ierr = VecAssemblyEnd(snes_f); CHKERRQ(ierr);
          ierr = VecZeroEntries(tangentFrontF); CHKERRQ(ierr);
          ierr = VecGhostUpdateBegin(tangentFrontF,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
          ierr = VecGhostUpdateEnd(tangentFrontF,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
          ierr = VecAssemblyBegin(tangentFrontF); CHKERRQ(ierr);
          ierr = VecAssemblyEnd(tangentFrontF); CHKERRQ(ierr);
          /*//resent tags - only for one proc analysis
          for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(mField,lambda_field_name,dit)) {
          EntityHandle ent = dit->getEnt();
          rval = mField.get_moab().tag_set_data(thFrontTangent,&ent,1,def); CHKERRQ_MOAB(rval);
        }*/
      } break;
      case CTX_SNESSETJACOBIAN: {
        ierr = MatAssemblyBegin(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
        ierr = MatAssemblyEnd(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
      } break;
      default: {
      } break;
    }
    PetscFunctionReturn(0);
  }

  PetscErrorCode operator()() {
    PetscFunctionBegin;
    EntityHandle face = numeredEntFiniteElementPtr->getEnt();
    Range tet;
    rval = mField.get_moab().get_adjacencies(&face,1,3,false,tet); CHKERRQ_MOAB(rval);
    try {
      ierr = getData(true,true); CHKERRQ(ierr);
    } catch (const std::exception& ex) {
      ostringstream ss;
      ss << "thorw in method: " << ex.what() << endl;
      SETERRQ(PETSC_COMM_SELF,1,ss.str().c_str());
    }
    try {
      ublas::vector<double,ublas::bounded_array<double,9> > ELEM_CONSTRAIN1(9);
      ierr = calcDirevatives(
        &*diffNTRI.data().begin(),&*dofs_X.data().begin(),
        NULL,NULL,NULL,&*ELEM_CONSTRAIN1.data().begin(),NULL
      ); CHKERRQ(ierr);
      //take in account face orientation in respect crack surface
      Tag th_interface_side;
      rval = moab.tag_get_handle("INTERFACE_SIDE",th_interface_side); CHKERRQ_MOAB(rval);
      int side;
      rval = moab.tag_get_data(th_interface_side,&face,1,&side); CHKERRQ_MOAB(rval);
      if(side == 1) {
        ELEM_CONSTRAIN1 *= -1;
      }
      //quality
      ublas::vector<double,ublas::bounded_array<double,9> > F_FRONT_MESH_SMOOTHING(9);
      ublas::noalias(F_FRONT_MESH_SMOOTHING) = ublas::zero_vector<double>(9);
      double *f_front_mesh_array;
      ierr = VecGetArray(frontF,&f_front_mesh_array); CHKERRQ(ierr);
      for(int nn = 0;nn<3;nn++) {
        for(int dd = 0;dd<3;dd++) {
          if(local_disp_dofs_row_idx[3*nn+dd]==-1) continue;
          F_FRONT_MESH_SMOOTHING[3*nn+dd] = f_front_mesh_array[local_disp_dofs_row_idx[3*nn+dd]];
        }
      }
      ierr = VecRestoreArray(frontF,&f_front_mesh_array); CHKERRQ(ierr);
      //tangent
      if(snes_ctx == CTX_SNESSETJACOBIAN) {
        double center[3];
        tricircumcenter3d_tp(&coords.data()[0],&coords.data()[3],&coords.data()[6],center,NULL,NULL);
        cblas_daxpy(3,-1,&coords.data()[0],1,center,1);
        double r = cblas_dnrm2(3,center,1);
        for(int nn = 0;nn<3;nn++) {
          for(int dd = 0;dd<3;dd++) {
            // ---> calculate tangent starts here
            ublas::vector<double,ublas::bounded_array<double,9> > idofs_X(9,0);
            idofs_X[nn*3+dd] = r*eps;
            ublas::vector<double,ublas::bounded_array<double,9> > dELEM_CONSTRAIN1(9);
            ierr = calcDirevatives(
              &*diffNTRI.data().begin(),
              &*dofs_X.data().begin(),
              &*idofs_X.data().begin(),
              NULL,NULL,NULL,&*dELEM_CONSTRAIN1.data().begin()
            ); CHKERRQ(ierr);
            if(side == 1) {
              dELEM_CONSTRAIN1 /= -r*eps;
            } else {
              dELEM_CONSTRAIN1 /= +r*eps;
            }
            //dg -> C*q_quality
            double g[3] = {0,0,0};
            for(int nnn = 0;nnn<3;nnn++) {
              if(lambda_dofs_row_indx[nnn] == -1) continue;
              for(int ddd = 0;ddd<3;ddd++) {
                g[nnn] += dELEM_CONSTRAIN1[nnn*3+ddd]*F_FRONT_MESH_SMOOTHING[3*nnn+ddd];
              }
            }
            for(int nnn = 0;nnn<3;nnn++) {
              if(lambda_dofs_row_indx[nnn] == -1) continue;
              ierr = MatSetValues(snes_B,
                1,&lambda_dofs_row_indx[nnn],1,&disp_dofs_col_idx[3*nn+dd],
                &g[nnn],ADD_VALUES
              ); CHKERRQ(ierr);
            }
            //dCT_lambda
            for(int nnn = 0;nnn<3;nnn++) {
              for(int ddd = 0;ddd<3;ddd++) {
                dELEM_CONSTRAIN1[nnn*3+ddd] *= lambda[nnn];
              }
            }
            for(int nnn = 0;nnn<3;nnn++) {
              if(lambda_dofs_row_indx[nnn] == -1) continue;
              ierr = MatSetValues(snes_B,
                3,&disp_dofs_row_idx[3*nnn],
                1,&disp_dofs_col_idx[3*nn+dd],
                &dELEM_CONSTRAIN1[3*nnn],ADD_VALUES
              ); CHKERRQ(ierr);
            }
            // ---> calculate tangent end here
          }
        }
      }
      switch(snes_ctx) {
        case CTX_SNESSETFUNCTION: {
          ublas::vector<double,ublas::bounded_array<double,3> > g(3);
          for(int nn = 0;nn<3;nn++) {
            g[nn] = cblas_ddot(3,&ELEM_CONSTRAIN1[3*nn],1,&F_FRONT_MESH_SMOOTHING[3*nn],1);
          }
          //cerr << "g : " << g << endl;
          ierr = VecSetValues(snes_f,3,&*lambda_dofs_row_indx.data().begin(),&*g.data().begin(),ADD_VALUES); CHKERRQ(ierr);
          ierr = VecSetValues(tangentFrontF,9,&disp_dofs_row_idx[0],&*ELEM_CONSTRAIN1.data().begin(),ADD_VALUES); CHKERRQ(ierr);
          ublas::vector<double,ublas::bounded_array<double,9> > f(9);
          for(int nn = 0;nn<3;nn++) {
            for(int dd = 0;dd<3;dd++) {
              f[nn*3+dd] = lambda[nn]*ELEM_CONSTRAIN1[3*nn+dd];
            }
          }
          //cerr << "f : " << f << endl;
          ierr = VecSetValues(snes_f,9,&disp_dofs_row_idx[0],&*f.data().begin(),ADD_VALUES); CHKERRQ(ierr);
          /*//TAG - only for one proc analysis
          for(int nn = 0;nn<3;nn++) {
          EntityHandle ent = lambda_dofs_row_ents[nn];
          if(ent == no_handle) continue;
          double *t;
          rval = mField.get_moab().tag_get_by_ptr(thFrontTangent,&ent,1,(const void **)&t); CHKERRQ_MOAB(rval);
          cblas_daxpy(3,+1,&ELEM_CONSTRAIN1[3*nn],1,t,1);
        }*/
      } break;
      case CTX_SNESSETJACOBIAN: {
        for(int nn = 0;nn<3;nn++) {
          int lambda_dof_idx = lambda_dofs_col_indx[nn];
          ierr = MatSetValues(snes_B,3,&disp_dofs_row_idx[3*nn],1,&lambda_dof_idx,&ELEM_CONSTRAIN1[3*nn],ADD_VALUES); CHKERRQ(ierr);
        }
      } break;
      default:
      break;
    }
  } catch (const std::exception& ex) {
    ostringstream ss;
    ss << "thorw in method: " << ex.what() << endl;
    SETERRQ(PETSC_COMM_SELF,1,ss.str().c_str());
  }
  PetscFunctionReturn(0);
  }

  PetscErrorCode postProcess() {
    PetscFunctionBegin;
    switch(snes_ctx) {
      case CTX_SNESSETFUNCTION: {
        ierr = VecAssemblyBegin(tangentFrontF); CHKERRQ(ierr);
        ierr = VecAssemblyEnd(tangentFrontF); CHKERRQ(ierr);
        ierr = VecGhostUpdateBegin(tangentFrontF,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
        ierr = VecGhostUpdateEnd(tangentFrontF,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
        ierr = VecGhostUpdateBegin(tangentFrontF,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
        ierr = VecGhostUpdateEnd(tangentFrontF,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      } break;
      case CTX_SNESSETJACOBIAN: {
        ierr = MatAssemblyBegin(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
        ierr = MatAssemblyEnd(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
      } break;
      default: {
      } break;
    }
    PetscFunctionReturn(0);
  }

};

struct BothSurfaceConstrains: public FEMethod {

  MoFEM::Interface& mField;
  BothSurfaceConstrains(MoFEM::Interface& m_field): mField(m_field) {}

  PetscErrorCode preProcess() {
    PetscFunctionBegin;
    PetscErrorCode ierr;
    switch (ts_ctx) {
      case CTX_TSSETIFUNCTION: {
        snes_ctx = CTX_SNESSETFUNCTION;
        snes_f = ts_F;
        break;
      }
      case CTX_TSSETIJACOBIAN: {
        snes_ctx = CTX_SNESSETJACOBIAN;
        snes_B = ts_B;
        break;
      }
      default:
      break;
    }
    switch(snes_ctx) {
      case CTX_SNESSETFUNCTION: {
        ierr = VecAssemblyBegin(snes_f); CHKERRQ(ierr);
        ierr = VecAssemblyEnd(snes_f); CHKERRQ(ierr);
      }
      break;
      case CTX_SNESSETJACOBIAN:
      ierr = MatAssemblyBegin(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
      ierr = MatAssemblyEnd(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
      break;
      default:
      break;
    }
    PetscFunctionReturn(0);
  }
  PetscErrorCode operator()() {
    PetscFunctionBegin;
    //cerr << "AAAAAAAA\n";
    PetscErrorCode ierr;
    vector<int> lambda_dofs(9,-1);
    vector<double> lambda_vals(9,0);
    for(_IT_GET_FEROW_BY_TYPE_DOFS_FOR_LOOP_(this,"LAMBDA_BOTH_SIDES",MBVERTEX,it)) {
      lambda_dofs[3*it->get()->sideNumberPtr->side_number+it->get()->getEntDofIdx()] = it->get()->getPetscGlobalDofIdx();
      lambda_vals[3*it->get()->sideNumberPtr->side_number+it->get()->getEntDofIdx()] = it->get()->getFieldData();
      //cerr << "l "
      //<< 3*it->sideNumberPtr->side_number+it->getEntDofIdx() << " "
      //<< lambda_dofs[3*it->sideNumberPtr->side_number+it->getEntDofIdx()] << " "
      //<< lambda_vals[3*it->sideNumberPtr->side_number+it->getEntDofIdx()]
      //<< endl;
    }
    vector<int> positions_dofs(18,-1);
    vector<double> positions_vals(18,0);
    for(_IT_GET_FEROW_BY_TYPE_DOFS_FOR_LOOP_(this,"MESH_NODE_POSITIONS",MBVERTEX,it)) {
      int dd = 3*it->get()->sideNumberPtr->side_number+it->get()->getEntDofIdx();
      if(lambda_dofs[dd>8 ? dd -9 : dd] == -1) continue;
      positions_dofs[dd] = it->get()->getPetscGlobalDofIdx();
      positions_vals[dd] = it->get()->getFieldData();
      //cerr << "p " << dd << " " << positions_dofs[dd] << " " << positions_vals[dd] << endl;
    }
    //cerr << endl;
    const double alpha = 0;
    const double betha = 1e2;
    switch(snes_ctx) {
      case CTX_SNESSETFUNCTION:
      for(int ii = 0;ii<9;ii++) {
        if(lambda_dofs[ii] == -1) continue;
        double val1 = betha*(positions_vals[0+ii] - positions_vals[9+ii]) + alpha*lambda_vals[ii];
        ierr = VecSetValue(snes_f,lambda_dofs[ii],val1,INSERT_VALUES); CHKERRQ(ierr);
        double val2 = betha*lambda_vals[ii];
        ierr = VecSetValue(snes_f,positions_dofs[0+ii],+val2,INSERT_VALUES); CHKERRQ(ierr);
        ierr = VecSetValue(snes_f,positions_dofs[9+ii],-val2,INSERT_VALUES); CHKERRQ(ierr);
      }
      break;
      case CTX_SNESSETJACOBIAN:
      for(int ii = 0;ii<9;ii++) {
        if(lambda_dofs[ii] == -1) continue;
        ierr = MatSetValue(snes_B,lambda_dofs[ii],positions_dofs[0+ii],+1*betha,INSERT_VALUES); CHKERRQ(ierr);
        ierr = MatSetValue(snes_B,lambda_dofs[ii],positions_dofs[9+ii],-1*betha,INSERT_VALUES); CHKERRQ(ierr);
        ierr = MatSetValue(snes_B,positions_dofs[0+ii],lambda_dofs[ii],+1*betha,INSERT_VALUES); CHKERRQ(ierr);
        ierr = MatSetValue(snes_B,positions_dofs[9+ii],lambda_dofs[ii],-1*betha,INSERT_VALUES); CHKERRQ(ierr);
        ierr = MatSetValue(snes_B,lambda_dofs[ii],lambda_dofs[ii],alpha,INSERT_VALUES); CHKERRQ(ierr);
      }
      break;
      default:
      break;
    }

    PetscFunctionReturn(0);
  }
  PetscErrorCode postProcess() {
    PetscFunctionBegin;
    PetscErrorCode ierr;
    switch(snes_ctx) {
      case CTX_SNESSETFUNCTION: {
        ierr = VecAssemblyBegin(snes_f); CHKERRQ(ierr);
        ierr = VecAssemblyEnd(snes_f); CHKERRQ(ierr);
      }
      break;
      case CTX_SNESSETJACOBIAN: {
        ierr = MatAssemblyBegin(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
        ierr = MatAssemblyEnd(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
      }
      break;
      default:
      break;
    }
    PetscFunctionReturn(0);
  }

};

struct MyPrePostProcess: public FEMethod {

  MoFEM::Interface& m_field;
  ArcLengthCtx *arcPtr;

  MyPrePostProcess(MoFEM::Interface& _m_field,
    ArcLengthCtx *_arc_ptr):
    m_field(_m_field),arcPtr(_arc_ptr) {}

  PetscErrorCode ierr;

  PetscErrorCode preProcess() {
    PetscFunctionBegin;

    switch (ts_ctx) {
	case CTX_TSSETIFUNCTION: {
	  snes_ctx = CTX_SNESSETFUNCTION;
	  snes_f = ts_F;
	  break;
	}
	case CTX_TSSETIJACOBIAN: {
	  snes_ctx = CTX_SNESSETJACOBIAN;
	  snes_B = ts_B;
	  break;
	}
	default:
	break;
    }

    //PetscAttachDebugger();
    switch(snes_ctx) {
      case CTX_SNESSETFUNCTION: {
        ierr = VecZeroEntries(snes_f); CHKERRQ(ierr);
        ierr = VecGhostUpdateBegin(snes_f,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
        ierr = VecGhostUpdateEnd(snes_f,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
        ierr = VecZeroEntries(arcPtr->F_lambda); CHKERRQ(ierr);
        ierr = VecGhostUpdateBegin(arcPtr->F_lambda,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
        ierr = VecGhostUpdateEnd(arcPtr->F_lambda,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      }
      break;
      default:
        SETERRQ(PETSC_COMM_SELF,1,"not implemented");
    }

    PetscFunctionReturn(0);
  }

  PetscErrorCode postProcess() {
    PetscFunctionBegin;
    switch(snes_ctx) {
      case CTX_SNESSETFUNCTION: {
	  //snes_f
        ierr = VecGhostUpdateBegin(snes_f,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
        ierr = VecGhostUpdateEnd(snes_f,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
        ierr = VecAssemblyBegin(snes_f); CHKERRQ(ierr);
        ierr = VecAssemblyEnd(snes_f); CHKERRQ(ierr);
      }
      break;
      default:
        SETERRQ(PETSC_COMM_SELF,1,"not implemented");
    }
    PetscFunctionReturn(0);
  }

};

struct AssembleLambda: public FEMethod {

  MoFEM::Interface& m_field;
  ArcLengthCtx *arcPtr;

  SpatialPositionsBCFEMethodPreAndPostProc *bC;

  AssembleLambda(MoFEM::Interface& _m_field,
    ArcLengthCtx *_arc_ptr,SpatialPositionsBCFEMethodPreAndPostProc *bc):
    m_field(_m_field),arcPtr(_arc_ptr),bC(bc) {}

  PetscErrorCode ierr;

  PetscErrorCode preProcess() {
    PetscFunctionBegin;
    switch (ts_ctx) {
      case CTX_TSSETIFUNCTION: {
        snes_ctx = CTX_SNESSETFUNCTION;
        snes_f = ts_F;
        break;
      }
      case CTX_TSSETIJACOBIAN: {
        snes_ctx = CTX_SNESSETJACOBIAN;
        snes_B = ts_B;
        break;
      }
      default:
      break;
    }
    PetscFunctionReturn(0);
  }
  PetscErrorCode operator()() {
    PetscFunctionBegin;
    PetscFunctionReturn(0);
  }
  PetscErrorCode postProcess() {
    PetscFunctionBegin;
    switch(snes_ctx) {
      case CTX_SNESSETFUNCTION: {
        //F_lambda
        ierr = VecGhostUpdateBegin(arcPtr->F_lambda,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
        ierr = VecGhostUpdateEnd(arcPtr->F_lambda,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
        ierr = VecAssemblyBegin(arcPtr->F_lambda); CHKERRQ(ierr);
        ierr = VecAssemblyEnd(arcPtr->F_lambda); CHKERRQ(ierr);
        for(vector<int>::iterator vit = bC->dofsIndices.begin();vit!=bC->dofsIndices.end();vit++) {
          ierr = VecSetValue(arcPtr->F_lambda,*vit,0,INSERT_VALUES); CHKERRQ(ierr);
        }
        ierr = VecAssemblyBegin(arcPtr->F_lambda); CHKERRQ(ierr);
        ierr = VecAssemblyEnd(arcPtr->F_lambda); CHKERRQ(ierr);
        ierr = VecDot(arcPtr->F_lambda,arcPtr->F_lambda,&arcPtr->F_lambda2); CHKERRQ(ierr);
        PetscPrintf(PETSC_COMM_WORLD,"\tFlambda2 = %6.4e\n",arcPtr->F_lambda2);
        //add F_lambda
        ierr = VecAXPY(snes_f,arcPtr->getFieldData(),arcPtr->F_lambda); CHKERRQ(ierr);
        PetscPrintf(PETSC_COMM_WORLD,"\tlambda = %6.4e\n",arcPtr->getFieldData());
        double fnorm;
        ierr = VecNorm(snes_f,NORM_2,&fnorm); CHKERRQ(ierr);
        PetscPrintf(PETSC_COMM_WORLD,"\tfnorm = %6.4e\n",fnorm);
      }
      break;
      default:
      SETERRQ(PETSC_COMM_SELF,1,"not implemented");
    }
    PetscFunctionReturn(0);
  }

};

PetscErrorCode ConfigurationalFractureMechanics::set_material_fire_wall(MoFEM::Interface& m_field) {
  PetscFunctionBegin;
  ErrorCode rval;
  BitRefLevel def_bit_level = 0;
  rval = m_field.get_moab().tag_get_handle("_Materiar_FireWall",sizeof(Material_FirelWall_def),MB_TYPE_OPAQUE,
    th_MaterialFireWall,MB_TAG_CREAT|MB_TAG_SPARSE|MB_TAG_BYTES,&def_bit_level);
  const EntityHandle root_meshset = m_field.get_moab().get_root_set();
  rval = m_field.get_moab().tag_get_by_ptr(th_MaterialFireWall,&root_meshset,1,(const void**)&material_FirelWall); CHKERRQ_MOAB(rval);
  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::thermal_field(MoFEM::Interface& m_field) {
  PetscFunctionBegin;
  if(material_FirelWall->operator[](FW_thermal_field)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_thermal_field);

  PetscErrorCode ierr;

  ierr = m_field.add_field("TEMPERATURE",H1,1,MF_ZERO); CHKERRQ(ierr);

  Range level_tets;
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTET,level_tets); CHKERRQ(ierr);
  ierr = m_field.add_ents_to_field_by_TETs(level_tets,"TEMPERATURE"); CHKERRQ(ierr);

  ierr = m_field.set_field_order(0,MBTET,"TEMPERATURE",1); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBTRI,"TEMPERATURE",1); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBEDGE,"TEMPERATURE",1); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBVERTEX,"TEMPERATURE",1); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::spatial_problem_definition(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  if(material_FirelWall->operator[](FW_spatial_problem_definition)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_spatial_problem_definition);

  PetscErrorCode ierr;
  ErrorCode rval;

  //Fields
  ierr = m_field.add_field("SPATIAL_POSITION",H1,3,MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_field("MESH_NODE_POSITIONS",H1,3,MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_field("SPATIAL_DISPLACEMENT",H1,3,MF_ZERO); CHKERRQ(ierr);

  //FE
  ierr = m_field.add_finite_element("ELASTIC",MF_ZERO); CHKERRQ(ierr);

  //Define rows/cols and element data
  ierr = m_field.modify_finite_element_add_field_row("ELASTIC","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("ELASTIC","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("ELASTIC","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("ELASTIC","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  if(m_field.check_field("TEMPERATURE")) {
    ierr = m_field.modify_finite_element_add_field_data("ELASTIC","TEMPERATURE"); CHKERRQ(ierr);
  }

  //add neumman finite elemnets to add static boundary conditions
  ierr = m_field.add_finite_element("NEUAMNN_FE",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_row("NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("NEUAMNN_FE","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  Range level_tris;
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,NODESET|FORCESET,it)) {
    Range tris;
    rval = m_field.get_moab().get_entities_by_type(it->meshset,MBTRI,tris,true); CHKERRQ_MOAB(rval);
    tris = intersect(level_tris,tris);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(tris,"NEUAMNN_FE"); CHKERRQ(ierr);
  }
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,SIDESET|PRESSURESET,it)) {
    Range tris;
    rval = m_field.get_moab().get_entities_by_type(it->meshset,MBTRI,tris,true); CHKERRQ_MOAB(rval);
    tris = intersect(level_tris,tris);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(tris,"NEUAMNN_FE"); CHKERRQ(ierr);
  }

  #ifdef __ANALITICAL_TRACTION__
  {
    MeshsetsManager *meshset_manager_ptr;
    ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
    if(meshset_manager_ptr->checkMeshset("ANALITICAL_TRACTION")) {
      const CubitMeshSets *cubit_meshset_ptr;
      meshset_manager_ptr->getCubitMeshsetPtr("ANALITICAL_TRACTION",&cubit_meshset_ptr);
      Range tris;
      ierr = meshset_manager_ptr->getEntitiesByDimension(
        cubit_meshset_ptr->getMeshsetId(),BLOCKSET,2,tris,true
      ); CHKERRQ(ierr);
      ierr = m_field.add_ents_to_finite_element_by_TRIs(tris,"NEUAMNN_FE"); CHKERRQ(ierr);
    }
  }
  #endif //__ANALITICAL_TRACTION__

  #ifdef __ANALITICAL_DISPLACEMENT__
  {
    MeshsetsManager *meshset_manager_ptr;
    ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
    if(meshset_manager_ptr->checkMeshset("ANALITICAL_DISP")) {
      const CubitMeshSets *cubit_meshset_ptr;
      meshset_manager_ptr->getCubitMeshsetPtr("ANALITICAL_DISP",&cubit_meshset_ptr);
      Range tris;
      ierr = meshset_manager_ptr->getEntitiesByDimension(
        cubit_meshset_ptr->getMeshsetId(),BLOCKSET,2,tris,true
      ); CHKERRQ(ierr);
      if(tris.size()>0) {
        ierr = m_field.add_problem("BC_PROBLEM",MF_ZERO); CHKERRQ(ierr);
        ierr = m_field.modify_problem_ref_level_add_bit("BC_PROBLEM",*ptr_bit_level0); CHKERRQ(ierr);
        ierr = m_field.add_finite_element("BC_FE",MF_ZERO); CHKERRQ(ierr);
        ierr = m_field.modify_finite_element_add_field_row("BC_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
        ierr = m_field.modify_finite_element_add_field_col("BC_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
        ierr = m_field.modify_finite_element_add_field_data("BC_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
        ierr = m_field.modify_finite_element_add_field_data("BC_FE","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
        ierr = m_field.add_ents_to_finite_element_by_TRIs(tris,"BC_FE"); CHKERRQ(ierr);
        ierr = m_field.modify_problem_add_finite_element("BC_PROBLEM","BC_FE"); CHKERRQ(ierr);
      }
    }
  }
  #endif //__ANALITICAL_DISPLACEMENT__

  //define problems
  ierr = m_field.add_problem("ELASTIC_MECHANICS",MF_ZERO); CHKERRQ(ierr);
  //set finite elements for problems
  ierr = m_field.modify_problem_add_finite_element("ELASTIC_MECHANICS","ELASTIC"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("ELASTIC_MECHANICS","NEUAMNN_FE"); CHKERRQ(ierr);

  //add nodal force element
  ierr = MetaNodalForces::addElement(m_field,"SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = MetaEdgeForces::addElement(m_field,"SPATIAL_POSITION"); CHKERRQ(ierr);

  ierr = m_field.modify_problem_add_finite_element("ELASTIC_MECHANICS","FORCE_FE"); CHKERRQ(ierr);

  Range level_tets;
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTET,level_tets); CHKERRQ(ierr);
  ierr = m_field.add_ents_to_field_by_TETs(level_tets,"SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.add_ents_to_field_by_TETs(level_tets,"SPATIAL_DISPLACEMENT"); CHKERRQ(ierr);
  ierr = m_field.add_ents_to_field_by_TETs(level_tets,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  //ierr = m_field.add_ents_to_field_by_TETs(level_tets,"MESH_NODE_DISPLACEMENT"); CHKERRQ(ierr);
  Range blocked_tets;
  BitRefLevel bit_to_block = BitRefLevel().set(BITREFLEVEL_SIZE-2);
  ierr = m_field.get_entities_by_type_and_ref_level(bit_to_block,BitRefLevel().set(),MBTET,blocked_tets); CHKERRQ(ierr);
  ierr = m_field.add_ents_to_field_by_TETs(blocked_tets,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  //ierr = m_field.add_ents_to_field_by_TETs(blocked_tets,"MESH_NODE_DISPLACEMENT"); CHKERRQ(ierr);

  PetscInt order;
  PetscBool flg = PETSC_TRUE;
  ierr = PetscOptionsGetInt(PETSC_NULL,PETSC_NULL,"-my_order",&order,&flg); CHKERRQ(ierr);
  if(flg != PETSC_TRUE) {
    Tag th_set_order;
    rval = m_field.get_moab().tag_get_handle("_SET_ORDER",th_set_order); CHKERRQ_MOAB(rval);
    const EntityHandle root_meshset = m_field.get_moab().get_root_set();
    rval = m_field.get_moab().tag_get_data(th_set_order,&root_meshset,1,&order); CHKERRQ_MOAB(rval);
  }

  //set app. order
  ierr = m_field.set_field_order(0,MBTET,"SPATIAL_POSITION",order); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBTRI,"SPATIAL_POSITION",order); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBEDGE,"SPATIAL_POSITION",order); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBVERTEX,"SPATIAL_POSITION",1); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBTET,"SPATIAL_DISPLACEMENT",order); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBTRI,"SPATIAL_DISPLACEMENT",order); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBEDGE,"SPATIAL_DISPLACEMENT",order); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBVERTEX,"SPATIAL_DISPLACEMENT",1); CHKERRQ(ierr);
  //NOTE: always order should be 1
  ierr = m_field.set_field_order(0,MBVERTEX,"MESH_NODE_POSITIONS",1); CHKERRQ(ierr);
  //ierr = m_field.set_field_order(0,MBVERTEX,"MESH_NODE_DISPLACEMENT",1); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


PetscErrorCode ConfigurationalFractureMechanics::material_problem_definition(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  if(material_FirelWall->operator[](FW_material_problem_definition)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_material_problem_definition);

  PetscErrorCode ierr;
  ErrorCode rval;

  //Fields
  ierr = m_field.add_field("MESH_NODE_POSITIONS",H1,3,MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_field("MATERIAL_FORCE",H1,3,MF_ZERO); CHKERRQ(ierr);

  //FE
  ierr = m_field.add_finite_element("MATERIAL",MF_ZERO); CHKERRQ(ierr);

  //Define rows/cols and element data
  ierr = m_field.modify_finite_element_add_field_data("ELASTIC","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  //
  ierr = m_field.modify_finite_element_add_field_row("MATERIAL","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("MATERIAL","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_off_field_row("MATERIAL","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("MATERIAL","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("MATERIAL","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  if(m_field.check_field("TEMPERATURE")) {
    ierr = m_field.modify_finite_element_add_field_data("MATERIAL","TEMPERATURE"); CHKERRQ(ierr);
  }

  {
    ierr = m_field.add_finite_element("GRIFFITH_FORCE_ELEMENT",MF_ZERO); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_row("GRIFFITH_FORCE_ELEMENT","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col("GRIFFITH_FORCE_ELEMENT","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data("GRIFFITH_FORCE_ELEMENT","MESH_NODE_POSITIONS"); CHKERRQ(ierr);


    Range crack_surfaces_faces;
    ierr = m_field.get_cubit_msId_entities_by_dimension(200,SIDESET,2,crack_surfaces_faces,true); CHKERRQ(ierr);
    Range crack_front_edges;
    ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRQ(ierr);
    Range crack_front_nodes;
    rval = m_field.get_moab().get_connectivity(crack_front_edges,crack_front_nodes,true); CHKERRQ_MOAB(rval);
    Range crack_surfaces_edge_faces;
    rval = m_field.get_moab().get_adjacencies(
      crack_front_nodes,2,false,crack_surfaces_edge_faces,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    crack_surfaces_edge_faces = crack_surfaces_edge_faces.subset_by_type(MBTRI);
    crack_surfaces_edge_faces = intersect(crack_surfaces_edge_faces,crack_surfaces_faces);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(crack_surfaces_edge_faces,"GRIFFITH_FORCE_ELEMENT"); CHKERRQ(ierr);
  }

  //define problems
  ierr = m_field.add_problem("MATERIAL_MECHANICS",MF_ZERO); CHKERRQ(ierr);

  //set finite elements for problems
  ierr = m_field.modify_problem_add_finite_element("MATERIAL_MECHANICS","MATERIAL"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("MATERIAL_MECHANICS","GRIFFITH_FORCE_ELEMENT"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::coupled_problem_definition(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  if(material_FirelWall->operator[](FW_coupled_problem_definition)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_coupled_problem_definition);

  PetscErrorCode ierr;

  //Fields
  //ierr = m_field.add_field("SPATIAL_POSITION",H1,3,MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_field("MESH_NODE_POSITIONS",H1,3,MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_field("MATERIAL_FORCE",H1,3,MF_ZERO); CHKERRQ(ierr);
  //FE
  ierr = m_field.add_finite_element("ELASTIC_COUPLED",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_finite_element("MATERIAL_COUPLED",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_finite_element("MESH_SMOOTHER",MF_ZERO); CHKERRQ(ierr);

  //fes definitions
  ierr = m_field.modify_finite_element_add_field_row("ELASTIC_COUPLED","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_row("ELASTIC_COUPLED","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("ELASTIC_COUPLED","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("ELASTIC_COUPLED","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("ELASTIC_COUPLED","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("ELASTIC_COUPLED","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  //if(m_field.check_field("TEMPERATURE")) {
    //ierr = m_field.modify_finite_element_add_field_data("ELASTIC_COUPLED","TEMPERATURE"); CHKERRQ(ierr);
  //}


  //
  ierr = m_field.modify_finite_element_add_field_row("MATERIAL_COUPLED","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_row("MATERIAL_COUPLED","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("MATERIAL_COUPLED","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("MATERIAL_COUPLED","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("MATERIAL_COUPLED","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("MATERIAL_COUPLED","SPATIAL_POSITION"); CHKERRQ(ierr);
  //if(m_field.check_field("TEMPERATURE")) {
    //ierr = m_field.modify_finite_element_add_field_data("MATERIAL_COUPLED","TEMPERATURE"); CHKERRQ(ierr);
  //}

  //
  ierr = m_field.modify_finite_element_add_field_row("MESH_SMOOTHER","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("MESH_SMOOTHER","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("MESH_SMOOTHER","MESH_NODE_POSITIONS"); CHKERRQ(ierr);

  //define problems
  ierr = m_field.add_problem("COUPLED_PROBLEM",MF_ZERO); CHKERRQ(ierr);
  //set finite elements for problems
  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","ELASTIC_COUPLED"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","MATERIAL_COUPLED"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","NEUAMNN_FE"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","FORCE_FE"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","MESH_SMOOTHER"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","GRIFFITH_FORCE_ELEMENT"); CHKERRQ(ierr);

  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","CandCT_SURFACE_ELEM"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","BOTH_SIDE_OF_CRACK"); CHKERRQ(ierr);

  bool cs = true;
  if(cs) {
    ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","CandCT_CRACK_SURFACE_ELEM"); CHKERRQ(ierr);
  }
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss2;
    ss2 << "CandCT_SURFACE_ELEM_msId_" << msId;
    ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM",ss2.str()); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::arclength_problem_definition(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  ErrorCode rval;

  if(material_FirelWall->operator[](FW_arc_lenhghat_definition)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_arc_lenhghat_definition);

  PetscErrorCode ierr;

  if(!m_field.check_field("LAMBDA")) {
    ierr = m_field.add_field("LAMBDA",NOFIELD,1); CHKERRQ(ierr);
    ierr = m_field.add_finite_element("ARC_LENGTH"); CHKERRQ(ierr);
    // Setting up LAMBDA field and ARC_LENGTH interface
    {
      //Add dummy no-field vertex
      EntityHandle no_field_vertex;
      {
        const double coords[] = { 0, 0, 0};
        rval = m_field.get_moab().create_vertex(coords,no_field_vertex); CHKERRQ_MOAB(rval);
        Range range_no_field_vertex;
        range_no_field_vertex.insert(no_field_vertex);
        ierr = m_field.seed_ref_level(range_no_field_vertex,BitRefLevel().set()); CHKERRQ(ierr);
        EntityHandle lambda_meshset = m_field.get_field_meshset("LAMBDA");
        rval = m_field.get_moab().add_entities(lambda_meshset,range_no_field_vertex); CHKERRQ_MOAB(rval);
      }
      //this entity will carray data for this finite element
      EntityHandle meshset_fe_arc_length;
      {
        rval = m_field.get_moab().create_meshset(MESHSET_SET|MESHSET_TRACK_OWNER,meshset_fe_arc_length); CHKERRQ_MOAB(rval);
        rval = m_field.get_moab().add_entities(meshset_fe_arc_length,&no_field_vertex,1); CHKERRQ_MOAB(rval);
        ierr = m_field.seed_ref_level_MESHSET(meshset_fe_arc_length,BitRefLevel().set()); CHKERRQ(ierr);
      }
      //finally add created meshset to the ARC_LENGTH finite element
      ierr = m_field.add_ents_to_finite_element_by_MESHSET(meshset_fe_arc_length,"ARC_LENGTH",false); CHKERRQ(ierr);
    }
  } else {
    EntityHandle lambda_meshset = m_field.get_field_meshset("LAMBDA");
    Range ents;
    rval = m_field.get_moab().get_entities_by_handle(lambda_meshset,ents); CHKERRQ_MOAB(rval);
    ierr = m_field.seed_ref_level(ents,BitRefLevel().set()); CHKERRQ(ierr);
  }

  //Define rows/cols and element data
  ierr = m_field.modify_finite_element_add_field_row("ARC_LENGTH","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("ARC_LENGTH","LAMBDA"); CHKERRQ(ierr);
  //elem data
  ierr = m_field.modify_finite_element_add_field_data("ARC_LENGTH","LAMBDA"); CHKERRQ(ierr);

  ierr = m_field.modify_finite_element_add_field_row("ELASTIC_COUPLED","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("ELASTIC_COUPLED","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("ELASTIC_COUPLED","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_row("MATERIAL_COUPLED","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("MATERIAL_COUPLED","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("ELASTIC_COUPLED","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_row("MESH_SMOOTHER","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("MESH_SMOOTHER","LAMBDA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("MESH_SMOOTHER","LAMBDA"); CHKERRQ(ierr);

  ierr = m_field.modify_problem_add_finite_element("COUPLED_PROBLEM","ARC_LENGTH"); CHKERRQ(ierr);


  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::constrains_problem_definition(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  if(material_FirelWall->operator[](FW_constrains_problem_definition)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_constrains_problem_definition);

  ErrorCode rval;
  PetscErrorCode ierr;

  bool cs = true;

  //Fields
  ierr = m_field.add_field("LAMBDA_SURFACE",H1,1,MF_ZERO); CHKERRQ(ierr);
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss;
    ss << "LAMBDA_SURFACE_msId_" << msId;
    ierr = m_field.add_field(ss.str(),H1,1,MF_ZERO); CHKERRQ(ierr);
  }
  //CRACK
  if(cs) {
    ierr = m_field.add_field("LAMBDA_CRACK_SURFACE",H1,1,MF_ZERO); CHKERRQ(ierr);
    ierr = m_field.add_field("LAMBDA_BOTH_SIDES",H1,3,MF_ZERO); CHKERRQ(ierr);
  }

  //FE
  ierr = m_field.add_finite_element("C_SURFACE_ELEM",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_finite_element("CTC_SURFACE_ELEM",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_finite_element("CandCT_SURFACE_ELEM",MF_ZERO); CHKERRQ(ierr);
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss0;
    ss0 << "C_SURFACE_ELEM_msId_" << msId;
    ostringstream ss1;
    ss1 << "CTC_SURFACE_ELEM_msId_" << msId;
    ostringstream ss2;
    ss2 << "CandCT_SURFACE_ELEM_msId_" << msId;
    ierr = m_field.add_finite_element(ss0.str(),MF_ZERO); CHKERRQ(ierr);
    ierr = m_field.add_finite_element(ss1.str(),MF_ZERO); CHKERRQ(ierr);
    ierr = m_field.add_finite_element(ss2.str(),MF_ZERO); CHKERRQ(ierr);
  }

  //CRACK
  if(cs) {
    //ierr = m_field.add_finite_element("C_CRACK_SURFACE_ELEM",MF_ZERO); CHKERRQ(ierr);
    //ierr = m_field.add_finite_element("CTC_CRACK_SURFACE_ELEM",MF_ZERO); CHKERRQ(ierr);
    ierr = m_field.add_finite_element("CandCT_CRACK_SURFACE_ELEM",MF_ZERO); CHKERRQ(ierr);
    ierr = m_field.add_finite_element("BOTH_SIDE_OF_CRACK",MF_ZERO); CHKERRQ(ierr);
  }

  //Define rows/cols and element data
  ierr = m_field.modify_finite_element_add_field_row("C_SURFACE_ELEM","LAMBDA_SURFACE"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("C_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("C_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("C_SURFACE_ELEM","LAMBDA_SURFACE"); CHKERRQ(ierr);

  ierr = m_field.modify_finite_element_add_field_row("CTC_SURFACE_ELEM","LAMBDA_SURFACE"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("CTC_SURFACE_ELEM","LAMBDA_SURFACE"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("CTC_SURFACE_ELEM","LAMBDA_SURFACE"); CHKERRQ(ierr);

  ierr = m_field.modify_finite_element_add_field_row("CandCT_SURFACE_ELEM","LAMBDA_SURFACE"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_row("CandCT_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("CandCT_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("CandCT_SURFACE_ELEM","LAMBDA_SURFACE"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("CandCT_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("CandCT_SURFACE_ELEM","LAMBDA_SURFACE"); CHKERRQ(ierr);

  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss;
    ss << "LAMBDA_SURFACE_msId_" << msId;
    ostringstream ss0;
    ss0 << "C_SURFACE_ELEM_msId_" << msId;
    ierr = m_field.modify_finite_element_add_field_row(ss0.str(),ss.str()); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col(ss0.str(),"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data(ss0.str(),"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data(ss0.str(),ss.str()); CHKERRQ(ierr);
    ostringstream ss1;
    ss1 << "CTC_SURFACE_ELEM_msId_" << msId;
    ierr = m_field.modify_finite_element_add_field_row(ss1.str(),ss.str()); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col(ss1.str(),ss.str()); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data(ss1.str(),ss.str()); CHKERRQ(ierr);
    ostringstream ss2;
    ss2 << "CandCT_SURFACE_ELEM_msId_" << msId;
    ierr = m_field.modify_finite_element_add_field_row(ss2.str(),ss.str()); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_row(ss2.str(),"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col(ss2.str(),"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col(ss2.str(),ss.str()); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data(ss2.str(),"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data(ss2.str(),ss.str()); CHKERRQ(ierr);
  }

  //CRACK
  if(cs) {
    /*ierr = m_field.modify_finite_element_add_field_row("C_CRACK_SURFACE_ELEM","LAMBDA_CRACK_SURFACE"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col("C_CRACK_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data("C_CRACK_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data("C_CRACK_SURFACE_ELEM","LAMBDA_CRACK_SURFACE"); CHKERRQ(ierr);

    ierr = m_field.modify_finite_element_add_field_row("CTC_CRACK_SURFACE_ELEM","LAMBDA_CRACK_SURFACE"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col("CTC_CRACK_SURFACE_ELEM","LAMBDA_CRACK_SURFACE"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data("CTC_CRACK_SURFACE_ELEM","LAMBDA_CRACK_SURFACE"); CHKERRQ(ierr);*/

    ierr = m_field.modify_finite_element_add_field_row("CandCT_CRACK_SURFACE_ELEM","LAMBDA_CRACK_SURFACE"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_row("CandCT_CRACK_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col("CandCT_CRACK_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col("CandCT_CRACK_SURFACE_ELEM","LAMBDA_CRACK_SURFACE"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data("CandCT_CRACK_SURFACE_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data("CandCT_CRACK_SURFACE_ELEM","LAMBDA_CRACK_SURFACE"); CHKERRQ(ierr);

    ierr = m_field.modify_finite_element_add_field_row("BOTH_SIDE_OF_CRACK","LAMBDA_BOTH_SIDES"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_row("BOTH_SIDE_OF_CRACK","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col("BOTH_SIDE_OF_CRACK","LAMBDA_BOTH_SIDES"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col("BOTH_SIDE_OF_CRACK","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data("BOTH_SIDE_OF_CRACK","LAMBDA_BOTH_SIDES"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_data("BOTH_SIDE_OF_CRACK","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  }

  //define problems
  ierr = m_field.add_problem("CCT_ALL_MATRIX",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_problem("C_ALL_MATRIX",MF_ZERO); CHKERRQ(ierr);

  //set finite elements for problems
  ierr = m_field.modify_problem_add_finite_element("C_ALL_MATRIX","C_SURFACE_ELEM"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("CCT_ALL_MATRIX","CTC_SURFACE_ELEM"); CHKERRQ(ierr);
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss0;
    ss0 << "C_SURFACE_ELEM_msId_" << msId;
    ostringstream ss1;
    ss1 << "CTC_SURFACE_ELEM_msId_" << msId;
    ierr = m_field.modify_problem_add_finite_element("C_ALL_MATRIX",ss0.str()); CHKERRQ(ierr);
    ierr = m_field.modify_problem_add_finite_element("CCT_ALL_MATRIX",ss1.str()); CHKERRQ(ierr);
  }

  //CRACK //this for testing only
  /*if(cs) {
    ierr = m_field.modify_problem_add_finite_element("CCT_ALL_MATRIX","CTC_CRACK_SURFACE_ELEM"); CHKERRQ(ierr);
    ierr = m_field.modify_problem_add_finite_element("C_ALL_MATRIX","C_CRACK_SURFACE_ELEM"); CHKERRQ(ierr);
  }*/

  Range level_tets,level_tris,level_edges,level_nodes;
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTET,level_tets); CHKERRQ(ierr);
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBVERTEX,level_nodes); CHKERRQ(ierr);
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBEDGE,level_edges); CHKERRQ(ierr);

  MeshsetsManager *meshset_manager_ptr;
  ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Range corners_edges,corners_nodes,surfaces_faces;
  if(meshset_manager_ptr->checkMeshset(100,SIDESET)) {
    ierr = meshset_manager_ptr->getEntitiesByDimension(100,SIDESET,1,corners_edges,true); CHKERRQ(ierr);
  }
  corners_edges = intersect(corners_edges,level_edges);
  if(meshset_manager_ptr->checkMeshset(101,NODESET)) {
    ierr = meshset_manager_ptr->getEntitiesByDimension(101,NODESET,0,corners_nodes,true); CHKERRQ(ierr);
  }
  corners_nodes = intersect(corners_nodes,level_nodes);
  ierr = meshset_manager_ptr->getEntitiesByDimension(102,SIDESET,2,surfaces_faces,true); CHKERRQ(ierr);
  Skinner skin(&m_field.get_moab());
  Range skin_faces;
  rval = skin.find_skin(0,level_tets,false,skin_faces); CHKERR_MOAB(rval);
  surfaces_faces = intersect(surfaces_faces,skin_faces);

  ierr = PetscPrintf(PETSC_COMM_WORLD,"number of SIDESET 100 = %d\n",corners_edges.size()); CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"number of NODESET 101 = %d\n",corners_nodes.size()); CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"number of SIDESET 102 = %d\n",surfaces_faces.size()); CHKERRQ(ierr);

  Range crack_surfaces_faces,crack_front_edges;
  ierr = m_field.get_cubit_msId_entities_by_dimension(200,SIDESET,2,crack_surfaces_faces,true); CHKERRQ(ierr);
  ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRQ(ierr);
  crack_surfaces_faces = intersect(crack_surfaces_faces,level_tris);
  crack_front_edges = intersect(crack_front_edges,level_edges);

  Range crack_front_nodes;
  rval = m_field.get_moab().get_connectivity(crack_front_edges,crack_front_nodes,true); CHKERRQ_MOAB(rval);
  Range corners_edges_nodes;
  rval = m_field.get_moab().get_connectivity(corners_edges,corners_edges_nodes,true); CHKERRQ_MOAB(rval);
  corners_nodes.insert(corners_edges_nodes.begin(),corners_edges_nodes.end());

  //FE Couple nodes on two sides of crack
  Range one_side_crack_surface_faces;
  vector<int> sides(crack_surfaces_faces.size());
  Tag th_interface_side;
  const int def_side[] = {0};
  rval = m_field.get_moab().tag_get_handle("INTERFACE_SIDE",1,MB_TYPE_INTEGER,
    th_interface_side,MB_TAG_CREAT|MB_TAG_SPARSE,def_side); CHKERRQ_MOAB(rval);
  rval = m_field.get_moab().tag_get_data(th_interface_side,crack_surfaces_faces,&*sides.begin()); CHKERRQ_MOAB(rval);
  int nn = 0;
  for(Range::iterator fit = crack_surfaces_faces.begin();fit!=crack_surfaces_faces.end();fit++,nn++) {
    if(sides[nn] == 0) {
      one_side_crack_surface_faces.insert(*fit);
    }
  }
  ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Crack Surfaces Faces = %d\n",crack_surfaces_faces.size()); CHKERRQ(ierr);
  ierr = PetscPrintf(
    PETSC_COMM_WORLD,"number of Crack Surfaces Faces On One Side = %d\n",one_side_crack_surface_faces.size()
  ); CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Crack Front Edges = %d\n",crack_front_edges.size()); CHKERRQ(ierr);

  moab::Interface& moab = m_field.get_moab();

  ierr = m_field.seed_finite_elements(surfaces_faces); CHKERRQ(ierr);
  ierr = m_field.add_ents_to_finite_element_by_TRIs(surfaces_faces,"C_SURFACE_ELEM"); CHKERRQ(ierr);
  ierr = m_field.add_ents_to_finite_element_by_TRIs(surfaces_faces,"CTC_SURFACE_ELEM"); CHKERRQ(ierr);
  ierr = m_field.add_ents_to_finite_element_by_TRIs(surfaces_faces,"CandCT_SURFACE_ELEM"); CHKERRQ(ierr);
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss0;
    ss0 << "C_SURFACE_ELEM_msId_" << msId;
    ostringstream ss1;
    ss1 << "CTC_SURFACE_ELEM_msId_" << msId;
    ostringstream ss2;
    ss2 << "CandCT_SURFACE_ELEM_msId_" << msId;
    Range surfaces_faces_msId;
    ierr = m_field.get_cubit_msId_entities_by_dimension(msId,SIDESET,2,surfaces_faces_msId,true); CHKERRQ(ierr);
    surfaces_faces_msId = intersect(surfaces_faces_msId,skin_faces);
    ierr = m_field.seed_finite_elements(surfaces_faces_msId); CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"number of SIDESET ( mdId =  %d ) = %d\n",msId,surfaces_faces_msId.size()); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(surfaces_faces_msId,ss0.str()); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(surfaces_faces_msId,ss1.str()); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(surfaces_faces_msId,ss2.str()); CHKERRQ(ierr);
  }

  if(cs) {
    ierr = m_field.seed_finite_elements(crack_surfaces_faces); CHKERRQ(ierr);
    //ierr = m_field.add_ents_to_finite_element_by_TRIs(one_side_crack_surface_faces,"C_CRACK_SURFACE_ELEM"); CHKERRQ(ierr);
    //ierr = m_field.add_ents_to_finite_element_by_TRIs(one_side_crack_surface_faces,"CTC_CRACK_SURFACE_ELEM"); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(one_side_crack_surface_faces,"CandCT_CRACK_SURFACE_ELEM"); CHKERRQ(ierr);
  }

  Range surfaces_nodes;
  rval = moab.get_connectivity(surfaces_faces,surfaces_nodes,true); CHKERRQ_MOAB(rval);
  surfaces_nodes = subtract(surfaces_nodes,corners_nodes);
  //add entitities (by tets) to the field
  ierr = m_field.add_ents_to_field_by_VERTICEs(surfaces_nodes,"LAMBDA_SURFACE"); CHKERRQ(ierr);
  //NOTE: always order should be 1
  ierr = m_field.set_field_order(0,MBVERTEX,"LAMBDA_SURFACE",1); CHKERRQ(ierr);

  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    Range surfaces_faces_msId;
    ierr = m_field.get_cubit_msId_entities_by_dimension(msId,SIDESET,2,surfaces_faces_msId,true); CHKERRQ(ierr);
    surfaces_faces_msId = intersect(surfaces_faces_msId,level_tris);
    Range surfaces_nodes_msId;
    rval = m_field.get_moab().get_connectivity(surfaces_faces_msId,surfaces_nodes_msId,true); CHKERRQ_MOAB(rval);
    surfaces_nodes_msId = subtract(surfaces_nodes_msId,corners_nodes);
    ostringstream ss;
    ss << "LAMBDA_SURFACE_msId_" << msId;
    ierr = m_field.add_ents_to_field_by_VERTICEs(surfaces_nodes_msId,ss.str()); CHKERRQ(ierr);
    ierr = m_field.set_field_order(0,MBVERTEX,ss.str(),1); CHKERRQ(ierr);
  }

  //CRCAK
  if(cs) {

    Range one_side_crack_surface_faces_nodes;
    rval = m_field.get_moab().get_connectivity(
      one_side_crack_surface_faces,one_side_crack_surface_faces_nodes,true
    ); CHKERRQ_MOAB(rval);
    one_side_crack_surface_faces_nodes = subtract(one_side_crack_surface_faces_nodes,corners_nodes);
    Range one_side_crack_surface_faces_nodes_without_crack_front;
    one_side_crack_surface_faces_nodes_without_crack_front = subtract(one_side_crack_surface_faces_nodes,crack_front_nodes);

    Range other_side_nodes = one_side_crack_surface_faces_nodes_without_crack_front;
    PetscBool flg_other_side = PETSC_FALSE;
    ierr = PetscOptionsGetBool(
      PETSC_NULL,PETSC_NULL,"-my_other_side_constrain",&flg_other_side,PETSC_NULL
    ); CHKERRQ(ierr);

    if(!flg_other_side) {
      ierr = m_field.remove_ents_from_field(
        "LAMBDA_SURFACE",one_side_crack_surface_faces_nodes_without_crack_front
      ); CHKERRQ(ierr);
      for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
        int msId = it->getMeshsetId();
        if((msId < 10200)||(msId >= 10300)) continue;
        ostringstream ss;
        ss << "LAMBDA_SURFACE_msId_" << msId;
        ierr = m_field.remove_ents_from_field(
          ss.str(),one_side_crack_surface_faces_nodes_without_crack_front
        ); CHKERRQ(ierr);
      }
      ierr = m_field.add_ents_to_field_by_VERTICEs(
        one_side_crack_surface_faces_nodes_without_crack_front,"LAMBDA_CRACK_SURFACE"
      ); CHKERRQ(ierr);
    } else {
      Skinner skin(&m_field.get_moab());
      Range crack_surface_skin_edges;
      rval = skin.find_skin(0,crack_surfaces_faces,false,crack_surface_skin_edges); CHKERRQ_MOAB(rval);
      Range surfaces;
      Range surfaces_edges;
      ierr = m_field.get_cubit_msId_entities_by_dimension(102,SIDESET,2,surfaces,true); CHKERRQ(ierr);
      rval = skin.find_skin(0,surfaces,false,surfaces_edges); CHKERRQ_MOAB(rval);
      surfaces_edges = intersect(surfaces_edges,crack_surface_skin_edges);
      Range surfaces_edges_nodes;
      rval = m_field.get_moab().get_connectivity(
        surfaces_edges,surfaces_edges_nodes,true
      ); CHKERRQ_MOAB(rval);
      Range crack_surface_nodes = one_side_crack_surface_faces_nodes_without_crack_front;
      crack_surface_nodes = subtract(
        crack_surface_nodes,surfaces_edges_nodes
      );
      ierr = m_field.remove_ents_from_field(
        "LAMBDA_SURFACE",
        subtract(
          one_side_crack_surface_faces_nodes_without_crack_front,surfaces_edges_nodes
        )
      ); CHKERRQ(ierr);
      for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
        int msId = it->getMeshsetId();
        if((msId < 10200)||(msId >= 10300)) continue;
        ostringstream ss;
        ss << "LAMBDA_SURFACE_msId_" << msId;
        EntityHandle meshset = it->getMeshset();
        surfaces.clear();
        rval = m_field.get_moab().get_entities_by_type(meshset,MBTRI,surfaces,true); CHKERRQ_MOAB(rval);
        surfaces_edges.clear();
        rval = skin.find_skin(0,surfaces,false,surfaces_edges); CHKERRQ_MOAB(rval);
        surfaces_edges = intersect(surfaces_edges,crack_surface_skin_edges);
        surfaces_edges_nodes.clear();
        rval = m_field.get_moab().get_connectivity(
          surfaces_edges,surfaces_edges_nodes,true
        ); CHKERRQ_MOAB(rval);
        crack_surface_nodes = subtract(
          crack_surface_nodes,surfaces_edges_nodes
        );
        ierr = m_field.remove_ents_from_field(
          ss.str(),
          subtract(
            one_side_crack_surface_faces_nodes_without_crack_front,surfaces_edges_nodes
          )
        ); CHKERRQ(ierr);
      }
      ierr = m_field.add_ents_to_field_by_VERTICEs(
        crack_surface_nodes,"LAMBDA_CRACK_SURFACE"
      ); CHKERRQ(ierr);
    }
    ierr = m_field.set_field_order(0,MBVERTEX,"LAMBDA_CRACK_SURFACE",1); CHKERRQ(ierr);

    ierr = m_field.add_ents_to_field_by_VERTICEs(
      one_side_crack_surface_faces_nodes_without_crack_front,"LAMBDA_BOTH_SIDES"
    ); CHKERRQ(ierr);
    ierr = m_field.set_field_order(0,MBVERTEX,"LAMBDA_BOTH_SIDES",1); CHKERRQ(ierr);

    Tag th_my_ref_level;
    rval = m_field.get_moab().tag_get_handle("_MY_REFINEMENT_LEVEL",th_my_ref_level);
    const EntityHandle root_meshset = m_field.get_moab().get_root_set();
    BitRefLevel *ptr_bit_level0;
    rval = m_field.get_moab().tag_get_by_ptr(th_my_ref_level,&root_meshset,1,(const void**)&ptr_bit_level0); CHKERRQ_MOAB(rval);
    BitRefLevel& bit_level0 = *ptr_bit_level0;

    Range prisms;
    ierr = m_field.get_entities_by_type_and_ref_level(bit_level0,BitRefLevel().set(),MBPRISM,prisms); CHKERRQ(ierr);
    ierr = m_field.seed_ref_level(prisms,bit_level0); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_PRISMs(prisms,"BOTH_SIDE_OF_CRACK"); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::constrains_crack_front_problem_definition(MoFEM::Interface& m_field,string problem) {
  PetscFunctionBegin;

  if(material_FirelWall->operator[](FW_constrains_crack_front_problem_definition)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_constrains_crack_front_problem_definition);

  PetscErrorCode ierr;
  ErrorCode rval;

  Range level_tets,level_tris,level_edges,level_nodes;
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTET,level_tets); CHKERRQ(ierr);
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBVERTEX,level_nodes); CHKERRQ(ierr);
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBEDGE,level_edges); CHKERRQ(ierr);

  Range body_surface_nodes;
  {
    Range surfaces;
    ierr = m_field.get_cubit_msId_entities_by_dimension(102,SIDESET,2,surfaces,true); CHKERRQ(ierr);
    surfaces = intersect(surfaces,level_tris);
    for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
      int msId = it->getMeshsetId();
      if((msId < 10200)||(msId >= 10300)) continue;
      Range surf;
      rval = m_field.get_moab().get_entities_by_type(it->getMeshset(),MBTRI,surf,true); CHKERRQ_MOAB(rval);
      surf = intersect(surf,level_tris);
      surfaces.merge(surf);
    }
    rval = m_field.get_moab().get_connectivity(surfaces,body_surface_nodes,true); CHKERRQ_MOAB(rval);
  }


  //Fields
  ierr = m_field.add_field("LAMBDA_CRACKFRONT_AREA",H1,1,MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_field("LAMBDA_CRACK_TANGENT_CONSTRAIN",H1,1,MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_field("GRIFFITH_FORCE",H1,3,MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_field("GRIFFITH_FORCE_TANGENT",H1,3,MF_ZERO); CHKERRQ(ierr);

  //FE
  ierr = m_field.add_finite_element("C_CRACKFRONT_AREA_ELEM",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_finite_element("CTC_CRACKFRONT_AREA_ELEM",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_finite_element("dCT_CRACKFRONT_AREA_ELEM",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_finite_element("C_TANGENT_ELEM",MF_ZERO); CHKERRQ(ierr);

  ierr = m_field.modify_finite_element_add_field_row("C_CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("C_CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("C_CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("C_CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);

  ierr = m_field.modify_finite_element_add_field_row("CTC_CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("CTC_CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("CTC_CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("CTC_CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);

  ierr = m_field.modify_finite_element_add_field_row("dCT_CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("dCT_CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("dCT_CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("dCT_CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);

  ierr = m_field.modify_finite_element_add_field_row("C_TANGENT_ELEM","LAMBDA_CRACK_TANGENT_CONSTRAIN"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_row("C_TANGENT_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("C_TANGENT_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_col("C_TANGENT_ELEM","LAMBDA_CRACK_TANGENT_CONSTRAIN"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("C_TANGENT_ELEM","LAMBDA_CRACK_TANGENT_CONSTRAIN"); CHKERRQ(ierr);
  ierr = m_field.modify_finite_element_add_field_data("C_TANGENT_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);

  //Problem
  ierr = m_field.add_problem("C_CRACKFRONT_MATRIX",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.add_problem("CTC_CRACKFRONT_MATRIX",MF_ZERO); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("C_CRACKFRONT_MATRIX","C_CRACKFRONT_AREA_ELEM"); CHKERRQ(ierr);
  ierr = m_field.modify_problem_add_finite_element("CTC_CRACKFRONT_MATRIX","CTC_CRACKFRONT_AREA_ELEM"); CHKERRQ(ierr);

  {
    Range crack_surfaces_faces,crack_front_edges;
    ierr = m_field.get_cubit_msId_entities_by_dimension(200,SIDESET,2,crack_surfaces_faces,true); CHKERRQ(ierr);
    ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRQ(ierr);
    crack_surfaces_faces = intersect(crack_surfaces_faces,level_tris);
    crack_front_edges = intersect(crack_front_edges,level_edges);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Crack Faces = %d\n",crack_surfaces_faces.size()); CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Crack Front = %d\n",crack_front_edges.size()); CHKERRQ(ierr);
    Range crack_front_nodes;
    rval = m_field.get_moab().get_connectivity(crack_front_edges,crack_front_nodes,true); CHKERRQ_MOAB(rval);
    crack_front_nodes = intersect(crack_front_nodes,level_nodes);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Front Nodes = %d\n",crack_front_nodes.size()); CHKERRQ(ierr);
    Range tangent_crack_front_nodes = subtract(crack_front_nodes,body_surface_nodes);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Tangent Crack Front Nodes = %d\n",tangent_crack_front_nodes.size()); CHKERRQ(ierr);

    Range crack_surfaces_edge_faces;
    rval = m_field.get_moab().get_adjacencies(
      crack_front_nodes,2,false,crack_surfaces_edge_faces,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    crack_surfaces_edge_faces = crack_surfaces_edge_faces.subset_by_type(MBTRI);
    crack_surfaces_edge_faces = intersect(crack_surfaces_edge_faces,crack_surfaces_faces);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Front Faces = %d\n",crack_surfaces_edge_faces.size()); CHKERRQ(ierr);

    Range level_tris;
    ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
    crack_surfaces_edge_faces = intersect(crack_surfaces_edge_faces,level_tris);

    ierr = m_field.seed_finite_elements(crack_surfaces_edge_faces); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(crack_surfaces_edge_faces,"C_CRACKFRONT_AREA_ELEM"); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(crack_surfaces_edge_faces,"CTC_CRACKFRONT_AREA_ELEM"); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(crack_surfaces_edge_faces,"dCT_CRACKFRONT_AREA_ELEM"); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_finite_element_by_TRIs(crack_surfaces_edge_faces,"C_TANGENT_ELEM"); CHKERRQ(ierr);

    Range level_edges;
    ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBEDGE,level_edges); CHKERRQ(ierr);
    crack_front_edges = intersect(crack_front_edges,level_edges);
    ierr = m_field.seed_finite_elements(crack_front_edges); CHKERRQ(ierr);

    //add entitities (by tets) to the field
    ierr = m_field.add_ents_to_field_by_VERTICEs(crack_front_nodes,"LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
    ierr = m_field.add_ents_to_field_by_VERTICEs(tangent_crack_front_nodes,"LAMBDA_CRACK_TANGENT_CONSTRAIN"); CHKERRQ(ierr);

  }

  //NOTE: always order should be 1
  ierr = m_field.set_field_order(0,MBVERTEX,"LAMBDA_CRACKFRONT_AREA",1); CHKERRQ(ierr);
  ierr = m_field.set_field_order(0,MBVERTEX,"LAMBDA_CRACK_TANGENT_CONSTRAIN",1); CHKERRQ(ierr);

  //set finite elements for problems
  ierr = m_field.modify_problem_add_finite_element(problem,"dCT_CRACKFRONT_AREA_ELEM"); CHKERRQ(ierr);

  if(problem == "COUPLED_PROBLEM" || problem == "COUPLED_DYNAMIC") {
    ierr = m_field.modify_problem_add_finite_element(problem,"C_TANGENT_ELEM"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_row("MESH_SMOOTHER","LAMBDA_CRACK_TANGENT_CONSTRAIN"); CHKERRQ(ierr);
    ierr = m_field.modify_finite_element_add_field_col("MESH_SMOOTHER","LAMBDA_CRACK_TANGENT_CONSTRAIN"); CHKERRQ(ierr);
  }



  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::spatial_partition_problems(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  PetscErrorCode ierr;
  // MoABErrorCode rval;

  //rval = m_field.get_moab().write_file("test.h5m"); CHKERRQ_MOAB(rval);

  if(m_field.check_problem("BC_PROBLEM")) {
    ierr = m_field.partition_problem("BC_PROBLEM",1); CHKERRQ(ierr);
    ParallelComm* pcomm = ParallelComm::get_pcomm(&m_field.get_moab(),MYPCOMM_INDEX);
    ierr = m_field.partition_finite_elements("BC_PROBLEM",false,0,pcomm->size()); CHKERRQ(ierr);
    ierr = m_field.partition_ghost_dofs("BC_PROBLEM"); CHKERRQ(ierr);
  }

  //partition
  ierr = m_field.partition_problem("ELASTIC_MECHANICS",1); CHKERRQ(ierr);
  ParallelComm* pcomm = ParallelComm::get_pcomm(&m_field.get_moab(),MYPCOMM_INDEX);
  ierr = m_field.partition_finite_elements("ELASTIC_MECHANICS",false,0,pcomm->size()); CHKERRQ(ierr);
  ierr = m_field.partition_ghost_dofs("ELASTIC_MECHANICS"); CHKERRQ(ierr);


  // ierr = PetscBarrier(PETSC_NULL); CHKERRQ(ierr);

  // ierr = m_field.partition_check_matrix_fill_in("ELASTIC_MECHANICS",13152,3939,0); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::material_partition_problems(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  PetscErrorCode ierr;

  //partition MATERIAL_MECHANICS
  ierr = m_field.partition_problem("MATERIAL_MECHANICS"); CHKERRQ(ierr);
  ierr = m_field.partition_finite_elements("MATERIAL_MECHANICS"); CHKERRQ(ierr);
  ierr = m_field.partition_ghost_dofs("MATERIAL_MECHANICS"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::coupled_partition_problems(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  PetscErrorCode ierr;

  //partition
  ierr = m_field.partition_problem("COUPLED_PROBLEM"); CHKERRQ(ierr);
  ierr = m_field.partition_finite_elements("COUPLED_PROBLEM"); CHKERRQ(ierr);
  ierr = m_field.partition_ghost_dofs("COUPLED_PROBLEM"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::constrains_partition_problems(MoFEM::Interface& m_field,string problem) {
  PetscFunctionBegin;

  PetscErrorCode ierr;

  //partition
  ierr = m_field.partition_simple_problem("CCT_ALL_MATRIX",0); CHKERRQ(ierr);
  ierr = m_field.partition_finite_elements("CCT_ALL_MATRIX"); CHKERRQ(ierr);
  ierr = m_field.partition_ghost_dofs("CCT_ALL_MATRIX"); CHKERRQ(ierr);
  //partition
  ierr = m_field.partition_compose_problem("C_ALL_MATRIX","CCT_ALL_MATRIX",false,problem,true); CHKERRQ(ierr);
  ierr = m_field.partition_finite_elements("C_ALL_MATRIX"); CHKERRQ(ierr);
  ierr = m_field.partition_ghost_dofs("C_ALL_MATRIX"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::crackfront_partition_problems(MoFEM::Interface& m_field,string problem) {
  PetscFunctionBegin;

  PetscErrorCode ierr;

  //partition
  ierr = m_field.partition_simple_problem("CTC_CRACKFRONT_MATRIX",0); CHKERRQ(ierr);
  ierr = m_field.partition_finite_elements("CTC_CRACKFRONT_MATRIX"); CHKERRQ(ierr);
  ierr = m_field.partition_ghost_dofs("CTC_CRACKFRONT_MATRIX"); CHKERRQ(ierr);
  //partition
  ierr = m_field.partition_compose_problem("C_CRACKFRONT_MATRIX","CTC_CRACKFRONT_MATRIX",false,problem,true); CHKERRQ(ierr);
  ierr = m_field.partition_finite_elements("C_CRACKFRONT_MATRIX"); CHKERRQ(ierr);
  ierr = m_field.partition_ghost_dofs("C_CRACKFRONT_MATRIX"); CHKERRQ(ierr);

  PetscFunctionReturn(0);

}

PetscErrorCode ConfigurationalFractureMechanics::set_spatial_positions(MoFEM::Interface& m_field) {
  PetscFunctionBegin;
  if(material_FirelWall->operator[](FW_set_spatial_positions)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_set_spatial_positions);

  ErrorCode rval;

  EntityHandle node = 0;
  double coords[3];
  for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(m_field,"SPATIAL_POSITION",dof_ptr)) {
    if(dof_ptr->get()->getEntType()!=MBVERTEX) {
      double &fval = dof_ptr->get()->getFieldData();
      fval = 0;
      continue;
    }
    EntityHandle ent = dof_ptr->get()->getEnt();
    int dof_rank = dof_ptr->get()->getDofCoeffIdx();
    double &fval = dof_ptr->get()->getFieldData();
    if(node!=ent) {
      rval = m_field.get_moab().get_coords(&ent,1,coords); CHKERRQ_MOAB(rval);
      node = ent;
    }
    fval = coords[dof_rank];
  }

  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::set_material_positions(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  if(material_FirelWall->operator[](FW_set_material_positions)) PetscFunctionReturn(0);
  material_FirelWall->set(FW_set_material_positions);

  ErrorCode rval;

  EntityHandle node = 0;
  double coords[3];
  for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(m_field,"MESH_NODE_POSITIONS",dof_ptr)) {
    if(dof_ptr->get()->getEntType()!=MBVERTEX) {
      double &fval = dof_ptr->get()->getFieldData();
      fval = 0;
      continue;
    }
    EntityHandle ent = dof_ptr->get()->getEnt();
    int dof_rank = dof_ptr->get()->getDofCoeffIdx();
    double &fval = dof_ptr->get()->getFieldData();
    if(node!=ent) {
      rval = m_field.get_moab().get_coords(&ent,1,coords); CHKERRQ_MOAB(rval);
      node = ent;
    }
    fval = coords[dof_rank];
  }

  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::set_coordinates_from_material_solution(MoFEM::Interface& m_field,bool only_crack_front) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  ErrorCode rval;

  Range crack_front_edges,crack_front_edges_nodes;
  if(only_crack_front) {
    ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRQ(ierr);
    rval = m_field.get_moab().get_connectivity(crack_front_edges,crack_front_edges_nodes,true); CHKERRQ_MOAB(rval);
  }

  double coords[3];
  for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(m_field,"MESH_NODE_POSITIONS",dof_ptr)) {
    if(dof_ptr->get()->getEntType()!=MBVERTEX) continue;
    if(only_crack_front) {
      if(
        crack_front_edges_nodes.find(dof_ptr->get()->getEnt())
        ==crack_front_edges_nodes.end()) {
          continue;
        }
      }
    EntityHandle ent = dof_ptr->get()->getEnt();
    int dof_rank = dof_ptr->get()->getDofCoeffIdx();
    double fval = dof_ptr->get()->getFieldData();
    rval = m_field.get_moab().get_coords(&ent,1,coords); CHKERRQ_MOAB(rval);
    coords[dof_rank] = fval;
    rval = m_field.get_moab().set_coords(&ent,1,coords); CHKERRQ_MOAB(rval);
  }
  for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(m_field,"LAMBDA_SURFACE",dof_ptr)) {
    dof_ptr->get()->getFieldData() = 0;
  }
  for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(m_field,"LAMBDA_CRACK_SURFACE",dof_ptr)) {
    dof_ptr->get()->getFieldData() = 0;
  }
  for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(m_field,"LAMBDA_CRACK_TANGENT_CONSTRAIN",dof_ptr)) {
    dof_ptr->get()->getFieldData() = 0;
  }
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss;
    ss << "LAMBDA_SURFACE_msId_" << msId;
    for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(m_field,ss.str(),dof_ptr)) {
      dof_ptr->get()->getFieldData() = 0;
    }
  }
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::solve_spatial_problem(MoFEM::Interface& m_field,SNES *snes,bool postproc) {
  PetscFunctionBegin;

  PetscErrorCode ierr;
  ErrorCode rval;

  //set_PhysicalEquationNumber(hooke);

  //create matrices
  Vec F;
  ierr = m_field.VecCreateGhost("ELASTIC_MECHANICS",COL,&F); CHKERRQ(ierr);
  Vec D;
  ierr = VecDuplicate(F,&D); CHKERRQ(ierr);
  Mat Aij;
  ierr = m_field.MatCreateMPIAIJWithArrays("ELASTIC_MECHANICS",&Aij); CHKERRQ(ierr);

  PetscBool flg;
  PetscReal my_tol;
  ierr = PetscOptionsGetReal(PETSC_NULL,PETSC_NULL,"-my_tol",&my_tol,&flg); CHKERRQ(ierr);
  if(flg == PETSC_TRUE) {
    PetscReal atol,rtol,stol;
    PetscInt maxit,maxf;
    ierr = SNESGetTolerances(*snes,&atol,&rtol,&stol,&maxit,&maxf); CHKERRQ(ierr);
    atol = my_tol;
    rtol = atol*1e2;
    ierr = SNESSetTolerances(*snes,atol,rtol,stol,maxit,maxf); CHKERRQ(ierr);
  }

  SnesCtx snes_ctx(m_field,"ELASTIC_MECHANICS");

  ierr = SNESSetApplicationContext(*snes,&snes_ctx); CHKERRQ(ierr);
  ierr = SNESSetFunction(*snes,F,SnesRhs,&snes_ctx); CHKERRQ(ierr);
  ierr = SNESSetJacobian(*snes,Aij,Aij,SnesMat,&snes_ctx); CHKERRQ(ierr);

  NonlinearElasticElement spatial_fe(m_field,-1);

  boost::shared_ptr<Hooke<adouble> > hooke_adouble(new Hooke<adouble>);
  boost::shared_ptr<Hooke<double> > hooke_double(new Hooke<double>);
  //NeoHookean<adouble> hooke_adouble;
  //NeoHookean<double> hooke_double;

  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,BLOCKSET|MAT_ELASTICSET,it)) {
    Mat_Elastic mydata;
    ierr = it->getAttributeDataStructure(mydata); CHKERRQ(ierr);
    int id = it->getMeshsetId();
    EntityHandle meshset = it->getMeshset();
    rval = m_field.get_moab().get_entities_by_type(meshset,MBTET,spatial_fe.setOfBlocks[id].tEts,true); CHKERRQ_MOAB(rval);
    spatial_fe.setOfBlocks[id].iD = id;
    spatial_fe.setOfBlocks[id].E = mydata.data.Young;
    spatial_fe.setOfBlocks[id].PoissonRatio = mydata.data.Poisson;
    spatial_fe.setOfBlocks[id].materialDoublePtr = hooke_double;
    spatial_fe.setOfBlocks[id].materialAdoublePtr = hooke_adouble;
  }

  spatial_fe.commonData.spatialPositions = "SPATIAL_POSITION";
  spatial_fe.commonData.meshPositions = "MESH_NODE_POSITIONS";


  /// Rhs
  spatial_fe.feRhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("SPATIAL_POSITION",spatial_fe.commonData)
  );
  if(m_field.check_field("MESH_NODE_POSITIONS")) {
    spatial_fe.feRhs.getOpPtrVector().push_back(new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",spatial_fe.commonData));
  }
  map<int,NonlinearElasticElement::BlockData>::iterator sit = spatial_fe.setOfBlocks.begin();
  for(;sit!=spatial_fe.setOfBlocks.end();sit++) {
    spatial_fe.feRhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpJacobianPiolaKirchhoffStress("SPATIAL_POSITION",sit->second,spatial_fe.commonData,1,false,true,false)
    );
    spatial_fe.feRhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpRhsPiolaKirchhoff("SPATIAL_POSITION",sit->second,spatial_fe.commonData)
    );
  }
  spatial_fe.feRhs.meshPositionsFieldName = "NONE";
  spatial_fe.feRhs.addToRule = 0;

  /// Lhs
  spatial_fe.feLhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("SPATIAL_POSITION",spatial_fe.commonData)
  );
  spatial_fe.feLhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",spatial_fe.commonData)
  );
  sit = spatial_fe.setOfBlocks.begin();
  for(;sit!=spatial_fe.setOfBlocks.end();sit++) {
    spatial_fe.feLhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpJacobianPiolaKirchhoffStress("SPATIAL_POSITION",sit->second,spatial_fe.commonData,1,true,true,false)
    );
    spatial_fe.feLhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpLhsPiolaKirchhoff_dx("SPATIAL_POSITION","SPATIAL_POSITION",sit->second,spatial_fe.commonData)
    );
  }
  spatial_fe.feLhs.meshPositionsFieldName = "NONE";
  spatial_fe.feLhs.addToRule = 0;

  NeummanForcesSurfaceComplexForLazy neumann_forces(m_field,Aij,F);
  NeummanForcesSurfaceComplexForLazy::MyTriangleSpatialFE &fe_forces = neumann_forces.getLoopSpatialFe();
  fe_forces.typeOfForces = NeummanForcesSurfaceComplexForLazy::MyTriangleSpatialFE::NONCONSERVATIVE;
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,NODESET|FORCESET,it)) {
    ierr = fe_forces.addForce(it->getMeshsetId()); CHKERRQ(ierr);
  }
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,SIDESET|PRESSURESET,it)) {
    ierr = fe_forces.addPreassure(it->getMeshsetId()); CHKERRQ(ierr);
  }

  SpatialPositionsBCFEMethodPreAndPostProc my_dirichlet_bc(m_field,"SPATIAL_POSITION",Aij,D,F);
  #ifdef __ANALITICAL_DISPLACEMENT__
  AnalyticalDirichletBC::DirichletBC analytical_ditihlet_bc(m_field,"SPATIAL_POSITION",Aij,D,F);
  if(m_field.check_problem("BC_PROBLEM")) {
    //solve for ditihlet bc dofs
    AnalyticalDirichletBC analytical_bc(m_field);
    ierr = analytical_bc.setUpProblem(m_field,"BC_PROBLEM"); CHKERRQ(ierr);
    boost::shared_ptr<AnaliticalDisp> testing_function = boost::shared_ptr<AnaliticalDisp>(new AnaliticalDisp);
    EntityHandle fe_meshset = m_field.get_finite_element_meshset("BC_FE");
    ierr = analytical_bc.setApproxOps(m_field,"SPATIAL_POSITION",testing_function,0); CHKERRQ(ierr);
    ierr = analytical_bc.solveProblem(m_field,"BC_PROBLEM","BC_FE",analytical_ditihlet_bc); CHKERRQ(ierr);
    ierr = analytical_bc.destroyProblem(); CHKERRQ(ierr);
  }
  #endif //__ANALITICAL_DISPLACEMENT__

  SnesCtx::loops_to_do_type& loops_to_do_Rhs = snes_ctx.get_loops_to_do_Rhs();
  snes_ctx.get_preProcess_to_do_Rhs().push_back(&my_dirichlet_bc);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(m_field.check_problem("BC_PROBLEM")) {
    snes_ctx.get_preProcess_to_do_Rhs().push_back(&analytical_ditihlet_bc);
  }
  #endif //__ANALITICAL_DISPLACEMENT__

  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("ELASTIC",&spatial_fe.feRhs));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("NEUAMNN_FE",&fe_forces));
  //nodal forces
  boost::ptr_map<string,NodalForce> nodal_forces;
  string fe_name_str ="FORCE_FE";
  nodal_forces.insert(fe_name_str,new NodalForce(m_field));
  nodal_forces.at(fe_name_str).methodsOp.push_back(new MetaNodalForces::TagForceScale(m_field));
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,NODESET|FORCESET,it)) {
    ierr = nodal_forces.at(fe_name_str).addForce("SPATIAL_POSITION",F,it->getMeshsetId(),true);  CHKERRQ(ierr);
  }
  {
    boost::ptr_map<string,NodalForce>::iterator fit = nodal_forces.begin();
    for(;fit!=nodal_forces.end();fit++) {
      loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type(fit->first,&fit->second->getLoopFe()));
    }
  }
  //edge forces
  boost::ptr_map<string,EdgeForce> edge_forces;
  edge_forces.insert(fe_name_str,new EdgeForce(m_field));
  edge_forces.at(fe_name_str).methodsOp.push_back(new MetaNodalForces::TagForceScale(m_field));
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,NODESET|FORCESET,it)) {
    ierr = edge_forces.at(fe_name_str).addForce("SPATIAL_POSITION",F,it->getMeshsetId(),true);  CHKERRQ(ierr);
  }
  {
    boost::ptr_map<string,EdgeForce>::iterator fit = edge_forces.begin();
    for(;fit!=edge_forces.end();fit++) {
      loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type(fit->first,&fit->second->getLoopFe()));
    }
  }
  // Analytical forces
  #ifdef __ANALITICAL_TRACTION__
  boost::ptr_map<string,NeummanForcesSurface> analitical_surface_element;
  {
    MeshsetsManager *meshset_manager_ptr;
    ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
    if(meshset_manager_ptr->checkMeshset("ANALITICAL_TRACTION")) {
      string fe_name_str = "NEUAMNN_FE";
      analitical_surface_element.insert(fe_name_str,new NeummanForcesSurface(m_field));
      const CubitMeshSets *cubit_meshset_ptr;
      meshset_manager_ptr->getCubitMeshsetPtr("ANALITICAL_TRACTION",&cubit_meshset_ptr);
      Range faces;
      ierr = meshset_manager_ptr->getEntitiesByDimension(
        cubit_meshset_ptr->getMeshsetId(),BLOCKSET,2,faces,true
      ); CHKERRQ(ierr);
      analitical_surface_element.at(fe_name_str).methodsOp.push_back(
        new MetaNodalForces::TagForceScale(m_field)
      );
      analitical_surface_element.at(fe_name_str).analyticalForceOp.push_back(
        new AnalyticalForces()
      );
      analitical_surface_element.at(fe_name_str).fe.getOpPtrVector().push_back(
        new NeummanForcesSurface::OpNeumannForceAnalytical(
          "SPATIAL_POSITION",
          PETSC_NULL,
          faces,
          analitical_surface_element.at(fe_name_str).methodsOp,
          analitical_surface_element.at(fe_name_str).analyticalForceOp
        )
      );
      loops_to_do_Rhs.push_back(
        SnesCtx::loop_pair_type(
          fe_name_str,&analitical_surface_element.at(fe_name_str).fe
        )
      );
    }
  }
  #endif //__ANALITICAL_TRACTION__

  //postproc
  snes_ctx.get_postProcess_to_do_Rhs().push_back(&my_dirichlet_bc);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(m_field.check_problem("BC_PROBLEM")) {
    snes_ctx.get_postProcess_to_do_Rhs().push_back(&analytical_ditihlet_bc);
  }
  #endif //__ANALITICAL_DISPLACEMENT__

  SnesCtx::loops_to_do_type& loops_to_do_Mat = snes_ctx.get_loops_to_do_Mat();
  snes_ctx.get_preProcess_to_do_Mat().push_back(&my_dirichlet_bc);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(m_field.check_problem("BC_PROBLEM")) {
    snes_ctx.get_preProcess_to_do_Mat().push_back(&analytical_ditihlet_bc);
  }
  #endif //__ANALITICAL_DISPLACEMENT__
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("ELASTIC",&spatial_fe.feLhs));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("NEUAMNN_FE",&fe_forces));
  snes_ctx.get_postProcess_to_do_Mat().push_back(&my_dirichlet_bc);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(m_field.check_problem("BC_PROBLEM")) {
    snes_ctx.get_postProcess_to_do_Mat().push_back(&analytical_ditihlet_bc);
  }
  #endif //__ANALITICAL_DISPLACEMENT__

  ierr = m_field.set_local_ghost_vector("ELASTIC_MECHANICS",COL,D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  ierr = SNESSolve(*snes,PETSC_NULL,D); CHKERRQ(ierr);
  int its;
  ierr = SNESGetIterationNumber(*snes,&its); CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Newton iterations = %D\n",its); CHKERRQ(ierr);

  ierr = VecGhostUpdateBegin(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  //Save data on mesh
  ierr = m_field.set_global_ghost_vector("ELASTIC_MECHANICS",COL,D,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  //ierr = VecView(F,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);

  if(postproc) {

    PostProcVolumeOnRefinedMesh post_proc(m_field);
    ierr = post_proc.generateReferenceElementMesh(); CHKERRQ(ierr);
    ierr = post_proc.addFieldValuesPostProc("SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = post_proc.addFieldValuesPostProc("MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = post_proc.addFieldValuesGradientPostProc("SPATIAL_POSITION"); CHKERRQ(ierr);

    map<int,NonlinearElasticElement::BlockData>::iterator sit = spatial_fe.setOfBlocks.begin();
    for (; sit != spatial_fe.setOfBlocks.end(); sit++) {
      post_proc.getOpPtrVector().push_back(
        new PostPorcStress(
          post_proc.postProcMesh,
          post_proc.mapGaussPts,
          "SPATIAL_POSITION",
          sit->second,
          post_proc.commonData
        )
      );

    }

    ierr = m_field.loop_finite_elements("ELASTIC_MECHANICS","ELASTIC",post_proc); CHKERRQ(ierr);
    ierr = post_proc.writeFile("out_stress.h5m"); CHKERRQ(ierr);

  }

  //detroy matrices
  ierr = VecDestroy(&F); CHKERRQ(ierr);
  ierr = VecDestroy(&D); CHKERRQ(ierr);
  ierr = MatDestroy(&Aij); CHKERRQ(ierr);

  ierr = PetscPrintf(PETSC_COMM_WORLD,"Spatial problem solved\n"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::surface_projection_data(MoFEM::Interface& m_field,string problem) {
  PetscFunctionBegin;

  PetscErrorCode ierr;

  ierr = delete_surface_projection_data(m_field); CHKERRQ(ierr);

  //C_ALL_MATRIX is problem used to constrain surface constrain matrices
  if(projSurfaceCtx==NULL) {
    projSurfaceCtx = new ConstrainMatrixCtx(m_field,problem,"C_ALL_MATRIX");
    ierr = m_field.MatCreateMPIAIJWithArrays("C_ALL_MATRIX",&projSurfaceCtx->C); CHKERRQ(ierr);
  }

  MeshsetsManager *meshset_manager_ptr;
  ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Range corners_edges,corners_nodes;
  if(meshset_manager_ptr->checkMeshset(100,SIDESET)) {
    ierr = meshset_manager_ptr->getEntitiesByDimension(100,SIDESET,1,corners_edges,true); CHKERRQ(ierr);
  }
  if(meshset_manager_ptr->checkMeshset(101,NODESET)) {
    ierr = meshset_manager_ptr->getEntitiesByDimension(101,NODESET,0,corners_nodes,true); CHKERRQ(ierr);
  }
  // ierr = m_field.get_cubit_msId_entities_by_dimension(100,SIDESET,1,corners_edges,true); CHKERRQ(ierr);
  // ierr = m_field.get_cubit_msId_entities_by_dimension(101,NODESET,0,corners_nodes,true); CHKERRQ(ierr);
  ErrorCode rval;
  Range corners_edges_nodes;
  rval = m_field.get_moab().get_connectivity(corners_edges,corners_edges_nodes,true); CHKERRQ_MOAB(rval);
  corners_nodes.merge(corners_edges_nodes);
  Range nodes_to_block;
  BitRefLevel bit_to_block = BitRefLevel().set(BITREFLEVEL_SIZE-2);
  ierr = m_field.get_entities_by_type_and_ref_level(bit_to_block,BitRefLevel().set(),MBVERTEX,nodes_to_block); CHKERRQ(ierr);
  corners_nodes.merge(nodes_to_block);

  ierr = MatSetOption(projSurfaceCtx->C,MAT_NEW_NONZERO_LOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);
  ierr = MatSetOption(projSurfaceCtx->C,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);
  ierr = MatZeroEntries(projSurfaceCtx->C); CHKERRQ(ierr);

  {
    FaceOrientation orientation(false),orientation_crack(true);
    SurfaceSlidingConstrains surface_constrain(m_field,orientation);
    SurfaceSlidingConstrains surface_crack_constrain(m_field,orientation_crack);
    boost::ptr_map<int,SurfaceSlidingConstrains> surface_constrains_map;
    for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
      int msId = it->getMeshsetId();
      surface_constrains_map.insert(msId,new SurfaceSlidingConstrains(m_field,orientation));
    }

    surface_constrain.setOperatorsCOnly("LAMBDA_SURFACE","MESH_NODE_POSITIONS");
    surface_crack_constrain.setOperatorsCOnly("LAMBDA_CRACK_SURFACE","MESH_NODE_POSITIONS");
    for(
      boost::ptr_map<int,SurfaceSlidingConstrains>::iterator it = surface_constrains_map.begin();
      it!=surface_constrains_map.end();
      it++
    ) {
      int msId =  it->first;
      if((msId < 10200)||(msId >= 10300)) continue;
      ostringstream ss;
      ss << "LAMBDA_SURFACE_msId_" << msId;
      ierr = it->second->setOperatorsCOnly(ss.str(),"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    }

    surface_constrain.feLhs.snes_B = projSurfaceCtx->C;
    ierr = m_field.loop_finite_elements(
      "C_ALL_MATRIX","C_SURFACE_ELEM",surface_constrain.feLhs
    ); CHKERRQ(ierr);

    surface_crack_constrain.feLhs.snes_B = projSurfaceCtx->C;
    ierr = m_field.loop_finite_elements(
      "C_ALL_MATRIX","C_CRACK_SURFACE_ELEM",surface_crack_constrain.feLhs,MF_ZERO
    );  CHKERRQ(ierr);
    for(
      boost::ptr_map<int,SurfaceSlidingConstrains>::iterator it = surface_constrains_map.begin();
      it!=surface_constrains_map.end();
      it++
    ) {
      ostringstream ss;
      ss << "C_SURFACE_ELEM_msId_" << it->first;
      it->second->feLhs.snes_B = projSurfaceCtx->C;
      ierr = m_field.loop_finite_elements("C_ALL_MATRIX",ss.str(),it->second->feLhs,MF_ZERO);  CHKERRQ(ierr);
    }
  }

  /*
  //Loops over body surface (CFE_SURFACE) and crack surface (CFE_CRACK_SURFACE)
  ConstrainSurfacGeometry CFE_SURFACE(m_field,projSurfaceCtx->C);
  ConstrainSurfacGeometry CFE_CRACK_SURFACE(m_field,projSurfaceCtx->C,"LAMBDA_CRACK_SURFACE");
  map<int,ConstrainSurfacGeometry*> CFE_SURFACE_msId_ptr;
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss;
    ss << "LAMBDA_SURFACE_msId_" << msId;
    CFE_SURFACE_msId_ptr[msId] = new ConstrainSurfacGeometry(m_field,projSurfaceCtx->C,ss.str());
  }

  ierr = m_field.loop_finite_elements("C_ALL_MATRIX","C_SURFACE_ELEM",CFE_SURFACE);  CHKERRQ(ierr);
  //ierr = m_field.loop_finite_elements("C_ALL_MATRIX","C_CRACK_SURFACE_ELEM",CFE_CRACK_SURFACE);  CHKERRQ(ierr);
  for(map<int,ConstrainSurfacGeometry*>::iterator mit = CFE_SURFACE_msId_ptr.begin();
    mit!=CFE_SURFACE_msId_ptr.end();mit++) {
    ostringstream ss0;
    ss0 << "C_SURFACE_ELEM_msId_" << mit->first;
    ierr = m_field.loop_finite_elements("C_ALL_MATRIX",ss0.str(),*(mit->second));  CHKERRQ(ierr);
  }
  */

  ierr = MatAssemblyBegin(projSurfaceCtx->C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(projSurfaceCtx->C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);

  /*for(map<int,ConstrainSurfacGeometry*>::iterator mit = CFE_SURFACE_msId_ptr.begin();
    mit!=CFE_SURFACE_msId_ptr.end();mit++) {
    delete mit->second;
  }*/

  /*{
    //Matrix View
    MatView(projSurfaceCtx->C,PETSC_VIEWER_DRAW_WORLD);//PETSC_VIEWER_STDOUT_WORLD);
    std::string wait;
    std::cin >> wait;
  }*/

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::delete_surface_projection_data(MoFEM::Interface& m_field) {
  PetscFunctionBegin;
  PetscErrorCode ierr;

  if(projSurfaceCtx!=NULL) {
    ierr = projSurfaceCtx->destroyQorP(); CHKERRQ(ierr);
    ierr = projSurfaceCtx->destroyQTKQ(); CHKERRQ(ierr);
    ierr = MatDestroy(&projSurfaceCtx->C); CHKERRQ(ierr);
    delete projSurfaceCtx;
    projSurfaceCtx = NULL;
  }

  PetscFunctionReturn(0);
}


PetscErrorCode ConfigurationalFractureMechanics::project_force_vector(MoFEM::Interface& m_field,string problem) {
  PetscFunctionBegin;

  PetscErrorCode ierr;
  //ErrorCode rval;

  Vec F_Material;
  ierr = m_field.VecCreateGhost(problem,ROW,&F_Material); CHKERRQ(ierr);
  ierr = m_field.set_other_global_ghost_vector(problem,"MESH_NODE_POSITIONS","MATERIAL_FORCE",ROW,F_Material,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  int M,m;
  ierr = VecGetSize(F_Material,&M); CHKERRQ(ierr);
  ierr = VecGetLocalSize(F_Material,&m); CHKERRQ(ierr);
  Mat Q;
  ierr = MatCreateShell(PETSC_COMM_WORLD,m,m,M,M,projSurfaceCtx,&Q); CHKERRQ(ierr);
  ierr = MatShellSetOperation(Q,MATOP_MULT,(void(*)(void))PorjectionMatrixMultOpQ); CHKERRQ(ierr);

  Vec QTF_Material;
  ierr = VecDuplicate(F_Material,&QTF_Material); CHKERRQ(ierr);
  ierr = MatMult(Q,F_Material,QTF_Material); CHKERRQ(ierr);

  PostProcVertexMethod ent_method_material(m_field.get_moab(),"MESH_NODE_POSITIONS",QTF_Material,"MATERIAL_FORCE_PROJECTED");
  ierr = m_field.loop_dofs(problem,"MESH_NODE_POSITIONS",ROW,ent_method_material); CHKERRQ(ierr);
  ierr = m_field.set_other_global_ghost_vector(problem,"MESH_NODE_POSITIONS","MATERIAL_FORCE",ROW,QTF_Material,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);

  double nrm2_material;
  ierr = VecNorm(QTF_Material,NORM_2,&nrm2_material);   CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"nrm2_QTF_Material = %6.4e\n",nrm2_material); CHKERRQ(ierr);

  ierr = VecDestroy(&F_Material); CHKERRQ(ierr);
  ierr = MatDestroy(&Q); CHKERRQ(ierr);
  ierr = VecDestroy(&QTF_Material); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


PetscErrorCode ConfigurationalFractureMechanics::project_form_th_projection_tag(MoFEM::Interface& m_field,string problem,bool do_not_project) {
  PetscFunctionBegin;

  PetscErrorCode ierr;
  ErrorCode rval;

  Vec D,dD;
  ierr = m_field.VecCreateGhost(problem,COL,&D); CHKERRQ(ierr);
  ierr = VecDuplicate(D,&dD); CHKERRQ(ierr);
  ierr = VecZeroEntries(dD); CHKERRQ(ierr);
  ierr = VecSetOption(dD,VEC_IGNORE_NEGATIVE_INDICES,PETSC_TRUE);  CHKERRQ(ierr);

  ierr = m_field.set_local_ghost_vector(problem,COL,D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = m_field.set_global_ghost_vector(problem,COL,D,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);

  Tag th_projection;
  rval = m_field.get_moab().tag_get_handle("PROJECTION_CRACK_SURFACE",th_projection); CHKERRQ_MOAB(rval);

  const MoFEMProblem *problemPtr;
  ierr = m_field.get_problem(problem,&problemPtr); CHKERRQ(ierr);

  ParallelComm* pcomm = ParallelComm::get_pcomm(&m_field.get_moab(),MYPCOMM_INDEX);

  Range crack_front_edges;
  ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRQ(ierr);
  Range crack_front_edges_nodes;
  rval = m_field.get_moab().get_connectivity(crack_front_edges,crack_front_edges_nodes,true); CHKERRQ_MOAB(rval);
  Range::iterator nit = crack_front_edges_nodes.begin();
  for(;nit!=crack_front_edges_nodes.end();nit++) {

    NumeredDofEntityByNameEntAndPart::iterator dit,hi_dit;
    dit = problemPtr->numered_dofs_cols->get<Composite_Name_Ent_And_Part_mi_tag>().lower_bound(
      boost::make_tuple("MESH_NODE_POSITIONS",*nit,pcomm->rank()));
    hi_dit = problemPtr->numered_dofs_cols->get<Composite_Name_Ent_And_Part_mi_tag>().upper_bound(
      boost::make_tuple("MESH_NODE_POSITIONS",*nit,pcomm->rank()));
    if(distance(dit,hi_dit)==0) continue;
    ublas::vector<PetscInt,ublas::bounded_array<PetscInt,3> > indices;
    indices.resize(3);
    fill(indices.begin(),indices.end(),-1);
    ublas::vector<double,ublas::bounded_array<double,3> > new_coords;
    new_coords.resize(3);
    rval = m_field.get_moab().tag_get_data(th_projection,&*nit,1,&*new_coords.data().begin()); CHKERRQ_MOAB(rval);
    ublas::vector<double,ublas::bounded_array<double,3> > delta;
    delta.resize(3);
    for(;dit!=hi_dit;dit++) {
      indices[dit->get()->getDofCoeffIdx()] = dit->get()->getPetscGlobalDofIdx();
      delta[dit->get()->getDofCoeffIdx()] = new_coords[dit->get()->getDofCoeffIdx()] - dit->get()->getFieldData();
    }
    ierr = VecSetValues(dD,3,&*indices.data().begin(),&*delta.data().begin(),INSERT_VALUES); CHKERRQ(ierr);

  }

  ierr = VecAssemblyBegin(dD); CHKERRQ(ierr);
  ierr = VecAssemblyEnd(dD); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(dD,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(dD,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  int M,m;
  ierr = VecGetSize(dD,&M); CHKERRQ(ierr);
  ierr = VecGetLocalSize(dD,&m); CHKERRQ(ierr);
  Mat Q;
  ierr = MatCreateShell(PETSC_COMM_WORLD,m,m,M,M,projSurfaceCtx,&Q); CHKERRQ(ierr);
  ierr = MatShellSetOperation(Q,MATOP_MULT,(void(*)(void))PorjectionMatrixMultOpQ); CHKERRQ(ierr);

  Vec QTdD;
  ierr = VecDuplicate(D,&QTdD); CHKERRQ(ierr);
  ierr = MatMult(Q,dD,QTdD); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(QTdD,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(QTdD,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  ierr = VecAXPY(D,1.,QTdD); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  //not project nodes on crack surface
  if(do_not_project) {
    ierr = m_field.set_local_ghost_vector(problem,COL,QTdD,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  }
  ierr = m_field.set_global_ghost_vector(problem,COL,D,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);

  PostProcVertexMethod ent_method(m_field.get_moab(),"MESH_NODE_POSITIONS",D,"PROJECTION_CRACK_SURFACE");
  ierr = m_field.loop_dofs(problem,"MESH_NODE_POSITIONS",COL,ent_method); CHKERRQ(ierr);

  if(do_not_project) {
    ierr = m_field.set_global_ghost_vector(problem,COL,QTdD,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  }

  VecDestroy(&D);
  VecDestroy(&dD);
  VecDestroy(&QTdD);
  MatDestroy(&Q);

  PetscFunctionReturn(0);
}


PetscErrorCode ConfigurationalFractureMechanics::front_projection_data(MoFEM::Interface& m_field,string problem) {
  PetscFunctionBegin;

  PetscErrorCode ierr;

  ierr = delete_front_projection_data(m_field); CHKERRQ(ierr);

  if(projFrontCtx==NULL) {
    projFrontCtx = new ConstrainMatrixCtx(m_field,problem,"C_CRACKFRONT_MATRIX");
    ierr = m_field.MatCreateMPIAIJWithArrays("C_CRACKFRONT_MATRIX",&projFrontCtx->C); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::delete_front_projection_data(MoFEM::Interface& m_field) {
  PetscFunctionBegin;

  PetscErrorCode ierr;

  if(projFrontCtx!=NULL) {
    ierr = projFrontCtx->destroyQorP(); CHKERRQ(ierr);
    ierr = projFrontCtx->destroyQTKQ(); CHKERRQ(ierr);
    ierr = MatDestroy(&projFrontCtx->C); CHKERRQ(ierr);
    delete projFrontCtx;
    projFrontCtx = NULL;
  }

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::calculate_griffith_foces(MoFEM::Interface& m_field,string problem) {
  PetscFunctionBegin;

  PetscErrorCode ierr;
  ErrorCode rval;

  double gc;
  PetscBool flg;
  ierr = PetscOptionsGetReal(PETSC_NULL,PETSC_NULL,"-my_gc",&gc,&flg); CHKERRQ(ierr);

  {
    Vec GriffithForceVec;
    ierr = m_field.VecCreateGhost(problem,COL,&GriffithForceVec); CHKERRQ(ierr);
    ierr = VecSetOption(GriffithForceVec,VEC_IGNORE_NEGATIVE_INDICES,PETSC_TRUE);  CHKERRQ(ierr);
    {
      GriffithForceElement griffith_force_element(m_field);

      griffith_force_element.blockData[0].gc = gc;
      ierr = m_field.get_cubit_msId_entities_by_dimension(
        201,SIDESET,1,griffith_force_element.blockData[0].frontEdges,true
      ); CHKERRQ(ierr);
      rval = m_field.get_moab().get_connectivity(
        griffith_force_element.blockData[0].frontEdges,griffith_force_element.blockData[0].frontNodes,true
      ); CHKERRQ_MOAB(rval);
      griffith_force_element.feRhs.getOpPtrVector().push_back(
        new GriffithForceElement::OpJacobian(
          3,griffith_force_element.blockData[0],griffith_force_element.commonData
        )
      );
      griffith_force_element.feRhs.getOpPtrVector().push_back(
        new GriffithForceElement::OpRhs(
          3,griffith_force_element.blockData[0],griffith_force_element.commonData
        )
      );
      griffith_force_element.feRhs.snes_f = GriffithForceVec;
      ierr = VecZeroEntries(GriffithForceVec); CHKERRQ(ierr);
      ierr = VecGhostUpdateBegin(GriffithForceVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = VecGhostUpdateEnd(GriffithForceVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = m_field.loop_finite_elements(
        problem,"GRIFFITH_FORCE_ELEMENT",griffith_force_element.feRhs
      );  CHKERRQ(ierr);
      ierr = VecAssemblyBegin(GriffithForceVec); CHKERRQ(ierr);
      ierr = VecAssemblyEnd(GriffithForceVec); CHKERRQ(ierr);
      ierr = VecGhostUpdateBegin(GriffithForceVec,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
      ierr = VecGhostUpdateEnd(GriffithForceVec,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
      ierr = VecGhostUpdateBegin(GriffithForceVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = VecGhostUpdateEnd(GriffithForceVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    }

    // projection matrix
    Mat Q;
    {
      int M,m;
      ierr = VecGetSize(GriffithForceVec,&M); CHKERRQ(ierr);
      ierr = VecGetLocalSize(GriffithForceVec,&m); CHKERRQ(ierr);
      ierr = MatCreateShell(PETSC_COMM_WORLD,m,m,M,M,projSurfaceCtx,&Q); CHKERRQ(ierr);
      ierr = MatShellSetOperation(Q,MATOP_MULT,(void(*)(void))PorjectionMatrixMultOpQ); CHKERRQ(ierr);
    }

    // project griffith force
    {
      Vec QTGriffithForceVec;
      ierr = VecDuplicate(GriffithForceVec,&QTGriffithForceVec); CHKERRQ(ierr);
      ierr = MatMult(Q,GriffithForceVec,QTGriffithForceVec); CHKERRQ(ierr);
      ierr = VecGhostUpdateBegin(QTGriffithForceVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = VecGhostUpdateEnd(QTGriffithForceVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = m_field.set_other_global_ghost_vector(
        problem,"MESH_NODE_POSITIONS","GRIFFITH_FORCE",ROW,QTGriffithForceVec,INSERT_VALUES,SCATTER_REVERSE
      ); CHKERRQ(ierr);
      PostProcVertexMethod ent_method_area(m_field.get_moab(),"MESH_NODE_POSITIONS",QTGriffithForceVec,"GRIFFITH_FORCE");
      ierr = m_field.loop_dofs(problem,"MESH_NODE_POSITIONS",COL,ent_method_area); CHKERRQ(ierr);
      double nrm2_griffith_force;
      ierr = VecNorm(QTGriffithForceVec,NORM_2,&nrm2_griffith_force);   CHKERRQ(ierr);
      ierr = PetscPrintf(PETSC_COMM_WORLD,"nrm2_QTGriffithForceVec = %6.4e\n",nrm2_griffith_force); CHKERRQ(ierr);
      ierr = VecDestroy(&QTGriffithForceVec); CHKERRQ(ierr);
    }

    ierr = MatDestroy(&Q); CHKERRQ(ierr);
    ierr = VecDestroy(&GriffithForceVec); CHKERRQ(ierr);

  }

  //tangent front froce
  {
    Vec GriffithForceVec;
    ierr = m_field.VecCreateGhost("C_CRACKFRONT_MATRIX",COL,&GriffithForceVec); CHKERRQ(ierr);
    Vec LambdaVec;
    ierr = m_field.VecCreateGhost("C_CRACKFRONT_MATRIX",ROW,&LambdaVec); CHKERRQ(ierr);
    if(flg == PETSC_TRUE) {
      ierr = VecSet(LambdaVec,gc); CHKERRQ(ierr);
    } else {
      ierr = VecSet(LambdaVec,1); CHKERRQ(ierr);
    }
    ConstrainMatrixCtx projFrontCtx_tangent(m_field,problem,"C_CRACKFRONT_MATRIX");
    ierr = m_field.MatCreateMPIAIJWithArrays("C_CRACKFRONT_MATRIX",&projFrontCtx_tangent.C); CHKERRQ(ierr);
    C_FRONT_TANGENT C_TANGENT_ELEM(m_field,projFrontCtx_tangent.C,PETSC_NULL,"LAMBDA_CRACKFRONT_AREA");
    ierr = MatZeroEntries(projFrontCtx_tangent.C); CHKERRQ(ierr);
    ierr = m_field.loop_finite_elements("C_CRACKFRONT_MATRIX","C_CRACKFRONT_AREA_ELEM",C_TANGENT_ELEM);  CHKERRQ(ierr);
    ierr = MatAssemblyBegin(projFrontCtx_tangent.C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
    ierr = MatAssemblyEnd(projFrontCtx_tangent.C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
    ierr = MatMultTranspose(projFrontCtx_tangent.C,LambdaVec,GriffithForceVec); CHKERRQ(ierr);
    ierr = VecGhostUpdateBegin(GriffithForceVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = VecGhostUpdateEnd(GriffithForceVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = m_field.set_other_global_ghost_vector(
      problem,"MESH_NODE_POSITIONS","GRIFFITH_FORCE_TANGENT",ROW,GriffithForceVec,INSERT_VALUES,SCATTER_REVERSE
    ); CHKERRQ(ierr);
    PostProcVertexMethod ent_method_tangent(
      m_field.get_moab(),"MESH_NODE_POSITIONS",GriffithForceVec,"GRIFFITH_TANGENT_FORCE");
      ierr = m_field.loop_dofs(problem,"MESH_NODE_POSITIONS",COL,ent_method_tangent
    ); CHKERRQ(ierr);
    ierr = MatDestroy(&projFrontCtx_tangent.C); CHKERRQ(ierr);
    ierr = VecDestroy(&LambdaVec); CHKERRQ(ierr);
    ierr = VecDestroy(&GriffithForceVec); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::calculate_griffith_g(MoFEM::Interface& m_field,string problem) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  PetscFunctionBegin;

  Vec F_Material;
  ierr = m_field.VecCreateGhost(problem,ROW,&F_Material); CHKERRQ(ierr);
  ierr = m_field.set_other_global_ghost_vector(problem,"MESH_NODE_POSITIONS","MATERIAL_FORCE",ROW,F_Material,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  Vec F_Griffith;
  ierr = m_field.VecCreateGhost(problem,ROW,&F_Griffith); CHKERRQ(ierr);
  ierr = m_field.set_other_global_ghost_vector(problem,"MESH_NODE_POSITIONS","GRIFFITH_FORCE",ROW,F_Griffith,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  Vec F_Griffith_Tangent;
  ierr = m_field.VecCreateGhost(problem,ROW,&F_Griffith_Tangent); CHKERRQ(ierr);
  ierr = m_field.set_other_global_ghost_vector(
    problem,"MESH_NODE_POSITIONS","GRIFFITH_FORCE_TANGENT",ROW,F_Griffith_Tangent,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  Vec LambdaVec,LambdaVec_Tangent;
  ierr = m_field.VecCreateGhost("C_CRACKFRONT_MATRIX",ROW,&LambdaVec); CHKERRQ(ierr);
  ierr = VecDuplicate(LambdaVec,&LambdaVec_Tangent); CHKERRQ(ierr);

  Mat Q;
  {
    int M,m;
    ierr = VecGetSize(F_Material,&M); CHKERRQ(ierr);
    ierr = VecGetLocalSize(F_Material,&m); CHKERRQ(ierr);
    ierr = MatCreateShell(PETSC_COMM_WORLD,m,m,M,M,projSurfaceCtx,&Q); CHKERRQ(ierr);
    ierr = MatShellSetOperation(Q,MATOP_MULT,(void(*)(void))PorjectionMatrixMultOpQ); CHKERRQ(ierr);
  }

  // ErrorCode rval;
  // Range corners_edges,corners_nodes;
  // ierr = m_field.get_cubit_msId_entities_by_dimension(100,SIDESET,1,corners_edges,true); CHKERRQ(ierr);
  // ierr = m_field.get_cubit_msId_entities_by_dimension(101,NODESET,0,corners_nodes,true); CHKERRQ(ierr);

  MeshsetsManager *meshset_manager_ptr;
  ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Range corners_edges,corners_nodes;
  if(meshset_manager_ptr->checkMeshset(100,SIDESET)) {
    ierr = meshset_manager_ptr->getEntitiesByDimension(100,SIDESET,1,corners_edges,true); CHKERRQ(ierr);
  }
  if(meshset_manager_ptr->checkMeshset(101,NODESET)) {
    ierr = meshset_manager_ptr->getEntitiesByDimension(101,NODESET,0,corners_nodes,true); CHKERRQ(ierr);
  }


  Range corners_edges_nodes;
  rval = m_field.get_moab().get_connectivity(corners_edges,corners_edges_nodes,true); CHKERRQ_MOAB(rval);
  corners_nodes.insert(corners_edges_nodes.begin(),corners_edges_nodes.end());
  Range blocked_nodes;
  BitRefLevel bit_to_block = BitRefLevel().set(BITREFLEVEL_SIZE-2);
  ierr = m_field.get_entities_by_type_and_ref_level(bit_to_block,BitRefLevel().set(),MBVERTEX,blocked_nodes); CHKERRQ(ierr);
  corners_nodes.merge(blocked_nodes);

  C_CONSTANT_AREA C_AREA_ELEM(m_field,projFrontCtx->C,Q,"LAMBDA_CRACKFRONT_AREA");

  ierr = MatSetOption(projFrontCtx->C,MAT_NEW_NONZERO_LOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);
  ierr = MatSetOption(projFrontCtx->C,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);

  ierr = MatZeroEntries(projFrontCtx->C); CHKERRQ(ierr);
  ierr = m_field.loop_finite_elements("C_CRACKFRONT_MATRIX","C_CRACKFRONT_AREA_ELEM",C_AREA_ELEM);  CHKERRQ(ierr);
  ierr = MatAssemblyBegin(projFrontCtx->C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(projFrontCtx->C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);

  /*{
    //Matrix View
    MatView(projFrontCtx->C,PETSC_VIEWER_DRAW_WORLD);//PETSC_VIEWER_STDOUT_WORLD);
    std::string wait;
    std::cin >> wait;
  }*/

  // R = CT*(CC^T)^(-1) [ unit m/m^(-2) = 1/m ] [ 0.5/0.25 = 2 ]
  // R^T = (CC^T)^(-T)C [ unit m/m^(-2) = 1/m ]
  Mat RT;
  {
    int N,n;
    int M,m;
    ierr = VecGetSize(F_Material,&N); CHKERRQ(ierr);
    ierr = VecGetLocalSize(F_Material,&n); CHKERRQ(ierr);
    ierr = VecGetSize(LambdaVec,&M); CHKERRQ(ierr);
    ierr = VecGetLocalSize(LambdaVec,&m); CHKERRQ(ierr);
    ierr = MatCreateShell(PETSC_COMM_WORLD,m,n,M,N,projFrontCtx,&RT); CHKERRQ(ierr);
    ierr = MatShellSetOperation(RT,MATOP_MULT,(void(*)(void))ConstrainMatrixMultOpRT); CHKERRQ(ierr);
  }

  //calculate tangent griffith force
  ConstrainMatrixCtx projFrontCtx_tangent(m_field,problem,"C_CRACKFRONT_MATRIX");
  ierr = m_field.MatCreateMPIAIJWithArrays("C_CRACKFRONT_MATRIX",&projFrontCtx_tangent.C); CHKERRQ(ierr);
  C_FRONT_TANGENT C_TANGENT_ELEM(m_field,projFrontCtx_tangent.C,Q,"LAMBDA_CRACKFRONT_AREA");
  ierr = MatZeroEntries(projFrontCtx_tangent.C); CHKERRQ(ierr);
  ierr = m_field.loop_finite_elements("C_CRACKFRONT_MATRIX","C_CRACKFRONT_AREA_ELEM",C_TANGENT_ELEM);  CHKERRQ(ierr);
  ierr = MatAssemblyBegin(projFrontCtx_tangent.C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(projFrontCtx_tangent.C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);

  Mat RT_Tangent;
  {
    int N,n;
    int M,m;
    ierr = VecGetSize(F_Material,&N); CHKERRQ(ierr);
    ierr = VecGetLocalSize(F_Material,&n); CHKERRQ(ierr);
    ierr = VecGetSize(LambdaVec_Tangent,&M); CHKERRQ(ierr);
    ierr = VecGetLocalSize(LambdaVec_Tangent,&m); CHKERRQ(ierr);
    ierr = MatCreateShell(PETSC_COMM_WORLD,m,n,M,N,&projFrontCtx_tangent,&RT_Tangent); CHKERRQ(ierr);
    ierr = MatShellSetOperation(RT_Tangent,MATOP_MULT,(void(*)(void))ConstrainMatrixMultOpRT); CHKERRQ(ierr);
  }

  //clualte griffith forces
  double gc = 1;
  PetscBool flg_gc;
  ierr = PetscOptionsGetReal(PETSC_NULL,PETSC_NULL,"-my_gc",&gc,&flg_gc); CHKERRQ(ierr);

  ierr = projFrontCtx->initializeQorP(F_Material); CHKERRQ(ierr);
  ierr = projFrontCtx_tangent.initializeQorP(F_Material); CHKERRQ(ierr);

  // unit of LambdaVec [ N * 1/m = N*m/m^2 = J/m^2 ]
  ierr = VecScale(F_Material,-1./gc); CHKERRQ(ierr);
  ierr = MatMult(RT,F_Material,LambdaVec); CHKERRQ(ierr);
  ierr = MatMult(RT_Tangent,F_Material,LambdaVec_Tangent); CHKERRQ(ierr);
  ierr = VecScale(F_Material,gc); CHKERRQ(ierr);
  ierr = VecScale(LambdaVec,gc); CHKERRQ(ierr);
  ierr = VecScale(LambdaVec_Tangent,gc); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(LambdaVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(LambdaVec,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(LambdaVec_Tangent,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(LambdaVec_Tangent,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  //get pointer to problem
  const MoFEMProblem *problemPtr;
  ierr = m_field.get_problem("C_CRACKFRONT_MATRIX",&problemPtr); CHKERRQ(ierr);
  //make maps to save griffith forces
  ierr = m_field.set_global_ghost_vector("C_CRACKFRONT_MATRIX",ROW,LambdaVec,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  map_ent_g.clear();
  for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_NAME_FOR_LOOP_(problemPtr,"LAMBDA_CRACKFRONT_AREA",diit)) {
    map_ent_g[diit->get()->getEnt()] = diit->get()->getFieldData();
  }
  ierr = m_field.set_global_ghost_vector("C_CRACKFRONT_MATRIX",ROW,LambdaVec_Tangent,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  map<EntityHandle,double> g_tangent_map;
  for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_NAME_FOR_LOOP_(problemPtr,"LAMBDA_CRACKFRONT_AREA",diit)) {
    g_tangent_map[diit->get()->getEnt()] = diit->get()->getFieldData();
  }

  //vector of matrerial force magnitudes
  Vec JVec;
  ierr = VecDuplicate(LambdaVec,&JVec); CHKERRQ(ierr);
  ParallelComm* pcomm = ParallelComm::get_pcomm(&m_field.get_moab(),MYPCOMM_INDEX);

  ublas::vector<double> coords;
  coords.resize(3);
  PetscPrintf(PETSC_COMM_WORLD,"\n\ngriffith force\n\n");

  Tag th_g,th_g_tangent;
  const double def_val = 0;
  rval = m_field.get_moab().tag_get_handle(
    "G1",1,MB_TYPE_DOUBLE,th_g,MB_TAG_CREAT|MB_TAG_SPARSE,&def_val); MOAB_THROW(rval);
  rval = m_field.get_moab().tag_get_handle(
    "G3",1,MB_TYPE_DOUBLE,th_g_tangent,MB_TAG_CREAT|MB_TAG_SPARSE,&def_val); MOAB_THROW(rval);

  map_ent_j.clear();
  for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_NAME_FOR_LOOP_(problemPtr,"LAMBDA_CRACKFRONT_AREA",diit)) {
    EntityHandle ent = diit->get()->getEnt();
    double g_val = map_ent_g[ent];
    double g_tangent_val = fabs(g_tangent_map[ent]);
    rval = m_field.get_moab().tag_set_data(th_g,&ent,1,&g_val); CHKERRQ_MOAB(rval);
    rval = m_field.get_moab().tag_set_data(th_g_tangent,&ent,1,&g_tangent_val); CHKERRQ_MOAB(rval);
    rval = m_field.get_moab().get_coords(&ent,1,&*coords.data().begin()); CHKERRQ_MOAB(rval);
    int dd = 0;
    ublas::vector<double,ublas::bounded_array<double,9> > material_force(3);
    for(_IT_GET_DOFS_FIELD_BY_NAME_AND_ENT_FOR_LOOP_(m_field,"MATERIAL_FORCE",diit->get()->getEnt(),diiit)) {
      material_force[diiit->get()->getDofCoeffIdx()] = diiit->get()->getFieldData();
      dd++;
    }
    if(dd != 3) SETERRQ1(PETSC_COMM_SELF,1,"can not find material force vector at node %ld",diit->get()->getEnt());
    ublas::vector<double,ublas::bounded_array<double,9> > griffith_force(3);
    for(_IT_GET_DOFS_FIELD_BY_NAME_AND_ENT_FOR_LOOP_(m_field,"GRIFFITH_FORCE",diit->get()->getEnt(),diiit)) {
      griffith_force[diiit->get()->getDofCoeffIdx()] = diiit->get()->getFieldData();
      dd++;
    }
    if(dd != 6) SETERRQ1(PETSC_COMM_SELF,1,"can not find griffith force vector at node %ld",diit->get()->getEnt());
    double j = norm_2(material_force)/(norm_2(griffith_force)/gc);

    j = copysign(j,g_val);
    map_ent_j[ent] = j;
    if(diit->get()->getPart()==pcomm->rank()) {
      ierr = VecSetValue(JVec,diit->get()->getPetscGlobalDofIdx(),j,INSERT_VALUES); CHKERRQ(ierr);
    }
    ostringstream ss;
    ss << "griffith force at ent ";
    ss << setw(5) << ent;
    //coords
    ss << "\tcoords";
    ss << " " << setw(10) << setprecision(4) << coords[0];
    ss << " " << setw(10) << setprecision(4) << coords[1];
    ss << " " << setw(10) << setprecision(4) << coords[2];
    //g
    ss << "\t\tg1 " << scientific << setprecision(4) << g_val;
    ss << " / " << scientific << setprecision(4) << j;
    ss << " ( " << scientific << setprecision(4) << g_val/j << " )";
    //g tangent
    ss << "\t\tg3 " << scientific << setprecision(4) << g_tangent_val;
    ss << " ( " << scientific << setprecision(4) << g_tangent_val/j << " )";
    if(flg_gc == PETSC_TRUE) {
      ss << "\t relative error (gc-g)/gc ";
      ss << scientific << setprecision(4) << (gc-g_val)/gc;
      ss << " / " << scientific << setprecision(4) << (gc-j)/gc;
    }
    ss << endl;
    PetscPrintf(PETSC_COMM_WORLD,"%s",ss.str().c_str());
  }

  ierr = VecSum(LambdaVec,&ave_g); CHKERRQ(ierr);
  ierr = VecMin(LambdaVec,PETSC_NULL,&min_g); CHKERRQ(ierr);
  ierr = VecMax(LambdaVec,PETSC_NULL,&max_g); CHKERRQ(ierr);
  ierr = VecAssemblyBegin(JVec); CHKERRQ(ierr);
  ierr = VecAssemblyEnd(JVec); CHKERRQ(ierr);
  ierr = VecSum(JVec,&ave_j); CHKERRQ(ierr);
  ierr = VecMin(JVec,PETSC_NULL,&min_j); CHKERRQ(ierr);
  ierr = VecMax(JVec,PETSC_NULL,&max_j); CHKERRQ(ierr);
  ierr = VecDestroy(&JVec); CHKERRQ(ierr);

  Range crackSurfacesFaces;
  ierr = m_field.get_cubit_msId_entities_by_dimension(200,SIDESET,2,crackSurfacesFaces,true); CHKERRQ(ierr);
  Range level_tris;
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  crackSurfacesFaces = intersect(crackSurfacesFaces,level_tris);

  double aRea = 0;
  ublas::vector<double,ublas::bounded_array<double,6> > diffNTRI(6);
  ShapeDiffMBTRI(&diffNTRI.data()[0]);
  for(Range::iterator fit = crackSurfacesFaces.begin();fit!=crackSurfacesFaces.end();fit++) {
    const EntityHandle* conn;
    int num_nodes;
    rval = m_field.get_moab().get_connectivity(*fit,conn,num_nodes,true); CHKERRQ_MOAB(rval);
    ublas::vector<double,ublas::bounded_array<double,9> > dofsX(9);
    ublas::vector<double,ublas::bounded_array<double,3> > normal(3);
    for(int nn = 0;nn<num_nodes; nn++) {
      for(_IT_GET_DOFS_FIELD_BY_NAME_AND_ENT_FOR_LOOP_(m_field,"MESH_NODE_POSITIONS",conn[nn],dit)) {
        dofsX[nn*3+dit->get()->getDofCoeffIdx()] = dit->get()->getFieldData();
      }
    }
    ierr = ShapeFaceNormalMBTRI(&diffNTRI[0],&dofsX.data()[0],&normal.data()[0]); CHKERRQ(ierr);
    //crack surface area is a half of crack top and bottom body surface
    aRea += norm_2(normal)*0.25;
  }

  {
    int N;
    ierr = VecGetSize(LambdaVec,&N); CHKERRQ(ierr);
    ave_g /= N;
    ave_j /= N;
  }
  PetscPrintf(PETSC_COMM_WORLD,"\naverage griffith force %6.4e / %6.4e Crack surface area %6.4e\n",ave_g,ave_j,aRea);
  PetscPrintf(PETSC_COMM_WORLD,"\n\n");

  PostProcVertexMethod ent_method(m_field.get_moab(),"LAMBDA_CRACKFRONT_AREA");
  ierr = m_field.loop_dofs("C_CRACKFRONT_MATRIX","LAMBDA_CRACKFRONT_AREA",ROW,ent_method); CHKERRQ(ierr);

  ierr = MatDestroy(&Q); CHKERRQ(ierr);
  ierr = MatDestroy(&RT); CHKERRQ(ierr);
  ierr = MatDestroy(&RT_Tangent); CHKERRQ(ierr);
  ierr = VecDestroy(&F_Griffith); CHKERRQ(ierr);
  ierr = VecDestroy(&F_Griffith_Tangent); CHKERRQ(ierr);
  ierr = VecDestroy(&F_Material); CHKERRQ(ierr);
  ierr = VecDestroy(&LambdaVec); CHKERRQ(ierr);
  ierr = VecDestroy(&LambdaVec_Tangent); CHKERRQ(ierr);

  ierr = projFrontCtx_tangent.destroyQorP(); CHKERRQ(ierr);
  ierr = projFrontCtx_tangent.destroyQTKQ(); CHKERRQ(ierr);
  ierr = MatDestroy(&projFrontCtx_tangent.C); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode MySnesConvernceTest(SNES snes,int it,double xnorm,double gnorm,double fnorm,SNESConvergedReason *reason,void *void_ctx) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  ierr = SNESConvergedDefault(snes,it,xnorm,gnorm,fnorm,reason,PETSC_NULL); CHKERRQ(ierr);
  const PetscReal div = 10e3;
  if(fnorm > div) {
    *reason = SNES_DIVERGED_FUNCTION_DOMAIN;
  }
  PetscFunctionReturn(0);
}

PetscErrorCode MySnesConvernceTest_SNESLINESEARCHBT(SNES snes,int it,double xnorm,double gnorm,double fnorm,SNESConvergedReason *reason,void *void_ctx) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  ierr = SNESConvergedDefault(snes,it,xnorm,gnorm,fnorm,reason,PETSC_NULL); CHKERRQ(ierr);
  const PetscReal div = 10e3;
  if(fnorm > div) {
    *reason = SNES_DIVERGED_FUNCTION_DOMAIN;
  }
  if(it>0) {
    SNESLineSearch linesearch;
    ierr = SNESGetLineSearch(snes,&linesearch); CHKERRQ(ierr);
    ierr = SNESLineSearchSetType(linesearch,SNESLINESEARCHBT); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
PetscErrorCode MySnesConvernceTest_SNESLINESEARCHL2(SNES snes,int it,double xnorm,double gnorm,double fnorm,SNESConvergedReason *reason,void *void_ctx) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  ierr = SNESConvergedDefault(snes,it,xnorm,gnorm,fnorm,reason,PETSC_NULL); CHKERRQ(ierr);
  const PetscReal div = 10e3;
  if(fnorm > div) {
    *reason = SNES_DIVERGED_FUNCTION_DOMAIN;
  }
  if(it>0) {
    SNESLineSearch linesearch;
    ierr = SNESGetLineSearch(snes,&linesearch); CHKERRQ(ierr);
    ierr = SNESLineSearchSetType(linesearch,SNESLINESEARCHL2); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::fix_all_but_one(MoFEM::Interface& m_field,double da,Range &fix_nodes,const double fraction_treshold) {
  PetscFunctionBegin;

  PetscErrorCode ierr;
  double gc;
  PetscBool flg;
  ierr = PetscOptionsGetReal(PETSC_NULL,PETSC_NULL,"-my_gc",&gc,&flg); CHKERRQ(ierr);
  if(flg != PETSC_TRUE) {
      SETERRQ(PETSC_COMM_SELF,1,"*** ERROR -my_gc (what is the fracture energy ?)");
  }

  ErrorCode rval;
  Tag th_freez;
  const int def_order = 0;
  rval = m_field.get_moab().tag_get_handle(
    "FROZEN_NODE",1,MB_TYPE_INTEGER,th_freez,MB_TAG_CREAT|MB_TAG_SPARSE,&def_order
  ); CHKERRQ_MOAB(rval);

  ierr = PetscPrintf(PETSC_COMM_WORLD,"freeze front nodes:\n");
  map<EntityHandle,double>::iterator max_mit = map_ent_j.begin();
  //cout << "A " << max_mit->second << endl;
  for(
    map<EntityHandle,double>::iterator mit = max_mit;
    mit!=map_ent_j.end();mit++
  ) {
    if(max_mit->second!=max_mit->second) continue; // its not a number
    if(max_mit->second<mit->second) {
      max_mit = mit;
      //cout << "A " << max_mit->second << endl;
    }
  }
  double max_g_j = 0;
  EntityHandle max_g_j_ent = 0;
  for(
    map<EntityHandle,double>::iterator mit = map_ent_j.begin();
    mit!=map_ent_j.end();mit++
  ) {
    double fraction = (max_j-mit->second)/max_j;
    //double fraction_gc = (gc-map_ent_g[mit->first])/gc;
    double step_work_of_fracture = 0;
    if(!map_ent_work.empty()) {
      step_work_of_fracture = map_ent_work[mit->first];
    }
    double g_j = map_ent_g[mit->first]/mit->second;
    ierr = PetscPrintf(PETSC_COMM_WORLD,
      "front node = %ld max_j = %6.4e j = %6.4e (%6.4e) g/j = %4.3f step work of fracture = %2.1g",
      mit->first,max_j,mit->second,fraction,g_j,step_work_of_fracture
    ); CHKERRQ(ierr);
    //cout << endl;
    //cout << fraction <<  " " << (fraction > fraction_treshold) << " " << (mit!=max_mit) << " " << max_j << " " << max_mit->second <<  endl;
    //cout << endl;
    bool freez_or_not_to_freez;
    if( (fraction > fraction_treshold) && (mit!=max_mit) ) {
      freez_or_not_to_freez = true;
    } else {
      freez_or_not_to_freez = false;
      if(max_g_j_ent==0) {
        max_g_j = g_j;
        max_g_j_ent = mit->first;
      } else {
        if(g_j < max_g_j) {
          max_g_j = g_j;
          max_g_j_ent = mit->first;
        }
      }
    }
    if(freeze_all_but_one) {
      freez_or_not_to_freez = freeze_all_but_one;
    }
    if(freez_or_not_to_freez) {
      ierr = PetscPrintf(PETSC_COMM_WORLD," freeze\n");
      fix_nodes.insert(mit->first);
      int freez = 1;
      EntityHandle node = mit->first;
      rval = m_field.get_moab().tag_set_data(th_freez,&node,1,&freez); CHKERRQ_MOAB(rval);
    } else {
      int freez;
      EntityHandle node = mit->first;
      rval = m_field.get_moab().tag_get_data(th_freez,&node,1,&freez); CHKERRQ_MOAB(rval);
      if(freez == 0) {
        ierr = PetscPrintf(PETSC_COMM_WORLD,"\n");
      } else if(freez == 1) {
        if((gc<map_ent_g[mit->first])||(mit==max_mit)) {
          ierr = PetscPrintf(PETSC_COMM_WORLD," unfreeze\n");
          int freez = 0;
          rval = m_field.get_moab().tag_set_data(th_freez,&node,1,&freez); CHKERRQ_MOAB(rval);
        } else {
          ierr = PetscPrintf(PETSC_COMM_WORLD," freeze < gc\n");
          fix_nodes.insert(mit->first);
        }
      }
    }
  }
  if(freeze_all_but_one) {
    if(max_g_j_ent == 0) {
      SETERRQ(PETSC_COMM_SELF,1,"Impossible case, at least one node should not break");
    }
    ierr = PetscPrintf(PETSC_COMM_WORLD,"\nunfreez %ld g/j = %4.3f\n",max_g_j_ent,max_g_j);
    int freez = 0;
    rval = m_field.get_moab().tag_set_data(th_freez,&max_g_j_ent,1,&freez); CHKERRQ_MOAB(rval);
    fix_nodes.erase(max_g_j_ent);
  } else {
    Range crack_front_edges,crack_front_nodes;
    ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRQ(ierr);
    rval = m_field.get_moab().get_connectivity(crack_front_edges,crack_front_nodes,true); CHKERRQ_MOAB(rval);
    Range nodes_in_set;
    unsigned int nb_nodes;
    nodes_in_set.insert(max_mit->first);
    do  {
      nb_nodes = nodes_in_set.size();
      Range edges;
      ierr = m_field.get_moab().get_adjacencies(nodes_in_set,1,false,edges,moab::Interface::UNION); CHKERRQ(ierr);
      edges = intersect(edges,crack_front_edges);
      Range nodes;
      rval = m_field.get_moab().get_connectivity(edges,nodes,true); CHKERRQ_MOAB(rval);
      nodes_in_set.merge(subtract(nodes,fix_nodes));
    } while (nb_nodes!=nodes_in_set.size());
    fix_nodes.merge(subtract(crack_front_nodes,nodes_in_set));
    vector<int> fix_nodes_vec(fix_nodes.size(),1);
    if(!fix_nodes_vec.empty()) {
      rval = m_field.get_moab().tag_set_data(th_freez,fix_nodes,&fix_nodes_vec[0]); CHKERRQ_MOAB(rval);
    }
    ierr = PetscPrintf(
      PETSC_COMM_WORLD,
      "final number of fixed nodes %u out of %u\n",
      fix_nodes.size(),crack_front_nodes.size()
    );
  }

  ierr = PetscPrintf(PETSC_COMM_WORLD,"\n");
  ierr = PetscPrintf(PETSC_COMM_WORLD,"\n");

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::solve_coupled_problem(
  MoFEM::Interface& m_field,
  SNES *snes,
  double da,
  const double penalty,
  const double fraction_treshold,
  SNESConvergedReason &reason
) {
  PetscFunctionBegin;

  PetscErrorCode ierr;
  ErrorCode rval;
  PetscBool flg;

  //set_PhysicalEquationNumber(hooke);

  ierr = front_projection_data(m_field,"COUPLED_PROBLEM"); CHKERRQ(ierr);

  //create matrices
  Mat K;
  ierr = m_field.MatCreateMPIAIJWithArrays("COUPLED_PROBLEM",&K); CHKERRQ(ierr);
  //create vectors
  Vec F;
  ierr = m_field.VecCreateGhost("COUPLED_PROBLEM",ROW,&F); CHKERRQ(ierr);
  Vec D;
  ierr = m_field.VecCreateGhost("COUPLED_PROBLEM",COL,&D); CHKERRQ(ierr);

  ierr = VecSetOption(F,VEC_IGNORE_NEGATIVE_INDICES,PETSC_TRUE);  CHKERRQ(ierr);

  if(material_FirelWall->operator[](FW_arc_lenhghat_definition)) {
  } else {
    SETERRQ(PETSC_COMM_SELF,1,"arc length not initialised)");
  }

  Range corners_edges,corners_nodes;
  ierr = m_field.get_cubit_msId_entities_by_dimension(100,SIDESET,1,corners_edges,true); CHKERRQ(ierr);
  ierr = m_field.get_cubit_msId_entities_by_dimension(101,NODESET,0,corners_nodes,true); CHKERRQ(ierr);
  Range corners_edgesNodes;
  rval = m_field.get_moab().get_connectivity(corners_edges,corners_edgesNodes,true); CHKERRQ_MOAB(rval);
  corners_nodes.insert(corners_edgesNodes.begin(),corners_edgesNodes.end());
  Range nodes_to_block;
  BitRefLevel bit_to_block = BitRefLevel().set(BITREFLEVEL_SIZE-2);
  ierr = m_field.get_entities_by_type_and_ref_level(bit_to_block,BitRefLevel().set(),MBVERTEX,nodes_to_block); CHKERRQ(ierr);
  corners_nodes.merge(nodes_to_block);

  Range crack_front_edges,crack_front_nodes;
  ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRQ(ierr);
  rval = m_field.get_moab().get_connectivity(crack_front_edges,crack_front_nodes,true); CHKERRQ_MOAB(rval);
  Range level_nodes;
  ierr = m_field.get_entities_by_type_and_ref_level(*ptr_bit_level0,BitRefLevel().set(),MBVERTEX,level_nodes); CHKERRQ(ierr);
  crack_front_nodes = intersect(crack_front_nodes,level_nodes);
  //corners_nodes.merge(subtract(level_nodes,crack_front_nodes));

  Range fix_nodes;
  ierr = fix_all_but_one(m_field,da,fix_nodes,fraction_treshold); CHKERRQ(ierr);
  //da *= fabs(crack_front_nodes.size()-fix_nodes.size())/(double)crack_front_nodes.size();
  da *= fabs(crack_front_nodes.size()-fix_nodes.size());///(double)crack_front_nodes.size();
  corners_nodes.merge(fix_nodes);

  // Arc elem
  ArcLengthCtx arc_ctx(m_field,"COUPLED_PROBLEM");
  ArcLengthSnesCtx arc_snes_ctx(m_field,"COUPLED_PROBLEM",&arc_ctx);
  ArcLengthMatShell arc_mat_ctx(K,&arc_ctx,"COUPLED_PROBLEM");
  PCArcLengthCtx pc_ctx(K,K,&arc_ctx);
  FrontAreaArcLengthControl arc_elem(m_field,this,&arc_ctx);

  // Spatial and material forces
  NonlinearElasticElement spatial_fe(m_field,-1);
  NonlinearElasticElement material_fe(m_field,-1);

  boost::shared_ptr<Hooke<adouble> > hooke_adouble(new Hooke<adouble>);
  boost::shared_ptr<Hooke<double> > hooke_double(new Hooke<double>);

  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,BLOCKSET|MAT_ELASTICSET,it)) {
    Mat_Elastic mydata;
    ierr = it->getAttributeDataStructure(mydata); CHKERRQ(ierr);
    int id = it->getMeshsetId();
    EntityHandle meshset = it->getMeshset();
    rval = m_field.get_moab().get_entities_by_type(
      meshset,MBTET,spatial_fe.setOfBlocks[id].tEts,true
    ); CHKERRQ_MOAB(rval);

    spatial_fe.setOfBlocks[id].iD = id;
    spatial_fe.setOfBlocks[id].E = mydata.data.Young;
    spatial_fe.setOfBlocks[id].PoissonRatio = mydata.data.Poisson;
    spatial_fe.setOfBlocks[id].materialDoublePtr = hooke_double;
    spatial_fe.setOfBlocks[id].materialAdoublePtr = hooke_adouble;

    material_fe.setOfBlocks[id].tEts = spatial_fe.setOfBlocks[id].tEts;
    material_fe.setOfBlocks[id].iD = id;
    material_fe.setOfBlocks[id].E = mydata.data.Young;
    material_fe.setOfBlocks[id].PoissonRatio = mydata.data.Poisson;
    material_fe.setOfBlocks[id].materialDoublePtr = hooke_double;
    material_fe.setOfBlocks[id].materialAdoublePtr = hooke_adouble;
    material_fe.setOfBlocks[id].forcesOnlyOnEntitiesRow.merge(crack_front_edges);
    material_fe.setOfBlocks[id].forcesOnlyOnEntitiesRow.merge(crack_front_nodes);
  }

  spatial_fe.commonData.spatialPositions = "SPATIAL_POSITION";
  spatial_fe.commonData.meshPositions = "MESH_NODE_POSITIONS";
  material_fe.commonData.spatialPositions = "SPATIAL_POSITION";
  material_fe.commonData.meshPositions = "MESH_NODE_POSITIONS";

  /// Rhs spatial
  spatial_fe.feRhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("SPATIAL_POSITION",spatial_fe.commonData)
  );
  if(m_field.check_field("MESH_NODE_POSITIONS")) {
    spatial_fe.feRhs.getOpPtrVector().push_back(new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",spatial_fe.commonData));
  }
  map<int,NonlinearElasticElement::BlockData>::iterator sit = spatial_fe.setOfBlocks.begin();
  for(;sit!=spatial_fe.setOfBlocks.end();sit++) {
    spatial_fe.feRhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpJacobianPiolaKirchhoffStress("SPATIAL_POSITION",sit->second,spatial_fe.commonData,1,false,true,false)
    );
    spatial_fe.feRhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpRhsPiolaKirchhoff("SPATIAL_POSITION",sit->second,spatial_fe.commonData)
    );
  }
  spatial_fe.feRhs.meshPositionsFieldName = "NONE";
  spatial_fe.feRhs.addToRule = 0;

  // Rhs material
  material_fe.feRhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("SPATIAL_POSITION",material_fe.commonData)
  );
  material_fe.feRhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",material_fe.commonData)
  );
  sit = material_fe.setOfBlocks.begin();
  for(;sit!=material_fe.setOfBlocks.end();sit++) {
    material_fe.feRhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpJacobianEshelbyStress("SPATIAL_POSITION",sit->second,material_fe.commonData,2,false,true)
    );
    material_fe.feRhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpRhsEshelbyStrees("MESH_NODE_POSITIONS",sit->second,material_fe.commonData)
    );
  }
  material_fe.feRhs.meshPositionsFieldName = "NONE";
  material_fe.feRhs.addToRule = 0;

  /// Lhs spatial
  spatial_fe.feLhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("SPATIAL_POSITION",spatial_fe.commonData)
  );
  spatial_fe.feLhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",spatial_fe.commonData)
  );
  sit = spatial_fe.setOfBlocks.begin();
  for(;sit!=spatial_fe.setOfBlocks.end();sit++) {
    spatial_fe.feLhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpJacobianPiolaKirchhoffStress("SPATIAL_POSITION",sit->second,spatial_fe.commonData,1,true,true,false)
    );
    spatial_fe.feLhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpLhsPiolaKirchhoff_dx("SPATIAL_POSITION","SPATIAL_POSITION",sit->second,spatial_fe.commonData)
    );
    spatial_fe.feLhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpLhsPiolaKirchhoff_dX("SPATIAL_POSITION","MESH_NODE_POSITIONS",sit->second,spatial_fe.commonData)
    );
  }
  spatial_fe.feLhs.meshPositionsFieldName = "NONE";
  spatial_fe.feLhs.addToRule = 0;

  /// Lhs material
  material_fe.feLhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("SPATIAL_POSITION",material_fe.commonData)
  );
  material_fe.feLhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",material_fe.commonData)
  );
  sit = material_fe.setOfBlocks.begin();
  for(;sit!=material_fe.setOfBlocks.end();sit++) {
    material_fe.feLhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpJacobianEshelbyStress("SPATIAL_POSITION",sit->second,material_fe.commonData,2,true,true)
    );
    material_fe.feLhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpLhsEshelby_dX("MESH_NODE_POSITIONS","MESH_NODE_POSITIONS",sit->second,material_fe.commonData)
    );
    material_fe.feLhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpLhsEshelby_dx("MESH_NODE_POSITIONS","SPATIAL_POSITION",sit->second,material_fe.commonData)
    );
  }
  material_fe.feLhs.meshPositionsFieldName = "NONE";
  material_fe.feLhs.addToRule = 0;

  // Griffith force element
  GriffithForceElement griffith_force_element(m_field);

  double gc;
  ierr = PetscOptionsGetReal(PETSC_NULL,PETSC_NULL,"-my_gc",&gc,&flg); CHKERRQ(ierr);
  griffith_force_element.blockData[0].penalty = penalty;

  ierr = m_field.get_cubit_msId_entities_by_dimension(
    201,SIDESET,1,griffith_force_element.blockData[0].frontEdges,true
  ); CHKERRQ(ierr);

  griffith_force_element.blockData[0].gc = gc;
  ierr = m_field.get_cubit_msId_entities_by_dimension(
    201,SIDESET,1,griffith_force_element.blockData[0].frontEdges,true
  ); CHKERRQ(ierr);
  rval = m_field.get_moab().get_connectivity(
    griffith_force_element.blockData[0].frontEdges,griffith_force_element.blockData[0].frontNodes,true
  ); CHKERRQ_MOAB(rval);

  griffith_force_element.feRhs.getOpPtrVector().push_back(
    new GriffithForceElement::OpJacobian(
      3,griffith_force_element.blockData[0],griffith_force_element.commonData
    )
  );
  griffith_force_element.feRhs.getOpPtrVector().push_back(
    new GriffithForceElement::OpRhs(
      3,griffith_force_element.blockData[0],griffith_force_element.commonData
    )
  );
  griffith_force_element.feLhs.getOpPtrVector().push_back(
    new GriffithForceElement::OpJacobian(
      3,griffith_force_element.blockData[0],griffith_force_element.commonData
    )
  );
  griffith_force_element.feLhs.getOpPtrVector().push_back(
    new GriffithForceElement::OpLhs(
      3,griffith_force_element.blockData[0],griffith_force_element.commonData
    )
  );

  if(griffith_force_element.blockData[0].penalty!=0) {

    griffith_force_element.feRhs.getOpPtrVector().push_back(
      new GriffithForceElement::OpJacobianPenalty(
        6,griffith_force_element.blockData[0],griffith_force_element.commonData
      )
    );
    griffith_force_element.feRhs.getOpPtrVector().push_back(
      new GriffithForceElement::OpPenaltyRhs(
        6,griffith_force_element.blockData[0],griffith_force_element.commonData
      )
    );
    griffith_force_element.feLhs.getOpPtrVector().push_back(
      new GriffithForceElement::OpJacobianPenalty(
        6,griffith_force_element.blockData[0],griffith_force_element.commonData
      )
    );
    griffith_force_element.feLhs.getOpPtrVector().push_back(
      new GriffithForceElement::OpPenaltyLhs(
        6,griffith_force_element.blockData[0],griffith_force_element.commonData
      )
    );

  }

  // Crack front
  //Snes_CTgc_CONSTANT_AREA ct_gc(m_field,projFrontCtx,"COUPLED_PROBLEM","LAMBDA_CRACKFRONT_AREA");
  //Snes_dCTgc_CONSTANT_AREA dct_gc(m_field,K,"LAMBDA_CRACKFRONT_AREA");
  Snes_CalculateC_NeededForCrackAreaArcLengthControl calculate_C_for_arclength(
    m_field,projFrontCtx,"COUPLED_PROBLEM","LAMBDA_CRACKFRONT_AREA"
  );

  // Meshs smoothing

  Vec front_f,tangent_front_f;
  ierr = VecDuplicate(F,&front_f); CHKERRQ(ierr);
  ierr = VecSetOption(front_f,VEC_IGNORE_NEGATIVE_INDICES,PETSC_TRUE); CHKERRQ(ierr);
  ierr = VecDuplicate(F,&tangent_front_f); CHKERRQ(ierr);
  ierr = VecSetOption(tangent_front_f,VEC_IGNORE_NEGATIVE_INDICES,PETSC_TRUE); CHKERRQ(ierr);

  /*MyMeshSmoothing smoother(m_field);
  ierr = smoother.initCrackFrontData(m_field); CHKERRQ(ierr);
  set_qual_ver(3);
  smoother.frontF = front_f;
  smoother.tangentFrontF = tangent_front_f;
  //smoother.stabilise = true;*/

  TangentWithMeshSmoothingFrontConstrain tangent_constrain(m_field,"LAMBDA_CRACK_TANGENT_CONSTRAIN");

  tangent_constrain.frontF = front_f;
  tangent_constrain.tangentFrontF = tangent_front_f;

  PetscBool flg_alpha2,flg_gamma;
  double alpha2,gamma;
  ierr = PetscOptionsGetReal(PETSC_NULL,"","-my_alpha2",&alpha2,&flg_alpha2); CHKERRQ(ierr);
  ierr = PetscOptionsGetReal(PETSC_NULL,"","-my_gamma",&gamma,&flg_gamma); CHKERRQ(ierr);
	if(!flg_alpha2) SETERRQ(PETSC_COMM_SELF,1,"-my_alpha2 is not set");
	if(!flg_gamma) SETERRQ(PETSC_COMM_SELF,1,"-my_gamma is not set");

  boost::shared_ptr<VolumeLengthQuality<adouble> > volume_length_adouble(
    new VolumeLengthQuality<adouble>(
      BARRIER_AND_CHANGE_QUALITY_SCALED_BY_VOLUME,
      alpha2,
      gamma
    )
  );
  boost::shared_ptr<VolumeLengthQuality<double> > volume_length_double(
    new VolumeLengthQuality<double>(
      BARRIER_AND_CHANGE_QUALITY_SCALED_BY_VOLUME,
      alpha2,
      gamma
    )
  );

  Smoother smoother_fe(m_field);

  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,BLOCKSET|MAT_ELASTICSET,it)) {
    int id = it->getMeshsetId();
    EntityHandle meshset = it->getMeshset();
    rval = m_field.get_moab().get_entities_by_type(
      meshset,MBTET,smoother_fe.setOfBlocks[id].tEts,true
    ); CHKERRQ_MOAB(rval);
    smoother_fe.setOfBlocks[id].iD = id;
    smoother_fe.setOfBlocks[id].materialDoublePtr = volume_length_double;
    smoother_fe.setOfBlocks[id].materialAdoublePtr = volume_length_adouble;
    smoother_fe.setOfBlocks[id].forcesOnlyOnEntitiesRow.merge(crack_front_edges);
    smoother_fe.setOfBlocks[id].forcesOnlyOnEntitiesRow.merge(crack_front_nodes);
  }

  smoother_fe.commonData.spatialPositions = "MESH_NODE_POSITIONS";
  smoother_fe.commonData.meshPositions = "NONE";
  smoother_fe.smootherData.sTabilised = false;
  smoother_fe.smootherData.frontF = front_f;
  smoother_fe.smootherData.tangentFrontF = tangent_front_f;

  smoother_fe.feRhs.meshPositionsFieldName = "NONE";
  smoother_fe.feLhs.meshPositionsFieldName = "NONE";
  smoother_fe.feRhs.addToRule = 0;
  smoother_fe.feLhs.addToRule = 0;

  // Smoother Lhs
  smoother_fe.feLhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",smoother_fe.commonData)
  );
  sit = smoother_fe.setOfBlocks.begin();
  for(;sit!=smoother_fe.setOfBlocks.end();sit++) {
    smoother_fe.feLhs.getOpPtrVector().push_back(
      new Smoother::OpJacobianSmoother("MESH_NODE_POSITIONS",sit->second,smoother_fe.commonData,4,true)
    );
    smoother_fe.feLhs.getOpPtrVector().push_back(
      new Smoother::OpLhsSmoother(
        "MESH_NODE_POSITIONS","MESH_NODE_POSITIONS",sit->second,smoother_fe.commonData,smoother_fe.smootherData
      )
    );
  }

  // Smoother Rhs
  smoother_fe.feRhs.getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",smoother_fe.commonData)
  );
  sit = smoother_fe.setOfBlocks.begin();
  for(;sit!=smoother_fe.setOfBlocks.end();sit++) {
    smoother_fe.feRhs.getOpPtrVector().push_back(
      new Smoother::OpJacobianSmoother("MESH_NODE_POSITIONS",sit->second,smoother_fe.commonData,4,false)
    );
    smoother_fe.feRhs.getOpPtrVector().push_back(
      new Smoother::OpRhsSmoother(
        "MESH_NODE_POSITIONS",sit->second,smoother_fe.commonData,smoother_fe.smootherData
      )
    );
  }

  // Sourface constrains
  /*SnesConstrainSurfacGeometry constrain_body_surface(m_field,"LAMBDA_SURFACE");
  constrain_body_surface.nonlinear = true;
  SnesConstrainSurfacGeometry constrain_crack_surface(m_field,"LAMBDA_CRACK_SURFACE");
  constrain_crack_surface.nonlinear = true;
  map<int,SnesConstrainSurfacGeometry*> other_body_surface_constrains;
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss;
    ss << "LAMBDA_SURFACE_msId_" << msId;
    other_body_surface_constrains[msId] = new SnesConstrainSurfacGeometry(m_field,ss.str());
    other_body_surface_constrains[msId]->nonlinear = true;
  }*/

  FaceOrientation orientation(false),orientation_crack(true);
  SurfaceSlidingConstrains surface_constrain(m_field,orientation);
  SurfaceSlidingConstrains surface_crack_constrain(m_field,orientation_crack);
  boost::ptr_map<int,SurfaceSlidingConstrains> surface_constrains_map;
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    surface_constrains_map.insert(msId,new SurfaceSlidingConstrains(m_field,orientation));
  }

  surface_constrain.setOperatorsWithLinearGeometry("LAMBDA_SURFACE","MESH_NODE_POSITIONS",true,true);
  surface_crack_constrain.setOperatorsWithLinearGeometry("LAMBDA_CRACK_SURFACE","MESH_NODE_POSITIONS",true,true);
  for(
    boost::ptr_map<int,SurfaceSlidingConstrains>::iterator it = surface_constrains_map.begin();
    it!=surface_constrains_map.end();
    it++
  ) {
    int msId =  it->first;
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss;
    ss << "LAMBDA_SURFACE_msId_" << msId;
    ierr = it->second->setOperatorsWithLinearGeometry(ss.str(),"MESH_NODE_POSITIONS",true,true); CHKERRQ(ierr);
  }

  // Bothsieds constrains
  BothSurfaceConstrains both_sides_constrains(m_field);
  // Dirichlet constrains
  FixBcAtEntities fix_material_pts(m_field,"MESH_NODE_POSITIONS",corners_nodes);
  fix_material_pts.fieldNames.push_back("LAMBDA_SURFACE");
  fix_material_pts.fieldNames.push_back("LAMBDA_CRACK_SURFACE");
  fix_material_pts.fieldNames.push_back("LAMBDA_CRACK_SURFACE_WITH_CRACK_FRONT");
  fix_material_pts.fieldNames.push_back("LAMBDA_CRACK_TANGENT_CONSTRAIN");
  fix_material_pts.fieldNames.push_back("LAMBDA_BOTH_SIDES");
  for(_IT_CUBITMESHSETS_BY_SET_TYPE_FOR_LOOP_(m_field,SIDESET,it)) {
    int msId = it->getMeshsetId();
    if((msId < 10200)||(msId >= 10300)) continue;
    ostringstream ss;
    ss << "LAMBDA_SURFACE_msId_" << msId;
    fix_material_pts.fieldNames.push_back(ss.str());
  }
  SpatialPositionsBCFEMethodPreAndPostProc my_dirichlet_bc(m_field,"SPATIAL_POSITION",K,D,F);
  my_dirichlet_bc.fixFields.push_back("MESH_NODE_POSITIONS");
  my_dirichlet_bc.fixFields.push_back("LAMBDA_SURFACE");
  my_dirichlet_bc.fixFields.push_back("LAMBDA_CRACK_SURFACE_WITH_CRACK_FRONT");
  my_dirichlet_bc.fixFields.push_back("LAMBDA_BOTH_SIDES");
  // Boundary conditions
  Tag th_scale;
  const EntityHandle root_meshset = m_field.get_moab().get_root_set();
  rval = m_field.get_moab().tag_get_handle("_LoadFactor_Scale_",th_scale); CHKERRQ_MOAB(rval);
  double *force_scale;
  rval = m_field.get_moab().tag_get_by_ptr(th_scale,&root_meshset,1,(const void**)&force_scale); CHKERRQ_MOAB(rval);
  arc_ctx.getFieldData() = (*force_scale);
  double scaled_reference_load = 1;
  double *scale_lhs = &(arc_ctx.getFieldData());
  double *scale_rhs = &(scaled_reference_load);
  NeummanForcesSurfaceComplexForLazy neumann_forces(m_field,K,arc_ctx.F_lambda,scale_lhs,scale_rhs);
  NeummanForcesSurfaceComplexForLazy::MyTriangleSpatialFE &fe_forces = neumann_forces.getLoopSpatialFe();
  fe_forces.typeOfForces = NeummanForcesSurfaceComplexForLazy::MyTriangleSpatialFE::NONCONSERVATIVE;
  fe_forces.uSeF = true;
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,NODESET|FORCESET,it)) {
    ierr = fe_forces.addForce(it->getMeshsetId()); CHKERRQ(ierr);
  }
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,SIDESET|PRESSURESET,it)) {
    ierr = fe_forces.addPreassure(it->getMeshsetId()); CHKERRQ(ierr);
  }
  // PortsProc
  MyPrePostProcess pre_post_method(m_field,&arc_ctx);
  AssembleLambda assemble_F_lambda(m_field,&arc_ctx,&my_dirichlet_bc);

  // Rhs
  arc_snes_ctx.get_preProcess_to_do_Rhs().push_back(&pre_post_method);
  arc_snes_ctx.get_preProcess_to_do_Rhs().push_back(&my_dirichlet_bc);
  arc_snes_ctx.get_preProcess_to_do_Rhs().push_back(&fix_material_pts);
  //arc_snes_ctx.get_preProcess_to_do_Rhs().push_back(&ct_gc);
  arc_snes_ctx.get_preProcess_to_do_Rhs().push_back(&calculate_C_for_arclength);
  SnesCtx::loops_to_do_type& loops_to_do_Rhs = arc_snes_ctx.get_loops_to_do_Rhs();
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("BOTH_SIDE_OF_CRACK",&both_sides_constrains));

  /*loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("CandCT_SURFACE_ELEM",&constrain_body_surface));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("CandCT_CRACK_SURFACE_ELEM",&constrain_crack_surface));
  for(map<int,SnesConstrainSurfacGeometry*>::iterator mit = other_body_surface_constrains.begin();
    mit!=other_body_surface_constrains.end();mit++) {
    ostringstream ss2;
    ss2 << "CandCT_SURFACE_ELEM_msId_" << mit->first;
    loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type(ss2.str(),mit->second));
  }*/

  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("CandCT_SURFACE_ELEM",&surface_constrain.feRhs));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("CandCT_CRACK_SURFACE_ELEM",&surface_crack_constrain.feRhs));
  for(
    boost::ptr_map<int,SurfaceSlidingConstrains>::iterator it = surface_constrains_map.begin();
    it!=surface_constrains_map.end();
    it++
  ) {
    ostringstream ss;
    ss << "CandCT_SURFACE_ELEM_msId_" << it->first;
    loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type(ss.str(),&it->second->feRhs));
  }

  // Loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("MESH_SMOOTHER",&smoother));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("MESH_SMOOTHER",&smoother_fe.feRhs));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("C_TANGENT_ELEM",&tangent_constrain));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("ELASTIC_COUPLED",&spatial_fe.feRhs));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("MATERIAL_COUPLED",&material_fe.feRhs));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("GRIFFITH_FORCE_ELEMENT",&griffith_force_element.feRhs));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("NEUAMNN_FE",&fe_forces));
  // Nodal forces
  boost::ptr_map<string,NodalForce> nodal_forces;
  string fe_name_str ="FORCE_FE";
  nodal_forces.insert(fe_name_str,new NodalForce(m_field));
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,NODESET|FORCESET,it)) {
    ierr = nodal_forces.at(fe_name_str).addForce("SPATIAL_POSITION",arc_ctx.F_lambda,it->getMeshsetId(),false);  CHKERRQ(ierr);
  }
  {
    boost::ptr_map<string,NodalForce>::iterator fit = nodal_forces.begin();
    for(;fit!=nodal_forces.end();fit++) {
      loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type(fit->first,&fit->second->getLoopFe()));
    }
  }

  // Edge force
  boost::ptr_map<string,EdgeForce> edge_forces;
  edge_forces.insert(fe_name_str,new EdgeForce(m_field));
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,NODESET|FORCESET,it)) {
    ierr = edge_forces.at(fe_name_str).addForce("SPATIAL_POSITION",arc_ctx.F_lambda,it->getMeshsetId(),false);  CHKERRQ(ierr);
  }
  {
    boost::ptr_map<string,NodalForce>::iterator fit = edge_forces.begin();
    for(;fit!=edge_forces.end();fit++) {
      loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type(fit->first,&fit->second->getLoopFe()));
    }
  }
  // Arc length
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("NoNE",&assemble_F_lambda));
  loops_to_do_Rhs.push_back(SnesCtx::loop_pair_type("ARC_LENGTH",&arc_elem));

  arc_snes_ctx.get_postProcess_to_do_Rhs().push_back(&fix_material_pts);
  arc_snes_ctx.get_postProcess_to_do_Rhs().push_back(&my_dirichlet_bc);
  arc_snes_ctx.get_postProcess_to_do_Rhs().push_back(&pre_post_method);

  // Lhs
  arc_snes_ctx.get_preProcess_to_do_Mat().push_back(&my_dirichlet_bc);
  arc_snes_ctx.get_preProcess_to_do_Mat().push_back(&fix_material_pts);
  SnesCtx::loops_to_do_type& loops_to_do_Mat = arc_snes_ctx.get_loops_to_do_Mat();
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("BOTH_SIDE_OF_CRACK",&both_sides_constrains));
  //loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("dCT_CRACKFRONT_AREA_ELEM",&dct_gc));

  /*loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("CandCT_SURFACE_ELEM",&constrain_body_surface));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("CandCT_CRACK_SURFACE_ELEM",&constrain_crack_surface));
  for(map<int,SnesConstrainSurfacGeometry*>::iterator mit = other_body_surface_constrains.begin();
    mit!=other_body_surface_constrains.end();mit++) {
    ostringstream ss2;
    ss2 << "CandCT_SURFACE_ELEM_msId_" << mit->first;
    loops_to_do_Mat.push_back(SnesCtx::loop_pair_type(ss2.str(),mit->second));
  }*/

  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("CandCT_SURFACE_ELEM",&surface_constrain.feLhs));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("CandCT_CRACK_SURFACE_ELEM",&surface_crack_constrain.feLhs));
  for(
    boost::ptr_map<int,SurfaceSlidingConstrains>::iterator it = surface_constrains_map.begin();
    it!=surface_constrains_map.end();
    it++
  ) {
    ostringstream ss;
    ss << "CandCT_SURFACE_ELEM_msId_" << it->first;
    loops_to_do_Mat.push_back(SnesCtx::loop_pair_type(ss.str(),&it->second->feLhs));
  }

  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("C_TANGENT_ELEM",&tangent_constrain));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("MESH_SMOOTHER",&smoother_fe.feLhs));
  //loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("MESH_SMOOTHER",&smoother));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("ELASTIC_COUPLED",&spatial_fe.feLhs));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("MATERIAL_COUPLED",&material_fe.feLhs));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("GRIFFITH_FORCE_ELEMENT",&griffith_force_element.feLhs));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("NEUAMNN_FE",&fe_forces));
  loops_to_do_Mat.push_back(SnesCtx::loop_pair_type("ARC_LENGTH",&arc_elem));
  arc_snes_ctx.get_postProcess_to_do_Mat().push_back(&fix_material_pts);
  arc_snes_ctx.get_postProcess_to_do_Mat().push_back(&my_dirichlet_bc);

  PetscInt M,N;
  ierr = MatGetSize(K,&M,&N); CHKERRQ(ierr);
  PetscInt m,n;
  MatGetLocalSize(K,&m,&n);
  Mat ShellK;
  ierr = MatCreateShell(PETSC_COMM_WORLD,m,n,M,N,(void*)&arc_mat_ctx,&ShellK); CHKERRQ(ierr);
  ierr = MatShellSetOperation(ShellK,MATOP_MULT,(void(*)(void))ArcLengthMatMultShellOp); CHKERRQ(ierr);

  ierr = SNESSetApplicationContext(*snes,&arc_snes_ctx); CHKERRQ(ierr);
  ierr = SNESSetFunction(*snes,F,SnesRhs,&arc_snes_ctx); CHKERRQ(ierr);
  ierr = SNESSetJacobian(*snes,ShellK,K,SnesMat,&arc_snes_ctx); CHKERRQ(ierr);
  PetscReal my_tol;
  ierr = PetscOptionsGetReal(PETSC_NULL,PETSC_NULL,"-my_tol",&my_tol,&flg); CHKERRQ(ierr);
  if(flg == PETSC_TRUE) {
    PetscReal atol,rtol,stol;
    PetscInt maxit,maxf;
    ierr = SNESGetTolerances(*snes,&atol,&rtol,&stol,&maxit,&maxf); CHKERRQ(ierr);
    atol = my_tol;
    rtol = atol*1e2;
    ierr = SNESSetTolerances(*snes,atol,rtol,stol,maxit,maxf); CHKERRQ(ierr);
  }

  KSP ksp;
  ierr = SNESGetKSP(*snes,&ksp); CHKERRQ(ierr);
  if(flg == PETSC_TRUE) {
    PetscReal rtol,atol,dtol;
    PetscInt maxits;
    ierr = KSPGetTolerances(ksp,&rtol,&atol,&dtol,&maxits); CHKERRQ(ierr);
    atol = my_tol*1e-2;
    rtol = atol*1e-2;
    ierr = KSPSetTolerances(ksp,rtol,atol,dtol,maxits); CHKERRQ(ierr);
  }

  PC pc;
  ierr = KSPGetPC(ksp,&pc); CHKERRQ(ierr);
  ierr = PCSetType(pc,PCSHELL); CHKERRQ(ierr);
  ierr = PCShellSetContext(pc,&pc_ctx); CHKERRQ(ierr);
  ierr = PCShellSetApply(pc,PCApplyArcLength); CHKERRQ(ierr);
  ierr = PCShellSetSetUp(pc,PCSetupArcLength); CHKERRQ(ierr);

  ierr = m_field.set_local_ghost_vector("COUPLED_PROBLEM",COL,D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecZeroEntries(F); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(F,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(F,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  ierr = m_field.get_problem("COUPLED_PROBLEM",&(arc_elem.problemPtr)); CHKERRQ(ierr);
  ierr = arc_elem.set_dlambda_to_x(D,0); CHKERRQ(ierr);
  ierr = VecCopy(D,arc_ctx.x0); CHKERRQ(ierr);
  ierr = arc_elem.get_dlambda(D); CHKERRQ(ierr);

  double alpha = 1;
  ierr = PetscOptionsGetReal(PETSC_NULL,PETSC_NULL,"-my_arc_alpha",&alpha,&flg); CHKERRQ(ierr);
  // Calculate rhs and F_lambda
  ierr = arc_ctx.setS(0); CHKERRQ(ierr);
  ierr = arc_ctx.setAlphaBeta(0,1); CHKERRQ(ierr);
  ierr = arc_elem.calculate_lambda_int(); CHKERRQ(ierr);
  arc_snes_ctx.bH = MF_ZERO;
  ierr = SnesRhs(*snes,D,F,&arc_snes_ctx); CHKERRQ(ierr);
  // Set s
  ierr = arc_ctx.setAlphaBeta(alpha,0); CHKERRQ(ierr);
  ierr = arc_elem.calculate_lambda_int(); CHKERRQ(ierr);
  ierr = arc_ctx.setS(arc_elem.lambda_int+alpha*da/arc_elem.aRea0); CHKERRQ(ierr);

  SNESLineSearch linesearch;
  ierr = SNESGetLineSearch(*snes,&linesearch); CHKERRQ(ierr);
  PetscReal atol,rtol,stol;
  PetscInt maxit,maxf;
  ierr = SNESGetTolerances(*snes,&atol,&rtol,&stol,&maxit,&maxf); CHKERRQ(ierr);

  Vec D0;
  ierr = m_field.VecCreateGhost("COUPLED_PROBLEM",COL,&D0); CHKERRQ(ierr);
  ierr = m_field.set_local_ghost_vector("COUPLED_PROBLEM",COL,D0,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  ierr = SNESLineSearchSetType(linesearch,SNESLINESEARCHBASIC); CHKERRQ(ierr);
  ierr = SNESSetConvergenceTest(*snes,MySnesConvernceTest_SNESLINESEARCHBT,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
  ierr = SNESSolve(*snes,PETSC_NULL,D); CHKERRQ(ierr);
  int its;
  ierr = SNESGetIterationNumber(*snes,&its); CHKERRQ(ierr);
  total_its = its;

  ierr = SNESGetConvergedReason(*snes,&reason); CHKERRQ(ierr);
  if(reason < 0) {
    ierr = SNESLineSearchSetType(linesearch,SNESLINESEARCHL2); CHKERRQ(ierr);
    ierr = SNESSetConvergenceTest(*snes,MySnesConvernceTest_SNESLINESEARCHL2,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
    ierr = SNESSolve(*snes,PETSC_NULL,D); CHKERRQ(ierr);
    ierr = SNESGetIterationNumber(*snes,&its); CHKERRQ(ierr);
    total_its += its;

    SNESConvergedReason reason0;
    ierr = SNESGetConvergedReason(*snes,&reason0); CHKERRQ(ierr);

    int its;
    ierr = SNESGetIterationNumber(*snes,&its); CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"number of Newton iterations = %D\n",its); CHKERRQ(ierr);
    if(reason0 < 0) {

      ierr = m_field.set_global_ghost_vector("COUPLED_PROBLEM",COL,D0,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
      ierr = VecCopy(D0,D); CHKERRQ(ierr);
      //smoother.stabilise = true;
      smoother_fe.smootherData.sTabilised = true;
      griffith_force_element.blockData[0].penalty = 0;
      ierr = PetscPrintf(PETSC_COMM_WORLD,"Smoother On Everyware\n"); CHKERRQ(ierr);
      for(int ii = 1;ii<=3;ii++) {
        ierr = VecCopy(D,D0); CHKERRQ(ierr);
        //smoother.alpha22 = exp(-ii);
        volume_length_adouble->aLpha = alpha2*exp(-ii);
        volume_length_double->aLpha = alpha2*exp(-ii);

        ierr = SNESLineSearchSetType(linesearch,SNESLINESEARCHL2); CHKERRQ(ierr);
        ierr = SNESSetConvergenceTest(*snes,MySnesConvernceTest_SNESLINESEARCHL2,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
        ierr = SNESSolve(*snes,PETSC_NULL,D); CHKERRQ(ierr);
        ierr = SNESGetIterationNumber(*snes,&its); CHKERRQ(ierr);
        total_its += its;
        ierr = SNESGetConvergedReason(*snes,&reason0); CHKERRQ(ierr);
        if(reason0 < 0) {
          ierr = m_field.set_global_ghost_vector("COUPLED_PROBLEM",COL,D0,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
          ierr = VecCopy(D0,D); CHKERRQ(ierr);
          break;
        } else {
          ierr = m_field.set_global_ghost_vector("COUPLED_PROBLEM",COL,D0,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
          ierr = set_coordinates_from_material_solution(m_field,false); CHKERRQ(ierr);
        }

      }
      //smoother.stabilise = false;
      smoother_fe.smootherData.sTabilised = false;
    }
  }
  ierr = VecGhostUpdateBegin(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(D,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  *force_scale = (arc_ctx.getFieldData());
  ierr = VecDestroy(&D0); CHKERRQ(ierr);

  // Save data on mesh
  ierr = m_field.set_global_ghost_vector("COUPLED_PROBLEM",COL,D,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  aRea = arc_elem.aRea;

  ierr = m_field.set_field(0,MBVERTEX,"SPATIAL_DISPLACEMENT"); CHKERRQ(ierr);
  ierr = m_field.set_field(0,MBEDGE,"SPATIAL_DISPLACEMENT"); CHKERRQ(ierr);
  ierr = m_field.set_field(0,MBTRI,"SPATIAL_DISPLACEMENT"); CHKERRQ(ierr);
  ierr = m_field.set_field(0,MBTET,"SPATIAL_DISPLACEMENT"); CHKERRQ(ierr);
  ierr = m_field.field_axpy(+1.,"SPATIAL_POSITION","SPATIAL_DISPLACEMENT"); CHKERRQ(ierr);
  ierr = m_field.field_axpy(-1.,"MESH_NODE_POSITIONS","SPATIAL_DISPLACEMENT"); CHKERRQ(ierr);

  Vec DISP;
  ierr = VecDuplicate(D,&DISP); CHKERRQ(ierr);
  ierr = m_field.set_other_global_ghost_vector(
    "COUPLED_PROBLEM","SPATIAL_POSITION","SPATIAL_DISPLACEMENT",COL,DISP,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecDot(DISP,arc_ctx.F_lambda,&energy); CHKERRQ(ierr);
  lambda = arc_ctx.getFieldData();
  energy = 0.5*fabs(lambda)*fabs(energy);

  int verb = 1;
  if(reason>=0) {

    // Save field on mesh
    if(verb>0) {
      ierr = PetscPrintf(PETSC_COMM_WORLD,"Save SPATIAL_POSITION on tags\n"); CHKERRQ(ierr);
    }
    PostProcVertexMethod ent_method_spatial(m_field.get_moab(),"SPATIAL_POSITION");
    ierr = m_field.loop_dofs("COUPLED_PROBLEM","SPATIAL_POSITION",COL,ent_method_spatial); CHKERRQ(ierr);
    if(verb>0) {
      ierr = PetscPrintf(PETSC_COMM_WORLD,"Save MESH_NODE_POSITIONS on tags\n"); CHKERRQ(ierr);
    }
    PostProcVertexMethod ent_method_material(m_field.get_moab(),"MESH_NODE_POSITIONS");
    ierr = m_field.loop_dofs("COUPLED_PROBLEM","MESH_NODE_POSITIONS",COL,ent_method_material); CHKERRQ(ierr);
    if(verb>2) {
      ierr = PetscPrintf(PETSC_COMM_WORLD,"Save MATERIAL_RESIDUAL on tags\n"); CHKERRQ(ierr);
      PostProcVertexMethod ent_method_res_mat(m_field.get_moab(),"MESH_NODE_POSITIONS",F,"MATERIAL_RESIDUAL");
      ierr = m_field.loop_dofs("COUPLED_PROBLEM","MESH_NODE_POSITIONS",COL,ent_method_res_mat); CHKERRQ(ierr);
      ierr = PetscPrintf(PETSC_COMM_WORLD,"Save SPATIAL_RESIDUAL on tags\n"); CHKERRQ(ierr);
      PostProcVertexMethod ent_method_res_spat(m_field.get_moab(),"SPATIAL_POSITION",F,"SPATIAL_RESIDUAL");
      ierr = m_field.loop_dofs("COUPLED_PROBLEM","SPATIAL_POSITION",COL,ent_method_res_spat); CHKERRQ(ierr);
    }
    if(verb>0) {
      ierr = PetscPrintf(PETSC_COMM_WORLD,"Save SPATIAL_DISPLACEMENT on tags\n"); CHKERRQ(ierr);
      PostProcVertexMethod ent_method_disp_spat(m_field.get_moab(),"SPATIAL_POSITION",DISP,"SPATIAL_DISPLACEMENT");
      ierr = m_field.loop_dofs("COUPLED_PROBLEM","SPATIAL_POSITION",COL,ent_method_disp_spat); CHKERRQ(ierr);
    }

  }

  ierr = VecZeroEntries(F);  CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(F,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(F,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  //ct_gc.snes_f = F;
  //ierr = ct_gc.preProcess(); CHKERRQ(ierr);
  griffith_force_element.feRhs.snes_f = F;
  ierr = m_field.loop_finite_elements(
    "COUPLED_PROBLEM","GRIFFITH_FORCE_ELEMENT",griffith_force_element.feRhs
  );  CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(F,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(F,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(F,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(F,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = m_field.set_other_global_ghost_vector(
    "COUPLED_PROBLEM","MESH_NODE_POSITIONS","GRIFFITH_FORCE",ROW,F,INSERT_VALUES,SCATTER_REVERSE
  ); CHKERRQ(ierr);

  //calculate work of fracture
  const MoFEMProblem *problem_ptr;
  ierr = m_field.get_problem("C_CRACKFRONT_MATRIX",&problem_ptr); CHKERRQ(ierr);
  map_ent_work.clear();
  ublas::vector<double> mesh_node_disp(3),coords(3),griffith_force(3);
  for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_NAME_FOR_LOOP_(problem_ptr,"LAMBDA_CRACKFRONT_AREA",diit)) {
    EntityHandle ent = diit->get()->getEnt();
    for(_IT_GET_DOFS_FIELD_BY_NAME_AND_ENT_FOR_LOOP_(m_field,"GRIFFITH_FORCE",diit->get()->getEnt(),diiit)) {
      griffith_force[diiit->get()->getDofCoeffIdx()] = diiit->get()->getFieldData();
    }
    for(_IT_GET_DOFS_FIELD_BY_NAME_AND_ENT_FOR_LOOP_(m_field,"MESH_NODE_POSITIONS",diit->get()->getEnt(),diiit)) {
      mesh_node_disp[diiit->get()->getDofCoeffIdx()] = diiit->get()->getFieldData();
    }
    rval = m_field.get_moab().get_coords(&ent,1,&coords[0]); CHKERRQ_MOAB(rval);
    //cerr << "coords " << coords << endl;
    //cerr << "mesh_node_disp " << mesh_node_disp << endl;
    //cerr << "griffith_force " << griffith_force << endl;
    mesh_node_disp -= coords;
    double w = inner_prod(mesh_node_disp,griffith_force);
    map_ent_work[ent] = w;
    ierr = PetscPrintf(PETSC_COMM_WORLD,"at ent %ld load step work of fracture = %6.4e\n",ent,w); CHKERRQ(ierr);
  }
  ierr = VecDestroy(&DISP); CHKERRQ(ierr);
  ierr = MatDestroy(&ShellK); CHKERRQ(ierr);
  ierr = MatDestroy(&K); CHKERRQ(ierr);
  ierr = VecDestroy(&D); CHKERRQ(ierr);
  ierr = VecDestroy(&F); CHKERRQ(ierr);
  ierr = VecDestroy(&front_f); CHKERRQ(ierr);
  ierr = VecDestroy(&tangent_front_f); CHKERRQ(ierr);

  /*for(map<int,SnesConstrainSurfacGeometry*>::iterator mit = other_body_surface_constrains.begin();
    mit!=other_body_surface_constrains.end();mit++) {
    delete mit->second;
  }*/

  ierr = delete_front_projection_data(m_field); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode ConfigurationalFractureMechanics::calculate_material_forces(MoFEM::Interface& m_field,string problem,string fe) {
  PetscFunctionBegin;

  PetscErrorCode ierr;

  MeshsetsManager *meshset_manager_ptr;
  ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Range corners_edges,corners_nodes;
  if(meshset_manager_ptr->checkMeshset(100,SIDESET)) {
    ierr = meshset_manager_ptr->getEntitiesByDimension(100,SIDESET,1,corners_edges,true); CHKERRQ(ierr);
  }
  if(meshset_manager_ptr->checkMeshset(101,NODESET)) {
    ierr = meshset_manager_ptr->getEntitiesByDimension(101,NODESET,0,corners_nodes,true); CHKERRQ(ierr);
  }
  // ierr = m_field.get_cubit_msId_entities_by_dimension(100,SIDESET,1,corners_edges,true); CHKERRQ(ierr);
  // ierr = m_field.get_cubit_msId_entities_by_dimension(101,NODESET,0,corners_nodes,true); CHKERRQ(ierr);
  ErrorCode rval;
  Range corners_edges_nodes;
  rval = m_field.get_moab().get_connectivity(corners_edges,corners_edges_nodes,true); CHKERRQ_MOAB(rval);
  corners_nodes.insert(corners_edges_nodes.begin(),corners_edges_nodes.end());
  Range nodes_to_block;
  BitRefLevel bit_to_block = BitRefLevel().set(BITREFLEVEL_SIZE-2);
  ierr = m_field.get_entities_by_type_and_ref_level(bit_to_block,BitRefLevel().set(),MBVERTEX,nodes_to_block); CHKERRQ(ierr);
  corners_nodes.merge(nodes_to_block);

  Vec F_Material;
  ierr = m_field.VecCreateGhost(problem,COL,&F_Material); CHKERRQ(ierr);
  ierr = VecSetOption(F_Material,VEC_IGNORE_NEGATIVE_INDICES,PETSC_TRUE);  CHKERRQ(ierr);

  ierr = VecZeroEntries(F_Material); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(F_Material,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(F_Material,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  {
    Range crack_front_edges,crack_front_nodes;
    ierr = m_field.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRABORT(PETSC_COMM_WORLD,ierr);
    m_field.get_moab().get_connectivity(crack_front_edges,crack_front_nodes,true);


    FixBcAtEntities fix_material_pts(m_field,"MESH_NODE_POSITIONS",corners_nodes);
    fix_material_pts.snes_x = PETSC_NULL;
    fix_material_pts.snes_f = F_Material;
    fix_material_pts.set_snes_ctx(SnesMethod::CTX_SNESSETFUNCTION);

    NonlinearElasticElement material_fe(m_field,-1);

    boost::shared_ptr<Hooke<adouble> > hooke_adouble(new Hooke<adouble>);
    boost::shared_ptr<Hooke<double> > hooke_double(new Hooke<double>);
    //NeoHookean<adouble> hooke_adouble;
    //NeoHookean<double> hooke_double;

    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(m_field,BLOCKSET|MAT_ELASTICSET,it)) {
      Mat_Elastic mydata;
      ierr = it->getAttributeDataStructure(mydata); CHKERRQ(ierr);
      int id = it->getMeshsetId();
      EntityHandle meshset = it->getMeshset();
      rval = m_field.get_moab().get_entities_by_type(
        meshset,MBTET,material_fe.setOfBlocks[id].tEts,true
      ); CHKERRQ_MOAB(rval);
      material_fe.setOfBlocks[id].iD = id;
      material_fe.setOfBlocks[id].E = mydata.data.Young;
      material_fe.setOfBlocks[id].PoissonRatio = mydata.data.Poisson;
      material_fe.setOfBlocks[id].materialDoublePtr = hooke_double;
      material_fe.setOfBlocks[id].materialAdoublePtr = hooke_adouble;

      material_fe.setOfBlocks[id].forcesOnlyOnEntitiesRow.merge(crack_front_edges);
      material_fe.setOfBlocks[id].forcesOnlyOnEntitiesRow.merge(crack_front_nodes);

    }

    //Rhs
    material_fe.commonData.spatialPositions = "SPATIAL_POSITION";
    material_fe.commonData.meshPositions = "MESH_NODE_POSITIONS";

    material_fe.feRhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpGetCommonDataAtGaussPts("SPATIAL_POSITION",material_fe.commonData)
    );
    material_fe.feRhs.getOpPtrVector().push_back(
      new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",material_fe.commonData)
    );
    map<int,NonlinearElasticElement::BlockData>::iterator sit = material_fe.setOfBlocks.begin();
    for(;sit!=material_fe.setOfBlocks.end();sit++) {
      material_fe.feRhs.getOpPtrVector().push_back(
        new NonlinearElasticElement::OpJacobianEshelbyStress(
          "SPATIAL_POSITION",sit->second,material_fe.commonData,2,false,true
        )
      );
      material_fe.feRhs.getOpPtrVector().push_back(
        new NonlinearElasticElement::OpRhsEshelbyStrees("MESH_NODE_POSITIONS",sit->second,material_fe.commonData)
      );

    }

    material_fe.feRhs.meshPositionsFieldName = "NONE";
    material_fe.feRhs.addToRule = 0;
    material_fe.feRhs.snes_f = F_Material;

    ierr = m_field.problem_basic_method_preProcess(problem,fix_material_pts); CHKERRQ(ierr);
    ierr = m_field.loop_finite_elements(problem,fe,material_fe.feRhs);  CHKERRQ(ierr);
    ierr = m_field.problem_basic_method_postProcess(problem,fix_material_pts); CHKERRQ(ierr);
  }

  ierr = VecGhostUpdateBegin(F_Material,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(F_Material,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  ierr = VecAssemblyBegin(F_Material); CHKERRQ(ierr);
  ierr = VecAssemblyEnd(F_Material); CHKERRQ(ierr);

  PostProcVertexMethod ent_method_material(m_field.get_moab(),"MESH_NODE_POSITIONS",F_Material,"MATERIAL_FORCE");
  ierr = m_field.loop_dofs(problem,"MESH_NODE_POSITIONS",COL,ent_method_material); CHKERRQ(ierr);

  double nrm2_material;
  ierr = VecNorm(F_Material,NORM_2,&nrm2_material);   CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"nrm2_F_Material = %6.4e\n",nrm2_material); CHKERRQ(ierr);

  //Fields
  ierr = m_field.set_other_global_ghost_vector(
    problem,"MESH_NODE_POSITIONS","MATERIAL_FORCE",
    ROW,F_Material,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);

  //detroy matrices
  ierr = VecDestroy(&F_Material); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

ConfigurationalFractureMechanics::FrontAreaArcLengthControl::FrontAreaArcLengthControl(
  MoFEM::Interface& m_field,ConfigurationalFractureMechanics *confProb,ArcLengthCtx *arc_ptr
):
  mField(m_field),
  confProb(confProb),
  arcPtr(arc_ptr),
  aRea(0),
  aRea0(1) {
    PetscErrorCode ierr;
    ErrorCode rval;
    //get crack surcace
    ierr = mField.get_cubit_msId_entities_by_dimension(200,SIDESET,2,crackSurfacesFaces,true); CHKERRABORT(PETSC_COMM_WORLD,ierr);
    Range level_tris;
    ierr = mField.get_entities_by_type_and_ref_level(*confProb->ptr_bit_level0,BitRefLevel().set(),MBTRI,level_tris); CHKERRABORT(PETSC_COMM_WORLD,ierr);
    crackSurfacesFaces = intersect(crackSurfacesFaces,level_tris);
    //get crack surface faces adjacent to crack front
    Range crack_front_edges;
    ierr = mField.get_cubit_msId_entities_by_dimension(201,SIDESET,1,crack_front_edges,true); CHKERRABORT(PETSC_COMM_WORLD,ierr);
    mField.get_moab().get_connectivity(crack_front_edges,crackFrontNodes,true);

    //get pointer to coupled problem
    ierr = mField.get_problem("COUPLED_PROBLEM",&problemPtr); CHKERRABORT(PETSC_COMM_WORLD,ierr);
    set<DofIdx> set_idx; //set of global dofs indexes for crack surface
    for(Range::iterator fit = crackSurfacesFaces.begin();fit!=crackSurfacesFaces.end();fit++) {
      const EntityHandle* conn;
      int num_nodes;
      rval = mField.get_moab().get_connectivity(*fit,conn,num_nodes,true);
      if (MB_SUCCESS != rval) {
        PetscTraceBackErrorHandler(
          PETSC_COMM_WORLD,
          __LINE__,PETSC_FUNCTION_NAME,__FILE__,
          MOFEM_DATA_INCONSISTENCY,PETSC_ERROR_INITIAL,"can not get connectibility",PETSC_NULL
        ); CHKERRABORT(PETSC_COMM_SELF,rval);
      }
      for(int nn = 0;nn<num_nodes; nn++) {
        for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_ENT_FOR_LOOP_(problemPtr,conn[nn],dit)) {
          if(dit->get()->getName()!="MESH_NODE_POSITIONS") continue;
          set_idx.insert(dit->get()->getPetscGlobalDofIdx());
        }
      }
    }

    //vector surfaceDofs keeps material positions for crack surface dofs
    ierr = PetscMalloc(set_idx.size()*sizeof(PetscInt),&isIdx); CHKERRABORT(PETSC_COMM_WORLD,ierr);
    copy(set_idx.begin(),set_idx.end(),isIdx);
    ierr = ISCreateGeneral(PETSC_COMM_SELF,set_idx.size(),isIdx,PETSC_USE_POINTER,&isSurface); CHKERRABORT(PETSC_COMM_WORLD,ierr);
    ierr = VecCreateSeq(PETSC_COMM_SELF,set_idx.size(),&surfaceDofs); CHKERRABORT(PETSC_COMM_WORLD,ierr);

    ierr = mField.VecCreateGhost("C_CRACKFRONT_MATRIX",ROW,&lambdaVec); CHKERRABORT(PETSC_COMM_WORLD,ierr);
    ierr = VecSet(lambdaVec,1); CHKERRABORT(PETSC_COMM_WORLD,ierr);

}
ConfigurationalFractureMechanics::FrontAreaArcLengthControl::~FrontAreaArcLengthControl() {
  PetscErrorCode ierr;
  ierr = VecScatterDestroy(&surfaceScatter); CHKERRABORT(PETSC_COMM_WORLD,ierr);
  ierr = VecDestroy(&surfaceDofs); CHKERRABORT(PETSC_COMM_WORLD,ierr);
  ierr = ISDestroy(&isSurface); CHKERRABORT(PETSC_COMM_WORLD,ierr);
  ierr = PetscFree(isIdx); CHKERRABORT(PETSC_COMM_WORLD,ierr);
  ierr = VecDestroy(&lambdaVec); CHKERRABORT(PETSC_COMM_WORLD,ierr);
}
PetscErrorCode ConfigurationalFractureMechanics::FrontAreaArcLengthControl::calculate_area() {
  PetscFunctionBegin;

  ErrorCode rval;
  PetscErrorCode ierr;

  //scatter from snes_x to surfaceDofs
  ierr = VecScatterCreate(snes_x,isSurface,surfaceDofs,PETSC_NULL,&surfaceScatter); CHKERRABORT(PETSC_COMM_SELF,ierr);
  ierr = VecScatterBegin(surfaceScatter,snes_x,surfaceDofs,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecScatterEnd(surfaceScatter,snes_x,surfaceDofs,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecScatterDestroy(&surfaceScatter); CHKERRABORT(PETSC_COMM_SELF,ierr);

  //get crack front nodal positions form surfaceDofs
  double *array;
  ierr = VecGetArray(surfaceDofs,&array); CHKERRQ(ierr);
  PetscInt size;
  ISGetSize(isSurface,&size);
  for(int ii = 0;ii<size;ii++) {
    problemPtr->numered_dofs_rows->get<PetscGlobalIdx_mi_tag>().find(isIdx[ii])->get()->getFieldData() = array[ii];
  }
  ierr = VecRestoreArray(surfaceDofs,&array); CHKERRABORT(PETSC_COMM_SELF,ierr);

  //calculate crack surface area
  aRea = 0;
  aRea0 = 0;
  ublas::vector<double,ublas::bounded_array<double,6> > diffNTRI(6);
  ShapeDiffMBTRI(&diffNTRI.data()[0]);
  for(Range::iterator fit = crackSurfacesFaces.begin();fit!=crackSurfacesFaces.end();fit++) {
    const EntityHandle* conn;
    int num_nodes;
    rval = mField.get_moab().get_connectivity(*fit,conn,num_nodes,true); CHKERRQ_MOAB(rval);
    ublas::vector<double,ublas::bounded_array<double,9> > dofsX(9);
    ierr = mField.get_moab().get_coords(conn,num_nodes,&*dofsX.data().begin()); CHKERRQ(ierr);
    ublas::vector<double,ublas::bounded_array<double,3> > normal(3);
    ierr = ShapeFaceNormalMBTRI(&diffNTRI[0],&dofsX.data()[0],&normal.data()[0]); CHKERRQ(ierr);
    aRea0 += norm_2(normal)*0.25;
    for(int nn = 0;nn<num_nodes; nn++) {
      if(find(crackFrontNodes.begin(),crackFrontNodes.end(),conn[nn])!=crackFrontNodes.end()) {
        for(_IT_GET_DOFS_FIELD_BY_NAME_AND_ENT_FOR_LOOP_(mField,"MESH_NODE_POSITIONS",conn[nn],dit)) {
          dofsX[nn*3+dit->get()->getDofCoeffIdx()] = dit->get()->getFieldData();
        }
      }
    }
    ierr = ShapeFaceNormalMBTRI(&diffNTRI[0],&dofsX.data()[0],&normal.data()[0]); CHKERRQ(ierr);
    //crack surface area is a half of crack top and bottom body surface
    aRea += norm_2(normal)*0.25;
  }
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::FrontAreaArcLengthControl::calculate_lambda_int() {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  lambda_int = 0;
  aRea = 1;
  aRea0 = 1;
  if(arcPtr->alpha != 0) {
    ierr = calculate_area(); CHKERRQ(ierr);
    lambda_int += arcPtr->alpha*aRea/aRea0;
  }
  if(arcPtr->beta != 0) {
    lambda_int += arcPtr->beta*arcPtr->dLambda*sqrt(arcPtr->F_lambda2);
  }
  //PetscPrintf(PETSC_COMM_WORLD,"\tsurface crack area = %6.4e lambda_int = %6.4e\n",aRea,lambda_int);
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::FrontAreaArcLengthControl::calculate_db() {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  if(arcPtr->alpha!=0) {
    ierr = MatMultTranspose(confProb->projFrontCtx->C,lambdaVec,arcPtr->db); CHKERRQ(ierr);
    if(aRea0 == 0) {
      SETERRQ(PETSC_COMM_SELF,MOFEM_DATA_INCONSISTENCY,"Crack area is 0");
    }
    ierr = VecScale(arcPtr->db,1./aRea0); CHKERRQ(ierr);
    if(arcPtr->alpha!=1) {
      ierr = VecScale(arcPtr->db,arcPtr->alpha); CHKERRQ(ierr);
    }
  } else {
    ierr = VecZeroEntries(arcPtr->db); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::FrontAreaArcLengthControl::set_dlambda_to_x(Vec x,double dlambda) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  //check if locl dof idx is non zero, i.e. that lambda is acessible from this processor
  if(arcPtr->getPetscLocalDofIdx()!=-1) {
    double *array;
    ierr = VecGetArray(x,&array); CHKERRQ(ierr);
    double lambda_old = array[arcPtr->getPetscLocalDofIdx()];
    if(!(dlambda == dlambda)) {
      ostringstream sss;
      sss << "s " << arcPtr->s << " " << arcPtr->beta << " " << arcPtr->F_lambda2;
      SETERRQ(PETSC_COMM_SELF,1,sss.str().c_str());
    }
    array[arcPtr->getPetscLocalDofIdx()] = lambda_old + dlambda;
    PetscPrintf(PETSC_COMM_WORLD,"\tload factor lambda_old,lambda_new/dlambda = %6.4e, %6.4e (%6.4e)\n",lambda_old, array[arcPtr->getPetscLocalDofIdx()], dlambda);
    ierr = VecRestoreArray(x,&array); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::FrontAreaArcLengthControl::get_dlambda(Vec x) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  //set dx
  ierr = VecCopy(x,arcPtr->dx); CHKERRQ(ierr);
  ierr = VecAXPY(arcPtr->dx,-1,arcPtr->x0); CHKERRQ(ierr);
  //if LAMBDA dof is on this partition
  if(arcPtr->getPetscLocalDofIdx()!=-1) {
    double *array;
    ierr = VecGetArray(arcPtr->dx,&array); CHKERRQ(ierr);
    arcPtr->dLambda = array[arcPtr->getPetscLocalDofIdx()];
    array[arcPtr->getPetscLocalDofIdx()] = 0;
    ierr = VecRestoreArray(arcPtr->dx,&array); CHKERRQ(ierr);
  }
  //brodcast dLambda
  ierr = VecGhostUpdateBegin(arcPtr->ghosTdLambda,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(arcPtr->ghosTdLambda,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  PetscPrintf(PETSC_COMM_WORLD,"\tload factor increment dlambda = %6.4e\n",arcPtr->dLambda);
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::FrontAreaArcLengthControl::preProcess() {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  switch (ts_ctx) {
    case CTX_TSSETIFUNCTION: {
      snes_ctx = CTX_SNESSETFUNCTION;
      snes_f = ts_F;
      break;
    }
    case CTX_TSSETIJACOBIAN: {
      snes_ctx = CTX_SNESSETJACOBIAN;
      snes_B = ts_B;
      break;
    }
    default:
    break;
  }
  switch(snes_ctx) {
    case CTX_SNESSETFUNCTION: {
      ierr = VecAssemblyBegin(snes_f); CHKERRQ(ierr);
      ierr = VecAssemblyEnd(snes_f); CHKERRQ(ierr);
      ierr = get_dlambda(snes_x); CHKERRQ(ierr);
      ierr = calculate_lambda_int(); CHKERRQ(ierr);
    }
    break;
    case CTX_SNESSETJACOBIAN:
    ierr = MatAssemblyBegin(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
    ierr = MatAssemblyEnd(snes_B,MAT_FLUSH_ASSEMBLY); CHKERRQ(ierr);
    ierr = calculate_db(); CHKERRQ(ierr);
    break;
    default:
    break;
  }
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::FrontAreaArcLengthControl::operator()() {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  //get dlambda dof
  switch(snes_ctx) {
    case CTX_SNESSETFUNCTION: {
      //calculate residual for arc length row
      arcPtr->res_lambda = lambda_int - arcPtr->s;
      ierr = VecSetValue(snes_f,arcPtr->getPetscGlobalDofIdx(),arcPtr->res_lambda,INSERT_VALUES); CHKERRQ(ierr);
      PetscPrintf(PETSC_COMM_SELF,"\n");
      PetscPrintf(PETSC_COMM_SELF,"\t** Arc-Length residual:\n");
      PetscPrintf(PETSC_COMM_SELF,
        "\t  residual of arc-length control res_lambda = %6.4e crack area/f_lambda_int = %6.4e ( %6.4e )\n"
        ,arcPtr->res_lambda,aRea,aRea/aRea0);
    } break;
    case CTX_SNESSETJACOBIAN: {
      //calculate diagonal therm
      arcPtr->dIag = arcPtr->beta*sqrt(arcPtr->F_lambda2);
      ierr = MatSetValue(
        snes_B,arcPtr->getPetscGlobalDofIdx(),arcPtr->getPetscGlobalDofIdx(),1,INSERT_VALUES
      ); CHKERRQ(ierr);
    } break;
    default:
    break;
  }
  PetscFunctionReturn(0);
}
PetscErrorCode ConfigurationalFractureMechanics::FrontAreaArcLengthControl::postProcess() {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  switch(snes_ctx) {
    case CTX_SNESSETFUNCTION: {
      ParallelComm* pcomm = ParallelComm::get_pcomm(&mField.get_moab(),MYPCOMM_INDEX);
      double res_nrm2[6];
      Vec res_nrm2_vec;
      if(pcomm->rank()==0) {
        ierr = VecCreateMPIWithArray(PETSC_COMM_WORLD,1,6,6,res_nrm2,&res_nrm2_vec); CHKERRQ(ierr);
      } else {
        ierr = VecCreateMPIWithArray(PETSC_COMM_WORLD,1,0,6,res_nrm2,&res_nrm2_vec); CHKERRQ(ierr);
      }
      //
      Range CrackFrontEdges;
      ierr = mField.get_cubit_msId_entities_by_dimension(201,SIDESET,1,CrackFrontEdges,true); CHKERRQ(ierr);
      Range CrackFrontNodes;
      ErrorCode rval;
      rval = mField.get_moab().get_connectivity(CrackFrontEdges,CrackFrontNodes,true); CHKERRQ_MOAB(rval);
      //
      double *array;
      ierr = VecAssemblyBegin(snes_f); CHKERRQ(ierr);
      ierr = VecAssemblyEnd(snes_f); CHKERRQ(ierr);
      ierr = VecGetArray(snes_f,&array); CHKERRQ(ierr);
      ierr = VecZeroEntries(res_nrm2_vec); CHKERRQ(ierr);
      for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_LOCIDX_FOR_LOOP_(problemPtr,dof)) {
        if(dof->get()->getPart()!=pcomm->rank()) continue;
        double val = pow(array[dof->get()->getPetscLocalDofIdx()],2);
        if(dof->get()->getName() == "SPATIAL_POSITION") {
          ierr = VecSetValue(res_nrm2_vec,0,val,ADD_VALUES); CHKERRQ(ierr);
        }
        if(dof->get()->getName() == "MESH_NODE_POSITIONS") {
          if(find(CrackFrontNodes.begin(),CrackFrontNodes.end(),dof->get()->getEnt())!=CrackFrontNodes.end()) {
            ierr = VecSetValue(res_nrm2_vec,1,val,ADD_VALUES); CHKERRQ(ierr);
          } else {
            ierr = VecSetValue(res_nrm2_vec,2,val,ADD_VALUES); CHKERRQ(ierr);
          }
        }
        if(dof->get()->getName().compare(0,14,"LAMBDA_SURFACE") == 0) {
          ierr = VecSetValue(res_nrm2_vec,3,val,ADD_VALUES); CHKERRQ(ierr);
        }
        if(dof->get()->getName() == "LAMBDA_CRACK_SURFACE") {
          ierr = VecSetValue(res_nrm2_vec,4,val,ADD_VALUES); CHKERRQ(ierr);
        }
        if(dof->get()->getName() == "LAMBDA_CRACK_TANGENT_CONSTRAIN") {
          ierr = VecSetValue(res_nrm2_vec,5,val,ADD_VALUES); CHKERRQ(ierr);
        }
      }
      ierr = VecAssemblyBegin(res_nrm2_vec); CHKERRQ(ierr);
      ierr = VecAssemblyEnd(res_nrm2_vec); CHKERRQ(ierr);
      if(pcomm->rank()==0) {
        resSpatialNrm2 = sqrt(res_nrm2[0]);
        resCrackFrontNrm2 = sqrt(res_nrm2[1]);
        PetscPrintf(PETSC_COMM_SELF,"\t** Equilibrium Residuals:\n");
        PetscPrintf(PETSC_COMM_SELF,"\t  equilibrium of physical forces res_spatial_nrm2 = %6.4e\n",resSpatialNrm2);
        PetscPrintf(PETSC_COMM_SELF,"\t  equilibrium of material forces res_crack_front_nrm2 = %6.4e\n",resCrackFrontNrm2);
        PetscPrintf(PETSC_COMM_SELF,"\t** Mesh Quality Residuals:\n");
        PetscPrintf(PETSC_COMM_SELF,"\t  residual of mesh smoother res_mesh_smoother_nrm2 = %6.4e\n",sqrt(res_nrm2[2]));
        PetscPrintf(PETSC_COMM_SELF,"\t  residual of surface constraint res_surface_constrain_nrm2 = %6.4e\n",sqrt(res_nrm2[3]));
        PetscPrintf(PETSC_COMM_SELF,"\t  residual of crack surface constraint res_crack_surface_constrain_nrm2 = %6.4e\n",sqrt(res_nrm2[4]));
        PetscPrintf(PETSC_COMM_SELF,"\t  residual of crack front mesh smoother tangential constraint res_crack_front_tangent_constrain_nrm2 = %6.4e\n",sqrt(res_nrm2[5]));
        PetscPrintf(PETSC_COMM_SELF,"\n");
      }
      ierr = VecRestoreArray(snes_f,&array); CHKERRQ(ierr);
      ierr = VecDestroy(&res_nrm2_vec); CHKERRQ(ierr);
    } break;
    case CTX_SNESSETJACOBIAN: {
      ierr = VecGhostUpdateBegin(arcPtr->ghostDiag,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = VecGhostUpdateEnd(arcPtr->ghostDiag,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      //PetscPrintf(PETSC_COMM_WORLD,"\tdiag = %6.4e\n",arcPtr->diag);
      /*ierr = MatAssemblyBegin(snes_B,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
      ierr = MatAssemblyEnd(snes_B,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
      //Matrix View
      MatView(snes_B,PETSC_VIEWER_DRAW_WORLD);//PETSC_VIEWER_STDOUT_WORLD);
      std::string wait;
      std::cin >> wait;*/
    } break;
    default:
    break;
  }

  PetscFunctionReturn(0);
}

#include <impl/ConfigurationalFractureForDynamics.cpp>
