/* This file is part of MoFEM.
 * MoFEM is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * MoFEM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MoFEM. If not, see <http://www.gnu.org/licenses/>. */

#ifdef WITH_TETGEN

#include <tetgen.h>
#ifdef REAL
  #undef REAL
#endif

#endif

#include <moab/BVHTree.hpp>
#include <MoFEM.hpp>
using namespace MoFEM;
#include <BasicFiniteElements.hpp>
#include <NeoHookean.hpp>
#include <Hooke.hpp>

#include <CrackPropagation.hpp>

ErrorCode rval;
PetscErrorCode ierr;

static char help[] =
"Calculate crack release energy and crack propagation"
"\n\n";

int main(int argc, char *argv[]) {
  PetscInitialize(&argc,&argv,(char *)0,help);

  try {

    // Create mesh database

    PetscBool flg = PETSC_TRUE;
    char mesh_file_name[255];
    ierr = PetscOptionsGetString(PETSC_NULL,"","-my_file",mesh_file_name,255,&flg); CHKERRQ(ierr);

    moab::Core mb_instance;
    moab::Interface& moab = mb_instance;

    // Create separte MOAB communicator, it is duplciate of PETSc coomunicator.
    // NOTE That this should eliminate potential comminication problems between
    // MOAB and PETSC finctions.
    MPI_Comm moab_comm_world;
    MPI_Comm_dup(PETSC_COMM_WORLD,&moab_comm_world);
    ParallelComm* pcomm = ParallelComm::get_pcomm(&moab,MYPCOMM_INDEX);
    if(pcomm == NULL) pcomm =  new ParallelComm(&moab,moab_comm_world);

    if(flg) {
      const char *option;
      option = "";
      rval = moab.load_file(mesh_file_name,0,option); CHKERRQ_MOAB(rval);
    }

    // Create mofem database

    MoFEM::Core core(moab);
    MoFEM::Interface& m_field = core;

    // Create data structure for crack propagation
    CrackPropagation cp(m_field,3,2);

    // Configure meshes
    ierr = cp.getOptions(); CHKERRQ(ierr);
    if(cp.medFileName.size()>0) {
      ierr = cp.readMedFile(); CHKERRQ(ierr);
    }
    if(cp.medFileName.size()>0) {
      ierr = cp.readConfigFileOptions(); CHKERRQ(ierr);
    }
    if(flg != PETSC_TRUE && cp.medFileName.size() == 0) {
      SETERRQ(PETSC_COMM_SELF,1,"*** ERROR -my_file (MESH FILE NEEDED)");
    }
    if(cp.medFileName.size()>0) {
      ierr = m_field.rebuild_database(); CHKERRQ(ierr);
    }

    ProblemsManager *prb_mng_ptr;
    ierr = m_field.query_interface(prb_mng_ptr); CHKERRQ(ierr);
    m_field.get_basic_entity_data_ptr()->unSetDistributedMesh();

    // Register MOFEM discrete data managers
    {
     DMType dm_name = "MOFEM";
     ierr = DMRegister_MoFEM(dm_name); CHKERRQ(ierr);
    }
    {
      DMType dm_name = "MOFEM_MG";
      ierr = DMRegister_MGViaApproxOrders(dm_name); CHKERRQ(ierr);
    }

    // Get topological information about body skin, cracs, etc.
    // Seed BitRefLevels
    BitRefLevel bit0;
    bit0.set(0);
    // {
    //   Range ents;
    //   rval = moab.get_entities_by_handle(0,ents,true); CHKERRQ_MOAB(rval);
    //   ierr = m_field.seed_ref_level(ents,bit0,false); CHKERRQ(ierr);
    // }
    ierr = m_field.seed_ref_level_3D(0,bit0); CHKERRQ(ierr);
    ierr = cp.findBodySkin(bit0,102,1,false); CHKERRQ(ierr);
    ierr = cp.getCrack(bit0,200,201,1,false); CHKERRQ(ierr);
    if(cp.refAtCrackTip>0) {
      BitRefLevel bit1;
      bit1.set(1);
      ierr = cp.refineCrackTip(bit0,bit1,201,1,false); CHKERRQ(ierr);
      ierr = cp.findBodySkin(bit0,102,1,false); CHKERRQ(ierr);
      ierr = cp.getCrack(bit0,200,201,1,false); CHKERRQ(ierr);
    }

    BitRefLevel bit1;
    bit1.set(1);
    ierr = cp.splitFaces(bit0,bit1,200,1,false); CHKERRQ(ierr);
    ierr = cp.getElastcBlockTets(); CHKERRQ(ierr);

    ierr = cp.findBodySkin(bit1,102,1,false); CHKERRQ(ierr);
    ierr = cp.getCrack(bit1,200,201,1,false); CHKERRQ(ierr);
    ierr = cp.getFrontEdgesAndElements(bit1,0); CHKERRQ(ierr);

    // Partion mesh
    Range tets;
    tets.clear();
    ierr = m_field.get_entities_by_ref_level(bit1,BitRefLevel().set(),tets); CHKERRQ(ierr);
    if(cp.isPartitioned == PETSC_FALSE) {
      ierr = prb_mng_ptr->partitionMesh(
        tets.subset_by_type(MBTET),3,2,m_field.get_comm_size(),1
      ); CHKERRQ(ierr);
      cp.isPartitioned = PETSC_TRUE;
    }
    ierr = cp.resolveSharedBitRefLevel(bit1,1,false); CHKERRQ(ierr);

    // Build fields, elements and DMs for elastic analysis
    ierr = cp.buildArcLengthField(false); CHKERRQ(ierr);
    ierr = cp.buildElasticFields(bit1,BitRefLevel().set(),true,true); CHKERRQ(ierr);
    ierr = cp.declareElasticFE(bit1,BitRefLevel().set(),true,true); CHKERRQ(ierr);
    DM dm_elastic;
    ierr = cp.createElasticDM(&dm_elastic,"ELASTIC",bit1,BitRefLevel().set()); CHKERRQ(ierr);
    DM dm_bc;
    {
      MeshsetsManager *meshset_manager_ptr;
      ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
      if(meshset_manager_ptr->checkMeshset("ANALITICAL_DISP")) {
        ierr = cp.createBcDM(&dm_bc,"BC_PROBLEM",bit1,BitRefLevel().set()); CHKERRQ(ierr);
      }
    }

    // Set bit ref relvel for elements at crack front
    BitRefLevel bit2;
    bit2.set(2);
    ierr = cp.setCrackFrontBitLevel(bit1,bit2); CHKERRQ(ierr);
    ierr = cp.declareMaterialFE(bit2,BitRefLevel().set(),true,false); CHKERRQ(ierr);

    // Build fields, elements for crack front
    ierr = cp.buildSurfaceFields(bit2,BitRefLevel().set(),102,true,true,2,false); CHKERRQ(ierr);
    ierr = cp.declareSurfaceFE("SURFACE",bit2,BitRefLevel().set(),102,true,1,false); CHKERRQ(ierr);
    ierr = cp.buildCrackSurfaceFieldId(bit2,BitRefLevel().set(),200,true,true,2,false); CHKERRQ(ierr);
    ierr = cp.declareCrackSurfaceFE("CRACK_SURFACE",bit2,BitRefLevel().set(),200,true,1,false); CHKERRQ(ierr);

    // Create DMs for crack front problem
    std::vector<int> surface_ids;
    surface_ids.push_back(102);
    // surface_ids.push_back(200);
    std::vector<std::string> fe_surf_proj_list;
    fe_surf_proj_list.push_back("SURFACE");
    // fe_surf_proj_list.push_back("CRACK_SURFACE");

    DM dm_material;
    ierr = cp.createMaterialDM(
      &dm_material,"MATERIAL",bit2,BitRefLevel().set(),fe_surf_proj_list
    ); CHKERRQ(ierr);
    DM dm_material_forces;
    ierr = cp.createMaterialForcesDM(&dm_material_forces,dm_material,"MATERIAL_FORCES"); CHKERRQ(ierr);
    DM dm_surface_projection;
    ierr = cp.createSurfaceProjectionDM(
      &dm_surface_projection,dm_material,"SURFACE_PROJECTION",surface_ids,fe_surf_proj_list
    ); CHKERRQ(ierr);
    DM dm_crack_srf_area;
    ierr = cp.createCrackFrontAreaDM(
      &dm_crack_srf_area,dm_material,"CRACK_SURFACE_AREA"
    ); CHKERRQ(ierr);

    // Create vector of matrial forces and set opperators
    Vec q_material,f_material;
    ierr = DMCreateGlobalVector(dm_material_forces,&q_material); CHKERRQ(ierr);
    ierr = VecDuplicate(q_material,&f_material); CHKERRQ(ierr);
    // surface_ids.push_back(200);
    Vec f_material_proj,f_griffith,f_griffith_proj;
    ierr = VecDuplicate(f_material,&f_material_proj); CHKERRQ(ierr);
    ierr = VecDuplicate(f_material,&f_griffith); CHKERRQ(ierr);
    ierr = VecDuplicate(f_griffith,&f_griffith_proj); CHKERRQ(ierr);
    Vec f_lambda;
    {
      const MoFEM::MoFEMProblem *prb_ptr;
      ierr = DMMoFEMGetProblemPtr(dm_crack_srf_area,&prb_ptr); CHKERRQ(ierr);
      // Need to use directly MoFEM interface to create vector, matrix for non suqre matrix
      ierr = m_field.VecCreateGhost(prb_ptr->getName(),ROW,&f_lambda); CHKERRQ(ierr);
    }

    // Create matrix
    Mat m_elastic;
    ierr = DMCreateMatrix(dm_elastic,&m_elastic); CHKERRQ(ierr);
    ierr = cp.getArcLengthDof(); CHKERRQ(ierr);
    cp.arcLengthDof->getFieldData() = 1; // Set load factor to 1

    // Solve problem
    Vec q_elastic,f_elastic;
    ierr = DMCreateGlobalVector(dm_elastic,&q_elastic); CHKERRQ(ierr);
    ierr = VecDuplicate(q_elastic,&f_elastic); CHKERRQ(ierr);
    ierr = cp.setMatrialPositionFromCoords(); CHKERRQ(ierr);
    ierr = cp.setSpatialPositionFromCoords(); CHKERRQ(ierr);
    ierr = cp.setSingularDofs("SPATIAL_POSITION"); CHKERRQ(ierr);

    if(cp.geometryOrder>1) {
      ierr = cp.setSingularElementMatrialPositions(); CHKERRQ(ierr);
    }

    // Set opperators for elasticity abalysis
    ierr = cp.assembleElasticDM(
      dm_elastic,m_elastic,q_elastic,f_elastic,1,false
    ); CHKERRQ(ierr);

    // Set opperators for material forxes
    ierr = cp.assembleMaterialForcesDM(dm_material_forces); CHKERRQ(ierr);

    // Calculate projection matrices
    ierr = cp.calculateSurfaceProjectionMatrix(
      dm_surface_projection,dm_material_forces,surface_ids,1,false
    ); CHKERRQ(ierr);
    ierr = cp.calculateFrontProjectionMatrix(
      dm_crack_srf_area,dm_material_forces,1,false
    ); CHKERRQ(ierr);

    // Calulate griffith force vector (material resistance)
    ierr = cp.calculateGriffithForce(
      dm_material_forces,cp.gC?cp.gC:1,f_griffith,1,false
    ); CHKERRQ(ierr);
    ierr = cp.projectGriffithForce(
      dm_material_forces,f_griffith,f_griffith_proj,1,false
    ); CHKERRQ(ierr);

    SNES snes;
    ierr = SNESCreate(PETSC_COMM_WORLD,&snes); CHKERRQ(ierr);
    {
      for(int ss = 0;ss!=cp.getNbLoadSteps();ss++) {
        const double load_factor = (ss+1)*cp.getLoadScale();
        PetscPrintf(PETSC_COMM_WORLD,"Step %d load factor %4.3f\n",ss,load_factor);
        cp.arcLengthDof->getFieldData() = load_factor;
        ierr = cp.solveElasticDM(
          dm_elastic,snes,m_elastic,q_elastic,f_elastic,(ss==0)
        ); CHKERRQ(ierr);
        ierr = cp.calculateElasticEnergy(dm_elastic); CHKERRQ(ierr);
        ierr = cp.calculateMaterialFocesDM(dm_material_forces,q_material,f_material,2,false); CHKERRQ(ierr);
        ierr = DMoFEMMeshToLocalVector(
          dm_material_forces,q_material,INSERT_VALUES,SCATTER_FORWARD
        ); CHKERRQ(ierr);
        ierr = cp.projectMaterialForcesDM(
          dm_material_forces,f_material,f_material_proj,1,false
        ); CHKERRQ(ierr);
        ierr = cp.calculateReleaseEneregy(
          dm_crack_srf_area,f_material_proj,f_griffith_proj,f_lambda,cp.gC,2,true
        ); CHKERRQ(ierr);
      }
    }
    ierr = SNESDestroy(&snes); CHKERRQ(ierr);

    ierr = cp.postProcessElasticDM(dm_elastic,0); CHKERRQ(ierr);
    ierr = cp.savePositionsOnCrackFrontDM(dm_elastic,q_elastic,2,false); CHKERRQ(ierr);
    if(cp.geometryOrder>1) {
      ierr = cp.unsetSingularElementMatrialPositions(); CHKERRQ(ierr);
    }

    ierr = VecDestroy(&q_elastic); CHKERRQ(ierr);
    ierr = VecDestroy(&f_elastic); CHKERRQ(ierr);
    ierr = MatDestroy(&m_elastic); CHKERRQ(ierr);

    {
      MeshsetsManager *meshset_manager_ptr;
      ierr = m_field.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
      if(meshset_manager_ptr->checkMeshset("ANALITICAL_DISP")) {
        ierr = DMDestroy(&dm_bc); CHKERRQ(ierr);
      }
    }
    ierr = DMDestroy(&dm_elastic); CHKERRQ(ierr);

    ierr = VecDestroy(&f_lambda); CHKERRQ(ierr);
    ierr = VecDestroy(&q_material); CHKERRQ(ierr);
    ierr = VecDestroy(&f_material); CHKERRQ(ierr);
    ierr = VecDestroy(&f_material_proj); CHKERRQ(ierr);
    ierr = VecDestroy(&f_griffith); CHKERRQ(ierr);
    ierr = VecDestroy(&f_griffith_proj); CHKERRQ(ierr);

    ierr = DMDestroy(&dm_crack_srf_area); CHKERRQ(ierr);
    ierr = DMDestroy(&dm_surface_projection); CHKERRQ(ierr);
    ierr = DMDestroy(&dm_material_forces); CHKERRQ(ierr);
    ierr = DMDestroy(&dm_material); CHKERRQ(ierr);

    PetscSynchronizedPrintf(PETSC_COMM_WORLD,"Done rank = %d\n",m_field.get_comm_rank());
    PetscSynchronizedFlush(PETSC_COMM_WORLD,PETSC_STDOUT);
    // PetscBarrier(PETSC_NULL);
    // ierr = cp.writeCrackFont(0); CHKERRQ(ierr);
    // PetscBarrier(PETSC_NULL);

    MPI_Comm_free(&moab_comm_world);

  } catch (MoFEMException const &e) {
    SETERRQ(PETSC_COMM_SELF,e.errorCode,e.errorMessage);
  }

  PetscFinalize();
}
