/** \file CrackPropagation.cpp
*/

/* This file is part of MoFEM.
 * MoFEM is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * MoFEM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MoFEM. If not, see <http://www.gnu.org/licenses/>. */

#ifdef WITH_TETGEN

#include <tetgen.h>
#ifdef REAL
#undef REAL
#endif

#endif

#define NDEBUG 1
#define BOOST_UBLAS_NDEBUG 1

// #include <moab/BVHTree.hpp>
#include <moab/SpatialLocator.hpp>
#include <moab/ElemEvaluator.hpp>

#include <MoFEM.hpp>
using namespace MoFEM;

#include <cholesky.hpp>

#include <BasicFiniteElements.hpp>
#include <Hooke.hpp>
#include <GriffithForceElement.hpp>
#include <AnalyticalFun.hpp>

// TODO implementation in ComplexConstArea is obsolete, not follows naming convention
// and code is unnecessarly complicated and in places diffiult to follow.
#include <ComplexConstArea.hpp>
#include <ConstantArea.hpp>

extern "C" {
  #include <phg-quadrule/quad.h>
}

#include <NeoHookean.hpp>
#include <Hooke.hpp>

#include <CrackPropagation.hpp>

namespace po = boost::program_options;

bool CrackPropagation::setSingularCoordinatesSet = false;

CrackPropagation::CrackPropagation(MoFEM::Interface &mField,const int approx_order,int geometry_order):
mField(mField),
approxOrder(approx_order),
geometryOrder(geometry_order),
gC(0),
isPartitioned(PETSC_FALSE),
refAtCrackTip(0),
refOrderAtTip(0),
residualStressBlock(-1),
rhoAtGaussPt(new VectorDouble),
rhoGradAtGaussPt(new MatrixDouble),
rHo0(1),
nBone(0) {
}

CrackPropagation::~CrackPropagation() {
}

enum materials {
  HOOKE,
  KIRCHOFF,
  NEOHOOKEAN,
  BONEHOOKE,
  BONENEOHOOKEAN,
  LASTOP
};

const char *materials_list[] = {
  "HOOKE",
  "KIRCHOFF",
  "NEOHOOKEAN",
  "BONEHOOKE",
  "BONENEOHOOKEAN"
};

template<>
PetscErrorCode CrackPropagation::BoneHooke<adouble>::setUserActiveVariables(int &nb_active_variables) {
  PetscFunctionBegin;
  this->rHo <<= (*rhoAtGaussPts)[this->gG];
  nbActiveVariables0 = nb_active_variables;
  nb_active_variables++;
  PetscFunctionReturn(0);
}

template<>
PetscErrorCode CrackPropagation::BoneHooke<double>::setUserActiveVariables(int &nb_active_variables) {
  PetscFunctionBegin;
  this->rHo = (*rhoAtGaussPts)[this->gG];
  nbActiveVariables0 = nb_active_variables;
  nb_active_variables++;
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::getOptions() {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscOptionsBegin(mField.get_comm(),"","Fracture options","none"); CHKERRQ(ierr);
  {
    ierr = PetscOptionsInt(
      "-my_order",
      "approximation order","",
      approxOrder,
      &approxOrder,
      PETSC_NULL
    ); CHKERRQ(ierr);
    ierr = PetscOptionsInt(
      "-my_geom_order",
      "approximation geometry order","",
      geometryOrder,
      &geometryOrder,
      PETSC_NULL
    ); CHKERRQ(ierr);
    ierr = PetscOptionsScalar(
      "-my_gc",
      "release energy","",
      gC,&gC,PETSC_NULL
    ); CHKERRQ(ierr);
    ierr = PetscOptionsBool(
      "-my_is_partitioned",
      "true if mesh is partitioned","",
      isPartitioned,&isPartitioned,NULL
    ); CHKERRQ(ierr);
    ierr = PetscOptionsInt(
      "-my_ref",
      "crack tip refinement level","",
      refAtCrackTip,
      &refAtCrackTip,
      PETSC_NULL
    ); CHKERRQ(ierr);
    ierr = PetscOptionsInt(
      "-my_ref_order",
      "crack tip refinement level","",
      refOrderAtTip,
      &refOrderAtTip,
      PETSC_NULL
    ); CHKERRQ(ierr);
    {
      PetscBool flg;
      char file_name[255] = "mwls.med";
      ierr = PetscOptionsString(
        "-my_mwls_approx_file",
        "file with data form med file (code-ster) with radiation data","",
        file_name,file_name,255,&flg
      ); CHKERRQ(ierr);
      if(flg==PETSC_TRUE) {
       mwlsApproxFile = std::string(file_name);
      }
    }
    ierr = PetscOptionsInt(
      "-my_residual_stress_block",
      "block to which residual stress is applied","",
      residualStressBlock,
      &residualStressBlock,
      PETSC_NULL
    );
    {
      PetscBool flg;
      char file_name[255] = "file.med";
      ierr = PetscOptionsString(
        "-my_med_file",
        "read med file","",
        file_name,file_name,255,&flg
      ); CHKERRQ(ierr);
      if(flg==PETSC_TRUE) {
        medFileName = std::string(file_name);
      }
    }
    {
      PetscBool flg;
      char file_name[255] = "add_cubit_meshsets.in";
      ierr = PetscOptionsString(
        "-meshsets_config",
        "meshsets config  file name","",
        file_name,file_name,255,&flg
      ); CHKERRQ(ierr);
      if(flg==PETSC_TRUE) {
        ifstream f(file_name);
        if(!f.good()) {
          SETERRQ1(
            mField.get_comm(),
            MOFEM_DATA_INCONSISTENCY,
            "File configuring meshsets ( %s ) can not be open\n",
            file_name
          );
        }
        configFile = file_name;
      }
    }
    {
      PetscBool flg;
      defaultMaterial = HOOKE;
      ierr = PetscOptionsEList(
        "-material","Material type","",materials_list,LASTOP,materials_list[HOOKE],&defaultMaterial,&flg
      );
    }
    // Set paramaters with bone density like material
    {
      ierr = PetscOptionsScalar(
        "-my_rho0",
        "release energy","",
        rHo0,&rHo0,PETSC_NULL
      ); CHKERRQ(ierr);
      ierr = PetscOptionsScalar(
        "-my_n_bone",
        "release energy","",
        nBone,&nBone,PETSC_NULL
      ); CHKERRQ(ierr);
    }
    {
      nbLoadSteps = 1;
      ierr = PetscOptionsInt(
        "-nb_load_steps",
        "number of load steps","",
        nbLoadSteps,
        &nbLoadSteps,
        PETSC_NULL
      );
      loadScale = 1;
      ierr = PetscOptionsScalar(
        "-load_scale",
        "load scale","",
        loadScale,&loadScale,PETSC_NULL
      ); CHKERRQ(ierr);
    }
  }
  ierr = PetscOptionsEnd(); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::readMedFile() {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  PetscFunctionBegin;

  #ifdef WITH_MED
  MedInterface *med_interface_ptr;
  ierr = mField.query_interface(med_interface_ptr); CHKERRQ(ierr);
  ierr = med_interface_ptr->readMed(medFileName); CHKERRQ(ierr);

  ierr = med_interface_ptr->medGetFieldNames(medFileName); CHKERRQ(ierr);
  for(
    std::map<std::string,MedInterface::FieldData>::iterator fit =
    med_interface_ptr->fieldNames.begin();
    fit!=med_interface_ptr->fieldNames.end();
    fit++
  ) {
    ierr = med_interface_ptr->readFields(
      med_interface_ptr->medFileName,fit->first,false,0
    ); CHKERRQ(ierr);
  }
  Range tets;
  rval = mField.get_moab().get_entities_by_type(0,MBTET,tets,true); CHKERRQ_MOAB(rval);
  Range edges;
  rval = mField.get_moab().get_adjacencies(tets,1,true,edges,moab::Interface::UNION); CHKERRQ_MOAB(rval);
  Range faces;
  rval = mField.get_moab().get_adjacencies(tets,2,true,faces,moab::Interface::UNION); CHKERRQ_MOAB(rval);
  #endif // WITH_MED

  #ifndef WITH_MED
  SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"MED file readed not installed");
  #endif

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::readConfigFileOptions() {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  MeshsetsManager *meshsets_interface_ptr;
  ierr = mField.query_interface(meshsets_interface_ptr); CHKERRQ(ierr);
  // Read additional adata
  meshsets_interface_ptr->getConfigFileOptionsPtr() =
  boost::shared_ptr<boost::program_options::options_description>(
    new boost::program_options::options_description()
  );
  meshsets_interface_ptr->getConfigFileOptionsPtr()->add_options()
  (
    "general.radiation_data_file",
    po::value<string>(&mwlsApproxFile)->default_value("mwls.med"),
    "MED field with stresses"
  )(
    "general.residual_stress_block",
    po::value<int>(&residualStressBlock)->default_value(-1),
    "Block where residual stresses are applied"
  )(
    "general.add_order_at_crack_front",
    po::value<int>(&refOrderAtTip)->default_value(0),
    "Add to rder at crack top"
  )(
    "refine at crack tip",
    po::value<int>(&refAtCrackTip)->default_value(0),
    "Refine at crack top"
  );

  ierr = meshsets_interface_ptr->setMeshsetFromFile(configFile,false); CHKERRQ(ierr);
  // for(
  //   CubitMeshSet_multiIndex::iterator cit = meshsets_interface_ptr->getBegin();
  //   cit!=meshsets_interface_ptr->getEnd(); cit++
  // ) {
  //   std::cout << *cit << endl;
  // }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::getElastcBlockTets() {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
    MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
  // rval = moab.get_entities_by_type(0,MBTET,elasticBlockTets); CHKERRQ_MOAB(rval);
  elasticBlockTets.clear();
  for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,BLOCKSET|MAT_ELASTICSET,it)) {
    EntityHandle meshset = it->getMeshset();
    Range tets;
    rval = mField.get_moab().get_entities_by_type(meshset,MBTET,tets,true); CHKERRQ_MOAB(rval);
    elasticBlockTets.merge(tets);
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::resolveShared(
  const Range &tets,Range &proc_ents,const int verb,const bool debug
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  ParallelComm* pcomm = ParallelComm::get_pcomm(&moab,MYPCOMM_INDEX);
  PetscFunctionBegin;

  Tag part_tag = pcomm->part_tag();
  Range tagged_sets;
  rval = mField.get_moab().get_entities_by_type_and_tag(
    0,MBENTITYSET,&part_tag,NULL,1,tagged_sets,moab::Interface::UNION
  ); CHKERRQ_MOAB(rval);
  for(Range::iterator mit = tagged_sets.begin();mit!=tagged_sets.end();mit++) {
    int part;
    rval = moab.tag_get_data(part_tag,&*mit,1,&part); CHKERRQ_MOAB(rval);
    if(part==mField.get_comm_rank()) {
      rval = moab.get_entities_by_type(*mit,MBTET,proc_ents,true); CHKERRQ_MOAB(rval);
    }
  }
  proc_ents = intersect(proc_ents,tets);

  // EntityHandle meshset;
  // rval = moab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
  // rval = moab.add_entities(meshset,tets.subset_by_dimension(3)); CHKERRQ_MOAB(rval);

  Skinner skin(&mField.get_moab());
  Range tets_skin;
  rval = skin.find_skin(0,tets.subset_by_dimension(3),false,tets_skin);
  // if(rval != MB_SUCCESS) {
  //   cerr << "Rank " << mField.get_comm_rank() << endl;
  //   cerr << tets << endl;
  //   cerr << tets_skin << endl;
  // }
  // CHKERRQ_MOAB(rval);
  Range proc_ents_skin[4];
  proc_ents_skin[3] = proc_ents;
  rval = skin.find_skin(0,proc_ents,false,proc_ents_skin[2]); CHKERRQ_MOAB(rval);
  proc_ents_skin[2] = subtract(proc_ents_skin[2],tets_skin);
  rval = moab.get_adjacencies(
    proc_ents_skin[2],1,false,proc_ents_skin[1],moab::Interface::UNION
  ); CHKERRQ_MOAB(rval);
  rval = moab.get_connectivity(proc_ents_skin[1],proc_ents_skin[0],true); CHKERRQ_MOAB(rval);
  proc_ents_skin[0].merge(crackFrontNodes);
  proc_ents_skin[1].merge(crackFront);

  if(mField.check_field("ARC_LENGTH_LAMBDA")) {
    EntityHandle lambda_meshset = mField.get_field_meshset("ARC_LENGTH_LAMBDA");
    Range arc_length_vertices;
    rval = moab.get_entities_by_type(lambda_meshset,MBVERTEX,arc_length_vertices,true); CHKERRQ_MOAB(rval);
    if(arc_length_vertices.size()!=1) {
      SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Should be one vertex in <ARC_LENGTH_LAMBDA> field");
    }
    arcLengthVertex = arc_length_vertices[0];
    proc_ents_skin[0].merge(arc_length_vertices);
  } else {
    const double coords[] = {0,0,0};
    rval = moab.create_vertex(coords,arcLengthVertex); CHKERRQ_MOAB(rval);
    proc_ents_skin[0].insert(arcLengthVertex);
  }

  if(debug) {
    std::ostringstream file_skin;
    file_skin << "out_skin_" << mField.get_comm_rank() << ".vtk";
    EntityHandle meshset_skin;
    rval = moab.create_meshset(MESHSET_SET,meshset_skin); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_skin,proc_ents_skin[2]); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_skin,proc_ents_skin[1]); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_skin,proc_ents_skin[0]); CHKERRQ_MOAB(rval);
    rval = moab.write_file(file_skin.str().c_str(),"VTK","",&meshset_skin,1); CHKERRQ_MOAB(rval);
    std::ostringstream file_owned;
    file_owned << "out_owned_" << mField.get_comm_rank() << ".vtk";
    EntityHandle meshset_owned;
    rval = moab.create_meshset(MESHSET_SET,meshset_owned); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_owned,proc_ents); CHKERRQ_MOAB(rval);
    rval = moab.write_file(file_owned.str().c_str(),"VTK","",&meshset_owned,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_owned,1); CHKERRQ_MOAB(rval);
  }

  rval = pcomm->resolve_shared_ents(0,proc_ents,3,-1,proc_ents_skin); CHKERRQ_MOAB(rval);

  if(debug) {
    Range shared_ents;
    // Get entities shared with all other processors
    rval = pcomm->get_shared_entities(-1,shared_ents);CHKERRQ_MOAB(rval);
    std::ostringstream file_shared_owned;
    file_shared_owned << "out_shared_owned_" << mField.get_comm_rank() << ".vtk";
    EntityHandle meshset_shared_owned;
    rval = moab.create_meshset(MESHSET_SET,meshset_shared_owned); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_shared_owned,shared_ents); CHKERRQ_MOAB(rval);
    rval = moab.write_file(file_shared_owned.str().c_str(),"VTK","",&meshset_shared_owned,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_shared_owned,1); CHKERRQ_MOAB(rval);
  }

  const RefEntity_multiIndex *refined_ents_ptr;
  ierr = mField.get_ref_ents(&refined_ents_ptr); CHKERRQ(ierr);
  for(
    RefEntity_multiIndex::iterator
    eit = refined_ents_ptr->begin();eit!=refined_ents_ptr->end();eit++
  ) {
    const_cast<RefEntity_multiIndex*>(refined_ents_ptr)->modify(eit,Entity_update_pcomm_data<RefEntity>());
    const_cast<RefEntity_multiIndex*>(refined_ents_ptr)->modify(eit,Entity_update_part_proc<RefEntity>());
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::resolveSharedBlockTest(const int verb,const bool debug) {
  return resolveShared(elasticBlockTets,elasticBlockProcTets,verb,debug);
}

PetscErrorCode CrackPropagation::resolveSharedBitRefLevel(
  const BitRefLevel bit,const int verb,const bool debug
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  moab::Interface &moab = mField.get_moab();
  PetscFunctionBegin;
  bitEnts.clear();
  bitProcEnts.clear();
  ierr = mField.get_entities_by_ref_level(bit,BitRefLevel().set(),bitEnts); CHKERRQ(ierr);
  if(debug) {
    std::ostringstream file_owned;
    file_owned << "out_bit_" << mField.get_comm_rank() << ".vtk";
    EntityHandle meshset;
    rval = moab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset,bitEnts.subset_by_dimension(3)); CHKERRQ_MOAB(rval);
    rval = moab.write_file(file_owned.str().c_str(),"VTK","",&meshset,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
  }
  ierr = resolveShared(bitEnts.subset_by_type(MBTET),bitProcEnts,verb,debug); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::findBodySkin(
  const BitRefLevel bit,const int id,const int verb,const bool debug
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  moab::Interface &moab = mField.get_moab();
  MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
  Range level_tris;
  ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  // Find body skin
  if(!meshset_manager_ptr->checkMeshset(id,SIDESET)) {
    Range tets;
    ierr = mField.get_entities_by_type_and_ref_level(
      bit,BitRefLevel().set(),MBTET,tets
    ); CHKERRQ(ierr);
    Range tets_faces;
    rval = moab.get_adjacencies(
      tets,2,true,tets_faces,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    Skinner skin(&mField.get_moab());
    rval = skin.find_skin(0,tets,false,bodySkin); CHKERRQ_MOAB(rval);
    bodySkin = intersect(level_tris,bodySkin);
    ierr = meshset_manager_ptr->addMeshset(SIDESET,id); CHKERRQ(ierr);
    EntityHandle meshset;
    ierr = meshset_manager_ptr->getMeshset(id,SIDESET,meshset); CHKERRQ(ierr);
    rval = moab.add_entities(meshset,bodySkin); CHKERRQ(ierr);
  }
  bodySkin.clear();
  ierr = meshset_manager_ptr->getEntitiesByDimension(id,SIDESET,2,bodySkin,true); CHKERRQ(ierr);
  bodySkin = intersect(bodySkin,level_tris);
  if(verb>1) {
    if(mField.get_comm_rank()==0) {
      cerr << bodySkin << endl;
    }
  }
  if(debug) {
    if(mField.get_comm_rank()==0) {
      EntityHandle meshset;
      rval = moab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
      rval = moab.add_entities(meshset,bodySkin); CHKERRQ_MOAB(rval);
      rval = moab.write_file("out_body_skin.vtk","VTK","",&meshset,1); CHKERRQ_MOAB(rval);
      rval = moab.delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
    }
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::getCrack(
  const BitRefLevel bit,
  const int crack_id,
  const int front_id,
  const int verb,
  const bool debug
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  moab::Interface &moab = mField.get_moab();
  MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
  crackFaces.clear();
  ierr = meshset_manager_ptr->getEntitiesByDimension(crack_id,SIDESET,2,crackFaces,true); CHKERRQ(ierr);
  Range level_tris;
  ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  crackFaces = intersect(level_tris,crackFaces);
  // Find crack fornt
  if(!meshset_manager_ptr->checkMeshset(front_id,SIDESET)) {
    Range skin_faces_edges;
    rval = moab.get_adjacencies(
      bodySkin,1,true,skin_faces_edges,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    Skinner skin(&mField.get_moab());
    rval = skin.find_skin(0,crackFaces,false,crackFront); CHKERRQ_MOAB(rval);
    crackFront = subtract(crackFront,skin_faces_edges);
    ierr = meshset_manager_ptr->addMeshset(SIDESET,front_id); CHKERRQ(ierr);
    EntityHandle meshset;
    ierr = meshset_manager_ptr->getMeshset(front_id,SIDESET,meshset); CHKERRQ(ierr);
    rval = moab.add_entities(meshset,crackFront); CHKERRQ(ierr);
  }
  crackFront.clear();
  ierr = meshset_manager_ptr->getEntitiesByDimension(
    front_id,SIDESET,1,crackFront,true
  ); CHKERRQ(ierr);
  if(verb>1) {
    if(mField.get_comm_rank()==0) {
      cerr << crackFront << endl;
    }
  }
  if(debug) {
    if(mField.get_comm_rank()==0) {
      EntityHandle meshset;
      rval = moab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
      // Range skin_faces_edges;
      // rval = moab.get_adjacencies(
      //   bodySkin,1,false,skin_faces_edges,moab::Interface::UNION
      // ); CHKERRQ_MOAB(rval);
      // rval = moab.add_entities(meshset,skin_faces_edges); CHKERRQ_MOAB(rval);
      rval = moab.add_entities(meshset,crackFront); CHKERRQ_MOAB(rval);
      rval = moab.add_entities(meshset,crackFaces); CHKERRQ_MOAB(rval);
      rval = moab.write_file("out_front.vtk","VTK","",&meshset,1); CHKERRQ_MOAB(rval);
      rval = moab.delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
    }
  }

  // FE Couple nodes on two sides of crack
  std::vector<int> sides(crackFaces.size());
  Tag th_interface_side;
  const int def_side[] = {0};
  rval = mField.get_moab().tag_get_handle(
    "INTERFACE_SIDE",1,MB_TYPE_INTEGER,
    th_interface_side,MB_TAG_CREAT|MB_TAG_SPARSE,def_side
  ); CHKERRQ_MOAB(rval);
  rval = mField.get_moab().tag_get_data(
    th_interface_side,crackFaces,&*sides.begin()
  ); CHKERRQ_MOAB(rval);
  oneSideCrackFaces.clear();
  int nn = 0;
  for(Range::iterator fit = crackFaces.begin();fit!=crackFaces.end();fit++,nn++) {
    if(sides[nn] == 0) {
      oneSideCrackFaces.insert(*fit);
    }
  }

  if(verb>1) {
    ierr = PetscPrintf(mField.get_comm(),"number of Crack Surfaces Faces = %d\n",crackFaces.size()); CHKERRQ(ierr);
    ierr = PetscPrintf(
      mField.get_comm(),"number of Crack Surfaces Faces On One Side = %d\n",oneSideCrackFaces.size()
    ); CHKERRQ(ierr);
    ierr = PetscPrintf(mField.get_comm(),"number of Crack Front Edges = %d\n",crackFront.size()); CHKERRQ(ierr);
  }

  if(debug) {
    if(mField.get_comm_rank()==0) {
      EntityHandle meshset;
      rval = moab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
      // Range skin_faces_edges;
      // rval = moab.get_adjacencies(
      //   bodySkin,1,false,skin_faces_edges,moab::Interface::UNION
      // ); CHKERRQ_MOAB(rval);
      // rval = moab.add_entities(meshset,skin_faces_edges); CHKERRQ_MOAB(rval);
      rval = moab.add_entities(meshset,oneSideCrackFaces); CHKERRQ_MOAB(rval);
      rval = moab.write_file("out_one_side.vtk","VTK","",&meshset,1); CHKERRQ_MOAB(rval);
      rval = moab.delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
    }
  }


  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::crackCut(
  const BitRefLevel bit1,
  const BitRefLevel bit2,
  const int verb,
  const bool debug
) {
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  PetscErrorCode ierr;
  CutMeshInterface *cut_mesh;
  PetscFunctionBegin;
  ierr = mField.query_interface(cut_mesh); CHKERRQ(ierr);
  ierr = cut_mesh->setSurface(crackFaces); CHKERRQ(ierr);
  ierr = cut_mesh->setVolume(elasticBlockTets); CHKERRQ(ierr);
  // build tree
  ierr = cut_mesh->buildTree(); CHKERRQ(ierr);

  if(verb>1) {
    cerr << crackFaces << endl;
    cerr << elasticBlockTets << endl;
  }

  // // find edges to cut
  ierr = cut_mesh->findEdgesToCut(0); CHKERRQ(ierr);
  if(verb>1) {
    cerr << cut_mesh->getVerticesOnSurface() << endl;
  }
  ierr = cut_mesh->cutEdgesInMiddle(bit1); CHKERRQ(ierr);
  ierr = cut_mesh->moveMidNodesOnCutEdges(); CHKERRQ(ierr);
  ierr = cut_mesh->findEdgesToTrim(NULL,1e-2); CHKERRQ(ierr);
  ierr = cut_mesh->trimEdgesInTheMiddle(bit2); CHKERRQ(ierr);
  ierr = cut_mesh->moveMidNodesOnTrimedEdges(); CHKERRQ(ierr);

  if(debug) {

    EntityHandle meshset_cut_edges;
    rval = moab.create_meshset(MESHSET_SET,meshset_cut_edges); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_cut_edges,cut_mesh->getCutEdges()); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_cut_edges.vtk","VTK","",&meshset_cut_edges,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_cut_edges,1); CHKERRQ_MOAB(rval);

    EntityHandle meshset_cut_vol;
    rval = moab.create_meshset(MESHSET_SET,meshset_cut_vol); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_cut_vol,cut_mesh->getCutVolumes()); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_cut_vol.vtk","VTK","",&meshset_cut_vol,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_cut_vol,1); CHKERRQ_MOAB(rval);

    EntityHandle meshset_cut_tets;
    rval = moab.create_meshset(MESHSET_SET,meshset_cut_tets); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_cut_tets,cut_mesh->getNewCutVolumes()); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_cut_tets.vtk","VTK","",&meshset_cut_tets,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_cut_tets,1); CHKERRQ_MOAB(rval);

    EntityHandle meshset_new_faces;
    rval = moab.create_meshset(MESHSET_SET,meshset_new_faces); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_new_faces,cut_mesh->getNewCutSurfaces()); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_cut_new_faces.vtk","VTK","",&meshset_new_faces,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_new_faces,1); CHKERRQ_MOAB(rval);

    EntityHandle meshset_trim_edges;
    rval = moab.create_meshset(MESHSET_SET,meshset_trim_edges); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_trim_edges,cut_mesh->getTrimEdges()); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_trim_edges.vtk","VTK","",&meshset_trim_edges,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_trim_edges,1); CHKERRQ_MOAB(rval);

    EntityHandle meshset_outside_edges;
    rval = moab.create_meshset(MESHSET_SET,meshset_outside_edges); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_outside_edges,cut_mesh->getOutsideEdges()); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_outside_edges.vtk","VTK","",&meshset_outside_edges,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_outside_edges,1); CHKERRQ_MOAB(rval);

    EntityHandle meshset_trim_new_surface;
    rval = moab.create_meshset(MESHSET_SET,meshset_trim_new_surface); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_trim_new_surface,cut_mesh->getNewTrimSurfaces()); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_trim_new_surface.vtk","VTK","",&meshset_trim_new_surface,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_trim_new_surface,1); CHKERRQ_MOAB(rval);

  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::splitFaces(
  BitRefLevel bit,
  BitRefLevel bit_split,
  const int crack_id,
  const int verb,
  const bool debug
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  PrismInterface *prism_interface;
  MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(prism_interface); CHKERRQ(ierr);
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Tag th_interface_side;
  rval = mField.get_moab().tag_get_handle("INTERFACE_SIDE",th_interface_side);
  if(rval == MB_SUCCESS) {
    rval = mField.get_moab().tag_delete(th_interface_side); CHKERRQ_MOAB(rval);
  }

  EntityHandle bit_meshset;
  rval = mField.get_moab().create_meshset(MESHSET_SET|MESHSET_TRACK_OWNER,bit_meshset); CHKERRQ_MOAB(rval);
  ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBTET,bit_meshset); CHKERRQ(ierr);

  EntityHandle meshset_interface;
  ierr = meshset_manager_ptr->getMeshset(crack_id,SIDESET,meshset_interface); CHKERRQ(ierr);
  ierr = prism_interface->getSides(meshset_interface,bit,true,0); CHKERRQ(ierr);
  ierr = prism_interface->splitSides(
    bit_meshset,bit_split,meshset_interface,true,true
  ); CHKERRQ(ierr);

  //add refined ent to cubit meshsets
  for(_IT_CUBITMESHSETS_FOR_LOOP_(mField,cubit_it)) {
    EntityHandle cubit_meshset = cubit_it->meshset;
    ierr = mField.update_meshset_by_entities_children(cubit_meshset,bit_split,cubit_meshset,MBVERTEX,true); CHKERRQ(ierr);
    ierr = mField.update_meshset_by_entities_children(cubit_meshset,bit_split,cubit_meshset,MBEDGE,true); CHKERRQ(ierr);
    ierr = mField.update_meshset_by_entities_children(cubit_meshset,bit_split,cubit_meshset,MBTRI,true); CHKERRQ(ierr);
    ierr = mField.update_meshset_by_entities_children(cubit_meshset,bit_split,cubit_meshset,MBTET,true); CHKERRQ(ierr);
  }

  if(isPartitioned) {
    ParallelComm* pcomm = ParallelComm::get_pcomm(&moab,MYPCOMM_INDEX);
    Tag part_tag = pcomm->part_tag();
    Range tagged_sets;
    rval = mField.get_moab().get_entities_by_type_and_tag(
      0,MBENTITYSET,&part_tag,NULL,1,tagged_sets,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    for(Range::iterator mit = tagged_sets.begin();mit!=tagged_sets.end();mit++) {
      int part;
      rval = moab.tag_get_data(part_tag,&*mit,1,&part); CHKERRQ_MOAB(rval);
      ierr = mField.update_meshset_by_entities_children(*mit,bit_split,*mit,MBTET,true); CHKERRQ(ierr);
      Range ents3d;
      rval = moab.get_entities_by_dimension(*mit,3,ents3d,true); CHKERRQ_MOAB(rval);
      for(Range::iterator eit = ents3d.begin();eit!=ents3d.end();eit++) {
        rval = moab.tag_set_data(part_tag,&*eit,1,&part); CHKERRQ_MOAB(rval);
      }
      for(int dd = 0;dd!=3;dd++) {
        Range ents;
        rval = moab.get_adjacencies(ents3d,dd,false,ents,moab::Interface::UNION); CHKERRQ_MOAB(rval);
        for(Range::iterator eit = ents.begin();eit!=ents.end();eit++) {
          rval = moab.tag_set_data(part_tag,&*eit,1,&part); CHKERRQ_MOAB(rval);
        }
      }
    }
  }

  if(debug) {
    Range tets_level;
    ierr = mField.get_entities_by_type_and_ref_level(
      bit_split,BitRefLevel().set(),MBTET,tets_level
    ); CHKERRQ(ierr);
    Range prisms_level;
    ierr = mField.get_entities_by_type_and_ref_level(
      bit_split,BitRefLevel().set(),MBPRISM,prisms_level
    ); CHKERRQ(ierr);
    EntityHandle meshset_level;
    rval = moab.create_meshset(MESHSET_SET,meshset_level); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_level,tets_level); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_tets_level.vtk","VTK","",&meshset_level,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_level,1); CHKERRQ_MOAB(rval);
    EntityHandle meshset_prims_level;
    rval = moab.create_meshset(MESHSET_SET,meshset_prims_level); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_prims_level,prisms_level); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_prisms_level.vtk","VTK","",&meshset_prims_level,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_prims_level,1); CHKERRQ_MOAB(rval);
  }

  if(debug) {
    Range tets_level;
    ierr = mField.get_entities_by_type_and_ref_level(
      bit_split,bit_split,MBTET,tets_level
    ); CHKERRQ(ierr);
    EntityHandle meshset_level;
    rval = moab.create_meshset(MESHSET_SET,meshset_level); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset_level,tets_level); CHKERRQ_MOAB(rval);
    rval = moab.write_file("out_tets_level_only.vtk","VTK","",&meshset_level,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset_level,1); CHKERRQ_MOAB(rval);
  }

  rval = moab.delete_entities(&bit_meshset,1);

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::refineCrackTip(
  BitRefLevel bit,BitRefLevel bit_ref,const int front_id,const int verb,const bool debug
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  MeshsetsManager *meshset_manager_ptr;
  MeshRefinement *mesh_refiner_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
  ierr = mField.query_interface(mesh_refiner_ptr); CHKERRQ(ierr);

  for(int ll = 0;ll!=refAtCrackTip;ll++) {

    Range tets_level;
    ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBTET,tets_level); CHKERRQ(ierr);
    Range edges_level;
    ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBEDGE,edges_level); CHKERRQ(ierr);
    Range crack_fornt;
    ierr = meshset_manager_ptr->getEntitiesByDimension(front_id,SIDESET,1,crack_fornt,true); CHKERRQ(ierr);
    crack_fornt = intersect(crack_fornt,edges_level);
    Range crack_front_nodes;
    rval = moab.get_connectivity(crack_fornt,crack_front_nodes,true); CHKERRQ_MOAB(rval);

    Range crack_front_nodes_tets;
    rval = mField.get_moab().get_adjacencies(
      crack_front_nodes,3,false,crack_front_nodes_tets,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    crack_front_nodes_tets = intersect(crack_front_nodes_tets,tets_level);
    Range edges_to_refine;
    rval = mField.get_moab().get_adjacencies(
      crack_front_nodes_tets,1,false,edges_to_refine,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    edges_to_refine = intersect(edges_to_refine,edges_level);
    if(edges_to_refine.empty()) continue;

    ierr = mesh_refiner_ptr->add_verices_in_the_middel_of_edges(edges_to_refine,bit_ref,2); CHKERRQ(ierr);
    ierr = mesh_refiner_ptr->refine_TET(tets_level,bit_ref,false); CHKERRQ(ierr);
    for(_IT_CUBITMESHSETS_FOR_LOOP_(mField,cubit_it)) {
      EntityHandle cubit_meshset = cubit_it->meshset;
      ierr = mField.update_meshset_by_entities_children(cubit_meshset,bit_ref,cubit_meshset,MBVERTEX,true); CHKERRQ(ierr);
      ierr = mField.update_meshset_by_entities_children(cubit_meshset,bit_ref,cubit_meshset,MBEDGE,true); CHKERRQ(ierr);
      ierr = mField.update_meshset_by_entities_children(cubit_meshset,bit_ref,cubit_meshset,MBTRI,true); CHKERRQ(ierr);
      ierr = mField.update_meshset_by_entities_children(cubit_meshset,bit_ref,cubit_meshset,MBTET,true); CHKERRQ(ierr);
    }

    if(isPartitioned) {
      ParallelComm* pcomm = ParallelComm::get_pcomm(&moab,MYPCOMM_INDEX);
      Tag part_tag = pcomm->part_tag();
      Range tagged_sets;
      rval = mField.get_moab().get_entities_by_type_and_tag(
        0,MBENTITYSET,&part_tag,NULL,1,tagged_sets,moab::Interface::UNION
      ); CHKERRQ_MOAB(rval);
      for(Range::iterator mit = tagged_sets.begin();mit!=tagged_sets.end();mit++) {
        int part;
        rval = moab.tag_get_data(part_tag,&*mit,1,&part); CHKERRQ_MOAB(rval);
        ierr = mField.update_meshset_by_entities_children(*mit,bit_ref,*mit,MBTET,true); CHKERRQ(ierr);
        Range ents3d;
        rval = moab.get_entities_by_dimension(*mit,3,ents3d,true); CHKERRQ_MOAB(rval);
        for(Range::iterator eit = ents3d.begin();eit!=ents3d.end();eit++) {
          rval = moab.tag_set_data(part_tag,&*eit,1,&part); CHKERRQ_MOAB(rval);
        }
        for(int dd = 0;dd!=3;dd++) {
          Range ents;
          rval = moab.get_adjacencies(ents3d,dd,false,ents,moab::Interface::UNION); CHKERRQ_MOAB(rval);
          for(Range::iterator eit = ents.begin();eit!=ents.end();eit++) {
            rval = moab.tag_set_data(part_tag,&*eit,1,&part); CHKERRQ_MOAB(rval);
          }
        }
      }
    }

    ierr = mField.shift_right_bit_ref(1); CHKERRQ(ierr);

    if(debug) {
      if(mField.get_comm_rank()==0) {
        std::string name;
        name = "out_crack_surface_level_"+boost::lexical_cast<std::string>(ll)+".vtk";
        MeshsetsManager *meshset_manager_ptr;
        ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
        // EntityHandle meshset;
        // ierr = meshset_manager_ptr->getMeshset(200,SIDESET,meshset); CHKERRQ(ierr);
        Range crack_faces;
        ierr = meshset_manager_ptr->getEntitiesByDimension(200,SIDESET,2,crack_faces,true); CHKERRQ(ierr);
        EntityHandle meshset;
        rval = mField.get_moab().create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
        rval = mField.get_moab().add_entities(meshset,crack_faces); CHKERRQ_MOAB(rval);
        rval = mField.get_moab().write_file(name.c_str(),"VTK","",&meshset,1); CHKERRQ_MOAB(rval);
        rval = mField.get_moab().delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
      }
    }

  }


  PetscFunctionReturn(0);
}



PetscErrorCode CrackPropagation::buildElasticFields(
  const BitRefLevel bit,
  const BitRefLevel mask,
  const bool proc_only,
  const bool build_fields
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  PetscFunctionBegin;

  Range tets_level;
  ierr = mField.get_entities_by_type_and_ref_level(bit,mask,MBTET,tets_level); CHKERRQ(ierr);
  if(proc_only) {
    tets_level = intersect(bitProcEnts,tets_level);
  }

  ierr = mField.add_field("SPATIAL_POSITION",H1,AINSWORTH_LEGENDRE_BASE,3,MB_TAG_SPARSE,MF_ZERO); CHKERRQ(ierr);
  ierr = mField.add_field("MESH_NODE_POSITIONS",H1,AINSWORTH_LEGENDRE_BASE,3,MB_TAG_SPARSE,MF_ZERO); CHKERRQ(ierr);
  // Add entities to field
  ierr = mField.add_ents_to_field_by_TETs(tets_level,"SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = mField.add_ents_to_field_by_VERTICEs(crackFrontNodes,"SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = mField.add_ents_to_field_by_EDGEs(crackFront,"SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = mField.add_ents_to_field_by_TETs(tets_level,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.add_ents_to_field_by_VERTICEs(crackFrontNodes,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.add_ents_to_field_by_EDGEs(crackFront,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);

  // Set approx order
  Range dofs_ents;
  dofs_ents.merge(tets_level);
  for(int dd = 1;dd!=3;dd++) {
    rval = moab.get_adjacencies(
      tets_level,dd,false,dofs_ents,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
  }
  Range dofs_nodes;
  rval = moab.get_connectivity(tets_level,dofs_nodes,true); CHKERRQ_MOAB(rval);
  dofs_nodes.merge(crackFrontNodes);
  dofs_ents.merge(crackFront);
  ierr = mField.set_field_order(dofs_ents,"SPATIAL_POSITION",approxOrder); CHKERRQ(ierr);
  ierr = mField.set_field_order(dofs_nodes,"SPATIAL_POSITION",1); CHKERRQ(ierr);
  ierr = mField.set_field_order(dofs_ents,"MESH_NODE_POSITIONS",geometryOrder); CHKERRQ(ierr);
  ierr = mField.set_field_order(dofs_nodes,"MESH_NODE_POSITIONS",1); CHKERRQ(ierr);

  if(refOrderAtTip>0) {
    map<int,Range> layers;
    Range ents = crackFrontNodes;
    for(int ll = 0;ll!=refOrderAtTip;ll++) {
      Range new_ents;
      rval = moab.get_adjacencies(ents,3,false,new_ents,moab::Interface::UNION); CHKERRQ(ierr);
      rval = moab.get_adjacencies(ents,2,false,new_ents,moab::Interface::UNION); CHKERRQ(ierr);
      rval = moab.get_adjacencies(ents,1,false,new_ents,moab::Interface::UNION); CHKERRQ(ierr);
      rval = moab.get_adjacencies(ents,0,false,new_ents,moab::Interface::UNION); CHKERRQ(ierr);
      layers[ll] = new_ents;
      ents = new_ents;
    }
    int poo = 1;
    for(map<int,Range>::reverse_iterator mit = layers.rbegin();mit!=layers.rend();mit++,poo++) {
      mit->second = intersect(mit->second,dofs_ents);
      ierr = mField.set_field_order(mit->second,"SPATIAL_POSITION",approxOrder+poo); CHKERRQ(ierr);
    }
  }

  bool set_rho = false;
  if(defaultMaterial == BONEHOOKE || defaultMaterial == BONENEOHOOKEAN) {
    if(!mField.check_field("RHO")) {
      ierr = mField.add_field("RHO",H1,AINSWORTH_LOBBATO_BASE,1); CHKERRQ(ierr);
      ierr = mField.add_ents_to_field_by_TETs(tets_level,"RHO"); CHKERRQ(ierr);
      ierr = mField.set_field_order(dofs_nodes,"RHO",1); CHKERRQ(ierr);
      set_rho = true;
    }
  }

  if(build_fields) {
    ierr = mField.build_fields(); CHKERRQ(ierr);
    if(set_rho) {
      ierr = mField.set_field(rHo0,MBVERTEX,"RHO");
    }
  } else if(set_rho) {
    SETERRQ(
      mField.get_comm(),
      MOFEM_DATA_INCONSISTENCY,
      "You need build fields to set density field"
    );
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::buildArcLengthField(const bool build_fields) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  PetscFunctionBegin;
  ierr = mField.add_field("ARC_LENGTH_LAMBDA",NOFIELD,NOBASE,1,MB_TAG_SPARSE,MF_ZERO); CHKERRQ(ierr);
  EntityHandle lambda_meshset = mField.get_field_meshset("ARC_LENGTH_LAMBDA");
  rval = mField.get_moab().add_entities(lambda_meshset,&arcLengthVertex,1); CHKERRQ_MOAB(rval);
  const RefEntity_multiIndex *refined_ents_ptr;
  ierr = mField.get_ref_ents(&refined_ents_ptr); CHKERRQ(ierr);
  // Add vertex to MoFEM database
  std::pair<RefEntity_multiIndex::iterator,bool> p_ent =
  const_cast<RefEntity_multiIndex*>(refined_ents_ptr)->insert(
    boost::shared_ptr<RefEntity>(new RefEntity(mField.get_basic_entity_data_ptr(),arcLengthVertex))
  );
  const_cast<RefEntity_multiIndex*>(refined_ents_ptr)->modify(
    p_ent.first,RefEntity_change_add_bit(BitRefLevel().set())
  );
  if(build_fields) {
    ierr = mField.build_fields(); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}


PetscErrorCode CrackPropagation::buildSurfaceFields(
  const BitRefLevel bit,
  const BitRefLevel mask,
  const int id,
  const bool proc_only,
  const bool build_fields,
  const int verb,
  const bool debug
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Range level_tris;
  ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  Range body_skin_proc_tris;
  body_skin_proc_tris = intersect(bodySkin,level_tris);

  if(proc_only) {
    Range proc_tris;
    rval = moab.get_adjacencies(
      bitProcEnts,2,false,proc_tris,moab::Interface::UNION
    ); CHKERRQ(ierr);
    body_skin_proc_tris = intersect(body_skin_proc_tris,proc_tris);
  }

  if(debug) {
    EntityHandle meshset;
    rval = mField.get_moab().create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
    rval = mField.get_moab().add_entities(meshset,body_skin_proc_tris); CHKERRQ_MOAB(rval);
    std::string field_name;
    field_name = "out_proc_body_skin_"+boost::lexical_cast<std::string>(mField.get_comm_rank())+".vtk";
    rval = mField.get_moab().write_file(field_name.c_str(),"VTK","",&meshset,1); CHKERRQ_MOAB(rval);
    rval = mField.get_moab().delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
  }

  std::string field_name = "LAMBDA_SURFACE"+boost::lexical_cast<std::string>(id);
  ierr = mField.add_field(
    field_name.c_str(),H1,AINSWORTH_LEGENDRE_BASE,1,MB_TAG_SPARSE,MF_ZERO
  ); CHKERRQ(ierr);

  Range dofs_nodes;
  rval = moab.get_connectivity(body_skin_proc_tris,dofs_nodes,true); CHKERRQ_MOAB(rval);
  dofs_nodes.merge(crackFrontNodes);
  ierr = mField.synchronise_entities(dofs_nodes); CHKERRQ(ierr);
  ierr = mField.add_ents_to_field_by_VERTICEs(dofs_nodes,field_name.c_str()); CHKERRQ(ierr);
  ierr = mField.set_field_order(dofs_nodes,field_name.c_str(),1); CHKERRQ(ierr);

  if(build_fields) {
    ierr = mField.build_fields(); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::buildCrackSurfaceFieldId(
  const BitRefLevel bit,
  const BitRefLevel mask,
  const int surf_id,
  const bool proc_only,
  const bool build_fields,
  const int verb,
  const bool debug
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;

  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Range crack_surface_ents = oneSideCrackFaces;
  Range level_tris;
  ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  crack_surface_ents = intersect(crack_surface_ents,level_tris);

  if(proc_only) {
    Range proc_tris;
    rval = moab.get_adjacencies(
      bitProcEnts,2,false,proc_tris,moab::Interface::UNION
    ); CHKERRQ(ierr);
    crack_surface_ents = intersect(crack_surface_ents,proc_tris);
  }

  if(debug) {
    EntityHandle meshset;
    rval = mField.get_moab().create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
    rval = mField.get_moab().add_entities(meshset,crack_surface_ents); CHKERRQ_MOAB(rval);
    std::string field_name;
    field_name = "out_proc_crack_"+boost::lexical_cast<std::string>(mField.get_comm_rank())+".vtk";
    rval = mField.get_moab().write_file(field_name.c_str(),"VTK","",&meshset,1); CHKERRQ_MOAB(rval);
    rval = mField.get_moab().delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
  }

  std::string field_name = "LAMBDA_SURFACE"+boost::lexical_cast<std::string>(surf_id);
  ierr = mField.add_field(
    field_name.c_str(),H1,AINSWORTH_LEGENDRE_BASE,1,MB_TAG_SPARSE,MF_ZERO
  ); CHKERRQ(ierr);

  Range dofs_nodes;
  rval = moab.get_connectivity(crack_surface_ents,dofs_nodes,true); CHKERRQ_MOAB(rval);
  dofs_nodes = subtract(dofs_nodes,crackFrontNodes);
  ierr = mField.add_ents_to_field_by_VERTICEs(dofs_nodes,field_name.c_str()); CHKERRQ(ierr);
  ierr = mField.set_field_order(dofs_nodes,field_name.c_str(),1); CHKERRQ(ierr);

  Range level_edges;
  ierr = mField.get_entities_by_type_and_ref_level(
    bit,BitRefLevel().set(),MBEDGE,level_edges
  ); CHKERRQ(ierr);
  Range crack_front_edges = intersect(crackFront,level_edges);
  // if(proc_only) {
  //   Range proc_edges;
  //   rval = moab.get_adjacencies(
  //     bitProcEnts,1,false,proc_edges,moab::Interface::UNION
  //   ); CHKERRQ(ierr);
  //   crack_front_edges = intersect(crack_front_edges,proc_edges);
  // }
  Range crack_front_nodes;
  rval = moab.get_connectivity(crack_front_edges,crack_front_nodes,true);

  ierr = mField.add_field(
    "LAMBDA_CRACKFRONT_AREA",H1,AINSWORTH_LEGENDRE_BASE,1,MB_TAG_SPARSE,MF_ZERO
  ); CHKERRQ(ierr);
  ierr = mField.add_ents_to_field_by_VERTICEs(crack_front_nodes,"LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
  ierr = mField.set_field_order(crack_front_nodes,"LAMBDA_CRACKFRONT_AREA",1); CHKERRQ(ierr);

  ierr = mField.add_field(
    "LAMBDA_CRACKFRONT_AREA_TANGENT",H1,AINSWORTH_LEGENDRE_BASE,1,MB_TAG_SPARSE,MF_ZERO
  ); CHKERRQ(ierr);
  Skinner skin(&mField.get_moab());
  Range skin_crack_front;
  rval = skin.find_skin(0,crack_front_edges,false,skin_crack_front);
  ierr = mField.add_ents_to_field_by_VERTICEs(
    subtract(crack_front_nodes,skin_crack_front),"LAMBDA_CRACKFRONT_AREA_TANGENT"
  ); CHKERRQ(ierr);
  ierr = mField.set_field_order(
    subtract(crack_front_nodes,skin_crack_front),"LAMBDA_CRACKFRONT_AREA_TANGENT",1
  ); CHKERRQ(ierr);

  if(build_fields) {
    ierr = mField.build_fields(); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}


PetscErrorCode CrackPropagation::declareElasticFE(
  const BitRefLevel bit,
  const BitRefLevel mask,
  const bool add_forces,
  const bool proc_only
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  PetscFunctionBegin;

  Range tets_level;
  ierr = mField.get_entities_by_type_and_ref_level(bit,mask,MBTET,tets_level); CHKERRQ(ierr);
  if(proc_only) {
    tets_level = intersect(bitProcEnts,tets_level);
  }
  Range adj_front;
  rval = moab.get_adjacencies(crackFrontNodes,3,false,adj_front,moab::Interface::UNION); CHKERRQ_MOAB(rval);
  tets_level.merge(adj_front);

  // Add finite elements
  ierr = mField.add_finite_element("ELASTIC",MF_ZERO); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row("ELASTIC","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col("ELASTIC","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data("ELASTIC","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data("ELASTIC","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.add_ents_to_finite_element_by_TETs(tets_level,"ELASTIC"); CHKERRQ(ierr);
  ierr = mField.build_finite_elements("ELASTIC",&tets_level); CHKERRQ(ierr);

  if(add_forces) {

    // Set approx order
    Range dofs_ents;
    dofs_ents.merge(tets_level);
    for(int dd = 1;dd!=3;dd++) {
      rval = moab.get_adjacencies(
        tets_level,dd,false,dofs_ents,moab::Interface::UNION
      ); CHKERRQ_MOAB(rval);
    }
    Range dofs_nodes;

    // Forces
    Range lower_dim_ents = unite(dofs_ents,dofs_nodes);
    ierr = mField.add_finite_element("NEUAMNN_FE",MF_ZERO); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_row("NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_col("NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_data("NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_data("NEUAMNN_FE","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,NODESET|FORCESET,it)) {
      Range tris;
      rval = mField.get_moab().get_entities_by_type(it->meshset,MBTRI,tris,true); CHKERRQ_MOAB(rval);
      tris = intersect(lower_dim_ents,tris);
      ierr = mField.add_ents_to_finite_element_by_TRIs(tris,"NEUAMNN_FE"); CHKERRQ(ierr);
    }
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,SIDESET|PRESSURESET,it)) {
      Range tris;
      rval = mField.get_moab().get_entities_by_type(it->meshset,MBTRI,tris,true); CHKERRQ_MOAB(rval);
      tris = intersect(lower_dim_ents,tris);
      ierr = mField.add_ents_to_finite_element_by_TRIs(tris,"NEUAMNN_FE"); CHKERRQ(ierr);
    }
    // ierr = mField.synchronise_entities(lower_dim_ents); CHKERRQ(ierr);
    ierr = mField.build_finite_elements("NEUAMNN_FE",&lower_dim_ents); CHKERRQ(ierr);

    #ifdef __ANALITICAL_TRACTION__
    {
      ierr = mField.add_finite_element("ANALITICAL_NEUAMNN_FE",MF_ZERO); CHKERRQ(ierr);
      ierr = mField.modify_finite_element_add_field_row("ANALITICAL_NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
      ierr = mField.modify_finite_element_add_field_col("ANALITICAL_NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
      ierr = mField.modify_finite_element_add_field_data("ANALITICAL_NEUAMNN_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
      ierr = mField.modify_finite_element_add_field_data("ANALITICAL_NEUAMNN_FE","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
      MeshsetsManager *meshset_manager_ptr;
      ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
      if(meshset_manager_ptr->checkMeshset("ANALITICAL_TRACTION")) {
        const CubitMeshSets *cubit_meshset_ptr;
        meshset_manager_ptr->getCubitMeshsetPtr("ANALITICAL_TRACTION",&cubit_meshset_ptr);
        Range tris;
        ierr = meshset_manager_ptr->getEntitiesByDimension(
          cubit_meshset_ptr->getMeshsetId(),BLOCKSET,2,tris,true
        ); CHKERRQ(ierr);
        tris = intersect(tris,lower_dim_ents);
        // ierr = mField.synchronise_entities(tris); CHKERRQ(ierr);
        ierr = mField.add_ents_to_finite_element_by_TRIs(tris,"ANALITICAL_NEUAMNN_FE"); CHKERRQ(ierr);
      }
      ierr = mField.build_finite_elements("ANALITICAL_NEUAMNN_FE",&lower_dim_ents); CHKERRQ(ierr);
    }
    #endif //__ANALITICAL_TRACTION__

    ierr = mField.add_finite_element("FORCE_FE",MF_ZERO); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_row("FORCE_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_col("FORCE_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_data("FORCE_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_data("FORCE_FE","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,NODESET|FORCESET,it)) {
      Range tris;
      rval = mField.get_moab().get_entities_by_type(it->meshset,MBTRI,tris,true); CHKERRQ_MOAB(rval);
      Range edges;
      rval = mField.get_moab().get_entities_by_type(it->meshset,MBEDGE,edges,true); CHKERRQ_MOAB(rval);
      Range tris_edges;
      rval = mField.get_moab().get_adjacencies(tris,1,false,tris_edges,moab::Interface::UNION); CHKERRQ_MOAB(rval);
      edges = subtract(edges,tris_edges);
      edges = intersect(lower_dim_ents,edges);
      // ierr = mField.synchronise_entities(edges); CHKERRQ(ierr);
      ierr = mField.add_ents_to_finite_element_by_EDGEs(edges,"FORCE_FE"); CHKERRQ(ierr);
    }
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,NODESET|FORCESET,it)) {
      Range tris;
      rval = mField.get_moab().get_entities_by_type(it->meshset,MBTRI,tris,true); CHKERRQ_MOAB(rval);
      Range edges;
      rval = mField.get_moab().get_entities_by_type(it->meshset,MBEDGE,edges,true); CHKERRQ_MOAB(rval);
      Range tris_nodes;
      rval = mField.get_moab().get_connectivity(tris,tris_nodes); CHKERRQ_MOAB(rval);
      Range edges_nodes;
      rval = mField.get_moab().get_connectivity(edges,edges_nodes); CHKERRQ_MOAB(rval);
      Range nodes;
      rval = mField.get_moab().get_entities_by_type(it->meshset,MBVERTEX,nodes,true); CHKERRQ_MOAB(rval);
      nodes = subtract(nodes,tris_nodes);
      nodes = subtract(nodes,edges_nodes);
      nodes = intersect(lower_dim_ents,nodes);
      // ierr = mField.synchronise_entities(nodes); CHKERRQ(ierr);
      ierr = mField.add_ents_to_finite_element_by_VERTICEs(nodes,"FORCE_FE"); CHKERRQ(ierr);
    }

    ierr = MetaNodalForces::addElement(mField,"SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = MetaEdgeForces::addElement(mField,"SPATIAL_POSITION"); CHKERRQ(ierr);
    ierr = mField.build_finite_elements("FORCE_FE",&lower_dim_ents); CHKERRQ(ierr);

    #ifdef __ANALITICAL_DISPLACEMENT__
    {
      MeshsetsManager *meshset_manager_ptr;
      ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
      if(meshset_manager_ptr->checkMeshset("ANALITICAL_DISP")) {
        const CubitMeshSets *cubit_meshset_ptr;
        meshset_manager_ptr->getCubitMeshsetPtr("ANALITICAL_DISP",&cubit_meshset_ptr);
        Range tris;
        ierr = meshset_manager_ptr->getEntitiesByDimension(
          cubit_meshset_ptr->getMeshsetId(),BLOCKSET,2,tris,true
        ); CHKERRQ(ierr);
        if(tris.size()>0) {
          ierr = mField.add_finite_element("BC_FE",MF_ZERO); CHKERRQ(ierr);
          ierr = mField.modify_finite_element_add_field_row("BC_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
          ierr = mField.modify_finite_element_add_field_col("BC_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
          ierr = mField.modify_finite_element_add_field_data("BC_FE","SPATIAL_POSITION"); CHKERRQ(ierr);
          ierr = mField.modify_finite_element_add_field_data("BC_FE","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
          tris = intersect(lower_dim_ents,tris);
          // ierr = mField.synchronise_entities(tris); CHKERRQ(ierr);
          ierr = mField.add_ents_to_finite_element_by_TRIs(tris,"BC_FE"); CHKERRQ(ierr);
          ierr = mField.build_finite_elements("BC_FE",&lower_dim_ents); CHKERRQ(ierr);
        }
      }
    }
    #endif //__ANALITICAL_DISPLACEMENT__

  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::declareMaterialFE(
  const BitRefLevel bit,const BitRefLevel mask,const bool add_forces,const bool proc_only
) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  PetscFunctionBegin;

  Range tets_level;
  ierr = mField.get_entities_by_type_and_ref_level(bit,mask,MBTET,tets_level); CHKERRQ(ierr);
  if(proc_only) {
    tets_level = intersect(bitProcEnts,tets_level);
  }

  // Add finite elements
  ierr = mField.add_finite_element("MATERIAL",MF_ZERO); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row("MATERIAL","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col("MATERIAL","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data("MATERIAL","SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data("MATERIAL","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  // ierr = mField.synchronise_entities(tets_level); CHKERRQ(ierr);
  ierr = mField.add_ents_to_finite_element_by_TETs(tets_level,"MATERIAL"); CHKERRQ(ierr);
  ierr = mField.build_finite_elements("MATERIAL",&tets_level); CHKERRQ(ierr);

  if(add_forces) {
    ierr = mField.add_finite_element("GRIFFITH_FORCE_ELEMENT",MF_ZERO); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_row("GRIFFITH_FORCE_ELEMENT","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_col("GRIFFITH_FORCE_ELEMENT","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    ierr = mField.modify_finite_element_add_field_data("GRIFFITH_FORCE_ELEMENT","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    Range crack_front_adj_tris;
    rval = mField.get_moab().get_adjacencies(
      crackFrontNodes,2,false,crack_front_adj_tris,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    crack_front_adj_tris = intersect(crack_front_adj_tris,crackFaces);
    Range tris_level;
    ierr = mField.get_entities_by_type_and_ref_level(bit,mask,MBTRI,tris_level); CHKERRQ(ierr);
    crack_front_adj_tris = intersect(crack_front_adj_tris,tris_level);
    // ierr = mField.synchronise_entities(crack_front_adj_tris); CHKERRQ(ierr);
    ierr = mField.add_ents_to_finite_element_by_TRIs(
      crack_front_adj_tris,"GRIFFITH_FORCE_ELEMENT"
    ); CHKERRQ(ierr);
    ierr = mField.build_finite_elements("GRIFFITH_FORCE_ELEMENT",&crack_front_adj_tris); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}


PetscErrorCode CrackPropagation::declareSurfaceFE(
  std::string fe_name,
  const BitRefLevel bit,
  const BitRefLevel mask,
  const int id,
  const bool proc_only,
  const int verb,const bool debug
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Range level_tris;
  ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  Range surface_ents = intersect(bodySkin,level_tris);

  Range body_skin_proc_tris;
  if(proc_only) {
    // Note that some elements are on this proc, but not owned by that proc, since
    // some tetrahedral share lower dimension entities
    Range proc_tris;
    rval = moab.get_adjacencies(
      bitProcEnts,2,false,proc_tris,moab::Interface::UNION
    ); CHKERRQ(ierr);
    surface_ents = intersect(surface_ents,proc_tris);
  }

  std::string field_name = "LAMBDA_SURFACE"+boost::lexical_cast<std::string>(id);

  for(_IT_GET_ENT_FIELD_BY_NAME_FOR_LOOP_(mField,field_name,dit)) {
    // cerr << dit->get()->getOwnerProc() << endl;
    if(dit->get()->getOwnerProc()==mField.get_comm_rank()) {
      EntityHandle ent = dit->get()->getEnt();
      Range adj_tris;
      rval = moab.get_adjacencies(&ent,1,2,false,adj_tris); CHKERRQ_MOAB(rval);
      adj_tris = intersect(adj_tris,bodySkin);
      // cerr << adj_tris << endl;
      surface_ents.merge(adj_tris);
    }
  }

  // Add finite elements
  ierr = mField.add_finite_element(fe_name.c_str(),MF_ZERO); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row(fe_name,field_name.c_str()); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row(fe_name,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col(fe_name,field_name.c_str()); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col(fe_name,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data(fe_name,field_name.c_str()); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data(fe_name,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.synchronise_entities(surface_ents); CHKERRQ(ierr);
  ierr = mField.add_ents_to_finite_element_by_TRIs(surface_ents,fe_name); CHKERRQ(ierr);
  ierr = mField.build_finite_elements(fe_name,&surface_ents); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::declareCrackSurfaceFE(
  std::string fe_name,
  const BitRefLevel bit,
  const BitRefLevel maskX,
  const int id,
  const bool proc_only,
  const int verb,
  const bool debug
) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  MoABErrorCode rval;
  moab::Interface &moab = mField.get_moab();
  MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);

  Range level_tris;
  ierr = mField.get_entities_by_type_and_ref_level(bit,BitRefLevel().set(),MBTRI,level_tris); CHKERRQ(ierr);
  Range crack_surface;
  crack_surface = intersect(oneSideCrackFaces,level_tris);
  if(proc_only) {
    Range proc_tris;
    rval = moab.get_adjacencies(
      bitProcEnts,2,false,proc_tris,moab::Interface::UNION
    ); CHKERRQ(ierr);
    crack_surface = intersect(oneSideCrackFaces,proc_tris);
  }

  Range crack_front_adj_tris;
  rval = mField.get_moab().get_adjacencies(
    crackFrontNodes,2,false,crack_front_adj_tris,moab::Interface::UNION
  ); CHKERRQ_MOAB(rval);
  crack_front_adj_tris = intersect(crack_front_adj_tris,crackFaces);
  crack_surface = subtract(crack_surface,crack_front_adj_tris);
  crack_front_adj_tris = intersect(crack_front_adj_tris,level_tris);

  std::string field_name = "LAMBDA_SURFACE"+boost::lexical_cast<std::string>(id);

  for(_IT_GET_ENT_FIELD_BY_NAME_FOR_LOOP_(mField,field_name,dit)) {
    // cerr << dit->get()->getOwnerProc() << endl;
    if(dit->get()->getOwnerProc()==mField.get_comm_rank()) {
      EntityHandle ent = dit->get()->getEnt();
      Range adj_tris;
      rval = moab.get_adjacencies(&ent,1,2,false,adj_tris); CHKERRQ_MOAB(rval);
      adj_tris = intersect(adj_tris,bodySkin);
      // cerr << adj_tris << endl;
      crack_surface.merge(adj_tris);
    }
  }

  // Add finite elements
  ierr = mField.add_finite_element(fe_name.c_str(),MF_ZERO); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row(fe_name,field_name.c_str()); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row(fe_name,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col(fe_name,field_name.c_str()); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col(fe_name,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data(fe_name,field_name.c_str()); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data(fe_name,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.add_ents_to_finite_element_by_TRIs(crack_surface,fe_name); CHKERRQ(ierr);
  ierr = mField.build_finite_elements(fe_name,&crack_surface); CHKERRQ(ierr);

  if(debug) {
    std::ostringstream file;
    file << "out_crack_faces_" << mField.get_comm_rank() << ".vtk";
    EntityHandle meshset;
    rval = moab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset,crack_surface); CHKERRQ_MOAB(rval);
    rval = moab.write_file(file.str().c_str(),"VTK","",&meshset,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
  }

  ierr = mField.add_finite_element("CRACKFRONT_AREA_ELEM",MF_ZERO); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row("CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row("CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col("CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col("CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data("CRACKFRONT_AREA_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data("CRACKFRONT_AREA_ELEM","LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
  ierr = mField.add_finite_element("CRACKFRONT_AREA_TANGENT_ELEM",MF_ZERO); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row("CRACKFRONT_AREA_TANGENT_ELEM","LAMBDA_CRACKFRONT_AREA_TANGENT"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_row("CRACKFRONT_AREA_TANGENT_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col("CRACKFRONT_AREA_TANGENT_ELEM","LAMBDA_CRACKFRONT_AREA_TANGENT"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_col("CRACKFRONT_AREA_TANGENT_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data("CRACKFRONT_AREA_TANGENT_ELEM","MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = mField.modify_finite_element_add_field_data("CRACKFRONT_AREA_TANGENT_ELEM","LAMBDA_CRACKFRONT_AREA_TANGENT"); CHKERRQ(ierr);

  ierr = mField.add_ents_to_finite_element_by_TRIs(
    crack_front_adj_tris,"CRACKFRONT_AREA_ELEM"
  ); CHKERRQ(ierr);
  ierr = mField.add_ents_to_finite_element_by_TRIs(
    crack_front_adj_tris,"CRACKFRONT_AREA_TANGENT_ELEM"
  ); CHKERRQ(ierr);

  if(debug) {
    std::ostringstream file;
    file << "out_crack_front_adj_tris_" << mField.get_comm_rank() << ".vtk";
    EntityHandle meshset;
    rval = moab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
    rval = moab.add_entities(meshset,crack_front_adj_tris); CHKERRQ_MOAB(rval);
    rval = moab.write_file(file.str().c_str(),"VTK","",&meshset,1); CHKERRQ_MOAB(rval);
    rval = moab.delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
  }



  ierr = mField.build_finite_elements("CRACKFRONT_AREA_ELEM",&crack_front_adj_tris); CHKERRQ(ierr);
  ierr = mField.build_finite_elements("CRACKFRONT_AREA_TANGENT_ELEM",&crack_front_adj_tris); CHKERRQ(ierr);


  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::createElasticDM(
  DM *dm,const std::string prb_name,const BitRefLevel bit,const BitRefLevel mask
) {
  PetscErrorCode ierr;
  ProblemsManager *prb_mng_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(prb_mng_ptr); CHKERRQ(ierr);
  ierr = mField.build_adjacencies(bit); CHKERRQ(ierr);
  ierr = DMCreate(PETSC_COMM_WORLD,dm);CHKERRQ(ierr);
  ierr = DMSetType(*dm,"MOFEM_MG");CHKERRQ(ierr);
  ierr = DMMoFEMCreateMoFEM(*dm,&mField,prb_name.c_str(),bit,mask); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"ELASTIC"); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"NEUAMNN_FE"); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"FORCE_FE"); CHKERRQ(ierr);
  #ifdef __ANALITICAL_TRACTION__
  ierr = DMMoFEMAddElement(*dm,"ANALITICAL_NEUAMNN_FE"); CHKERRQ(ierr);
  #endif
  ierr = DMMoFEMSetIsPartitioned(*dm,PETSC_TRUE); CHKERRQ(ierr);
  ierr = DMSetUp(*dm); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::createBcDM(
  DM *dm,const std::string prb_name,const BitRefLevel bit,const BitRefLevel mask
) {
  PetscErrorCode ierr;
  ProblemsManager *prb_mng_ptr;
  MeshsetsManager *meshset_manager_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(prb_mng_ptr); CHKERRQ(ierr);
  ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
  ierr = mField.build_adjacencies(bit); CHKERRQ(ierr);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(meshset_manager_ptr->checkMeshset("ANALITICAL_DISP")) {
    const CubitMeshSets *cubit_meshset_ptr;
    meshset_manager_ptr->getCubitMeshsetPtr("ANALITICAL_DISP",&cubit_meshset_ptr);
    Range tris;
    ierr = meshset_manager_ptr->getEntitiesByDimension(
      cubit_meshset_ptr->getMeshsetId(),BLOCKSET,2,tris,true
    ); CHKERRQ(ierr);
    if(tris.size()>0) {
      ierr = DMCreate(PETSC_COMM_WORLD,dm);CHKERRQ(ierr);
      ierr = DMSetType(*dm,"MOFEM");CHKERRQ(ierr);
      ierr = DMMoFEMCreateMoFEM(*dm,&mField,prb_name.c_str(),bit,mask); CHKERRQ(ierr);
      ierr = DMMoFEMAddElement(*dm,"BC_FE"); CHKERRQ(ierr);
      ierr = DMMoFEMSetIsPartitioned(*dm,PETSC_TRUE); CHKERRQ(ierr);
      ierr = DMSetUp(*dm); CHKERRQ(ierr);
    }
  }
  #endif //__ANALITICAL_DISPLACEMENT__
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::createMaterialDM(
  DM *dm,const std::string prb_name,const BitRefLevel bit,const BitRefLevel mask,const std::vector<std::string> fe_list
) {
  PetscErrorCode ierr;
  ProblemsManager *prb_mng_ptr;
  PetscFunctionBegin;
  ierr = mField.query_interface(prb_mng_ptr); CHKERRQ(ierr);
  ierr = mField.build_adjacencies(bit); CHKERRQ(ierr);
  ierr = DMCreate(PETSC_COMM_WORLD,dm);CHKERRQ(ierr);
  ierr = DMSetType(*dm,"MOFEM");CHKERRQ(ierr);
  ierr = DMMoFEMCreateMoFEM(*dm,&mField,prb_name.c_str(),bit,mask); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"MATERIAL"); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"GRIFFITH_FORCE_ELEMENT"); CHKERRQ(ierr);
  for(std::vector<std::string>::const_iterator it = fe_list.begin();it!=fe_list.end();it++) {
    ierr = DMMoFEMAddElement(*dm,it->c_str()); CHKERRQ(ierr);
  }
  ierr = DMMoFEMAddElement(*dm,"CRACKFRONT_AREA_ELEM"); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"CRACKFRONT_AREA_TANGENT_ELEM"); CHKERRQ(ierr);
  ierr = DMMoFEMSetIsPartitioned(*dm,PETSC_TRUE); CHKERRQ(ierr);
  ierr = DMSetUp(*dm); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::createMaterialForcesDM(DM *dm,DM dm_material,const std::string prb_name) {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = DMCreate(PETSC_COMM_WORLD,dm);CHKERRQ(ierr);
  ierr = DMSetType(*dm,"MOFEM");CHKERRQ(ierr);
  ierr = DMMoFEMCreateSubDM(*dm,dm_material,prb_name.c_str()); CHKERRQ(ierr);
  ierr = DMMoFEMAddSubFieldRow(*dm,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = DMMoFEMAddSubFieldCol(*dm,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"MATERIAL"); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"GRIFFITH_FORCE_ELEMENT"); CHKERRQ(ierr);
  ierr = DMMoFEMSetSquareProblem(*dm,PETSC_TRUE); CHKERRQ(ierr);
  ierr = DMMoFEMSetIsPartitioned(*dm,PETSC_TRUE); CHKERRQ(ierr);
  ierr = DMSetUp(*dm); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::createSurfaceProjectionDM(
  DM *dm,
  DM dm_material,
  const std::string prb_name,
  const std::vector<int> surface_ids,
  std::vector<std::string> fe_list
) {
  PetscFunctionBegin;
  PetscErrorCode ierr;
  // ProblemsManager *prb_mng_ptr;
  PetscFunctionBegin;
  // ierr = mField.query_interface(prb_mng_ptr); CHKERRQ(ierr);
  ierr = DMCreate(PETSC_COMM_WORLD,dm);CHKERRQ(ierr);
  ierr = DMSetType(*dm,"MOFEM");CHKERRQ(ierr);
  ierr = DMMoFEMCreateSubDM(*dm,dm_material,prb_name.c_str()); CHKERRQ(ierr);
  ierr = DMMoFEMSetSquareProblem(*dm,PETSC_FALSE); CHKERRQ(ierr);
  for(std::vector<int>::const_iterator id = surface_ids.begin();id!=surface_ids.end();id++) {
    ierr = DMMoFEMAddSubFieldRow(
      *dm,("LAMBDA_SURFACE"+boost::lexical_cast<std::string>(*id)).c_str()
    ); CHKERRQ(ierr);
  }
  ierr = DMMoFEMAddSubFieldCol(*dm,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  for(std::vector<std::string>::const_iterator it = fe_list.begin();it!=fe_list.end();it++) {
    ierr = DMMoFEMAddElement(*dm,it->c_str()); CHKERRQ(ierr);
  }
  ierr = DMMoFEMSetIsPartitioned(*dm,PETSC_TRUE); CHKERRQ(ierr);
  ierr = DMSetUp(*dm); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::createCrackFrontAreaDM(
  DM *dm,
  DM dm_material,
  const std::string prb_name
) {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = DMCreate(PETSC_COMM_WORLD,dm);CHKERRQ(ierr);
  ierr = DMSetType(*dm,"MOFEM");CHKERRQ(ierr);
  ierr = DMMoFEMCreateSubDM(*dm,dm_material,prb_name.c_str()); CHKERRQ(ierr);
  ierr = DMMoFEMAddSubFieldRow(*dm,"LAMBDA_CRACKFRONT_AREA"); CHKERRQ(ierr);
  ierr = DMMoFEMAddSubFieldRow(*dm,"LAMBDA_CRACKFRONT_AREA_TANGENT"); CHKERRQ(ierr);
  ierr = DMMoFEMAddSubFieldCol(*dm,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"CRACKFRONT_AREA_ELEM"); CHKERRQ(ierr);
  ierr = DMMoFEMAddElement(*dm,"CRACKFRONT_AREA_TANGENT_ELEM"); CHKERRQ(ierr);
  ierr = DMMoFEMSetSquareProblem(*dm,PETSC_FALSE); CHKERRQ(ierr);
  ierr = DMMoFEMSetIsPartitioned(*dm,PETSC_TRUE); CHKERRQ(ierr);
  ierr = DMSetUp(*dm); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::assembleElasticDM(
  DM dm,Mat m,Vec q,Vec f,const int verb,const bool debug
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  PetscFunctionBegin;

  if(!elasticFe) {
    elasticFe = boost::shared_ptr<NonlinearElasticElement>(new NonlinearElasticElement(mField,ELASTIC_TAG));
  }
  boost::shared_ptr<CrackFrontFE> feRhs = boost::shared_ptr<CrackFrontFE>(
    elasticFe,new CrackFrontFE(mField,crackFrontNodes,crackFrontNodesEdges,crackFrontElements)
  );
  boost::shared_ptr<CrackFrontFE> feLhs = boost::shared_ptr<CrackFrontFE>(
    elasticFe,new CrackFrontFE(mField,crackFrontNodes,crackFrontNodesEdges,crackFrontElements)
  );

  // Create operators to calculate finite element matrices for elastic element
  {
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,BLOCKSET|MAT_ELASTICSET,it)) {

      Mat_Elastic mydata;
      ierr = it->getAttributeDataStructure(mydata); CHKERRQ(ierr);
      int id = it->getMeshsetId();
      EntityHandle meshset = it->getMeshset();
      rval = mField.get_moab().get_entities_by_type(
        meshset,MBTET,elasticFe->setOfBlocks[id].tEts,true
      ); CHKERRQ_MOAB(rval);
      elasticFe->setOfBlocks[id].iD = id;
      elasticFe->setOfBlocks[id].E = mydata.data.Young;
      elasticFe->setOfBlocks[id].PoissonRatio = mydata.data.Poisson;

      // Set default material to elastic blocks
      switch(defaultMaterial) {
        case HOOKE:
        elasticFe->setOfBlocks[id].materialDoublePtr = boost::make_shared<Hooke<double> >();
        elasticFe->setOfBlocks[id].materialAdoublePtr = boost::make_shared<Hooke<adouble> >();
        break;
        case KIRCHOFF:
        elasticFe->setOfBlocks[id].materialDoublePtr = boost::make_shared<
        NonlinearElasticElement::FunctionsToCalculatePiolaKirchhoffI<double>
        >();
        elasticFe->setOfBlocks[id].materialAdoublePtr = boost::make_shared<
        NonlinearElasticElement::FunctionsToCalculatePiolaKirchhoffI<adouble>
        >();
        break;
        case NEOHOOKEAN:
        elasticFe->setOfBlocks[id].materialDoublePtr = boost::make_shared<NeoHookean<double> >();
        elasticFe->setOfBlocks[id].materialAdoublePtr = boost::make_shared<NeoHookean<adouble> >();
        break;
        case BONEHOOKE: {
          boost::shared_ptr<BoneHooke<double> > mat_double = boost::make_shared<BoneHooke<double> >();
          boost::shared_ptr<BoneHooke<adouble> > mat_adouble = boost::make_shared<BoneHooke<adouble> >();
          // double
          mat_double->rHo0 = rHo0;
          mat_double->n = nBone;
          mat_double->rhoAtGaussPts = rhoAtGaussPt;
          elasticFe->setOfBlocks[id].materialDoublePtr = mat_double;
          // adouble
          mat_adouble->rHo0 = rHo0;
          mat_adouble->n = nBone;
          mat_adouble->rhoAtGaussPts = rhoAtGaussPt;
          elasticFe->setOfBlocks[id].materialAdoublePtr = mat_adouble;
        }
        break;
        case BONENEOHOOKEAN:
        SETERRQ(mField.get_comm(),MOFEM_NOT_IMPLEMENTED,"Not yet implemented");
        break;
      }

    }


    elasticFe->commonData.spatialPositions = "SPATIAL_POSITION";
    elasticFe->commonData.meshPositions = "MESH_NODE_POSITIONS";

    /// Rhs
    feRhs->getOpPtrVector().push_back(
      new OpGetCrackFrontCommonDataAtGaussPts(
        "SPATIAL_POSITION",elasticFe->commonData,feRhs->singularElement,feRhs->invSJac
      )
    );
    feRhs->getOpPtrVector().push_back(
      new NonlinearElasticElement::OpGetCommonDataAtGaussPts(
        "MESH_NODE_POSITIONS",elasticFe->commonData
      )
    );
    if(defaultMaterial == BONEHOOKE || defaultMaterial == BONENEOHOOKEAN) {
      feRhs->getOpPtrVector().push_back(new OpCalculateScalarFieldValues("RHO",rhoAtGaussPt));
    }
    feRhs->getOpPtrVector().push_back(
      new OpTransfromSingularBaseFunctions(feRhs->singularElement,feRhs->invSJac)
    );
    map<int,NonlinearElasticElement::BlockData>::iterator sit = elasticFe->setOfBlocks.begin();
    for(;sit!=elasticFe->setOfBlocks.end();sit++) {
      feRhs->getOpPtrVector().push_back(
        new NonlinearElasticElement::OpJacobianPiolaKirchhoffStress(
          "SPATIAL_POSITION",sit->second,elasticFe->commonData,ELASTIC_TAG,false,true,false
        )
      );
      feRhs->getOpPtrVector().push_back(
        new NonlinearElasticElement::OpRhsPiolaKirchhoff(
          "SPATIAL_POSITION",sit->second,elasticFe->commonData
        )
      );
    }
    if(residualStressBlock!=-1) {
      if(!mwlsApprox) {
        mwlsApprox = boost::shared_ptr<MWLSApprox>(elasticFe,new MWLSApprox(mField));
        ierr = mwlsApprox->loadMWLSMesh(mwlsApproxFile.c_str());
        MeshsetsManager* block_manager_ptr;
        ierr = mField.query_interface(block_manager_ptr); CHKERRQ(ierr);
        ierr = block_manager_ptr->getEntitiesByDimension(
          residualStressBlock,BLOCKSET,3,mwlsApprox->tetsInBlock,true
        ); CHKERRQ(ierr);
        Tag th;
        rval = mwlsApprox->mwlsMoab.tag_get_handle("MED_SIGP",th); CHKERRQ_MOAB(rval);
        ierr = mwlsApprox->getValuesToNodes(th); CHKERRQ(ierr);
      }
      feRhs->getOpPtrVector().push_back(new MWLSApprox::OpMWLSStressAtGaussPts(mwlsApprox));
      feRhs->getOpPtrVector().push_back(new MWLSApprox::OpMWLSStressRhs(mwlsApprox));
    }

    /// Lhs
    feLhs->getOpPtrVector().push_back(
      new OpGetCrackFrontCommonDataAtGaussPts(
        "SPATIAL_POSITION",elasticFe->commonData,feLhs->singularElement,feLhs->invSJac
      )
    );
    feLhs->getOpPtrVector().push_back(
      new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",elasticFe->commonData)
    );
    if(defaultMaterial == BONEHOOKE || defaultMaterial == BONENEOHOOKEAN) {
      feLhs->getOpPtrVector().push_back(new OpCalculateScalarFieldValues("RHO",rhoAtGaussPt));
    }
    feLhs->getOpPtrVector().push_back(
      new OpTransfromSingularBaseFunctions(feLhs->singularElement,feLhs->invSJac)
    );
    sit = elasticFe->setOfBlocks.begin();
    for(;sit!=elasticFe->setOfBlocks.end();sit++) {
      feLhs->getOpPtrVector().push_back(
        new NonlinearElasticElement::OpJacobianPiolaKirchhoffStress(
          "SPATIAL_POSITION",sit->second,elasticFe->commonData,ELASTIC_TAG,true,true,false
        )
      );
      feLhs->getOpPtrVector().push_back(
        new NonlinearElasticElement::OpLhsPiolaKirchhoff_dx(
          "SPATIAL_POSITION","SPATIAL_POSITION",sit->second,elasticFe->commonData
        )
      );
    }
  }

  double *scale_lhs = &(arcLengthDof->getFieldData());
  double *scale_rhs = &(arcLengthDof->getFieldData());
  boost::shared_ptr<NeummanForcesSurfaceComplexForLazy> neumann_forces =
  boost::shared_ptr<NeummanForcesSurfaceComplexForLazy>(new NeummanForcesSurfaceComplexForLazy(
    mField,PETSC_NULL,PETSC_NULL,scale_lhs,scale_rhs)
  );
  // Calculate element vectors to calculate  forces on surface
  {
    NeummanForcesSurfaceComplexForLazy::MyTriangleSpatialFE &fe_forces = neumann_forces->getLoopSpatialFe();
    fe_forces.typeOfForces = NeummanForcesSurfaceComplexForLazy::MyTriangleSpatialFE::NONCONSERVATIVE;
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,NODESET|FORCESET,it)) {
      ierr = fe_forces.addForce(it->getMeshsetId()); CHKERRQ(ierr);
    }
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,SIDESET|PRESSURESET,it)) {
      ierr = fe_forces.addPreassure(it->getMeshsetId()); CHKERRQ(ierr);
    }
  }

  boost::shared_ptr<boost::ptr_map<string,NodalForce> > nodal_forces = boost::make_shared<boost::ptr_map<string,NodalForce> >();
  // Nodal forces
  {
    string fe_name_str ="FORCE_FE";
    nodal_forces->insert(fe_name_str,new NodalForce(mField));
    nodal_forces->at(fe_name_str).methodsOp.push_back(new MetaNodalForces::DofForceScale(arcLengthDof));
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,NODESET|FORCESET,it)) {
      ierr = nodal_forces->at(fe_name_str).addForce("SPATIAL_POSITION",f,it->getMeshsetId(),true);  CHKERRQ(ierr);
    }
  }
  // Edge forces
  boost::shared_ptr<boost::ptr_map<string,EdgeForce> > edge_forces = boost::make_shared<boost::ptr_map<string,EdgeForce> >();
  {
    string fe_name_str ="FORCE_FE";
    edge_forces->insert(fe_name_str,new EdgeForce(mField));
    edge_forces->at(fe_name_str).methodsOp.push_back(new MetaNodalForces::DofForceScale(arcLengthDof));
    for(_IT_CUBITMESHSETS_BY_BCDATA_TYPE_FOR_LOOP_(mField,NODESET|FORCESET,it)) {
      ierr = edge_forces->at(fe_name_str).addForce("SPATIAL_POSITION",f,it->getMeshsetId(),true);  CHKERRQ(ierr);
    }
  }

  // Kinematic boundary conditions
  boost::shared_ptr<FEMethod> my_dirichlet_bc = boost::shared_ptr<FEMethod>(
    new SpatialPositionsBCFEMethodPreAndPostProc(mField,"SPATIAL_POSITION",PETSC_NULL,PETSC_NULL,PETSC_NULL)
  );

  #ifdef __ANALITICAL_DISPLACEMENT__
  boost::shared_ptr<AnalyticalDirichletBC::DirichletBC> analytical_ditihlet_bc =
  boost::shared_ptr<AnalyticalDirichletBC::DirichletBC>(new AnalyticalDirichletBC::DirichletBC(mField,"SPATIAL_POSITION",m,q,f));
  if(mField.check_problem("BC_PROBLEM")) {
    //solve for ditihlet bc dofs
    AnalyticalDirichletBC analytical_bc(mField);
    ierr = analytical_bc.setUpProblem(mField,"BC_PROBLEM"); CHKERRQ(ierr);
    boost::shared_ptr<AnaliticalDisp> testing_function = boost::shared_ptr<AnaliticalDisp>(new AnaliticalDisp);
    EntityHandle fe_meshset  = mField.get_finite_element_meshset("BC_FE");
    ierr = analytical_bc.setApproxOps(mField,"SPATIAL_POSITION",testing_function,0,"MESH_NODE_POSITIONS"); CHKERRQ(ierr);
    analytical_bc.approxField.getLoopFeApprox().addToRule = 4;
    ierr = analytical_bc.solveProblem(mField,"BC_PROBLEM","BC_FE",*analytical_ditihlet_bc); CHKERRQ(ierr);
    ierr = analytical_bc.destroyProblem(); CHKERRQ(ierr);
  }
  #endif //__ANALITICAL_DISPLACEMENT__

  // Analytical forces
  #ifdef __ANALITICAL_TRACTION__
  boost::shared_ptr<boost::ptr_map<string,NeummanForcesSurface> >
  analitical_surface_element = boost::make_shared<boost::ptr_map<string,NeummanForcesSurface> >();
  {
    MeshsetsManager *meshset_manager_ptr;
    ierr = mField.query_interface(meshset_manager_ptr); CHKERRQ(ierr);
    if(meshset_manager_ptr->checkMeshset("ANALITICAL_TRACTION")) {
      string fe_name_str = "ANALITICAL_NEUAMNN_FE";
      analitical_surface_element->insert(fe_name_str,new NeummanForcesSurface(mField));
      analitical_surface_element->at(fe_name_str).fe.addToRule = 4;
      const CubitMeshSets *cubit_meshset_ptr;
      meshset_manager_ptr->getCubitMeshsetPtr("ANALITICAL_TRACTION",&cubit_meshset_ptr);
      Range faces;
      ierr = meshset_manager_ptr->getEntitiesByDimension(
        cubit_meshset_ptr->getMeshsetId(),BLOCKSET,2,faces,true
      ); CHKERRQ(ierr);
      analitical_surface_element->at(fe_name_str).analyticalForceOp.push_back(
        new AnalyticalForces()
      );
      analitical_surface_element->at(fe_name_str).fe.getOpPtrVector().push_back(
        new NeummanForcesSurface::OpNeumannForceAnalytical(
          "SPATIAL_POSITION",
          PETSC_NULL,
          faces,
          analitical_surface_element->at(fe_name_str).methodsOp,
          analitical_surface_element->at(fe_name_str).analyticalForceOp,
          true
        )
      );
    }
  }
  #endif //__ANALITICAL_TRACTION__

  boost::shared_ptr<FEMethod> null;

  // Add elements to SNES
  ierr = DMMoFEMSNESSetJacobian(dm,"ELASTIC",null,my_dirichlet_bc,null); CHKERRQ(ierr);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(mField.check_problem("BC_PROBLEM")) {
    ierr = DMMoFEMSNESSetJacobian(dm,"ELASTIC",null,analytical_ditihlet_bc,null); CHKERRQ(ierr);
  }
  #endif //__ANALITICAL_DISPLACEMENT__
  ierr = DMMoFEMSNESSetJacobian(dm,"ELASTIC",feLhs,null,null); CHKERRQ(ierr);
  ierr = DMMoFEMSNESSetJacobian(
    dm,"NEUAMNN_FE",boost::shared_ptr<FEMethod>(neumann_forces,&neumann_forces->getLoopSpatialFe()),null,null
  ); CHKERRQ(ierr);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(mField.check_problem("BC_PROBLEM")) {
    ierr = DMMoFEMSNESSetJacobian(dm,"ELASTIC",null,null,analytical_ditihlet_bc); CHKERRQ(ierr);
  }
  #endif //__ANALITICAL_DISPLACEMENT__
  ierr = DMMoFEMSNESSetJacobian(dm,"ELASTIC",null,null,my_dirichlet_bc); CHKERRQ(ierr);

  ierr = DMMoFEMSNESSetFunction(dm,"ELASTIC",null,my_dirichlet_bc,null); CHKERRQ(ierr);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(mField.check_problem("BC_PROBLEM")) {
    ierr = DMMoFEMSNESSetFunction(dm,"ELASTIC",null,analytical_ditihlet_bc,null); CHKERRQ(ierr);
  }
  #endif //__ANALITICAL_DISPLACEMENT__
  ierr = DMMoFEMSNESSetFunction(dm,"ELASTIC",feRhs,null,null); CHKERRQ(ierr);
  ierr = DMMoFEMSNESSetFunction(
    dm,"NEUAMNN_FE",boost::shared_ptr<FEMethod>(neumann_forces,&neumann_forces->getLoopSpatialFe()),null,null
  ); CHKERRQ(ierr);
  for(
    boost::ptr_map<string,NodalForce>::iterator
    fit = edge_forces->begin();fit!=edge_forces->end();fit++
  ) {
    ierr = DMMoFEMSNESSetFunction(
      dm,fit->first.c_str(),
      boost::shared_ptr<FEMethod>(edge_forces,&fit->second->getLoopFe()),
      null,null
    ); CHKERRQ(ierr);
  }
  for(
    boost::ptr_map<string,NodalForce>::iterator
    fit = nodal_forces->begin();fit!=nodal_forces->end();fit++
  ) {
    ierr = DMMoFEMSNESSetFunction(
      dm,fit->first.c_str(),
      boost::shared_ptr<FEMethod>(nodal_forces,&fit->second->getLoopFe()),
      null,null
    ); CHKERRQ(ierr);
  }
  #ifdef __ANALITICAL_TRACTION__
  for(
    boost::ptr_map<string,NeummanForcesSurface>::iterator
    fit = analitical_surface_element->begin();fit!=analitical_surface_element->end();fit++
  ) {
    ierr = DMMoFEMSNESSetFunction(
      dm,fit->first.c_str(),
      boost::shared_ptr<FEMethod>(analitical_surface_element,&fit->second->fe),null,null
    ); CHKERRQ(ierr);
  }
  #endif //__ANALITICAL_TRACTION__
  ierr = DMMoFEMSNESSetFunction(dm,"ELASTIC",null,null,my_dirichlet_bc); CHKERRQ(ierr);
  #ifdef __ANALITICAL_DISPLACEMENT__
  if(mField.check_problem("BC_PROBLEM")) {
    ierr = DMMoFEMSNESSetFunction(dm,"ELASTIC",null,null,analytical_ditihlet_bc); CHKERRQ(ierr);
  }
  #endif //__ANALITICAL_DISPLACEMENT__

  if(debug) {
    SNES snes;
    ierr = SNESCreate(PETSC_COMM_WORLD,&snes); CHKERRQ(ierr);
    SnesCtx *snes_ctx;
    ierr = DMMoFEMGetSnesCtx(dm,&snes_ctx); CHKERRQ(ierr);
    ierr = SNESSetFunction(snes,f,SnesRhs,snes_ctx); CHKERRQ(ierr);
    ierr = SNESSetJacobian(snes,m,m,SnesMat,snes_ctx); CHKERRQ(ierr);
    ierr = SnesMat(snes,q,m,m,snes_ctx); CHKERRQ(ierr);
    ierr = SnesRhs(snes,q,f,snes_ctx); CHKERRQ(ierr);
    ierr = SNESDestroy(&snes); CHKERRQ(ierr);
    MatView(m,PETSC_VIEWER_DRAW_WORLD);
    std::string wait;
    std::cin >> wait;
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::assembleMaterialForcesDM(
  DM dm,const int verb,const bool debug
) {
  PetscFunctionBegin;
  if(!elasticFe) {
    elasticFe = boost::shared_ptr<NonlinearElasticElement>(new NonlinearElasticElement(mField,ELASTIC_TAG));
  }
  feMaterialRhs = boost::shared_ptr<CrackFrontFE>(
    elasticFe,new CrackFrontFE(mField,crackFrontNodes,crackFrontNodesEdges,crackFrontElements)
  );
  feMaterialLhs = boost::shared_ptr<CrackFrontFE>(
    elasticFe,new CrackFrontFE(mField,crackFrontNodes,crackFrontNodesEdges,crackFrontElements)
  );

  /// Rhs
  feMaterialRhs->getOpPtrVector().push_back(
    new OpGetCrackFrontCommonDataAtGaussPts(
      "SPATIAL_POSITION",elasticFe->commonData,feMaterialRhs->singularElement,feMaterialRhs->invSJac
    )
  );
  feMaterialRhs->getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",elasticFe->commonData)
  );
  if(defaultMaterial == BONEHOOKE || defaultMaterial == BONENEOHOOKEAN) {
    feMaterialRhs->getOpPtrVector().push_back(new OpCalculateScalarFieldValues("RHO",rhoAtGaussPt));
    feMaterialRhs->getOpPtrVector().push_back(new OpCalculateScalarFieldGradient<3>("RHO",rhoGradAtGaussPt));
  }
  feMaterialRhs->getOpPtrVector().push_back(
    new OpTransfromSingularBaseFunctions(feMaterialRhs->singularElement,feMaterialRhs->invSJac)
  );
  map<int,NonlinearElasticElement::BlockData>::iterator sit = elasticFe->setOfBlocks.begin();
  for(;sit!=elasticFe->setOfBlocks.end();sit++) {
    feMaterialRhs->getOpPtrVector().push_back(
      new NonlinearElasticElement::OpJacobianEshelbyStress(
        "SPATIAL_POSITION",sit->second,elasticFe->commonData,ELASTIC_TAG,false,true
      )
    );
    feMaterialRhs->getOpPtrVector().push_back(
      new NonlinearElasticElement::OpRhsEshelbyStrees(
        "MESH_NODE_POSITIONS",sit->second,elasticFe->commonData
      )
    );
    if(defaultMaterial == BONEHOOKE || defaultMaterial == BONENEOHOOKEAN) {
      feMaterialRhs->getOpPtrVector().push_back(
        new NonlinearElasticElement::OpJacobianEnergy(
          "SPATIAL_POSITION",sit->second,elasticFe->commonData,EXTERIOR_DIREVATIVE_TAG,true,false,true,false
        )
      );
      feMaterialRhs->getOpPtrVector().push_back(
        new OpRhsBoneExplicitDerivarive(sit->second,elasticFe->commonData,rhoGradAtGaussPt)
      );

    }
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::calculateElasticEnergy(DM dm) {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if(!elasticFe) {
    elasticFe = boost::shared_ptr<NonlinearElasticElement>(new NonlinearElasticElement(mField,ELASTIC_TAG));
  }
  feEnergy = boost::shared_ptr<CrackFrontFE>(
    elasticFe,new CrackFrontFE(mField,crackFrontNodes,crackFrontNodesEdges,crackFrontElements)
  );
  feEnergy->getOpPtrVector().push_back(
    new OpGetCrackFrontCommonDataAtGaussPts(
      "SPATIAL_POSITION",elasticFe->commonData,feEnergy->singularElement,feEnergy->invSJac
    )
  );
  feEnergy->getOpPtrVector().push_back(
    new NonlinearElasticElement::OpGetCommonDataAtGaussPts("MESH_NODE_POSITIONS",elasticFe->commonData)
  );
  feEnergy->getOpPtrVector().push_back(
    new OpTransfromSingularBaseFunctions(feEnergy->singularElement,feEnergy->invSJac)
  );
  map<int,NonlinearElasticElement::BlockData>::iterator sit = elasticFe->setOfBlocks.begin();
  for(;sit!=elasticFe->setOfBlocks.end();sit++) {
    feEnergy->getOpPtrVector().push_back(new NonlinearElasticElement::OpEnergy(
      "SPATIAL_POSITION",sit->second,elasticFe->commonData,&feEnergy->V,false)
    );
  }
  feEnergy->snes_ctx = SnesMethod::CTX_SNESNONE;
  feEnergy->eNergy = 0;
  ierr = DMoFEMLoopFiniteElements(dm,"ELASTIC",feEnergy.get()); CHKERRQ(ierr);
  PetscPrintf(PETSC_COMM_WORLD,"Elastic energy %6.4e\n",feEnergy->eNergy);
  PetscFunctionReturn(0);
}


PetscErrorCode CrackPropagation::calculateMaterialFocesDM(
  DM dm,Vec q,Vec f,const int verb,const bool debug
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  feMaterialRhs->snes_f = f;
  feMaterialRhs->snes_ctx = FEMethod::CTX_SNESSETFUNCTION;
  ierr = VecZeroEntries(f); CHKERRQ(ierr);
  ierr = DMoFEMLoopFiniteElements(dm,"MATERIAL",feMaterialRhs.get()); CHKERRQ(ierr);
  ierr = VecAssemblyBegin(f); CHKERRQ(ierr);
  ierr = VecAssemblyEnd(f); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(f,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(f,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(f,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(f,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  if(debug) {
    ierr = VecView(f,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
  }
  const MoFEM::MoFEMProblem *problem_ptr;
  ierr = DMMoFEMGetProblemPtr(dm,&problem_ptr); CHKERRQ(ierr);
  PostProcVertexMethod ent_method_material_force(mField,f,"MATERIAL_FORCE");
  ierr = mField.loop_dofs(
    problem_ptr->getName(),"MESH_NODE_POSITIONS",ROW,ent_method_material_force
  ); CHKERRQ(ierr);
  PostProcVertexMethod ent_method_material_position(mField,q,"MESH_NODE_POSITIONS");
  ierr = mField.loop_dofs(
    problem_ptr->getName(),"MESH_NODE_POSITIONS",ROW,ent_method_material_position
  ); CHKERRQ(ierr);
  if(debug) {
    EntityHandle meshset;
    rval = mField.get_moab().create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
    rval = mField.get_moab().add_entities(meshset,crackFront); CHKERRQ_MOAB(rval);
    // cerr << crackFront << endl;
    std::ostringstream file_name;
    file_name << "out_crack_front.h5m";
    rval = mField.get_moab().write_file(
      file_name.str().c_str(),"MOAB","PARALLEL=WRITE_PART",&meshset,1
    ); CHKERRQ_MOAB(rval);
    rval = mField.get_moab().delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::calculateSurfaceProjectionMatrix(
  DM dm_surface,
  DM dm_project,
  std::vector<int> &ids,
  const int verb,
  const bool debug
) {
  PetscErrorCode ierr;
  const MoFEM::MoFEMProblem *prb_surf_ptr;
  const MoFEM::MoFEMProblem *prb_proj_ptr;
  PetscFunctionBegin;
  ierr = DMMoFEMGetProblemPtr(dm_surface,&prb_surf_ptr); CHKERRQ(ierr);
  ierr = DMMoFEMGetProblemPtr(dm_project,&prb_proj_ptr); CHKERRQ(ierr);

  projSurfaceCtx = boost::shared_ptr<ConstrainMatrixCtx>(
    new ConstrainMatrixCtx(mField,prb_proj_ptr->getName(),prb_surf_ptr->getName(),true,true)
  );
  projSurfaceCtx->cancelKSPMonitor = false;

  ierr = DMCreateMatrix(dm_surface,&projSurfaceCtx->C); CHKERRQ(ierr);
  ierr = MatSetOption(projSurfaceCtx->C,MAT_NEW_NONZERO_LOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);
  ierr = MatSetOption(projSurfaceCtx->C,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);
  ierr = MatZeroEntries(projSurfaceCtx->C); CHKERRQ(ierr);

  // Calulate and assemble constrain matrix
  for(std::vector<int>::iterator id = ids.begin();id!=ids.end();id++) {
    bool crack_surface = (*id == 200) ? true : false;
    FaceOrientation orientation(crack_surface);
    SurfaceSlidingConstrains surface_constrain(mField,orientation);
    surface_constrain.setOperatorsCOnly("LAMBDA_SURFACE"+boost::lexical_cast<std::string>(*id),"MESH_NODE_POSITIONS");
    std::string fe_name = (*id == 200) ? "CRACK_SURFACE" : "SURFACE";
    surface_constrain.feLhs.snes_B = projSurfaceCtx->C;
    ierr = DMoFEMLoopFiniteElements(dm_surface,fe_name.c_str(),&surface_constrain.feLhs); CHKERRQ(ierr);
  }

  ierr = MatAssemblyBegin(projSurfaceCtx->C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(projSurfaceCtx->C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  if(debug) {
    MatView(projSurfaceCtx->C,PETSC_VIEWER_DRAW_WORLD);
    std::string wait;
    std::cin >> wait;
    MatView(projSurfaceCtx->C,PETSC_VIEWER_STDOUT_WORLD);
  }


  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::calculateFrontProjectionMatrix(
  DM dm_front,
  DM dm_project,
  const int verb,
  const bool debug
) {
  PetscErrorCode ierr;
  const MoFEM::MoFEMProblem *prb_front_ptr;
  const MoFEM::MoFEMProblem *prb_proj_ptr;
  PetscFunctionBegin;
  ierr = DMMoFEMGetProblemPtr(dm_front,&prb_front_ptr); CHKERRQ(ierr);
  ierr = DMMoFEMGetProblemPtr(dm_project,&prb_proj_ptr); CHKERRQ(ierr);

  projFrontCtx = boost::shared_ptr<ConstrainMatrixCtx>(
    new ConstrainMatrixCtx(mField,prb_proj_ptr->getName(),prb_front_ptr->getName(),true,true)
  );
  projFrontCtx->cancelKSPMonitor = false;

  Mat Q;
  {
    int M,m;
    M = prb_proj_ptr->getNbDofsCol();
    m = prb_proj_ptr->getNbLocalDofsCol();
    ierr = MatCreateShell(mField.get_comm(),m,m,M,M,projSurfaceCtx.get(),&Q); CHKERRQ(ierr);
    ierr = MatShellSetOperation(Q,MATOP_MULT,(void(*)(void))PorjectionMatrixMultOpQ); CHKERRQ(ierr);
  }

  ierr = DMCreateMatrix(dm_front,&projFrontCtx->C); CHKERRQ(ierr);
  ierr = MatSetOption(projFrontCtx->C,MAT_NEW_NONZERO_LOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);
  ierr = MatSetOption(projFrontCtx->C,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);
  ierr = MatZeroEntries(projFrontCtx->C); CHKERRQ(ierr);

  // Set up const area constrain

  // Calculate constant area constrains
  ConstantArea constant_area(mField);
  constant_area.feLhsPtr = boost::shared_ptr<ConstantArea::MyTriangleFE>(
    new ConstantArea::MyTriangleFE(mField,constant_area.commonData)
  );
  constant_area.feLhsPtr->petscB = projFrontCtx->C;
  constant_area.feLhsPtr->petscQ = Q;

  constant_area.feLhsPtr->getOpPtrVector().push_back(
    new ConstantArea::OpAreaJacobian(CONSTANT_AREA_TAG,constant_area.commonData)
  );
  constant_area.feLhsPtr->getOpPtrVector().push_back(
    new ConstantArea::OpAreaC(CONSTANT_AREA_TAG,constant_area.commonData)
  );
  ierr = DMoFEMLoopFiniteElements(dm_front,"CRACKFRONT_AREA_ELEM",constant_area.feLhsPtr.get()); CHKERRQ(ierr);

  // Calculate tangent constrains
  ConstantArea tangent_constrain(mField);
  tangent_constrain.feLhsPtr = boost::shared_ptr<ConstantArea::MyTriangleFE>(
    new ConstantArea::MyTriangleFE(mField,tangent_constrain.commonData)
  );
  tangent_constrain.feLhsPtr->petscB = projFrontCtx->C;
  tangent_constrain.feLhsPtr->getOpPtrVector().push_back(
    new ConstantArea::OpTangentJacobian(FRONT_TANGENT,tangent_constrain.commonData)
  );
  tangent_constrain.feLhsPtr->getOpPtrVector().push_back(
    new ConstantArea::OpTangentC(FRONT_TANGENT,tangent_constrain.commonData)
  );
  ierr = DMoFEMLoopFiniteElements(dm_front,"CRACKFRONT_AREA_TANGENT_ELEM",tangent_constrain.feLhsPtr.get()); CHKERRQ(ierr);


  // ObosleteUsersModules::C_CONSTANT_AREA c_area_elem(mField,projFrontCtx->C,Q,"LAMBDA_CRACKFRONT_AREA");
  // ierr = DMoFEMLoopFiniteElements(dm_front,"CRACKFRONT_AREA_ELEM",&c_area_elem);  CHKERRQ(ierr);
  // ObosleteUsersModules::C_FRONT_TANGENT c_tangent_elem(mField,projFrontCtx->C,PETSC_NULL,"LAMBDA_CRACKFRONT_AREA_TANGENT");
  // ierr = DMoFEMLoopFiniteElements(dm_front,"CRACKFRONT_AREA_TANGENT_ELEM",&c_tangent_elem);  CHKERRQ(ierr);

  ierr = MatAssemblyBegin(projFrontCtx->C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(projFrontCtx->C,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  if(debug) {
    MatView(projFrontCtx->C,PETSC_VIEWER_DRAW_WORLD);
    std::string wait;
    std::cin >> wait;
    MatView(projFrontCtx->C,PETSC_VIEWER_STDOUT_WORLD);
  }

  ierr = MatDestroy(&Q); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::calculateReleaseEneregy(
  DM dm,
  Vec f_material_proj,
  Vec f_griffith_proj,
  Vec f_lambda,
  const double gc,
  const int verb,
  const bool debug
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  const MoFEM::MoFEMProblem *prb_ptr;
  PetscFunctionBegin;
  ierr = DMMoFEMGetProblemPtr(dm,&prb_ptr); CHKERRQ(ierr);

  // R = CT*(CC^T)^(-1) [ unit m/m^(-2) = 1/m ] [ 0.5/0.25 = 2 ]
  // R^T = (CC^T)^(-T)C [ unit m/m^(-2) = 1/m ]
  int N,n;
  int M,m;
  ierr = VecGetSize(f_material_proj,&N); CHKERRQ(ierr);
  ierr = VecGetLocalSize(f_material_proj,&n); CHKERRQ(ierr);
  ierr = VecGetSize(f_lambda,&M); CHKERRQ(ierr);
  ierr = VecGetLocalSize(f_lambda,&m); CHKERRQ(ierr);
  Mat RT;
  ierr = MatCreateShell(PETSC_COMM_WORLD,m,n,M,N,projFrontCtx.get(),&RT); CHKERRQ(ierr);
  ierr = MatShellSetOperation(RT,MATOP_MULT,(void(*)(void))ConstrainMatrixMultOpRT); CHKERRQ(ierr);

  ierr = projFrontCtx->initializeQorP(f_material_proj); CHKERRQ(ierr);
  ierr = MatMult(RT,f_material_proj,f_lambda); CHKERRQ(ierr);
  ierr = VecScale(f_lambda,-1); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(f_lambda,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(f_lambda,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  ierr = MatDestroy(&RT); CHKERRQ(ierr);

  if(debug) {
    ierr = VecView(f_lambda,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
  }

  PostProcVertexMethod ent_method_g1(mField,f_lambda,"G1");
  ierr = mField.loop_dofs(
    prb_ptr->getName(),"LAMBDA_CRACKFRONT_AREA",ROW,ent_method_g1
  ); CHKERRQ(ierr);
  PostProcVertexMethod ent_method_g3(mField,f_lambda,"G3");
  ierr = mField.loop_dofs(
    prb_ptr->getName(),"LAMBDA_CRACKFRONT_AREA_TANGENT",ROW,ent_method_g3
  ); CHKERRQ(ierr);

  if(verb>0) {
    map<EntityHandle,double> map_g1;
    map<EntityHandle,double> map_g3;
    map<EntityHandle,VectorDouble3> map_mat_force;
    map<EntityHandle,VectorDouble3> map_griffith;
    NumeredDofEntity_multiIndex::index<PetscLocalIdx_mi_tag>::type::iterator it,hi_it;
    {
      double *a_lambda;
      ierr = VecGetArray(f_lambda,&a_lambda); CHKERRQ(ierr);
      it=prb_ptr->numered_dofs_rows->get<PetscLocalIdx_mi_tag>().lower_bound(0);
      hi_it=prb_ptr->numered_dofs_rows->get<PetscLocalIdx_mi_tag>().upper_bound(prb_ptr->getNbLocalDofsRow()-1);
      for(;it!=hi_it;it++) {
        if(it->get()->getName()=="LAMBDA_CRACKFRONT_AREA") {
          map_g1[it->get()->getEnt()] = a_lambda[it->get()->getPetscLocalDofIdx()];
        } else if(it->get()->getName()=="LAMBDA_CRACKFRONT_AREA_TANGENT") {
          map_g3[it->get()->getEnt()] = a_lambda[it->get()->getPetscLocalDofIdx()];
        } else {
          SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Should not happen");
        }
      }
      ierr = VecRestoreArray(f_lambda,&a_lambda); CHKERRQ(ierr);
    }
    {
      Vec f_mat_scat,f_griff_scat;
      ierr = mField.VecCreateGhost(prb_ptr->getName(),COL,&f_mat_scat); CHKERRQ(ierr);
      ierr = VecDuplicate(f_mat_scat,&f_griff_scat); CHKERRQ(ierr);
      ierr = VecScatterBegin(projFrontCtx->sCatter,f_material_proj,f_mat_scat,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = VecScatterEnd(projFrontCtx->sCatter,f_material_proj,f_mat_scat,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = VecScatterBegin(projFrontCtx->sCatter,f_griffith_proj,f_griff_scat,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      ierr = VecScatterEnd(projFrontCtx->sCatter,f_griffith_proj,f_griff_scat,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
      double *a_material,*a_griffith;
      ierr = VecGetArray(f_mat_scat,&a_material); CHKERRQ(ierr);
      ierr = VecGetArray(f_griff_scat,&a_griffith); CHKERRQ(ierr);
      it=prb_ptr->numered_dofs_cols->get<PetscLocalIdx_mi_tag>().lower_bound(0);
      hi_it=prb_ptr->numered_dofs_cols->get<PetscLocalIdx_mi_tag>().upper_bound(prb_ptr->getNbLocalDofsCol()-1);
      for(;it!=hi_it;it++) {
        if(it->get()->getName()=="MESH_NODE_POSITIONS") {
          EntityHandle ent = it->get()->getEnt();
          map_mat_force[ent].resize(3);
          map_mat_force[ent][it->get()->getDofCoeffIdx()] = a_material[it->get()->getPetscLocalDofIdx()];
          map_griffith[ent].resize(3);
          map_griffith[ent][it->get()->getDofCoeffIdx()] = a_griffith[it->get()->getPetscLocalDofIdx()];
        } else {
          SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Should not happen");
        }
      }
      ierr = VecRestoreArray(f_mat_scat,&a_material); CHKERRQ(ierr);
      ierr = VecRestoreArray(f_griff_scat,&a_griffith); CHKERRQ(ierr);
      ierr = VecDestroy(&f_mat_scat); CHKERRQ(ierr);
      ierr = VecDestroy(&f_griff_scat); CHKERRQ(ierr);
    }
    for(
      map<EntityHandle,double>::iterator mit = map_g1.begin();
      mit!=map_g1.end();mit++
    ) {
      EntityHandle ent = mit->first;
      double g1 = mit->second;
      double g3 = map_g3[ent];
      double j = norm_2(map_mat_force[ent])/(norm_2(map_griffith[ent])/(gc>0?gc:1));
      double g2 = j*j-g1*g1-g3*g3;
      double coords[3];
      rval = mField.get_moab().get_coords(&ent,1,coords); CHKERRQ_MOAB(rval);
      ostringstream ss;
      ss << "griffith force at ent ";
      ss << setw(5) << ent;
      //coords
      ss << "\tcoords";
      ss << " " << setw(10) << setprecision(4) << coords[0];
      ss << " " << setw(10) << setprecision(4) << coords[1];
      ss << " " << setw(10) << setprecision(4) << coords[2];
      //g1
      ss << "\t\tg1 " << scientific << setprecision(4) << g1;
      ss << " / " << scientific << setprecision(4) << j;
      ss << " ( " << scientific << setprecision(4) << g1/j << " )";
      //g1
      ss << "\t\tg2 " << scientific << setprecision(4) << g2;
      ss << " ( " << scientific << setprecision(4) << g2/j << " )";
      //g3
      ss << "\t\tg3 " << scientific << setprecision(4) << g3;
      ss << " ( " << scientific << setprecision(4) << g3/j << " )";
      if(gc>0) {
        ss << "\t relative error (gc-g)/gc ";
        ss << scientific << setprecision(4) << (gc-g1)/gc;
        ss << " / " << scientific << setprecision(4) << (gc-j)/gc;
      }
      ss << endl;
      PetscSynchronizedPrintf(mField.get_comm(),"%s",ss.str().c_str());
    }
    PetscSynchronizedFlush(mField.get_comm(),PETSC_STDOUT);
  }

  PetscFunctionReturn(0);
}



PetscErrorCode CrackPropagation::projectMaterialForcesDM(
  DM dm_project,Vec f,Vec f_proj,const int verb,const bool debug
) {
  PetscErrorCode ierr;
  const MoFEM::MoFEMProblem *prb_proj_ptr;

  PetscFunctionBegin;
  ierr = DMMoFEMGetProblemPtr(dm_project,&prb_proj_ptr); CHKERRQ(ierr);

  if(debug) {
    ierr = VecView(f,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
    ierr = projSurfaceCtx->initializeQorP(f); CHKERRQ(ierr);
    ierr = VecScatterView(projSurfaceCtx->sCatter,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
    Vec x;
    ierr = mField.VecCreateGhost(projSurfaceCtx->yProblem,COL,&x); CHKERRQ(ierr);
    ierr = VecScatterBegin(projSurfaceCtx->sCatter,f,x,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = VecScatterEnd(projSurfaceCtx->sCatter,f,x,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = VecView(x,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
    Vec Cf;
    ierr = mField.VecCreateGhost(projSurfaceCtx->yProblem,ROW,&Cf); CHKERRQ(ierr);
    ierr = MatMult(projSurfaceCtx->C,x,Cf);  CHKERRQ(ierr);
    ierr = VecView(f,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
    ierr = VecDestroy(&Cf); CHKERRQ(ierr);
    ierr = VecDestroy(&x); CHKERRQ(ierr);
  }

  int M,m;
  ierr = VecGetSize(f,&M); CHKERRQ(ierr);
  ierr = VecGetLocalSize(f,&m); CHKERRQ(ierr);
  Mat Q;
  ierr = MatCreateShell(PETSC_COMM_WORLD,m,m,M,M,projSurfaceCtx.get(),&Q); CHKERRQ(ierr);
  ierr = MatShellSetOperation(Q,MATOP_MULT,(void(*)(void))PorjectionMatrixMultOpQ); CHKERRQ(ierr);
  ierr = MatMult(Q,f,f_proj); CHKERRQ(ierr);
  PostProcVertexMethod ent_method_spatial_position(mField,f_proj,"MATERIAL_FORCE_PROJECTED");
  ierr = mField.loop_dofs(
    prb_proj_ptr->getName(),"MESH_NODE_POSITIONS",ROW,ent_method_spatial_position
  ); CHKERRQ(ierr);
  ierr = MatDestroy(&Q); CHKERRQ(ierr);

  if(verb>1) {
    double *a_f,*a_f_proj;
    VecGetArray(f,&a_f);
    VecGetArray(f_proj,&a_f_proj);
    for(Range::iterator vit = crackFrontNodes.begin();vit!=crackFrontNodes.end();vit++) {
      for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_NAME_ENT_PART_FOR_LOOP_(
        prb_proj_ptr,"MESH_NODE_POSITIONS",*vit,mField.get_comm_rank(),dit
      )) {
        int idx = dit->get()->getPetscLocalDofIdx();
        ostringstream ss;
        ss << **dit << " -> " << a_f[idx] << " : " << a_f_proj[idx] << endl;
        PetscSynchronizedPrintf(mField.get_comm(),"%s",ss.str().c_str());
      }
    }
    PetscSynchronizedFlush(mField.get_comm(),PETSC_STDOUT);
    VecRestoreArray(f,&a_f);
    VecRestoreArray(f_proj,&a_f_proj);
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::calculateGriffithForce(
  DM dm,const double gc,Vec f_griffith,const int verb,const bool debug
) {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = VecSetOption(f_griffith,VEC_IGNORE_NEGATIVE_INDICES,PETSC_TRUE);  CHKERRQ(ierr);

  GriffithForceElement griffith_force_element(mField);
  griffith_force_element.blockData[0].gc = gc;
  griffith_force_element.blockData[0].frontEdges = crackFront;
  griffith_force_element.blockData[0].frontNodes = crackFrontNodes;
  griffith_force_element.feRhs.getOpPtrVector().push_back(
    new GriffithForceElement::OpJacobian(
      GRIFFITH_FORCE_TAG,griffith_force_element.blockData[0],griffith_force_element.commonData
    )
  );
  griffith_force_element.feRhs.getOpPtrVector().push_back(
    new GriffithForceElement::OpRhs(
      GRIFFITH_FORCE_TAG,griffith_force_element.blockData[0],griffith_force_element.commonData
    )
  );
  griffith_force_element.feRhs.snes_f = f_griffith;
  ierr = VecZeroEntries(f_griffith); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(f_griffith,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(f_griffith,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = DMoFEMLoopFiniteElements(dm,"GRIFFITH_FORCE_ELEMENT",&griffith_force_element.feRhs); CHKERRQ(ierr);
  ierr = VecAssemblyBegin(f_griffith); CHKERRQ(ierr);
  ierr = VecAssemblyEnd(f_griffith); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(f_griffith,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(f_griffith,ADD_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(f_griffith,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(f_griffith,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::projectGriffithForce(
  DM dm,
  Vec f_griffith,
  Vec f_griffith_proj,
  const int verb,
  const bool debug
) {
  PetscErrorCode ierr;
  const MoFEM::MoFEMProblem *problem_ptr;
  PetscFunctionBegin;

  ierr = DMMoFEMGetProblemPtr(dm,&problem_ptr); CHKERRQ(ierr);

  // projection matrix
  Mat Q;
  {
    int M,m;
    ierr = VecGetSize(f_griffith,&M); CHKERRQ(ierr);
    ierr = VecGetLocalSize(f_griffith,&m); CHKERRQ(ierr);
    ierr = MatCreateShell(mField.get_comm(),m,m,M,M,projSurfaceCtx.get(),&Q); CHKERRQ(ierr);
    ierr = MatShellSetOperation(Q,MATOP_MULT,(void(*)(void))PorjectionMatrixMultOpQ); CHKERRQ(ierr);
  }

  // project griffith force
  {
    if(debug) {
      // ierr = VecView(f_griffith,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
      double nrm2_griffith_force;
      ierr = VecNorm(f_griffith,NORM_2,&nrm2_griffith_force);   CHKERRQ(ierr);
      ierr = PetscPrintf(mField.get_comm(),"nrm2_f_griffith = %6.4e\n",nrm2_griffith_force); CHKERRQ(ierr);
    }
    ierr = MatMult(Q,f_griffith,f_griffith_proj); CHKERRQ(ierr);
    if(debug) {
      // ierr = VecView(f_griffith_proj,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
    }
    ierr = VecGhostUpdateBegin(f_griffith_proj,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = VecGhostUpdateEnd(f_griffith_proj,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    PostProcVertexMethod ent_method_area(mField,f_griffith_proj,"GRIFFITH_FORCE");
    ierr = mField.loop_dofs(problem_ptr->getName(),"MESH_NODE_POSITIONS",COL,ent_method_area); CHKERRQ(ierr);
    if(verb>1) {
      double nrm2_griffith_force;
      ierr = VecNorm(f_griffith_proj,NORM_2,&nrm2_griffith_force);   CHKERRQ(ierr);
      ierr = PetscPrintf(mField.get_comm(),"nrm2_f_griffith_proj = %6.4e\n",nrm2_griffith_force); CHKERRQ(ierr);
    }
    if(verb>1) {
      double *a_f_griffth,*a_f_griffth_proj;
      VecGetArray(f_griffith,&a_f_griffth);
      VecGetArray(f_griffith_proj,&a_f_griffth_proj);
      for(Range::iterator vit = crackFrontNodes.begin();vit!=crackFrontNodes.end();vit++) {
        for(_IT_NUMEREDDOFMOFEMENTITY_ROW_BY_NAME_ENT_PART_FOR_LOOP_(
          problem_ptr,"MESH_NODE_POSITIONS",*vit,mField.get_comm_rank(),dit
        )) {
          int idx = dit->get()->getPetscLocalDofIdx();
          ostringstream ss;
          ss << **dit << " -> " << a_f_griffth[idx] << " : " << a_f_griffth_proj[idx] << endl;
          PetscSynchronizedPrintf(mField.get_comm(),"%s",ss.str().c_str());
        }
      }
      PetscSynchronizedFlush(mField.get_comm(),PETSC_STDOUT);
      VecRestoreArray(f_griffith,&a_f_griffth);
      VecRestoreArray(f_griffith_proj,&a_f_griffth_proj);
    }
  }

  ierr = MatDestroy(&Q); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::savePositionsOnCrackFrontDM(DM dm,Vec q,const int verb,const bool debug) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = VecGhostUpdateBegin(q,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(q,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  const MoFEM::MoFEMProblem *problem_ptr;
  ierr = DMMoFEMGetProblemPtr(dm,&problem_ptr); CHKERRQ(ierr);
  PostProcVertexMethod ent_method_spatial_position(mField,q,"SPATIAL_POSITION");
  ierr = mField.loop_dofs(
    problem_ptr->getName(),"SPATIAL_POSITION",ROW,ent_method_spatial_position
  ); CHKERRQ(ierr);
  if(debug) {
    EntityHandle meshset;
    rval = mField.get_moab().create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
    rval = mField.get_moab().add_entities(meshset,crackFront); CHKERRQ_MOAB(rval);
    // cerr << crackFront << endl;
    std::ostringstream file_name;
    file_name << "out_crack_spatial_position.h5m";
    rval = mField.get_moab().write_file(
      file_name.str().c_str(),"MOAB","PARALLEL=WRITE_PART",&meshset,1
    ); CHKERRQ_MOAB(rval);
    rval = mField.get_moab().delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::writeCrackFont(const int step) {
  MoABErrorCode rval;
  PetscFunctionBegin;
  EntityHandle meshset;
  rval = mField.get_moab().create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
  rval = mField.get_moab().add_entities(meshset,crackFront); CHKERRQ_MOAB(rval);
  // cerr << crackFront << endl;
  std::ostringstream file_name;
  file_name << "out_crack_front_" << step << ".h5m";
  rval = mField.get_moab().write_file(
    file_name.str().c_str(),"MOAB","PARALLEL=WRITE_PART",&meshset,1
  ); CHKERRQ_MOAB(rval);
  rval = mField.get_moab().delete_entities(&meshset,1); CHKERRQ_MOAB(rval);
  PetscFunctionReturn(0);
}

static PetscErrorCode snes_mat(SNES snes,Vec x,Mat A,Mat B,void *ctx) {
  PetscErrorCode ierr;
  SnesCtx* snes_ctx = (SnesCtx*)ctx;
  PetscFunctionBegin;
  DM dm;
  ierr = SNESGetDM(snes,&dm); CHKERRQ(ierr);
  int level;
  ierr = DMGetCoarsenLevel(dm,&level); CHKERRQ(ierr);
  const DMMGViaApproxOrdersCtx *dm_ctx;
  ierr = DMMGViaApproxOrdersGetCtx(dm,&dm_ctx); CHKERRQ(ierr);
  // int m,n;
  // ierr = MatGetSize(B,&m,&n); CHKERRQ(ierr);
  // cerr << "level " << level << " " << m << " " << n << endl;
  if(level == 0) {
    // cerr << "calculate" << endl;
    ierr = SnesMat(snes,x,A,B,ctx); CHKERRQ(ierr);
    const DMMGViaApproxOrdersCtx *dm_ctx;
    ierr = DMMGViaApproxOrdersGetCtx(dm,&dm_ctx); CHKERRQ(ierr);
    // MatView(A,PETSC_VIEWER_DRAW_WORLD);
    // std::string wait;
    // std::cin >> wait;
  } else {
    int nb_levels = dm_ctx->kspOperators.size();
    // int is_g_size;
    // ierr = ISGetSize(dm_ctx->coarseningIS[nb_levels-level-1],&is_g_size); CHKERRQ(ierr);
    // PetscSynchronizedPrintf(
    //   PETSC_COMM_WORLD,
    //   "nb_levels %d level %d "
    //   "is_g_size %d"
    //   "\n",
    //   nb_levels,level,
    //   is_g_size
    // );
    PetscSynchronizedFlush(PETSC_COMM_WORLD,PETSC_STDOUT);

    ierr = MatGetSubMatrix(
      dm_ctx->kspOperators.back(),
      dm_ctx->coarseningIS[nb_levels-level-1],
      dm_ctx->coarseningIS[nb_levels-level-1],
      MAT_REUSE_MATRIX,
      &A
    ); CHKERRQ(ierr);
    // MatView(A,PETSC_VIEWER_DRAW_WORLD);
    // std::string wait;
    // std::cin >> wait;
  }
  PetscFunctionReturn(0);
};

PetscErrorCode CrackPropagation::solveElasticDM(
  DM dm,SNES snes,Mat m,Vec q,Vec f,bool snes_set_up
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  // SNES snes;
  // const MoFEMProblem *problem_ptr;
  PetscFunctionBegin;
  // ierr = DMMoFEMGetProblemPtr(dm,&problem_ptr); CHKERRQ(ierr);

  ierr = DMoFEMMeshToLocalVector(dm,q,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  KSP ksp;
  PC pc;
  PetscBool same = PETSC_FALSE;
  SnesCtx *snes_ctx;

  ierr = DMMoFEMGetSnesCtx(dm,&snes_ctx); CHKERRQ(ierr);
  if(snes_set_up) {
    ierr = SNESSetDM(snes,dm); CHKERRQ(ierr);
    ierr = SNESSetFunction(snes,f,SnesRhs,snes_ctx); CHKERRQ(ierr);
    ierr = SNESSetJacobian(snes,m,m,snes_mat,snes_ctx); CHKERRQ(ierr);
    ierr = SNESSetFromOptions(snes); CHKERRQ(ierr);
    ierr = SNESGetKSP(snes,&ksp); CHKERRQ(ierr);
    ierr = KSPGetPC(ksp,&pc); CHKERRQ(ierr);
    ierr = PCSetType(pc,PCMG); CHKERRQ(ierr);
    PetscObjectTypeCompare((PetscObject)pc,PCMG,&same);
    PCMGSetUpViaApproxOrdersCtx pc_ctx(dm,m);
    if(same) {
      ierr = pc_ctx.getOptions(); CHKERRQ(ierr);
      ierr = pc_ctx.buildProlongationOperator(true,1); CHKERRQ(ierr);
      ierr = PCMGSetGalerkin(pc,PETSC_FALSE); CHKERRQ(ierr);
      ierr = PCMGSetLevels(pc,pc_ctx.nbLevels,NULL);  CHKERRQ(ierr);
    } else {
      // Operators are already set, do not use DM for doing that
      ierr = KSPSetDMActive(ksp,PETSC_FALSE); CHKERRQ(ierr);
    }
    ierr = KSPSetDMActive(ksp,PETSC_FALSE); CHKERRQ(ierr);
    ierr = SNESSetUp(snes); CHKERRQ(ierr);
  }

  ierr = MatZeroEntries(m); CHKERRQ(ierr);
  ierr = VecZeroEntries(f); CHKERRQ(ierr);
  ierr = VecGhostUpdateBegin(f,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(f,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);

  ierr = SNESSolve(snes,PETSC_NULL,q); CHKERRQ(ierr);

  // FIXME: Ugly fix to destroy SNES proprly
  if(snes_set_up) {
    if(same) {
      int nb_levels;
      ierr = PCMGGetLevels(pc,&nb_levels); CHKERRQ(ierr);
      for(int ll = 0;ll!=nb_levels-1;ll++) {
        ierr = PCMGGetSmoother(pc,ll,&ksp); CHKERRQ(ierr);
        DM dmc;
        ierr = KSPGetDM(ksp,&dmc); CHKERRQ(ierr);
        ierr = PetscObjectReference((PetscObject)dmc); CHKERRQ(ierr);
      }
    }
  }

  ierr = VecGhostUpdateBegin(q,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  ierr = VecGhostUpdateEnd(q,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
  // ierr = VecAssemblyBegin(q); CHKERRQ(ierr);
  // ierr = VecAssemblyEnd(q); CHKERRQ(ierr);
  //Save data on mesh
  ierr = DMoFEMMeshToLocalVector(dm,q,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);

  double fnorm;
  ierr = VecNorm(q,NORM_2,&fnorm); CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"solution fnorm  = %9.8e\n",fnorm); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


PetscErrorCode CrackPropagation::postProcessElasticDM(DM dm,const int step) {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if(!elasticFe) {
    SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Pointer to elasticFe is NULL");
  }
  PostProcTemplateVolumeOnRefinedMesh<CrackFrontFE> post_proc(mField);
  post_proc.crackFrontNodes = crackFrontNodes;
  post_proc.crackFrontNodesEdges = crackFrontNodesEdges;
  post_proc.crackFrontElements = crackFrontElements;

  ierr = post_proc.generateReferenceElementMesh(); CHKERRQ(ierr);
  ierr = post_proc.addFieldValuesPostProc("MESH_NODE_POSITIONS"); CHKERRQ(ierr);
  post_proc.getOpPtrVector().push_back(
    new OpAddSingularCoords(
      post_proc.singularElement,
      post_proc.singularCoords,
      post_proc.postProcMesh,
      post_proc.mapGaussPts
    )
  );
  post_proc.getOpPtrVector().push_back(
    new OpTransfromSingularBaseFunctions(post_proc.singularElement,post_proc.invSJac)
  );
  ierr = post_proc.addFieldValuesPostProc("SPATIAL_POSITION"); CHKERRQ(ierr);
  ierr = post_proc.addFieldValuesGradientPostProc("SPATIAL_POSITION"); CHKERRQ(ierr);
  std::map<int,NonlinearElasticElement::BlockData>::iterator sit = elasticFe->setOfBlocks.begin();
  for(;sit!=elasticFe->setOfBlocks.end();sit++) {
    post_proc.getOpPtrVector().push_back(
      new PostPorcStress(
        post_proc.postProcMesh,
        post_proc.mapGaussPts,
        "SPATIAL_POSITION",
        sit->second,
        reinterpret_cast<PostProcVolumeOnRefinedMesh::CommonData&>(post_proc.commonData)
      )
    );
  }
  ierr = DMoFEMLoopFiniteElements(dm,"ELASTIC",&post_proc); CHKERRQ(ierr);
  ostringstream ss;
  ss << "out_spatial_" << step << ".h5m";
  ierr = post_proc.writeFile(ss.str()); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::setPositionFromCoords(const std::string field_name) {
  ErrorCode rval;
  EntityHandle node = 0;
  double coords[3];
  PetscFunctionBegin;
  for(_IT_GET_DOFS_FIELD_BY_NAME_FOR_LOOP_(mField,field_name.c_str(),dof_ptr)) {
    if(dof_ptr->get()->getEntType()!=MBVERTEX) {
      double &fval = dof_ptr->get()->getFieldData();
      fval = 0;
      continue;
    }
    EntityHandle ent = dof_ptr->get()->getEnt();
    int dof_rank = dof_ptr->get()->getDofCoeffIdx();
    double &fval = dof_ptr->get()->getFieldData();
    if(node!=ent) {
      rval = mField.get_moab().get_coords(&ent,1,coords); CHKERRQ_MOAB(rval);
      node = ent;
    }
    fval = coords[dof_rank];
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::setMatrialPositionFromCoords() {
  return setPositionFromCoords("MESH_NODE_POSITIONS");
}


PetscErrorCode CrackPropagation::setSpatialPositionFromCoords() {
  return setPositionFromCoords("SPATIAL_POSITION");
}

PetscErrorCode CrackPropagation::getFrontEdgesAndElements(const BitRefLevel bit,const int verb) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  moab::Interface &moab = mField.get_moab();
  PetscFunctionBegin;

  Range level_edges;
  ierr = mField.get_entities_by_type_and_ref_level(
    bit,BitRefLevel().set(),MBEDGE,level_edges
  ); CHKERRQ(ierr);
  rval = moab.get_connectivity(crackFront,crackFrontNodes,true);
  crackFrontNodesEdges.clear();
  rval = moab.get_adjacencies(
    crackFrontNodes,1,false,crackFrontNodesEdges,moab::Interface::UNION
  ); CHKERRQ_MOAB(rval);
  crackFrontNodesEdges = subtract(crackFrontNodesEdges,crackFront);
  crackFrontNodesEdges = intersect(crackFrontNodesEdges,level_edges);

  Range level_tets;
  ierr = mField.get_entities_by_type_and_ref_level(
    bit,BitRefLevel().set(),MBTET,level_tets
  ); CHKERRQ(ierr);

  crackFrontElements.clear();
  rval = moab.get_adjacencies(
    crackFrontNodes,3,false,crackFrontElements,moab::Interface::UNION
  ); CHKERRQ_MOAB(rval);
  crackFrontElements = intersect(crackFrontElements,level_tets);

  if(verb>1) {
    cerr << crackFrontNodes << endl;
    cerr << crackFrontNodesEdges << endl;
    cerr << crackFrontElements << endl;
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::setSingularDofs(const string field_name,const int verb) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  const DofEntity_multiIndex *dofs_ptr;
  moab::Interface &moab = mField.get_moab();

  PetscFunctionBegin;
  ierr = mField.get_dofs(&dofs_ptr); CHKERRQ(ierr);

  for(Range::iterator eit = crackFrontNodesEdges.begin();eit!=crackFrontNodesEdges.end();eit++) {
    int num_nodes;
    const EntityHandle* conn;
    rval = moab.get_connectivity(*eit,conn,num_nodes,false); CHKERRQ_MOAB(rval);
    double coords[6];
    rval = moab.get_coords(conn,num_nodes,coords); CHKERRQ_MOAB(rval);
    const double dir[3] = {
      coords[3]-coords[0],coords[4]-coords[1],coords[5]-coords[2]
    };
    double dof[3] = {0,0,0};
    if(crackFrontNodes.find(conn[0])!=crackFrontNodes.end()) {
      for(int dd = 0;dd!=3;dd++) {
        dof[dd] = -dir[dd];
      }
    } else if(crackFrontNodes.find(conn[1])!=crackFrontNodes.end()) {
      for(int dd = 0;dd!=3;dd++) {
        dof[dd] = +dir[dd];
      }
    }
    for(_IT_GET_DOFS_FIELD_BY_NAME_AND_ENT_FOR_LOOP_(mField,field_name,*eit,dit)) {
      const int idx = dit->get()->getEntDofIdx();
      if(idx>2) {
        dit->get()->getFieldData() = 0;
      } else {
        if(verb>1) {
          cerr << **dit << endl;
          cerr << dof[idx] << endl;
        }
        dit->get()->getFieldData() = dof[idx];
        // setSingularCoordinatesSet = true;
      }
    }
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::setSingularElementMatrialPositions(const int verb) {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = setSingularDofs("MESH_NODE_POSITIONS",verb); CHKERRQ(ierr);
  setSingularCoordinatesSet = true;
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::unsetSingularElementMatrialPositions() {
  PetscErrorCode ierr;
  const DofEntity_multiIndex *dofs_ptr;
  PetscFunctionBegin;
  ierr = mField.get_dofs(&dofs_ptr); CHKERRQ(ierr);
  for(Range::iterator eit = crackFrontNodesEdges.begin();eit!=crackFrontNodesEdges.end();eit++) {
    for(_IT_GET_DOFS_FIELD_BY_NAME_AND_ENT_FOR_LOOP_(mField,"MESH_NODE_POSITIONS",*eit,dit)) {
      dit->get()->getFieldData() = 0;
      setSingularCoordinatesSet = false;
    }
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::cleanSingularElementMatrialPositions() {
  PetscErrorCode ierr;
  const DofEntity_multiIndex *dofs_ptr;
  PetscFunctionBegin;
  ierr = mField.set_field_order(0,MBEDGE,"MESH_NODE_POSITIONS",1); CHKERRQ(ierr);
  // ierr = mField.clear_inactive_dofs(); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}


PetscErrorCode CrackPropagation::setCrackFrontBitLevel(
  BitRefLevel from_bit,BitRefLevel bit
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  const RefEntity_multiIndex *ref_ent_ptr;
  moab::Interface &moab = mField.get_moab();
  PetscFunctionBegin;
  ierr = mField.get_ref_ents(&ref_ent_ptr);
  Range bit_ents;
  ierr = mField.get_entities_by_ref_level(from_bit,BitRefLevel().set(),bit_ents); CHKERRQ(ierr);
  Range ents;
  rval = moab.get_connectivity(crackFrontElements,ents,true); CHKERRQ_MOAB(rval);
  ents = intersect(bit_ents,ents);
  Range adj_tets;
  rval = moab.get_adjacencies(ents,3,false,adj_tets,moab::Interface::UNION); CHKERRQ_MOAB(rval);
  adj_tets = intersect(bit_ents,adj_tets);
  rval = moab.get_connectivity(adj_tets,ents,true); CHKERRQ_MOAB(rval);
  rval = moab.get_adjacencies(adj_tets,1,false,ents,moab::Interface::UNION); CHKERRQ_MOAB(rval);
  rval = moab.get_adjacencies(adj_tets,2,false,ents,moab::Interface::UNION); CHKERRQ_MOAB(rval);
  ents.merge(adj_tets);
  ents = intersect(bit_ents,ents);
  for(Range::iterator eit = ents.begin();eit!=ents.end();eit++) {
    RefEntity_multiIndex::index<Ent_mi_tag>::type::iterator dit
    = ref_ent_ptr->get<Ent_mi_tag>().find(*eit);
    if(dit==ref_ent_ptr->get<Ent_mi_tag>().end()) continue;
    bool success = const_cast<RefEntity_multiIndex*>(ref_ent_ptr)->modify(
      ref_ent_ptr->project<0>(dit),RefEntity_change_add_bit(bit)
    );
    // if(mField.get_comm_rank()==0) {
    //   cerr << **dit << endl;
    //   cerr << dit->get()->getBitRefLevel() << endl;
    // }
    if(!success) {
      SETERRQ(mField.get_comm(),MOFEM_OPERATION_UNSUCCESSFUL,"operation unsuccessful");
    };
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::getArcLengthDof() {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  const DofEntity_multiIndex *dofs_ptr;
  ierr = mField.get_dofs(&dofs_ptr); CHKERRQ(ierr);
  DofEntity_multiIndex::index<FieldName_mi_tag>::type::iterator dit;
  dit = dofs_ptr->get<FieldName_mi_tag>().find("ARC_LENGTH_LAMBDA");
  if(dit==dofs_ptr->get<FieldName_mi_tag>().end()) {
    SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Arc length lambda field not defined");
  }
  arcLengthDof = *dit;
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::CrackFrontFE::getSpaceBaseAndOrderOnElement() {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = VolumeElementForcesAndSourcesCore::getSpaceBaseAndOrderOnElement(); CHKERRQ(ierr);
  // singularElement = false;
  // PetscFunctionReturn(0);

  // Determine if this element is singular
  EntityHandle ent = numeredEntFiniteElementPtr->getEnt();
  if(crackFrontElements.find(ent)!=crackFrontElements.end()) {
    singularElement = true;
    for(int nn = 0;nn!=4;nn++) {
      if(crackFrontNodes.find(conn[nn])!=crackFrontNodes.end()) {
        singularNodes[nn] = true;
      } else {
        singularNodes[nn] = false;
      }
    }
    for(int ee = 0;ee!=6;ee++) {
      EntityHandle edge;
      rval = mField.get_moab().side_element(ent,1,ee,edge); CHKERRQ_MOAB(rval);
      if(crackFrontNodesEdges.find(edge)!=crackFrontNodesEdges.end()) {
        singularEdges[ee] = true;
      } else {
        singularEdges[ee] = false;
      }
    }
  } else {
    singularElement = false;
  }
  // singularElement = false;

  // Set base and approximation order on edges
  if(singularElement) {
    dataH1.sPace.set(H1);
    dataH1.bAse.set(AINSWORTH_LEGENDRE_BASE);
    dataH1.spacesOnEntities[MBEDGE].set(H1);
    dataH1.basesOnEntities[MBEDGE].set(AINSWORTH_LEGENDRE_BASE);
    for(int ee = 0;ee!=6;ee++) {
      EntityHandle edge;
      rval = mField.get_moab().side_element(ent,1,ee,edge); CHKERRQ_MOAB(rval);
      int side_number,sense,offset;
      rval = mField.get_moab().side_number(ent,edge,side_number,sense,offset);
      if(side_number!=ee) {
        SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Data inconsistency");
      }
      dataH1.dataOnEntities[MBEDGE][ee].getSense() = sense;
      if(singularEdges[ee]) {
        if(dataH1.dataOnEntities[MBEDGE][ee].getDataOrder()<=1) {
          dataH1.dataOnEntities[MBEDGE][ee].getDataOrder() = 2;
        }
      }
    }
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::CrackFrontFE::setGaussPts(int order) {
  MoABErrorCode rval;
  PetscFunctionBegin;

  order = order ? order : 1;
  int rule = (singularElement) ? 2*(order-1) : 2*(order-1);

  if(rule<QUAD_3D_TABLE_SIZE) {
    if(QUAD_3D_TABLE[rule]->dim!=3) {
      SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"wrong dimension");
    }
    if(QUAD_3D_TABLE[rule]->order<rule) {
      SETERRQ2(
        mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"wrong order %d != %d",
        QUAD_3D_TABLE[rule]->order,rule
      );
    }
    nbGaussPts = QUAD_3D_TABLE[rule]->npoints;
    gaussPts.resize(4,nbGaussPts,false);
    cblas_dcopy(
      nbGaussPts,&QUAD_3D_TABLE[rule]->points[1],4,&gaussPts(0,0),1
    );
    cblas_dcopy(
      nbGaussPts,&QUAD_3D_TABLE[rule]->points[2],4,&gaussPts(1,0),1
    );
    cblas_dcopy(
      nbGaussPts,&QUAD_3D_TABLE[rule]->points[3],4,&gaussPts(2,0),1
    );
    cblas_dcopy(
      nbGaussPts,QUAD_3D_TABLE[rule]->weights,1,&gaussPts(3,0),1
    );
    dataH1.dataOnEntities[MBVERTEX][0].getN(NOBASE).resize(nbGaussPts,4,false);
    double *shape_ptr = dataH1.dataOnEntities[MBVERTEX][0].getN(NOBASE).data().begin();
    cblas_dcopy(
      4*nbGaussPts,QUAD_3D_TABLE[rule]->points,1,shape_ptr,1
    );
  }

  if(!singularElement) PetscFunctionReturn(0);

  // refine mesh on refine elements
  const int max_level = 2;
  if(1) {
    double base_coords[] = { 0,0,0, 1,0,0, 0,1,0, 0,0,1 };
    moab::Core core_ref;
    moab::Interface& moab_ref = core_ref;
    EntityHandle nodes[4];
    for(int nn = 0;nn!=4;nn++) {
      rval = moab_ref.create_vertex(&base_coords[3*nn],nodes[nn]); CHKERRQ_MOAB(rval);
    }
    EntityHandle tet;
    rval = moab_ref.create_element(MBTET,nodes,4,tet); CHKERRQ_MOAB(rval);
    MoFEM::Core m_core_ref(moab_ref,PETSC_COMM_SELF,-2);
    MoFEM::Interface& m_field_ref = m_core_ref;
    ierr = m_field_ref.seed_ref_level_3D(0,BitRefLevel().set(0)); CHKERRQ(ierr);
    Range singular_nodes;
    for(int nn = 0;nn!=4;nn++) {
      if(singularNodes[nn]) {
        EntityHandle ent;
        rval = moab_ref.side_element(tet,0,nn,ent); CHKERRQ_MOAB(rval);
        singular_nodes.insert(ent);
      }
    }
    EntityHandle meshset;
    if(singular_nodes.size()>1) {
      rval = moab_ref.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
      for(int ee = 0;ee!=6;ee++) {
        if(singularEdges[ee]) {
          EntityHandle ent;
          rval = moab_ref.side_element(tet,1,ee,ent); CHKERRQ_MOAB(rval);
          rval = moab_ref.add_entities(meshset,&ent,1); CHKERRQ_MOAB(rval);
        }
      }
    }
    MeshRefinement *m_ref;
    ierr = m_field_ref.query_interface(m_ref); CHKERRQ(ierr);
    for(int ll = 0;ll!=max_level;ll++) {
      // PetscPrintf(T::mField.get_comm(),"Refine Level %d\n",ll);
      Range edges;
      ierr = m_field_ref.get_entities_by_type_and_ref_level(
        BitRefLevel().set(ll),BitRefLevel().set(),MBEDGE,edges
      ); CHKERRQ(ierr);
      Range ref_edges;
      rval = moab_ref.get_adjacencies(
        singular_nodes,1,true,ref_edges,moab::Interface::UNION
      ); CHKERRQ_MOAB(rval);
      ref_edges = intersect(ref_edges,edges);
      if(singular_nodes.size()>1) {
        Range ents;
        rval = moab_ref.get_entities_by_type(
          meshset,MBEDGE,ents,true
        ); CHKERRQ_MOAB(rval);
        ref_edges = intersect(ref_edges,ents);
      }
      Range tets;
      ierr = m_field_ref.get_entities_by_type_and_ref_level(
        BitRefLevel().set(ll),BitRefLevel().set(),MBTET,tets
      ); CHKERRQ(ierr);
      //refine mesh
      ierr = m_ref->add_verices_in_the_middel_of_edges(
        ref_edges,BitRefLevel().set(ll+1)
      ); CHKERRQ(ierr);
      ierr = m_ref->refine_TET(tets,BitRefLevel().set(ll+1)); CHKERRQ(ierr);
      if(singular_nodes.size()>1) {
        ierr = m_field_ref.update_meshset_by_entities_children(
          meshset,BitRefLevel().set(ll+1),meshset,MBEDGE,true
        ); CHKERRQ(ierr);
      }
    }
    Range tets;
    ierr = m_field_ref.get_entities_by_type_and_ref_level(
      BitRefLevel().set(max_level),BitRefLevel().set(),MBTET,tets
    ); CHKERRQ(ierr);
    refCoords.resize(tets.size(),12,false);
    int tt = 0;
    for(Range::iterator tit = tets.begin();tit!=tets.end();tit++,tt++) {
      int num_nodes;
      const EntityHandle* conn;
      rval = moab_ref.get_connectivity(*tit,conn,num_nodes,false); CHKERRQ_MOAB(rval);
      rval = moab_ref.get_coords(conn,num_nodes,&refCoords(tt,0)); CHKERRQ_MOAB(rval);
    }
    // cerr << "refCoords" << endl;
    // cerr << refCoords << endl;
    if(0) {
      EntityHandle out_meshset;
      rval = moab_ref.create_meshset(MESHSET_SET|MESHSET_TRACK_OWNER,out_meshset); CHKERRQ_MOAB(rval);
      rval = moab_ref.add_entities(out_meshset,tets); CHKERRQ_MOAB(rval);
      std::string file_name = "ref_tet_"+boost::lexical_cast<std::string>(nInTheLoop)+".vtk";
      rval = moab_ref.write_file(file_name.c_str(),"VTK","",&out_meshset,1); CHKERRQ_MOAB(rval);
      rval = moab_ref.delete_entities(&out_meshset,1); CHKERRQ_MOAB(rval);
    }
  }

  // Set integration points
  {
    double diff_n[12];
    ierr = ShapeDiffMBTET(diff_n); CHKERRQ(ierr);
    refGuassCoords.resize(4,nbGaussPts*refCoords.size1());
    MatrixDouble &shape_n = dataH1.dataOnEntities[MBVERTEX][0].getN(NOBASE);
    int gg = 0;
    for(int tt = 0;tt!=refCoords.size1();tt++) {
      double *tet_coords = &refCoords(tt,0);
      double det = ShapeVolumeMBTET(diff_n,tet_coords);
      det *= 6;
      for(int ggg = 0;ggg!=nbGaussPts;ggg++,gg++) {
        for(int dd = 0;dd!=3;dd++) {
          refGuassCoords(dd,gg) =
          shape_n(ggg,0)*tet_coords[3*0+dd]+
          shape_n(ggg,1)*tet_coords[3*1+dd]+
          shape_n(ggg,2)*tet_coords[3*2+dd]+
          shape_n(ggg,3)*tet_coords[3*3+dd];
        }
        refGuassCoords(3,gg) = gaussPts(3,ggg)*det;///pow(8,max_level);
      }
    }
    gaussPts.swap(refGuassCoords);
    nbGaussPts = gaussPts.size2();
    dataH1.dataOnEntities[MBVERTEX][0].getN(NOBASE).resize(nbGaussPts,4);
    double *shape_ptr = dataH1.dataOnEntities[MBVERTEX][0].getN(NOBASE).data().begin();
    ierr = ShapeMBTET(
      shape_ptr,&gaussPts(0,0),&gaussPts(1,0),&gaussPts(2,0),nbGaussPts
    ); CHKERRQ(ierr);
  }

  const int edge_nodes[6][2] = {
    {0,1}, {1,2}, {2,0}, {0,3}, {1,3}, {2,3}
  };
  FTensor::Index<'i',3> i;
  FTensor::Index<'j',3> j;
  sJac.resize(nbGaussPts,9,false);
  double *s_jac_ptr = &sJac(0,0);
  {
    FTensor::Tensor2<double*,3,3> t_jac(
      s_jac_ptr,&s_jac_ptr[1],&s_jac_ptr[2],
      &s_jac_ptr[3],&s_jac_ptr[4],&s_jac_ptr[5],
      &s_jac_ptr[6],&s_jac_ptr[7],&s_jac_ptr[8],9
    );
    for(int gg = 0;gg!=nbGaussPts;gg++) {
      t_jac(i,j) = 0;
      t_jac(0,0) = 1;
      t_jac(1,1) = 1;
      t_jac(2,2) = 1;
      ++t_jac;
    }
  }

  const double coords[] = { 0,0,0, 1,0,0, 0,1,0, 0,0,1 };
  double diff_n[12];
  ierr = ShapeDiffMBTET(diff_n); CHKERRQ(ierr);
  MatrixDouble &shape_n = dataH1.dataOnEntities[MBVERTEX][0].getN(NOBASE);

  for(int ee = 0;ee!=6;ee++) {
    if(!singularEdges[ee]) continue;
    if(singularNodes[edge_nodes[ee][0]]&&singularNodes[edge_nodes[ee][1]]) {
      SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Data inconsistency");
    }
    const int i0 = edge_nodes[ee][0];
    const int i1 = edge_nodes[ee][1];
    FTensor::Tensor1<double,3> t_dir(
      coords[3*i1+0]-coords[3*i0+0],
      coords[3*i1+1]-coords[3*i0+1],
      coords[3*i1+2]-coords[3*i0+2]
    );
    if(singularNodes[edge_nodes[ee][0]]) { t_dir(i) *= -1; }
    else if(!singularNodes[edge_nodes[ee][1]]) {
      SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Data inconsistency");
    }
    FTensor::Tensor2<double*,3,3> t_jac(
      s_jac_ptr,&s_jac_ptr[1],&s_jac_ptr[2],
      &s_jac_ptr[3],&s_jac_ptr[4],&s_jac_ptr[5],
      &s_jac_ptr[6],&s_jac_ptr[7],&s_jac_ptr[8],9
    );
    FTensor::Tensor1<double*,3> t_integration_coords_ptr(
      &gaussPts(0,0),&gaussPts(1,0),&gaussPts(2,0),1
    );
    for(int gg = 0;gg!=nbGaussPts;gg++) {
      double t_n = shape_n(gg,i0)*shape_n(gg,i1);
      FTensor::Tensor1<double,3> t_diff_n(
        shape_n(gg,i0)*diff_n[3*i1+0]+shape_n(gg,i1)*diff_n[3*i0+0],
        shape_n(gg,i0)*diff_n[3*i1+1]+shape_n(gg,i1)*diff_n[3*i0+1],
        shape_n(gg,i0)*diff_n[3*i1+2]+shape_n(gg,i1)*diff_n[3*i0+2]
      );
      t_jac(i,j) += t_dir(i)*t_diff_n(j);
      t_integration_coords_ptr(i) += t_dir(i)*t_n;
      ++t_jac;
      ++t_integration_coords_ptr;
    }
  }

  FTensor::Tensor2<double*,3,3> t_jac(
    s_jac_ptr,&s_jac_ptr[1],&s_jac_ptr[2],
    &s_jac_ptr[3],&s_jac_ptr[4],&s_jac_ptr[5],
    &s_jac_ptr[6],&s_jac_ptr[7],&s_jac_ptr[8],9
  );
  for(int gg = 0;gg!=nbGaussPts;gg++) {
    double det;
    ierr = determinantTensor3by3(t_jac,det); CHKERRQ(ierr);
    gaussPts(3,gg) *= det;
    ++t_jac;
  }

  double *shape_ptr = dataH1.dataOnEntities[MBVERTEX][0].getN(NOBASE).data().begin();
  ierr = ShapeMBTET(
    shape_ptr,&gaussPts(0,0),&gaussPts(1,0),&gaussPts(2,0),nbGaussPts
  ); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::CrackFrontFE::calculateHoJacobian() {
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = VolumeElementForcesAndSourcesCore::calculateHoJacobian(); CHKERRQ(ierr);

  if(CrackPropagation::setSingularCoordinatesSet) {
    singularElement = false;
    PetscFunctionReturn(0);
  }

  if(singularElement) {

    const int edge_nodes[6][2] = {
      {0,1}, {1,2}, {2,0}, {0,3}, {1,3}, {2,3}
    };
    FTensor::Index<'i',3> i;
    FTensor::Index<'j',3> j;
    sJac.resize(nbGaussPts,9,false);
    singularCoords.resize(coordsAtGaussPts.size1(),coordsAtGaussPts.size2());
    singularCoords.clear();

    double *s_jac_ptr = &sJac(0,0);
    {
      FTensor::Tensor2<double*,3,3> t_jac(
        s_jac_ptr,&s_jac_ptr[1],&s_jac_ptr[2],
        &s_jac_ptr[3],&s_jac_ptr[4],&s_jac_ptr[5],
        &s_jac_ptr[6],&s_jac_ptr[7],&s_jac_ptr[8],9
      );
      for(int gg = 0;gg!=nbGaussPts;gg++) {
        t_jac(i,j) = 0;
        t_jac(0,0) = 1;
        t_jac(1,1) = 1;
        t_jac(2,2) = 1;
        ++t_jac;
      }
    }
    double diff_n[12];
    ierr = ShapeDiffMBTET(diff_n); CHKERRQ(ierr);
    MatrixDouble &shape_n = dataH1.dataOnEntities[MBVERTEX][0].getN(NOBASE);

    for(int ee = 0;ee!=6;ee++) {
      if(!singularEdges[ee]) continue;
      if(singularNodes[edge_nodes[ee][0]]&&singularNodes[edge_nodes[ee][1]]) {
        SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Data inconsistency");
      }
      const int i0 = edge_nodes[ee][0];
      const int i1 = edge_nodes[ee][1];
      FTensor::Tensor1<double,3> t_dir(
        coords[3*i1+0]-coords[3*i0+0],coords[3*i1+1]-coords[3*i0+1],coords[3*i1+2]-coords[3*i0+2]
      );

      // cerr << t_dir(0) <<  " " << t_dir(1)  << " "<< t_dir(2)  << endl;
      if(singularNodes[edge_nodes[ee][0]]) { t_dir(i) *= -1; }
      else if(!singularNodes[edge_nodes[ee][1]]) {
        SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"Data inconsistency");
      }

      FTensor::Tensor2<double*,3,3> t_jac(
        s_jac_ptr,&s_jac_ptr[1],&s_jac_ptr[2],
        &s_jac_ptr[3],&s_jac_ptr[4],&s_jac_ptr[5],
        &s_jac_ptr[6],&s_jac_ptr[7],&s_jac_ptr[8],9
      );
      // cerr << dataH1.dataOnEntities[MBEDGE][ee].getDiffN(AINSWORTH_LEGENDRE_BASE) << endl;
      // dataH1.dataOnEntities[MBEDGE][ee].getBase() = AINSWORTH_LEGENDRE_BASE;
      // const FieldApproximationBase base = AINSWORTH_LEGENDRE_BASE;
      FTensor::Tensor0<double*> t_n =
      dataH1.dataOnEntities[MBEDGE][ee].getFTensor0N(AINSWORTH_LEGENDRE_BASE,0);
      FTensor::Tensor1<double*,3> t_diff_n =
      dataH1.dataOnEntities[MBEDGE][ee].getFTensor1DiffN<3>(AINSWORTH_LEGENDRE_BASE,0);
      FTensor::Tensor1<double*,3> t_coords_at_gauss_ptr(
        &singularCoords(0,0),&singularCoords(0,1),&singularCoords(0,2),3
      );
      for(int gg = 0;gg!=nbGaussPts;gg++) {
        t_jac(i,j) += t_dir(i)*t_diff_n(j);
        t_coords_at_gauss_ptr(i) += t_dir(i)*t_n;
        ++t_n;
        ++t_diff_n;
        ++t_jac;
        ++t_coords_at_gauss_ptr;
      }
    }

    {
      FTensor::Tensor2<double*,3,3> t_jac(
        s_jac_ptr,&s_jac_ptr[1],&s_jac_ptr[2],
        &s_jac_ptr[3],&s_jac_ptr[4],&s_jac_ptr[5],
        &s_jac_ptr[6],&s_jac_ptr[7],&s_jac_ptr[8],9
      );
      invSJac.resize(nbGaussPts,9,false);
      double *t_inv_jac_ptr = &invSJac(0,0);
      FTensor::Tensor2<double*,3,3> t_inv_jac(
        t_inv_jac_ptr,&t_inv_jac_ptr[1],&t_inv_jac_ptr[2],
        &t_inv_jac_ptr[3],&t_inv_jac_ptr[4],&t_inv_jac_ptr[5],
        &t_inv_jac_ptr[6],&t_inv_jac_ptr[7],&t_inv_jac_ptr[8],9
      );

      // double sum_det = 0;
      for(int gg = 0;gg!=nbGaussPts;gg++) {
        double det;
        ierr = determinantTensor3by3(t_jac,det); CHKERRQ(ierr);
        ierr = invertTensor3by3(t_jac,det,t_inv_jac); CHKERRQ(ierr);
        gaussPts(3,gg) *= det;
        // cerr << det << endl;
        ++t_inv_jac;
        ++t_jac;
      }

      // cerr << sJac << endl;
      // cerr << sum_det << endl;
    }

  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::OpGetCrackFrontDataAtGaussPts::doWork(
  int side,EntityType type,DataForcesAndSurcesCore::EntData &data
) {
  PetscFunctionBegin;

  if(!singularElement) {
    return  NonlinearElasticElement::OpGetDataAtGaussPts::doWork(side,type,data);
  }

  try {

    const int nb_dofs = data.getFieldData().size();
    const int nb_base_functions = data.getN().size2();
    if(nb_dofs == 0) {
      PetscFunctionReturn(0);
    }
    const int nb_gauss_pts = data.getN().size1();
    const int rank = data.getFieldDofs()[0]->getNbOfCoeffs();

    //initialize
    if(type == zeroAtType) {
      valuesAtGaussPts.resize(nb_gauss_pts);
      gradientAtGaussPts.resize(nb_gauss_pts);
      for(int gg = 0;gg!=nb_gauss_pts;gg++) {
        valuesAtGaussPts[gg].resize(rank,false);
        gradientAtGaussPts[gg].resize(rank,3,false);
      }
      for(int gg = 0;gg!=nb_gauss_pts;gg++) {
        valuesAtGaussPts[gg].clear();
        gradientAtGaussPts[gg].clear();
      }
    }

    FTensor::Tensor0<double*> base_function = data.getFTensor0N();
    FTensor::Tensor1<double*,3> diff_base_functions = data.getFTensor1DiffN<3>();
    FTensor::Index<'i',3> i;
    FTensor::Index<'j',3> j;
    FTensor::Index<'k',3> k;

    double *inv_jac_ptr = invSJac.data().begin();
    FTensor::Tensor2<double*,3,3> t_inv_singular_jacbian(
      inv_jac_ptr,&inv_jac_ptr[1],&inv_jac_ptr[2],
      &inv_jac_ptr[3],&inv_jac_ptr[4],&inv_jac_ptr[5],
      &inv_jac_ptr[6],&inv_jac_ptr[7],&inv_jac_ptr[8],9
    );

    if(rank==3) {

      for(int gg = 0;gg!=nb_gauss_pts;gg++) {
        FTensor::Tensor1<double*,3> field_data = data.getFTensor1FieldData<3>();
        FTensor::Tensor1<double*,3> values(
          &valuesAtGaussPts[gg][0],&valuesAtGaussPts[gg][1],&valuesAtGaussPts[gg][2]
        );
        FTensor::Tensor2<double*,3,3> gradient(
          &gradientAtGaussPts[gg](0,0),&gradientAtGaussPts[gg](0,1),&gradientAtGaussPts[gg](0,2),
          &gradientAtGaussPts[gg](1,0),&gradientAtGaussPts[gg](1,1),&gradientAtGaussPts[gg](1,2),
          &gradientAtGaussPts[gg](2,0),&gradientAtGaussPts[gg](2,1),&gradientAtGaussPts[gg](2,2)
        );
        int bb = 0;
        for(;bb!=nb_dofs/3;bb++) {
          values(i) += base_function*field_data(i);
          gradient(i,j) += field_data(i)*(t_inv_singular_jacbian(k,j)*diff_base_functions(k));
          ++diff_base_functions;
          ++base_function;
          ++field_data;
        }
        for(;bb!=nb_base_functions;bb++) {
          ++diff_base_functions;
          ++base_function;
        }
        ++t_inv_singular_jacbian;
      }

    } else {

      SETERRQ1(PETSC_COMM_SELF,MOFEM_DATA_INCONSISTENCY,"Not implemented for rank  = %d",rank);

    }

  } catch (const std::exception& ex) {
    std::ostringstream ss;
    ss << "throw in method: " << ex.what() << std::endl;
    SETERRQ(PETSC_COMM_SELF,1,ss.str().c_str());
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::OpAddSingularCoords::doWork(
  int side,EntityType type,DataForcesAndSurcesCore::EntData &data
) {
  MoABErrorCode rval;
  PetscFunctionBegin;
  if(!singularElement || type!=MBVERTEX) {
    PetscFunctionReturn(0);
  }
  string tag_name = "SINGULAR_DISP";
  double def_VAL[3] = { 0,0,0 };
  Tag th;
  rval = postProcMesh.tag_get_handle(
    tag_name.c_str(),3,MB_TYPE_DOUBLE,th,MB_TAG_CREAT|MB_TAG_SPARSE,def_VAL
  ); CHKERRQ_MOAB(rval);
  for(int gg = 0;gg!=singularCoords.size1();gg++) {
    rval = postProcMesh.tag_set_data(th,&mapGaussPts[gg],1,&singularCoords(gg,0)); CHKERRQ_MOAB(rval);
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::OpTransfromSingularBaseFunctions::doWork(
  int side,EntityType type,DataForcesAndSurcesCore::EntData &data
) {
  FTensor::Index<'i',3> i;
  FTensor::Index<'j',3> j;
  PetscFunctionBegin;

  if(!singularElement) PetscFunctionReturn(0);

  try {

    FieldApproximationBase base = AINSWORTH_LEGENDRE_BASE;
    if(data.getDiffN(base).size2()==0) PetscFunctionReturn(0);

    unsigned int nb_gauss_pts = data.getN(base).size1();
    if(nb_gauss_pts==0) PetscFunctionReturn(0);
    unsigned int nb_base_functions = data.getN(base).size2();
    if(nb_base_functions==0) PetscFunctionReturn(0);

    if(invSJac.size2()!=9) {
      SETERRQ1(
        PETSC_COMM_SELF,
        MOFEM_DATA_INCONSISTENCY,
        "It looks that ho inverse of Jacobian is not calculated %d != 9",
        invSJac.size2()
      );
    }
    if(invSJac.size1()!=nb_gauss_pts) {
      SETERRQ2(
        PETSC_COMM_SELF,
        MOFEM_DATA_INCONSISTENCY,
        "It looks that ho inverse of Jacobian is not calculated %d != %d",
        invSJac.size1(),
        nb_gauss_pts
      );
    }
    double *t_inv_jac_ptr = &*invSJac.data().begin();
    FTensor::Tensor2<double*,3,3> t_inv_jac(
      t_inv_jac_ptr,&t_inv_jac_ptr[1],&t_inv_jac_ptr[2],
      &t_inv_jac_ptr[3],&t_inv_jac_ptr[4],&t_inv_jac_ptr[5],
      &t_inv_jac_ptr[6],&t_inv_jac_ptr[7],&t_inv_jac_ptr[8],9
    );

    diffNinvJac.resize(nb_gauss_pts,3*nb_base_functions,false);
    double *t_inv_n_ptr = &*diffNinvJac.data().begin();
    FTensor::Tensor1<double*,3> t_inv_diff_n(
      t_inv_n_ptr,&t_inv_n_ptr[1],&t_inv_n_ptr[2],3
    );

    switch (type) {
      case MBVERTEX:
      case MBEDGE:
      case MBTRI:
      case MBTET: {
        FTensor::Tensor1<double*,3> t_diff_n = data.getFTensor1DiffN<3>(base);
        for(unsigned int gg = 0;gg<nb_gauss_pts;gg++) {
          for(unsigned int bb = 0;bb!=nb_base_functions;bb++) {
            t_inv_diff_n(i) = t_diff_n(j)*t_inv_jac(j,i);
            ++t_diff_n;
            ++t_inv_diff_n;
          }
          ++t_inv_jac;
        }
      }
      break;
      default:
      SETERRQ(PETSC_COMM_SELF,MOFEM_NOT_IMPLEMENTED,"not implemented");
    }

    if(type == MBVERTEX) {
      data.getDiffN(base).resize(diffNinvJac.size1(),diffNinvJac.size2(),false);
    }
    data.getDiffN(base).data().swap(diffNinvJac.data());

  } catch (std::exception& ex) {
    std::ostringstream ss;
    ss << "thorw in method: " << ex.what() << " at line " << __LINE__ << " in file " << __FILE__;
    SETERRQ(PETSC_COMM_SELF,MOFEM_STD_EXCEPTION_THROW,ss.str().c_str());
  }

  PetscFunctionReturn(0);
}

CrackPropagation::OpRhsBoneExplicitDerivarive::OpRhsBoneExplicitDerivarive(
  NonlinearElasticElement::BlockData &data,
  NonlinearElasticElement::CommonData &common_data,
  boost::shared_ptr<MatrixDouble> rho_grad_at_gauss_pts
):
MoFEM::VolumeElementForcesAndSourcesCore::UserDataOperator(
  "MESH_NODE_POSITIONS",UserDataOperator::OPROW
),
dAta(data),
commonData(common_data),
rhoGradAtGaussPts(rho_grad_at_gauss_pts) {
}

PetscErrorCode CrackPropagation::OpRhsBoneExplicitDerivarive::doWork(
  int row_side,EntityType row_type,DataForcesAndSurcesCore::EntData &row_data
) {
  FTensor::Index<'i',3> i;
  PetscFunctionBegin;
  const int nb_dofs = row_data.getIndices().size();
  if(!nb_dofs) PetscFunctionReturn(0);
  // If this is not part of blockset
  if(dAta.tEts.find(getNumeredEntFiniteElementPtr()->getEnt()) == dAta.tEts.end()) {
    PetscFunctionReturn(0);
  }
  const int nb_gauss_pts = row_data.getN().size1();
  const int nb_base_functions = row_data.getN().size2();
  nF.resize(nb_dofs,false);
  nF.clear();
  FTensor::Tensor1<double*,3> t_rho_gradient = getTensor1FormData<3>(*rhoGradAtGaussPts);
  for(int gg = 0;gg!=nb_gauss_pts;gg++) {
    const double w = getVolume()*getGaussPts()(3,gg)*getHoGaussPtsDetJac()[gg];
    VectorDouble &energy_diff = commonData.jacEnergy[gg];
    // it is 18, since 8 element of h and H since F = hH^-1 for ale
    const double denergy_drho = energy_diff[18]; // derivative of energy in respect to density
    FTensor::Tensor0<double*> t_base = row_data.getFTensor0N(gg,0);
    FTensor::Tensor1<double*,3> t_nf(&nF[0],&nF[0],&nF[0],3);
    for(int bb = 0;bb!=nb_dofs/3;bb++) {
      t_nf(i) += t_base*(w*denergy_drho*t_rho_gradient(i));
      ++t_nf;
      ++t_base;
    }
    ++t_rho_gradient;
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::PostProcVertexMethod::preProcess() {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  PetscFunctionBegin;
  if(!fieldPtr) {
    SETERRQ(
      mField.get_comm(),
      MOFEM_DATA_INCONSISTENCY,
      "Null pointer, probably field not found"
    );
  }
  double def_val[fieldPtr->getNbOfCoeffs()];
  rval = mField.get_moab().tag_get_handle(
    tagName.c_str(),
    fieldPtr->getNbOfCoeffs(),
    MB_TYPE_DOUBLE,
    tH,
    MB_TAG_CREAT|MB_TAG_SPARSE,
    def_val
  );
  if(rval==MB_ALREADY_ALLOCATED) rval = MB_SUCCESS;
  CHKERRQ_MOAB(rval);
  ierr = VecGetArray(F,&aRray); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::PostProcVertexMethod::postProcess() {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = VecRestoreArray(F,&aRray); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::PostProcVertexMethod::operator()() {
  MoABErrorCode rval;
  PetscFunctionBegin;
  if(dofPtr->getEntType()!=MBVERTEX) PetscFunctionReturn(0);
  int local_idx = dofNumeredPtr->getPetscLocalDofIdx();
  if(local_idx==-1) PetscFunctionReturn(0);
  EntityHandle ent = dofNumeredPtr->getEnt();
  int dof_rank = dofNumeredPtr->getDofCoeffIdx();
  int rank = dofNumeredPtr->getNbOfCoeffs();
  double tag_val[rank];
  rval = mField.get_moab().tag_get_data(tH,&ent,1,tag_val); CHKERRQ_MOAB(rval);
  tag_val[dof_rank] = aRray[local_idx];
  rval = mField.get_moab().tag_set_data(tH,&ent,1,tag_val); CHKERRQ_MOAB(rval);
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::FaceOrientation::getElementOrientation(
  MoFEM::Interface &m_field,const FEMethod *fe_method_ptr
) {
  PetscFunctionBegin;
  elementOrientation = +1;
  ErrorCode rval;
  PetscErrorCode ierr;
  Range adj_side_elems;
  EntityHandle face = fe_method_ptr->numeredEntFiniteElementPtr->getEnt();
  BitRefLevel bit = fe_method_ptr->problemPtr->getBitRefLevel();
  ierr = m_field.get_adjacencies(bit,&face,1,3,adj_side_elems,moab::Interface::INTERSECT,0); CHKERRQ(ierr);
  adj_side_elems = adj_side_elems.subset_by_type(MBTET);
  if(adj_side_elems.size()==0) {
    Range adj_tets_on_surface;
    BitRefLevel bit_tet_on_surface;
    bit_tet_on_surface.set(BITREFLEVEL_SIZE-2);
    ierr = m_field.get_adjacencies(bit_tet_on_surface,&face,1,3,adj_tets_on_surface,moab::Interface::INTERSECT,0); CHKERRQ(ierr);
    adj_side_elems.insert(*adj_tets_on_surface.begin());
  }
  if(adj_side_elems.size()!=1) {
    adj_side_elems.clear();
    ierr = m_field.get_adjacencies(bit,&face,1,3,adj_side_elems,moab::Interface::INTERSECT,5); CHKERRQ(ierr);
    Range::iterator it = adj_side_elems.begin();
    for(;it!=adj_side_elems.end();it++) {
      Range nodes;
      rval = m_field.get_moab().get_connectivity(&*it,1,nodes,true); CHKERRQ_MOAB(rval);
      PetscPrintf(PETSC_COMM_WORLD,"Connectivity %lu %lu %lu %lu\n",nodes[0],nodes[1],nodes[2],nodes[3]);
    }
    int rank;
    MPI_Comm_rank(m_field.get_comm(),&rank);
    if(rank==0) {
      EntityHandle out_meshset;
      rval = m_field.get_moab().create_meshset(MESHSET_SET|MESHSET_TRACK_OWNER,out_meshset); CHKERRQ_MOAB(rval);
      rval = m_field.get_moab().add_entities(out_meshset,adj_side_elems); CHKERRQ_MOAB(rval);
      rval = m_field.get_moab().add_entities(out_meshset,&face,1); CHKERRQ_MOAB(rval);
      rval = m_field.get_moab().write_file("debug_error.vtk","VTK","",&out_meshset,1); CHKERRQ_MOAB(rval);
      rval = m_field.get_moab().delete_entities(&out_meshset,1); CHKERRQ_MOAB(rval);
    }
    SETERRQ1(PETSC_COMM_SELF,1,"Expect 1 tet but is %u",adj_side_elems.size());
  }
  EntityHandle side_elem = adj_side_elems[0];
  if(side_elem!=0) {
    int side_number,sense,offset;
    rval = m_field.get_moab().side_number(side_elem,face,side_number,sense,offset); CHKERRQ_MOAB(rval);
    if(sense == -1) {
      elementOrientation = -1;
    }
  }
  if(useProjectionFromCrackFront) {
    Tag th_interface_side;
    rval = m_field.get_moab().tag_get_handle("INTERFACE_SIDE",th_interface_side); CHKERRQ_MOAB(rval);
    int side;
    rval = m_field.get_moab().tag_get_data(th_interface_side,&face,1,&side); CHKERRQ_MOAB(rval);
    if(side == 1) {
      elementOrientation = -1;
    }
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::MWLSApprox::loadMWLSMesh(std::string file_name) {
  MoABErrorCode rval;
  PetscFunctionBegin;

  const char *option;
  option = "";
  rval = mwlsMoab.load_file(file_name.c_str(),0,option); CHKERRQ_MOAB(rval);
  // rval = mwlsMoab.get_entities_by_type(0,MBVERTEX,levelNodes); CHKERRQ_MOAB(rval);
  rval = mwlsMoab.get_entities_by_type(0,MBTET,levelTets); CHKERRQ_MOAB(rval);
  rval = mwlsMoab.get_adjacencies(levelTets,1,true,levelEdges,moab::Interface::UNION); CHKERRQ_MOAB(rval);

  // Create HO node on tetrahdrals
  EntityHandle meshset;
  rval = mwlsMoab.create_meshset(MESHSET_SET,meshset); CHKERRQ_MOAB(rval);
  rval = mwlsMoab.add_entities(meshset,levelTets); CHKERRQ_MOAB(rval);
  rval = mwlsMoab.convert_entities(meshset,false,false,true); CHKERRQ_MOAB(rval);
  rval = mwlsMoab.delete_entities(&meshset,1); CHKERRQ_MOAB(rval);

  // Store edge nodes coordinates in ftenor
  double edge_node_coords[6];
  FTensor::Tensor1<double*,3> t_node_edge[2] = {
    FTensor::Tensor1<double*,3>(edge_node_coords,&edge_node_coords[1],&edge_node_coords[2]),
    FTensor::Tensor1<double*,3>(&edge_node_coords[3],&edge_node_coords[4],&edge_node_coords[5])
  };

  // Get edge lengths
  maxEdgeL = 0;
  for(Range::iterator eit=levelEdges.begin();eit!=levelEdges.end();eit++) {
    int num_nodes;
    const EntityHandle* conn;
    rval = mwlsMoab.get_connectivity(*eit,conn,num_nodes,true); CHKERRQ_MOAB(rval);
    rval = mwlsMoab.get_coords(conn,num_nodes,edge_node_coords); CHKERRQ_MOAB(rval);
    t_node_edge[0](i) -= t_node_edge[1](i);
    double l = sqrt(t_node_edge[0](i)*t_node_edge[0](i));
    maxEdgeL = (maxEdgeL<l) ? l : maxEdgeL;
  }

  // Get HO nodes in TET center
  std::vector<double> edge_length;
  for(Range::iterator tit =levelTets.begin();tit!=levelTets.end();tit++) {
    int num_nodes;
    const EntityHandle* conn;
    rval = mwlsMoab.get_connectivity(*tit,conn,num_nodes,false); CHKERRQ_MOAB(rval);
    EntityHandle ho_node;
    rval = mwlsMoab.high_order_node(*tit,conn,MBTET,ho_node); CHKERRQ_MOAB(rval);
    levelNodes.insert(ho_node);
  }

  // Create tree and find maximal edge length. Maximal edge length will is used
  // to find neighbours nodes to material point.
  myTree = boost::shared_ptr<AdaptiveKDTree>(new AdaptiveKDTree(&mwlsMoab));
  PetscPrintf(mField.get_comm(),"Build tree ... ");
  rval = myTree->build_tree(levelTets,&treeRoot); CHKERRQ_MOAB(rval);
  boost::shared_ptr<ElemEvaluator> elemEval(myTree,new ElemEvaluator(&mwlsMoab));
  myTree->set_eval(elemEval.get());
  PetscPrintf(mField.get_comm(),"done\n");
  // myTree->print();

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::MWLSApprox::getValuesToNodes(Tag th) {
  MoABErrorCode rval;
  PetscFunctionBegin;
  int length;
  rval = mwlsMoab.tag_get_length(th,length); CHKERRQ_MOAB(rval);
  double vals[length];
  bzero(vals,length*sizeof(double));
  // clean tag on nodes
  rval = mwlsMoab.tag_clear_data(th,levelNodes,vals); CHKERRQ_MOAB(rval);
  for(Range::iterator tit = levelTets.begin();tit!=levelTets.end();tit++) {
    rval = mwlsMoab.tag_get_data(th,&*tit,1,vals); CHKERRQ_MOAB(rval);
    int num_nodes;
    const EntityHandle* conn;
    rval = mwlsMoab.get_connectivity(*tit,conn,num_nodes,false); CHKERRQ_MOAB(rval);
    EntityHandle ho_node;
    rval = mwlsMoab.high_order_node(*tit,conn,MBTET,ho_node); CHKERRQ_MOAB(rval);
    // double coords[3];
    // rval = mwlsMoab.get_coords(&ho_node,1,coords); CHKERRQ_MOAB(rval);
    // vals[0] = coords[0]*coords[0]*coords[0];
    rval = mwlsMoab.tag_set_data(th,&ho_node,1,vals); CHKERRQ_MOAB(rval);
  }
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::MWLSApprox::getBox(
  double material_coords[3],
  std::vector<double>& coords
) {
  PetscFunctionBegin;
  FTensor::Tensor1<double,3> t_material_coords(
    material_coords[0],material_coords[1],material_coords[2]
  );
  FTensor::Tensor1<double*,3> t_coords(
    &coords[0],&coords[1],&coords[2],3
  );
  FTensor::Index<'i',3> i;
  double max_xyz[] = {0,0,0};
  for(int bb = 0;bb!=coords.size()/3;bb++) {
    t_coords(i) -= t_material_coords(i);
    for(int dd = 0;dd!=3;dd++) {
      max_xyz[dd] = (max_xyz[dd]<fabs(t_coords(dd))) ? fabs(t_coords(dd)) : max_xyz[dd];
    }
    ++t_coords;
  }
  boxSize = (max_xyz[0]<max_xyz[1]) ? max_xyz[0] : max_xyz[1];
  boxSize = (boxSize<max_xyz[2]) ? boxSize : max_xyz[2];
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::MWLSApprox::getInfluenceNodes(
  double material_coords[3],bool search_point
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  PetscFunctionBegin;

  // Fining neighbours edge nodes
  Range tree_tets;
  Range nodes;
  std::vector<double> coords;

  if(myTree) {
    if(search_point) {
      // Find TET clase to location
      CartVect params(material_coords);
      EntityHandle leaf_out;
      rval = myTree->point_search(
        material_coords,leaf_out,maxEdgeL*1e-5,maxEdgeL*1e-3,NULL,NULL,&params
      ); CHKERRQ_MOAB(rval);
      // Get TETS in neighbourhood
      MeshTopoUtil mtu(&mwlsMoab);
      rval = mtu.get_bridge_adjacencies(leaf_out,0,3,tree_tets); CHKERRQ_MOAB(rval);
      // Get bounding box
      boxSize = 0;
      rval = mwlsMoab.get_connectivity(tree_tets,nodes,true); CHKERRQ_MOAB(rval);
      coords.resize(3*nodes.size());
      rval = mwlsMoab.get_coords(nodes,&coords[0]); CHKERRQ_MOAB(rval);
      ierr = getBox(material_coords,coords); CHKERRQ_MOAB(rval);
    } else {
      boxSize = 0;
      rval = mwlsMoab.get_connectivity(tree_tets,nodes,true); CHKERRQ_MOAB(rval);
      coords.resize(3*nodes.size());
      rval = mwlsMoab.get_coords(nodes,&coords[0]); CHKERRQ_MOAB(rval);
      ierr = getBox(material_coords,coords); CHKERRQ_MOAB(rval);
    }
  } else {
    SETERRQ(mField.get_comm(),MOFEM_DATA_INCONSISTENCY,"kd-three not build");
  }

  const double dm = maxEdgeL*dmFactor; // Influence radius
  while(boxSize<dm) {
    rval = mwlsMoab.get_adjacencies(
      nodes,3,false,tree_tets,moab::Interface::UNION
    ); CHKERRQ_MOAB(rval);
    rval = mwlsMoab.get_connectivity(tree_tets,nodes,true); CHKERRQ_MOAB(rval);
    coords.resize(3*nodes.size());
    rval = mwlsMoab.get_coords(nodes,&coords[0]); CHKERRQ_MOAB(rval);
    boxSize = 0;
    ierr = getBox(material_coords,coords); CHKERRQ_MOAB(rval);
  }

  influenceNodes.clear();
  rval = mwlsMoab.get_connectivity(tree_tets,nodes,false); CHKERRQ_MOAB(rval);
  influenceNodes = intersect(levelNodes,nodes);

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::MWLSApprox::caculateApproxFun(
  double material_coords[3],
  bool calc_base,
  bool calc_diff_base,
  bool trim_influence_nodes
) {
  PetscErrorCode ierr;
  MoABErrorCode rval;
  PetscFunctionBegin;

  FTensor::Tensor1<double,3> t_material_coords(
    material_coords[0],material_coords[1],material_coords[2]
  );

  // Node coordinates
  double node_coords[3];
  FTensor::Tensor1<double*,3> t_node_coords(node_coords,&node_coords[1],&node_coords[2]);

  // Determine points in influence volume
  const double dm = maxEdgeL*dmFactor; // influence radius
  Range& influence_nodes = influenceNodes;
  if(trim_influence_nodes) {
    int ii = 0;
    int jj = 0;
    for(Range::iterator nit = influence_nodes.begin();nit!=influence_nodes.end();nit++,jj++) {
      rval = mwlsMoab.get_coords(&*nit,1,node_coords); CHKERRQ_MOAB(rval);
      t_node_coords(i) -= t_material_coords(i);
      const double r = sqrt(t_node_coords(i)*t_node_coords(i))/dm;
      if(r>1) {
        nit = influence_nodes.erase(nit);
        if(nit==influence_nodes.end()) {
          break;
        }
      }
    }
  }


  A.resize(nbBasePolynomials,nbBasePolynomials,false);
  B.resize(nbBasePolynomials,influence_nodes.size(),false);
  A.clear();
  B.clear();
  for(int d = 0;d!=3;d++) {
    diffA[d].resize(nbBasePolynomials,nbBasePolynomials,false);
    diffB[d].resize(nbBasePolynomials,influence_nodes.size(),false);
    diffA[d].clear();
    diffB[d].clear();
  }

  // Local base functions
  baseFun.resize(nbBasePolynomials,false);
  baseFun.clear();
  baseFun[0] = 1;
  for(int d = 0;d!=3;d++) {
    diffBaseFun[d].resize(nbBasePolynomials,false);
    diffBaseFun[d].clear();
  }

  {
    int ii = 0;
    for(Range::iterator nit = influence_nodes.begin();nit!=influence_nodes.end();nit++,ii++) {
      rval = mwlsMoab.get_coords(&*nit,1,node_coords); CHKERRQ_MOAB(rval);
      t_node_coords(i) -= t_material_coords(i);
      t_node_coords(i) /= dm;
      const double r = sqrt(t_node_coords(i)*t_node_coords(i));
      // Weighths
      const double w = evalWeight(r);
      const double diff_w_r = evalDiffWeight(r)/dm;
      if(nbBasePolynomials>1) {
        baseFun[1] = t_node_coords(0);
        baseFun[2] = t_node_coords(1);
        baseFun[3] = t_node_coords(2);
      }
      if(nbBasePolynomials>4) {
        baseFun[4] = t_node_coords(0)*t_node_coords(0);
        baseFun[5] = t_node_coords(1)*t_node_coords(1);
        baseFun[6] = t_node_coords(2)*t_node_coords(2);
        baseFun[7] = t_node_coords(0)*t_node_coords(1);
        baseFun[8] = t_node_coords(0)*t_node_coords(2);
        baseFun[9] = t_node_coords(1)*t_node_coords(2);
      }
      for(int k = 0;k!=nbBasePolynomials;k++) {
        const double wp = w*baseFun[k];
        for(int j = 0;j<=k;j++) {
          A(k,j) += wp*baseFun[j];
        }
        B(k,ii) = wp;
        const double diff_wp_r = diff_w_r*baseFun[k];
        for(int d = 0;d!=3;d++) {
          const double diff_wp = diff_wp_r*copysign(1.,t_node_coords(d));
          for(int j = 0;j<=k;j++) {
            diffA[d](k,j) += diff_wp*baseFun[j];
          }
          diffB[d](k,ii) = diff_wp;
        }
      }
    }
  }

  invA.resize(nbBasePolynomials,nbBasePolynomials,false);
  approxFun.resize(influence_nodes.size(),false);
  for(int d = 0;d!=3;d++) {
    diffInvA[d].resize(nbBasePolynomials,nbBasePolynomials,false);
    diffApproxFun.resize(3,influence_nodes.size(),false);
  }

  // Invert martrix
  {
    L.resize(nbBasePolynomials,nbBasePolynomials,false);
    if(cholesky_decompose(A,L)) {
      cerr << A << endl;
      SETERRQ(mField.get_comm(),MOFEM_OPERATION_UNSUCCESSFUL,"Failed to invert matrix");
    }
    invA.clear();
    for(int k = 0;k!=nbBasePolynomials;k++) {
      ublas::matrix_column<ublas::matrix<double> > mr(invA,k);
      mr(k) = 1;
      cholesky_solve(L,mr,ublas::lower());
    }
  }

  for(int d = 0;d!=3;d++) {
    ublas::matrix<double> a = prod(diffA[d],invA);
    noalias(diffInvA[d]) = -prod(invA,a);
  }

  baseFun[0] = 1;
  for(int d = 0;d!=3;d++) {
    diffBaseFun[d][0] = 0;
  }
  if(nbBasePolynomials>1) {
    baseFun[1] = 0;//material_coords[0];
    baseFun[2] = 0;//material_coords[1];
    baseFun[3] = 0;//material_coords[2];
    // DX
    diffBaseFun[0][1] = 1./dm;
    diffBaseFun[0][2] = 0;
    diffBaseFun[0][3] = 0;
    // DY
    diffBaseFun[1][1] = 0;
    diffBaseFun[1][2] = 1./dm;
    diffBaseFun[1][3] = 0;
    // DZ
    diffBaseFun[2][1] = 0;
    diffBaseFun[2][2] = 0;
    diffBaseFun[2][3] = 1./dm;
  }

  if(nbBasePolynomials>4) {
    baseFun[4] = 0;//material_coords[0]*material_coords[0];
    baseFun[5] = 0;//material_coords[1]*material_coords[1];
    baseFun[6] = 0;//material_coords[2]*material_coords[2];
    baseFun[7] = 0;//material_coords[0]*material_coords[1];
    baseFun[8] = 0;//material_coords[0]*material_coords[2];
    baseFun[9] = 0;//material_coords[1]*material_coords[2];
    // DX
    diffBaseFun[0][4] = 0;//2*material_coords[0];
    diffBaseFun[0][5] = 0;//0;
    diffBaseFun[0][6] = 0;//0;
    diffBaseFun[0][7] = 0;//material_coords[1];
    diffBaseFun[0][8] = 0;//material_coords[2];
    diffBaseFun[0][9] = 0;//0;
    // DY
    diffBaseFun[1][4] = 0;//0;
    diffBaseFun[1][5] = 0;//2*material_coords[1];
    diffBaseFun[1][6] = 0;//0;
    diffBaseFun[1][7] = 0;//material_coords[0];
    diffBaseFun[1][8] = 0;//0;
    diffBaseFun[1][9] = 0;//material_coords[2];
    // DZ
    diffBaseFun[2][4] = 0;//0;
    diffBaseFun[2][5] = 0;//0;
    diffBaseFun[2][6] = 0;//2*material_coords[2];
    diffBaseFun[2][7] = 0;//0;
    diffBaseFun[2][8] = 0;//material_coords[0];
    diffBaseFun[2][9] = 0;//material_coords[1];
  }

  ublas::matrix<double> invAB = prod(invA,B);
  ublas::vector<double> baseFunInvA = prod(baseFun,invA);

  // Calulate approximation functions
  if(calc_base) {
    noalias(approxFun) = prod(baseFun,invAB);
  }
  // cerr << approxFun << endl;

  // Caluclate direvative of base functions
  ublas::vector<double> tmp;
  tmp.resize(nbBasePolynomials,false);
  if(calc_diff_base) {
    for(int d = 0;d!=3;d++) {
      ublas::matrix_row<ublas::matrix<double> > mr(diffApproxFun,d);
      noalias(mr) = prod(diffBaseFun[d],invAB);
      noalias(tmp) = prod(baseFun,diffInvA[d]);
      mr += prod(tmp,B);
      mr += prod(baseFunInvA,diffB[d]);
    }
  }

  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::MWLSApprox::approxTagData(Tag th) {
  MoABErrorCode rval;
  PetscFunctionBegin;
  int length;
  rval = mwlsMoab.tag_get_length(th,length); CHKERRQ_MOAB(rval);
  influenceNodesData.resize(influenceNodes.size(),length,false);
  rval = mwlsMoab.tag_get_data(th,influenceNodes,&influenceNodesData(0,0));
  outData.resize(length,false);
  outDiffData.resize(3,length,false);
  PetscFunctionReturn(0);
}

CrackPropagation::MWLSApprox::VecVals& CrackPropagation::MWLSApprox::getData() {
  noalias(outData) = prod(approxFun,influenceNodesData);
  return outData;
}

CrackPropagation::MWLSApprox::MatDiffVals& CrackPropagation::MWLSApprox::getDiffData() {
  noalias(outDiffData) = prod(diffApproxFun,influenceNodesData);
  return outDiffData;
}

PetscErrorCode CrackPropagation::MWLSApprox::OpMWLSStressAtGaussPts::doWork(
  int side,EntityType type,DataForcesAndSurcesCore::EntData &data
) {
  MoABErrorCode rval;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if(type!=MBVERTEX) {
    PetscFunctionReturn(0);
  }
  if(mwlsApprox->tetsInBlock.find(getFEEntityHandle())==mwlsApprox->tetsInBlock.end()) {
    PetscFunctionReturn(0);
  }
  const int nb_gauss_pts = data.getN().size1();
  double coords[3] = {0,0,0};
  for(int gg = 0;gg!=nb_gauss_pts;gg++) {
    for(int dd = 0;dd!=3;dd++) {
      coords[dd] += getHoCoordsAtGaussPts()(gg,dd);
    }
  }
  for(int dd = 0;dd!=3;dd++) {
    coords[dd] /= nb_gauss_pts;
  }
  ierr = mwlsApprox->getInfluenceNodes(coords,true); CHKERRQ(ierr);
  Tag th;
  rval = mwlsApprox->mwlsMoab.tag_get_handle(tagName.c_str(),th); CHKERRQ_MOAB(rval);
  ierr = mwlsApprox->approxTagData(th); CHKERRQ(ierr);
  // cerr << mwlsApprox->influenceNodesData << endl;
  // Find influence nodes only once and reuse it for every guauss point on element
  // influenceNodes = mwlsApprox->influenceNodes;

  MatrixDouble& stress = mwlsApprox->stressAtGaussPts;
  stress.resize(6,nb_gauss_pts);
  for(int gg = 0;gg!=nb_gauss_pts;gg++)  {
    // mwlsApprox->influenceNodes = influenceNodes;
    // ierr = mwlsApprox->getInfluenceNodes(&getHoCoordsAtGaussPts()(gg,0),false); CHKERRQ(ierr);
    ierr = mwlsApprox->caculateApproxFun(&getHoCoordsAtGaussPts()(gg,0),true,false,false); CHKERRQ(ierr);
    CrackPropagation::MWLSApprox::VecVals& vals = mwlsApprox->getData();
    for(int ii = 0;ii!=6;ii++) {
      stress(ii,gg) = vals[ii];
    }
  }
  // cerr << stress << endl;
  PetscFunctionReturn(0);
}

PetscErrorCode CrackPropagation::MWLSApprox::OpMWLSStressRhs::doWork(
  int side,EntityType type,DataForcesAndSurcesCore::EntData &data
) {
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if(mwlsApprox->tetsInBlock.find(getFEEntityHandle())==mwlsApprox->tetsInBlock.end()) {
    PetscFunctionReturn(0);
  }
  const int nb_dofs = data.getIndices().size();
  if(!nb_dofs) {
    PetscFunctionReturn(0);
  }
  FTensor::Tensor1<double*,3> t_diff_n = data.getFTensor1DiffN<3>();
  // SIXX SIYY SIZZ SIXY SIXZ SIYZ
  // T* d00, T* d01, T* d02, T* d11, T* d12, T* d22
  MatrixDouble& stress = mwlsApprox->stressAtGaussPts;
  FTensor::Tensor2_symmetric<double*,3> t_stress(
    &stress(0,0),&stress(3,0),&stress(4,0),
    &stress(1,0),&stress(5,0),
    &stress(2,0),
    1
  );
  FTensor::Index<'i',3> i;
  FTensor::Index<'j',3> j;
  nF.resize(nb_dofs,false);
  nF.clear();
  const int nb_gauss_pts = data.getN().size1();
  for(int gg = 0;gg!=nb_gauss_pts;gg++) {
    double val = getVolume()*getGaussPts()(3,gg);
    val *= getHoGaussPtsDetJac()[gg]; ///< higher order geometry
    FTensor::Tensor1<double*,3> rhs(&nF[0],&nF[1],&nF[2],3);
    int bb = 0;
    for(;bb!=nb_dofs/3;bb++) {
      rhs(i) += val*t_stress(i,j)*t_diff_n(j);
       ++t_diff_n;
       ++rhs;
    }
    for(;bb!=data.getN().size2();bb++) {
      ++t_diff_n;
    }
    ++t_stress;
  }
  ierr = VecSetValues(
    getFEMethod()->snes_f,
    nb_dofs,
    &data.getIndices()[0],
    &nF[0],
    ADD_VALUES
  ); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
